//
// Created by sylwe on 16.07.2024.
//

#include "Solver.h"
#include "Solver2.h"

#include <BridgesAndArtPoints.h>
#include <ConnectedComponents.h>
#include <GraphInducer.h>
#include <GraphUtils.h>
#include <StandardUtils.h>
#include <Stopwatch.h>
#include "../external/FAU.h"

#include "CollectionOperators.h"

// #include "CCH.h"
#include "Kernelizer.h"
#include "Solver2BT.h"
#include "Solver2FBS.h"
#include "Solver2MIM.h"

Solver2::Solver2(GraphSP2 _g, int _L, int _vs, int _vt) : g(_g), L(_L), vs(_vs), vt(_vt) {
    N = g.size();
    revg = g.reverseGraph();
}

void Solver2::kernelize() {
    Kernelizer2 kernelizer2( g, L, vs, vt );
    kernelizer2.cnf = cnf;
    auto core2 = kernelizer2.createCore2Graph();
    g = core2.G;
    L = core2.L;
    vs = core2.vs;
    vt = core2.vt;
    N = g.size();
    revg = g.reverseGraph();
}

pair<VI, VI> Solver2::findDistancesInclusive(GraphSP2 &g, int beg, Mask S) {
    assert( !S[beg] );
    int N = g.size();
    VI dst(N, inf);
    VI par(N,inf);
    dst[beg] = 0; par[beg] = -1;
    VI neigh; neigh.reserve(N);
    neigh.push_back(beg);
    VB in_queue(N);

    for( int i=0; i<neigh.size(); i++ ) {
        int v = neigh[i];
        in_queue[v] = false;

        for( auto & arc : g[v] ) {
            int d = arc.v;
            if( dst[v] + arc.minl() < dst[d] ) {
                dst[d] = dst[v] + arc.minl();
                par[d] = v;
                if( !S[d] && !in_queue[d] ) {
                    neigh.push_back(d);
                    in_queue[d] = true;
                }
            }
        }
    }

    return make_pair(dst,par);
}

pair<VI, VI> Solver2::findDistancesInclusive(GraphSP2 &g, int beg, VB& S) {
    assert( !S[beg] );
    int N = g.size();
    VI dst(N, inf);
    VI par(N,inf);
    dst[beg] = 0; par[beg] = -1;
    VI neigh; neigh.reserve(N);
    neigh.push_back(beg);
    VB in_queue(N);

    for( int i=0; i<neigh.size(); i++ ) {
        int v = neigh[i];
        in_queue[v] = false;

        for( auto & arc : g[v] ) {
            int d = arc.v;
            if( dst[v] + arc.minl() < dst[d] ) {
                dst[d] = dst[v] + arc.minl();
                par[d] = v;
                if( !S[d] && !in_queue[d] ) {
                    neigh.push_back(d);
                    in_queue[d] = true;
                }
            }
        }
    }

    return make_pair(dst,par);
}

pair<VI, VI> Solver2::findDistances(GraphSP2 &g, int beg, VB & S) {
    int N = g.size();
    VI dst(N, inf);
    VI par(N,inf);
    dst[beg] = 0; par[beg] = -1;
    VI neigh; neigh.reserve(N);
    neigh.push_back(beg);
    VB in_queue(N,false);

    for( int i=0; i<neigh.size(); i++ ) {
        int v = neigh[i];
        in_queue[v] = false;

        for( auto & arc : g[v] ) {
            int d = arc.v;
            if( dst[v] + arc.minl() < dst[d] && !S[d] ) {
                dst[d] = dst[v] + arc.minl();
                par[d] = v;
                if( !in_queue[d] ) {
                    neigh.push_back(d);
                    in_queue[d] = true;
                }
            }
        }
    }

    return make_pair(dst,par);
}

pair<VI, VI> Solver2::findDistances(GraphSP2 &g, int beg, Mask& S) {
    // if( S.popcnt() == 0 ) return findDistances(g,beg);
    int N = g.size();
    VI dst(N, inf);
    VI par(N,inf);
    dst[beg] = 0; par[beg] = -1;
    VI neigh; neigh.reserve(N);
    neigh.push_back(beg);
    VB in_queue(N,false);

    for( int i=0; i<neigh.size(); i++ ) {
        int v = neigh[i];
        in_queue[v] = false;

        for( auto & arc : g[v] ) {
            int d = arc.v;
            if( dst[v] + arc.minl() < dst[d] && !S[d] ) {
                dst[d] = dst[v] + arc.minl();
                par[d] = v;
                if( !in_queue[d] ) {
                    neigh.push_back(d);
                    in_queue[d] = true;
                }
            }
        }
    }

    return make_pair(dst,par);
}

pair<VI, VI> Solver2::findDistances(GraphSP2 &g, int beg) {
    int N = g.size();
    VI dst(N, inf);
    VI par(N,inf);
    dst[beg] = 0; par[beg] = -1;
    VI neigh; neigh.reserve(N);
    neigh.push_back(beg);
    VB in_queue(N);

    for( int i=0; i<neigh.size(); i++ ) {
        int v = neigh[i];
        in_queue[v] = false;

        for( auto & arc : g[v] ) {
            int d = arc.v;
            if( dst[v] + arc.minl() < dst[d] ) {
                // dst[d] = dst[v] + 1;
                dst[d] = dst[v] + arc.minl();
                par[d] = v;
                if(!in_queue[d]) {
                    neigh.push_back(d);
                    in_queue[d] = true;
                }
            }
        }
    }

    return make_pair(dst,par);
}

VI Solver2::findShortestPath(GraphSP2 &g, int beg, int end, Mask S) {
    auto [dst,par] = findDistances(g,beg,S);
    int u = end;
    VI path(1,u);
    while( u != beg && u != inf ) {
        u = par[u];
        path.push_back(u);
    }
    if(u == inf) path.clear();
    reverse(ALL(path));
    return path;
}

VBIG Solver2::solve() {
    assert(vs != vt);
    auto nores = VBIG(1,0);
    if( L == 0 ) { if( vs == vt ) return VBIG(1,1); return nores; }
    if( g.size() == 0 ) return nores;
    if( g.countArcs() == 0 ) return nores;
    if( findDistances(g, vs).first[vt] > L ) return nores; // here is also case of no path at all - then dst is inf

    {
        assert( g.isValid() );
        VBIG res = splitByDominators();
        auto res_lengths = Lengths::createFromVBIG(res);
        if(res_lengths.valid()) return res;
    }

    before_kernelization:
    if(cnf.use_kernelization) {
        clog << "Starting solver2::kernelize()";
        kernelize();

        if( g.size() == 0 ) return nores;
        if( g.countArcs() == 0 ) return nores;
        if( findDistances(g, vs).first[vt] > L ) return nores; // here is also case of no path at all - then dst is inf
    }



    clog << "After kernelization (or without), G.size(): " << g.size() << ", arcs: " << g.countArcs()
        << ", L: " << L << endl;



    assert( g.isValid() );
    VBIG res = splitByDominators();
    auto res_lengths = Lengths::createFromVBIG(res);
    if(res_lengths.valid()) return res;


    // bool use_mim = cnf.use_mim;
    // bool use_fbs = cnf.use_fbs;
    // bool use_bt = cnf.solver2_use_bt_merging;
    //
    // int mim_millis_limit = 1e9;
    // int fbs_millis_limit = 1e9;
    // int bt_millis_limit = 1e9;
    //
    // if( use_mim && use_fbs && cnf.select_mim_vs_fbs ) {
    //
    //     // selecting algorithm based on some graph characteristics...
    //     Solver2FBS s2fbs(g,L,vs,vt);
    //     s2fbs.cnf = cnf;
    //     auto [order,pathwidth] = s2fbs.createFBSNodeOrder();
    //
    //     if( pathwidth <= cnf.small_pathwidth_for_fbs ) {
    //         use_mim = false; // for small pathwidth always use FBS
    //         mim_millis_limit = bt_millis_limit = 0;
    //     }
    //     else if( pathwidth > cnf.max_pathwidth_for_fbs ) {
    //         use_fbs = false; // for large pathwidth don't even consider fbs
    //         fbs_millis_limit = 0;
    //     }
    //     else if( g.size() <= cnf.small_graph_size_for_mim || L <= cnf.small_path_length_for_mim
    //         || (L - findDistances(g, vs).first[vt] <= cnf.small_length_excess_for_mim) ) {
    //         use_fbs = false; // for short paths
    //         fbs_millis_limit = 0;
    //
    //         if( pathwidth <= cnf.max_pathwidth_for_fbs ) {
    //             use_fbs = true; // for short paths
    //             mim_millis_limit = 300 * 1000; // setting time limit to 5 minutes // #TEST
    //         }
    //     }
    //
    //     else use_mim = false; // in any other case, use FBS
    // }
    //
    // if( (use_mim && !use_fbs) && g.isSymmetric() && !cnf.use_all_reduction_rules_even_for_undirected_graphs ) {
    //     // if we would not use FBS, then it might be befenicial to rerun kernelization and use removing distant arcs
    //     cnf.use_all_reduction_rules_even_for_undirected_graphs = true;
    //     goto before_kernelization;
    // }
    //
    // if( !use_mim && !use_fbs && !cnf.solver2_use_bt_merging ) {
    //     use_mim = true;
    //     mim_millis_limit = 300 * 1000;
    //     bt_millis_limit = 300 * 1000;
    // }
    // assert( use_mim || use_fbs || cnf.solver2_use_bt_merging );
    //
    // if constexpr ( cnf.use_bt_instead_of_mim ) if( use_mim ) use_bt = !( use_mim = false ); // #TEST
    //
    // // if( use_mim ) assert( !use_fbs );
    // // if( use_fbs ) assert(!use_mim);
    //
    // if(use_mim) {
    //     Solver2MIM solver( g, L, vs, vt );
    //     solver.cnf = cnf;
    //     solver.timer.setLimit("mim3", mim_millis_limit );
    //     res = solver.solveMIM3();
    //     if(res.empty()) {
    //         use_fbs = true;
    //         clog << endl << "MIM3 time limit exceeded, proceeding to FBS" << endl;
    //     }
    //     return res;
    // }
    // if( use_fbs ) {
    //     Solver2FBS solver( g, L, vs, vt );
    //     solver.cnf = cnf;
    //     res = solver.solveFBS();
    //     if(res.empty()) {
    //         use_fbs = true;
    //         clog << endl << "FBS time limit exceeded..." << endl;
    //     }
    //     return res;
    // }
    // else if( use_bt ) {
    //     Solver2BT solver( g, L, vs, vt );
    //     solver.cnf = cnf;
    //     solver.timer.setLimit("bt",bt_millis_limit);
    //     return solver.solveBTStateMerging();
    //
    //     if( cnf.solver2_use_bt_merging ){
    //         return solver.solveBTStateMerging();
    //     }
    //
    //     // if(cnf.use_bt_cch) {
    //     //      res = solver.solveBTCCH();
    //     // }
    //     // else
    //     if( cnf.use_bt_caching ) {
    //         res = solver.solveBT<true>();
    //     }
    //     else {
    //         res = solver.solveBT<false>();
    //     }
    // }else {
    //     clog << "Solver algorithm not chosen, running MIM by default" << endl;
    //     Solver2MIM solver( g, L, vs, vt );
    //     solver.cnf = cnf;
    //     res = solver.solveMIM3();
    //     // exit(1);
    // }

    for( int i=0; i<N; i++ ) {
        for( auto & arc : g[i] ) {
            if(!arc.lengths.valid()) {
                DEBUG(g);
                DEBUG(g[i]);
                DEBUG(arc.lengths);
                assert( arc.lengths.valid() );
            }
        }
    }

    {
        constexpr int FBS = 1, BT = 2, MIM = 3;
        int K = 1000;
        VPII algos = { {FBS,10*K}, {BT,60*K}, { FBS, 1e9 }, {MIM,1e9} };

        auto removeAlg = [&](int alg) {
            int r = 0;
            for( int i=0; i<algos.size(); i++ ) {
                if( algos[i].first != alg ) algos[r++] = algos[i];
            }
            algos.resize(r);
        };

        bool use_fbs = cnf.use_fbs;
        bool use_bt = cnf.solver2_use_bt_merging;
        bool use_mim = cnf.use_mim;
        bool select_alg = cnf.select_algorithm;

        if( g.size() > cnf.max_graph_size_for_mim ) use_mim = false;

        if( !use_fbs ) removeAlg(FBS);
        if( !use_bt ) removeAlg(BT);
        if( !use_mim ) removeAlg(MIM);
        if(algos.empty()) select_alg = true;

        if( select_alg ) {
            algos.clear();
            bool fbs_usable = use_fbs;

            int pathwidth = 1e9;
            if( use_fbs ) {
                Solver2FBS s2fbs(g,L,vs,vt);
                s2fbs.cnf = cnf;
                auto [order,pw] = s2fbs.createFBSNodeOrder();
                pathwidth = pw;
                DEBUG(pathwidth);

                if( pathwidth <= cnf.small_pathwidth_for_fbs ) algos = { {FBS, 1e9} };
                else if( pathwidth > cnf.max_pathwidth_for_fbs ) fbs_usable = false;
            }

            if( use_mim ) {
                if( (!use_fbs || !fbs_usable) && !use_bt ) algos = {{MIM,1e9}};
                else if( g.size() <= cnf.small_graph_size_for_mim || L <= cnf.small_path_length_for_mim ) {
                    int millis = 300*K;
                    if( !use_fbs && !use_bt ) millis = 1e9;
                    algos.emplace_back( MIM, millis );
                }else if(  L <= cnf.small_path_length_for_mim + 4 ) {
                    int millis = 120*K;
                    if( !use_fbs && !use_bt ) millis = 1e9;
                    algos.emplace_back( MIM, millis );
                }
            }

            DEBUG(use_fbs);
            DEBUG(use_mim);
            DEBUG(use_bt);
            DEBUG(fbs_usable);

            if( use_bt ) {
                int millis = ( (algos.empty() && !fbs_usable ) ? 1e9 : 150*K );
                algos.emplace_back(BT,millis);
            }

            if( use_fbs && fbs_usable ) algos.emplace_back( FBS,1e9 );

            DEBUG(algos);

            if( algos.empty() ) {
                ENDL(5);
                clog << "ERROR" << endl << endl << "algorithm not chosen in Solver2::solve(), running bt..." << endl;
                ENDL(5);
                algos = {{BT,1e9}};
            }
        }

        bool single_method = true;
        for( int i=1; i<algos.size(); i++ ) if( algos[i].first != algos[i-1].first ) single_method = false;
        if( single_method ) algos = { { algos[0].first, 1e9 } };

        clog << endl << "Algorithm order for the instance: " << endl;
        for( auto [alg, millis] : algos ) {
            string alg_name;
            if( alg == FBS ) alg_name = "FBS";
            if( alg == MIM ) alg_name = "MIM";
            if( alg == BT ) alg_name = "BT";

            clog << "Run " << alg_name << " for at most " << (millis+999)/ 1000 << " seconds" << endl;
        }


        for( auto [alg, millis] : algos ) {
            string alg_name;
            if( alg == FBS ) alg_name = "FBS";
            if( alg == MIM ) alg_name = "MIM";
            if( alg == BT ) alg_name = "BT";

            clog << endl << "Running algorithm " << alg_name << " with time limit set to "
                 << (millis+999)/ 1000 << " seconds" << endl;

            if( alg == FBS ) {
                Solver2FBS s(g,L,vs,vt);
                s.timer.setLimit("fbs",millis);
                res = s.solveFBS();
            }

            if( alg == BT ) {
                Solver2BT s(g,L,vs,vt);
                s.timer.setLimit("bt",millis);
                res = s.solveBTStateMerging();
            }

            if( alg == MIM ) {
                Solver2MIM s(g, L, vs,vt);
                s.timer.setLimit("mim3",millis);
                res = s.solveMIM3();
            }


            if( !res.empty() ) {
                clog << endl << "Algorithm " << alg_name << " solved instance within "
                                 << (millis+999)/ 1000 << " seconds" << endl;
                break;
            }

            clog << endl << "Algorithm " << alg_name << " did not manage to solve instance within "
                                 << (millis+999)/ 1000 << " seconds" << endl;
        }
    }

    return res;
}

VBIG Solver2::splitByDominators() {
    constexpr bool debug = false;

    auto V = g.getStruture();
    VBIG res_for_length(L+1,0);

    auto dominators = Kernelizer::findDominators(V,vs,vt);


    assert( g.isValid() );

    if( !dominators.empty() ) {
        if constexpr (debug) DEBUG(N);
        if constexpr (debug) DEBUG(L);
        if constexpr (debug) DEBUG(dominators);
        VB was(N);
        for(int d : dominators) was[d] = true;
        dominators.clear();
        auto path = findShortestPath(g,vs,vt, Mask::zeros(N));
        int path_length = 0;
        for( int i=1; i<path.size(); i++ ) path_length += g.getArcMinl( path[i-1], path[i] );
        VI sp_lengths;

        int seg_len = 0;
        for( int i=0; i<path.size(); i++ ) {
            int v = path[i];
            if(i > 0) seg_len += g.getArcMinl( path[i-1], path[i] );
            if( was[v] ) { // v is a dominator
                dominators.push_back(v);
                sp_lengths.push_back(seg_len);
                seg_len = 0;
            }
        }
        for(int d : dominators) was[d] = false;

        sp_lengths.push_back(seg_len);
        if constexpr (debug) DEBUG(path.size());
        if constexpr (debug) DEBUG(dominators);
        if constexpr (debug) DEBUG(sp_lengths);

        VI split_nodes = {vs};
        split_nodes += dominators;
        split_nodes.push_back(vt);
        if constexpr (debug) DEBUG(split_nodes);

        if( split_nodes.size() > 2 ) {
            Lengths res( 0,0);
            res[0] = 1;

            int suml = accumulate(ALL(sp_lengths),0);
            assert(suml == path_length);

            if constexpr (debug) clog << "SP2: Splitting paths vs -> vt: " << vs << " -> " << vt << " with L: " << L << " into following cases:" << endl;
            for( int i=1; i<split_nodes.size(); i++ ) {
                int a = split_nodes[i-1], b = split_nodes[i], newL =  L - ( suml - sp_lengths[i-1] );
                newL = max(newL,0);
                if constexpr (debug) clog << "Path " << a << " -> " << b << " with L: " << newL << endl;

                was[b] = true;
                auto [dst,par] = findDistances( g, a, was );
                VI nodes(1,b);
                for( int j=0; j<N; j++ ) if( !was[j] && dst[j] != inf ) nodes.push_back(j);
                for( int d : nodes ) was[d] = true;
                was[b] = false;

                if constexpr (debug) clog << "\tInducing graph by nodes: " << nodes << endl;

                auto indg = Kernelizer2::InducedGraphSP2::induce(g, nodes);
                if constexpr (debug) {
                    if( g.isSymmetric() ) assert( indg.G.isSymmetric() );
                    DEBUG(indg.G);
                }
                assert(indg.G.isValid());
                Solver2 solver( indg.G, newL, indg.perm[a], indg.perm[b] );
                solver.cnf = cnf;
                // if constexpr (debug) DEBUG(indg);

                auto res_indg = solver.solve();
                if constexpr (debug) DEBUG(res_indg);
                Lengths lindg = Lengths::createFromVBIG(res_indg);
                if constexpr (debug) DEBUG(lindg);
                // Lengths lindg( 0, (int)res_indg.size()-1 );
                // for( int j=1; j<res_indg.size(); j++ ) lindg[j] = res_indg[j];
                res = Lengths::multiplyLengths( res, lindg );
            }

            if constexpr (debug) DEBUG(res);

            for(int i=res.minl(); i <= min(res.maxl(), (int)res_for_length.size()-1); i++) {
                res_for_length[i] += res[i];
            }
            return res_for_length;
        }

    }else {
        if constexpr (debug) {
            clog << "GraphSP2 cannot be split into disjoint cases, there are no dominators..." << endl;
        }
    }

    assert( g.isValid() );

    return VBIG();
}


bool Solver2::existsPath(GraphSP2 &g, int a, int b, Mask &S) {
    if( S.popcnt() == 0 ) return Solver2::findDistances( g, a ).first[b] != inf;

    VB was(g.size());
    function<bool(int)> dfs = [&](int num) {
        was[num] = true;
        if( num == b ) return true;
        for(auto & arc : g[num]) {
            int d = arc.v;
            if( !was[d] && !S[d] && dfs(d) ) return true;
        }
        return false;
    };
    return dfs(a);
}

bool Solver2::existsPath(GraphSP2 &g, int a, int b, VB &S) {
    VB was(g.size());
    function<bool(int)> dfs = [&](int num) {
        was[num] = true;
        if( num == b ) return true;
        for(auto & arc : g[num]) {
            int d = arc.v;
            if( !was[d] && !S[d] && dfs(d) ) return true;
        }
        return false;
    };
    return dfs(a);
}


