//
// Created by sylwester on 4/15/23.
//

#include <StandardUtils.h>
#include <GraphUtils.h>
#include "GVVisualizer.h"


GVVisualizer::_format GVVisualizer::default_node_specs = {
        {"color","black"},
        {"penwidth","1.5"},
        {"style","filled"},
        {"fillcolor","white"},
        {"shape","ellipse"}
};

GVVisualizer::_format GVVisualizer::default_edge_specs = {
        {"color","black"},
        {"style","solid"}, // e.g. "dashed", "dotted"
        {"dir","none"} // e.g. "forward", "backward", "both", "none"
};

GVVisualizer::_format GVVisualizer::default_subgraph_specs = {
//        {"label",     "Partition"},
        {"style",     "filled"},
        {"fillcolor", "gray"}
};


void GVVisualizer::visualizeBasic(VVI &V, GVConfig cnf, string file_name, bool open_when_built) {

    ofstream f( file_name + ".gv" );
    f << "digraph " + cnf.graph_name << " {" << endl;

    if( cnf.directed ) cnf.edge_dir = "forward";

    int N = V.size();

    for( int i=0; i<N; i++ ){ // write nodes
        f << i;

        f << "[shape=" << cnf.node_shape;
        f << ",style=" << cnf.node_style;
        f << ",color=" << cnf.node_color;
        if( cnf.node_style == "filled" ) f << ",fillcolor=" << cnf.node_fillcolor;
        f << ",label=" << i;

        f << "];";
        f << endl;
    }

    //*****************************

    for( int i=0; i<N; i++ ){ // write edges
        for( int d : V[i] ){
            f << i << " -> " << d;
            f << "[color=" << cnf.edge_color;
            f << ",style=" << cnf.edge_style;
            f << ",dir=" << cnf.edge_dir;
//            f << ",label=\"(" << i << "," << d << ")\"";
            f << "];" << endl;
        }
    }

    f << "}" << endl; // close digraph
    f.close(); // close the stream


    string command = "dot -Tpdf " + file_name + ".gv -o " + file_name + ".pdf";
    auto suppr = system( command.c_str() );

    if(open_when_built){
        command = "xreader " + file_name + ".pdf &";
        suppr = system( command.c_str() );
    }

}

void GVVisualizer::addNode(string name, map<string, string> specs) {
    nodes.emplace_back(name,specs);
}

void GVVisualizer::addEdge(string name1, string name2, GVVisualizer::_format specs) {
    edges.emplace_back( name1, name2, specs );
}

void GVVisualizer::markSubgraph(vector<string> nds, _format specs) {
    subgraphs.emplace_back(nds, specs);
}

void GVVisualizer::markSubgraph(VI nodes, GVVisualizer::_format specs) {
    vector<string> nds;
    for( auto & s : nodes ) nds.push_back(to_string(s));
    markSubgraph(nds, specs);
}

void GVVisualizer::visualize(string file_name, bool open_when_built) {
    ofstream f( file_name + ".gv" );
    f << "digraph {" << endl;
    f << "\tsize=\"" << "7,10\";" << endl; // mark graph sizes to fit into an A4 paper size


    // visualize nodes
    for( auto & [name,specs] : nodes ){
        f << name;
        if( specs.empty() ) f << ";" << endl;
        else{
            f << "\t[";
            int cnt = 0;
            for( auto & [key, value] : specs ){
                if( cnt++ > 0 ) f << ",";
                f << key << "=" << ( key == "label" ? "\"" + value + "\"" : value);
            }
            f << "];" << endl;
        }
    }
    f << endl;

    // visualize edges
    for( auto & [name1, name2, specs] : edges ){
        f << name1 << " -> " << name2;
        if( specs.empty() ) f << ";";
        else{
            f << "\t[";
            int cnt = 0;
            for( auto & [key, value] : specs ){
                if( cnt++ > 0 ) f << ",";
//                f << key << "=" << value;
                f << key << "=" << ( key == "label" ? "\"" + value + "\"" : value);
            }
            f << "];" << endl;
        }
    }

    // add clusters
    int subgraph_cnt = 0;
    for( auto & [nds,specs] : subgraphs ){
//        auto tokens = StandardUtils::split(specs["label"]);
//        string cluster_name = ""; for( auto xx : tokens ) cluster_name += xx;
//        f << "subgraph cluster" << cluster_name << " {" << endl;
        f << "subgraph cluster" << to_string(subgraph_cnt++) << "{" << endl;

        for( auto & [key, value] : specs ){
            f << "\t" << key << "=" << ( key == "label" ? "\"" + value + "\"" : value) << ";" << endl;
        }

        for( auto & name : nds ) f << "\t" << name << ";" << endl; // write nodes in the subgraph
        f << "}" << endl << endl;
    }

    f << "}" << endl; // close digraph
    f.close(); // close the stream

    // create pdf with given graph using GraphViz
    string command;
    if(out_format == PDF ) command = "dot -Tpdf " + file_name + ".gv -o " + file_name + ".pdf";
    else if(out_format == PNG ) command = "dot -Tpng " + file_name + ".gv -o " + file_name + ".png";
    auto suppr = system( command.c_str() ); // create the file

    if(open_when_built){
//        if(out_format == PDF ) command = "xreader " + file_name + ".pdf &";
        // if(out_format == PDF ) command = "google-chrome " + file_name + ".pdf &";
        if(out_format == PDF ) command = "chromium " + file_name + ".pdf &";
        // else if(out_format == PNG ) command = "google-chrome " + file_name + ".png &";
        else if(out_format == PNG ) command = "chromium " + file_name + ".png &";

        suppr = system( command.c_str() );
    }

}

void
GVVisualizer::preprocessInducedGraph( VVI & V, VI nds, bool directed, bool add_induced_visualization,
                                      _format node_specs,
                                      _format edge_specs,
                                      _format subgraph_specs){

    int N = V.size();
    set<int> zb(ALL(nds));
    if( directed ) edge_specs["dir"] = "forward";
    else edge_specs["dir"] = "none";

    auto all_edges = GraphUtils::getGraphEdges(V, directed);
    VPII in_edges;
    for( auto & [a,b] : all_edges ) if( zb.count(a) && zb.count(b) ) in_edges.emplace_back(a, b );
    VPII outsider_edges;


    set<int> outsiders;
    for( int i=0; i<N; i++ ){
        for( int d : V[i] ){
            if( zb.count(i) == 0 && zb.count(d) ){
                outsiders.insert(i);
                outsider_edges.emplace_back(i,d);
            }
            if( zb.count(d) == 0 && zb.count(i) ){
                outsiders.insert(d);
                outsider_edges.emplace_back(d,i);
            }
        }
    }

    for( int v : outsiders ){
        auto specs = node_specs;
        specs["color"] = "black";
        specs["shape"] = "box";
        specs["style"] = "rounded";
        addNode( v, specs );
    }

    for( int d : nds ) addNode( d, node_specs );
    for( auto & [a,b] : in_edges ) addEdge(a, b, edge_specs);

    auto outsider_edges_specs = edge_specs;
    outsider_edges_specs["color"] = "black";
    outsider_edges_specs["style"] = "dotted";
    for( auto & [a,b] : outsider_edges ) addEdge(a, b, outsider_edges_specs);

    markSubgraph( nds, subgraph_specs );


    if(add_induced_visualization){
        map<int,int> mapper;
        for(int i=0; i<nds.size(); i++) mapper[nds[i]] = i;

        for( auto & [name,specs] : nodes ){
            int v = stoi(name);
            if( zb.count(v) == 0 ) continue;
            if( specs["label"] == "" ) specs["label"] = name;
//            specs["label"] += " i(" + to_string(mapper[v]) + ")";
            specs["label"] += " | i" + to_string(mapper[v]);
        }
    }

}

void GVVisualizer::preprocessFullGraph(VVI &V, bool directed, GVVisualizer::_format node_specs,
                                       GVVisualizer::_format edge_specs, GVVisualizer::_format subgraph_specs) {
    VI nds(V.size());
    iota(ALL(nds),0);
    preprocessInducedGraph(V, nds, directed, false, node_specs, edge_specs, subgraph_specs);
    subgraphs.erase( subgraphs.begin() );
}





