//
// Created by sylwe on 07/08/2024.
//

#include "CCH.h"

#include <GraphUtils.h>
#include <routingkit/nested_dissection.h>

CCH::CCH(VVI V, VVI W)  {
    int maxw = -1;
    for( auto & v : W ) if( !v.empty() ) maxw = max(maxw, *max_element(ALL(v)));
    infty = V.size() * maxw;
    N = V.size();

    createTailsHeadsWeights(V,W);
}

int CCH::dst(int a, int b) {
    query.reset().add_source(a).add_target(b).run();
    return query.get_distance();
}

auto CCH::dst(auto sources, auto ends) {
    VVU res;
    for( auto s : sources ) {
        query.reset().add_source(s).pin_targets(ends);
        res.push_back( query.run_to_pinned_targets().get_distances_to_targets() );
    }
    return res;
}

CCH::VU CCH::dst(int s, VU ends) {
    query = CustomizableContractionHierarchyQuery(metric);

    query.reset().add_source(s).pin_targets(ends);
    return query.run_to_pinned_targets().get_distances_to_targets();
}

bool CCH::hasPath(int a, int b) {
    return dst(a,b) < infty;
}

void CCH::excludeNode(int v, VB & mask) {
    constexpr bool debug = false;
    partial_customization.reset();
    if constexpr (debug) clog << "Excluding node: " << v << endl;

    for(int i=first_of[v]; i < first_of[v+1]; i++) {
        int a = tail[i];
        int b = head[i];
        if( !mask[b] ) {
            if constexpr (debug) clog << "\texcluding arc " << a << " -> " << b << endl;
            weight[i] = infty;
            partial_customization.update_arc(i);
        }
    }

    for(int i=first_of_rev[v]; i < first_of_rev[v+1]; i++) {
        int a = tail_rev[i];
        int b = head_rev[i];
        if( !mask[b] ) {
            if constexpr (debug) clog << "\texcluding arc " << a << " -> " << b << endl;
            auto rev_id = rev_arc_id[i];
            weight[rev_id] = infty;
            partial_customization.update_arc(rev_id);
        }
    }

    partial_customization.customize(metric);
}

void CCH::includeNode(int v, VB &mask) {
    constexpr bool debug = false;
    if constexpr (debug) clog << "Including node: " << v << endl;

    partial_customization.reset();

    for(int i=first_of[v]; i < first_of[v+1]; i++) {
        int a = tail[i];
        int b = head[i];
        if( !mask[b] ) {
            if constexpr (debug) clog << "\tincluding arc " << a << " -> " << b << endl;
            weight[i] = init_weights[i];
            partial_customization.update_arc(i);
        }
    }

    for(int i=first_of_rev[v]; i < first_of_rev[v+1]; i++) {
        int a = tail_rev[i];
        int b = head_rev[i];
        if( !mask[b] ) {
            if constexpr (debug) clog << "\tincluding arc " << a << " -> " << b << endl;
            auto rev_id = rev_arc_id[i];
            weight[rev_id] = init_weights[rev_id];
            partial_customization.update_arc(rev_id);
        }
    }

    partial_customization.customize(metric);
}


void CCH::constructNodeOrderAndBuild() {

    constructNodeOrder();

    // auto logg = [](std::string msg){cerr << msg << endl;}; // #TEST - just for logging...
    auto logg = [](std::string msg){};
    cch = CustomizableContractionHierarchy(node_order, tail, head,  logg, true);

    metric = CustomizableContractionHierarchyMetric(cch, weight);
    metric.customize();

    query = CustomizableContractionHierarchyQuery(metric);
    partial_customization = CustomizableContractionHierarchyPartialCustomization(cch);
}

void CCH::createTailsHeadsWeights(VVI V, VVI W) {
    int M = GraphUtils::countEdges(V,true);
    vector<Triple<unsigned>> arcs;
    arcs.reserve(M);
    for( int i=0; i<V.size(); i++ ) {
        for(int j=0; j<V[i].size(); j++) {
            arcs.emplace_back( i, V[i][j], W[i][j] );
        }
    }

    auto createTailHeadWeight = [&arcs, N=N]( auto &tail, auto & head, auto & weight, auto & first_of, auto & init_weights ) {
        sort(ALL(arcs), [&](auto t1, auto t2) {
            if( t1.st != t2.st ) return t1.st < t2.st;
            if( t1.nd != t2.nd ) return t1.nd < t2.nd;
            return t1.rd < t2.rd;
        });

        head.resize(arcs.size());
        tail.resize(arcs.size());
        weight.resize(arcs.size());
        init_weights.resize(arcs.size());
        first_of = VI(N+1,-1);
        first_of[N] = arcs.size();

        map<PII,int> ids;

        for( int i=0; i<arcs.size(); i++ ) {
            tail[i] = arcs[i].st;
            head[i] = arcs[i].nd;
            weight[i] = arcs[i].rd;

            if( first_of[ tail[i] ] == -1 ) first_of[ tail[i] ] = i;
            init_weights[i] = arcs[i].rd;

            ids[ {tail[i], head[i]} ] = i;
        }

        for( int i=N-1; i>=0; i--) if( first_of[i] == -1 ) first_of[i] = first_of[i+1];
    };

    createTailHeadWeight(tail, head, weight, first_of, init_weights);




    for( auto & tr : arcs ) swap( tr.st, tr.nd ); // creating reverse arcs

    vector<Quadruple<int>> quads(arcs.size()); // for each reverse arc we want to find index of the original arc
    for( int i=0; i<arcs.size(); i++ ) quads[i] = Quadruple<int>(arcs[i].st, arcs[i].nd, arcs[i].rd, i);
    sort(ALL(quads), [&](auto t1, auto t2) {
       if( t1.st != t2.st ) return t1.st < t2.st;
       if( t1.nd != t2.nd ) return t1.nd < t2.nd;
       return t1.rd < t2.rd;
    });

    rev_arc_id = VI(arcs.size(),-1);
    for( int i=0; i<quads.size(); i++ ) {
        rev_arc_id[i] = quads[i].th;
        assert( PII(tail[ rev_arc_id[i] ], head[rev_arc_id[i]] ) == PII(quads[i].nd, quads[i].st) );
    }

    for( int i=0; i<quads.size(); i++ ) {
        arcs[i].st = quads[i].st;
        arcs[i].nd = quads[i].nd;
        arcs[i].rd = quads[i].rd;
    }

    createTailHeadWeight(tail_rev, head_rev, weight_rev, first_of_rev, init_weights_rev);


}

void CCH::constructNodeOrder() {
    // clog << "creating node order" << endl;
    vector<float> lat(N,0), lon(N,0); // #TEST - fnd node order using other method or assign valid coordinates
    node_order = compute_nested_node_dissection_order_using_inertial_flow(N, tail, head, lat,lon);

    // clog << "in constructNodeOrder(), node_order: " << node_order << endl;
}
