//
// Created by sylwe on 10/09/2024.
//

#include "HPP.h"

#include <GraphUtils.h>
#include <StandardUtils.h>

#include "CollectionOperators.h"
#include "Solver2MIM.h"

bool HPP::containsHP(int vs, int vt) {
    constexpr bool debug = true;

    int N2 = 2*N+4;

    VVI c(N2, VI(N2,0));
    constexpr int inf = 1e9;

    int s = 2*N;
    int t = 2*N+1;
    int ss = 2*N+2;
    int tt = 2*N+3;

    VVI W(N2);
    W[s].push_back(2*vs);
    W[2*vt+1].push_back(t);

    for( int i=0; i<N; i++ ) {
        W[2*i].push_back(2*i+1);
        for(int d : V[i]) {
            W[2*i+1].push_back(2*d);
            c[2*i+1][2*d] = 1;
        }
    }

    for(int i=0; i<N; i++) {
        W[ss].push_back(2*i+1);
        c[ss][2*i+1] = 1;
        W[2*i].push_back(tt);
        c[2*i][tt] = 1;
    }

    W[ss].push_back(2*vs);
    c[ss][2*vs] = 1;

    W[s].push_back(tt);
    c[s][tt] = 1;

    W[2*vt+1].push_back(tt);
    c[2*vt+1][tt] = 1;

    W[ss].push_back(t);
    c[ss][t] = 1;

    W[t].push_back(s);
    c[t][s] = inf;

    auto revW = GraphUtils::reverseGraph(W);
    for( int i=0; i<N2; i++ ) {
        W[i] += revW[i];
        StandardUtils::makeUnique(W[i]);
    }

    VB was(N,0);
    int depth = 0;
    VI path;
    auto indent = [&depth](){ for(int i=0; i<depth; i++) clog << "  "; };

    function<bool(int,int)> fun = [&]( int num, int par ) {
        if constexpr (debug){ indent(); clog << "entering " << num << endl; }

        if( num == tt ) {
            if constexpr (debug) {
                indent();
                path.push_back(num);
                clog << "Found augmentation, path: " << path << endl;
                path.pop_back();
            }
            return true;
        }

        depth++;
        was[num] = true;
        path.push_back(num);
        bool found = false;

        for( int d : W[num] ) {
            if( !was[d] && c[num][d] > 0 && fun(d,num) ) {
                c[num][d]--;
                c[d][num]++;
                found = true;
                break;
            }
        }

        depth--;
        was[num] = false;
        path.pop_back();

        if constexpr (debug) {
            indent();
            clog << "leaving " << num << ", going back to " << par << endl;
        }

        return found;
    };

    int flow = 0;
    while( fun(ss,-1) ) {
        flow++;
        if constexpr (debug) {
            ENDL(2);
            clog << "Starting " << flow+1 << "-th iteration" << endl;
            DEBUG(flow);
        }
    }
    DEBUG(flow);
    DEBUG( c[s][t] );
    assert( flow <= N+2 );

    return flow == N+2;
}

void HPP::test() {

    int N = 7;
    VVI V = {
        {1,3,4},
        {2},
        {3},
        {4,6},
        {5,6},
        {},
        {2,4,5}
    };


    HPP hpp(V);

    int vs = 0, vt = 5;
    DEBUG(PII(vs,vt));
    auto has_hp = hpp.containsHP(vs,vt);
    DEBUG(has_hp);
    ENDL(5);



    vs = 1, vt = 5;
    DEBUG(PII(vs,vt));
    has_hp = hpp.containsHP(vs,vt);
    DEBUG(has_hp);
    ENDL(5);


    vs = 0, vt = 6;
    DEBUG(PII(vs,vt));
    has_hp = hpp.containsHP(vs,vt);
    DEBUG(has_hp);
    ENDL(5);


    clog << "Now random tests: " << endl;

    int TESTS = 1000;

    N = 10;
    int A = N*N/3;
    V = VVI(N);

    UniformIntGenerator rnd(0,N-1);

    ofstream strout( "logging.txt" );

    int cur_a = 0;
    auto checkForV = [&]() {
        HPP hpp(V);

        for( int vs = 0; vs < N; vs++ ) {
            for( int vt =0; vt < N; vt++ ) {
                if( vs == vt ) continue;

                bool has_hp_fwd = hpp.containsHP(vs,vt);

                int L = N-1;
                Solver2MIM solver2(GraphSP2(V), L, vs,vt);
                auto res_mim3 = solver2.solveMIM3();
                bool has_hp_mim3 = ( res_mim3[L] > 0 );

                strout << "cur_a: " << cur_a << ", vs: " << vs << ", vt: " << vt << endl;
                strout << "has_hp_fwd: " << has_hp_fwd << endl;
                strout << "has_hp_mim3: " << has_hp_mim3 << endl;
                strout << "res_mim3: " << res_mim3 << endl << endl;

                if( has_hp_fwd != has_hp_mim3 ) {
                    DEBUG(has_hp_fwd);
                    DEBUG(has_hp_mim3);
                    DEBUG(res_mim3);

                    int hash = (1+rnd.rand()) * (1+rnd.rand()) * (1+rnd.rand());
                    ofstream str( "HP_failure_graph_" + to_string(hash) + ".col" );
                    GraphUtils::writeInstanceToFile( str, V, vs,vt,L, true );

                    assert( has_hp_fwd == has_hp_mim3 );
                }
            }
        }

        strout << endl << endl << "*************************************" << endl << endl;
    };

    for( int id = 0; id < TESTS; id++ ) {
        clog << "Considering test #" << id+1 << endl;
        strout << "Considering test #" << id+1 << endl;

        VVB was(N,VB(N));
        cur_a = 0;
        while( cur_a < A ) {
            clog << "\tadding arc #" << cur_a << endl;
            strout << endl << "\tadding arc #" << cur_a << endl;

            int a = 0, b = 0;
            while( a == b || was[a][b] ) {
                a = rnd.rand();
                b = rnd.rand();
            }
            cur_a++;
            V[a].push_back(b);
            was[a][b] = true;

            checkForV();
        }
    }
}
