//
// Created by sylwe on 15/09/2024.
//

#include "Solver2FBSD.h"

#include <FAU.h>
#include <Stopwatch.h>

#include "BiconnectedFinder.h"
#include "Kernelizer.h"
#include "Solver2.h"
#include "CollectionOperators.h"
#include "DistanceFinder2.h"
#include "Solver2FBS.h"


Solver2FBSD::Solver2FBSD(GraphSP2 _g, int _L, int _vs, int _vt) : g(_g), L(_L), vs(_vs), vt(_vt) {
    N = g.size();
    revg = g.reverseGraph();
}

VBIG Solver2FBSD::solveFBS3DForNodeOrder(VI order) {
    constexpr bool debug = false;
    {
        VB was(N);
        if( !Solver2::existsPath(g,vs,vt,was) ) return VBIG(L+1,0);
    }

    timer.start("fbs");

    if constexpr (debug) {
        clog << "Entering solveFBS3ForNodeOrder" << endl;
        DEBUG(g);
        DEBUG(order);
    }

    Stopwatch sw;
    sw.start("solveFBS3DForNodeOrder");

    VB was(N);
    VB in_neigh(N), out_neigh(N);
    set<int> S; // nodes on boundary - needed only for debugging


    VI S_nodes(32,-1);
    VI ind_in_S(N,-1);
    auto getFirstFreeSIndex = [&]() {
        int r = 0;
        while( S_nodes[r] != -1 ) r++;
        return r;
    };

    vector<unordered_map<int,int>> arc_id_mapper(N);
    VVI arc_id_mapper_tab;
    vector<Lengths> all_arcs_lengths;
    bool use_map_for_arc_id_mapper = g.countArcs() > 3'000;
    // bool use_map_for_arc_id_mapper = false;
    if(!use_map_for_arc_id_mapper) arc_id_mapper_tab.assign(N, VI(N,-1));
    for( int i=0; i<N; i++ ) {
        for( auto & arc : g[i] ) {
            if(use_map_for_arc_id_mapper) arc_id_mapper[i][ arc.v ] = all_arcs_lengths.size();
            else arc_id_mapper_tab[i][arc.v] = all_arcs_lengths.size();
            all_arcs_lengths.push_back(arc.lengths);
            all_arcs_lengths.back().minimize();
        }
    }
    auto getArcId = [&](int a, int b) {
        if( use_map_for_arc_id_mapper ) return arc_id_mapper[a][b];
        else return arc_id_mapper_tab[a][b];
    };

    vector<HashType> all_arcs_hashes; all_arcs_lengths.reserve( all_arcs_lengths.size() );
    for( auto & l : all_arcs_lengths ) all_arcs_hashes.push_back( l.hash() );



    DEBUG(PII(vs,vt));



    HashType cur_vis_hash = 0;
    HashType cur_vis_hash_immutable = 0;

    VVI underV = g.getStruture(); // here it is enough, as g is symmetric

    vector<Edge> trg_edges;
    vector<EdgeH> trg_edges_h;
    ULL total_masks_trimmed = 0;


    ULL expansion_iteration_cnt = 0;
    constexpr int expansion_iteration_shift_bits = 6; // up to 256 nodes
    constexpr bool include_iter_cnt_in_hash = true;
    auto updateHashWithIterationCnt = [&](HashType & h ) {
        if constexpr (include_iter_cnt_in_hash) {
            h.hash &= (ULLONG_MAX >> expansion_iteration_shift_bits);
            h.hash |= ( expansion_iteration_cnt << (64-expansion_iteration_shift_bits) );
        }
    };


    auto addEdge = [&]( HashType h1, HashType h2, int e1, int e2 = -1  ) {
        assert(h1 != h2);
        // if constexpr (debug) clog << "Adding to trg_edges_h arc " << h1 << " -> " << h2 << endl;
        trg_edges_h.emplace_back(h1,h2,e1,e2);
    };


    auto markNeighbors = [&](int v, const bool val) {
        for( auto & arc : g[v] ) if(was[arc.v] && S.contains(arc.v)) out_neigh[arc.v] = val;
        for( auto & arc : revg[v] ) if(was[arc.v] && S.contains(arc.v)) in_neigh[arc.v] = val;
    };

    int DST_MAX_MASK_SIZE = 5;
    int E = g.countArcs();
    if( E > 100 ) DST_MAX_MASK_SIZE = 4;
    if( E > 1'000 ) DST_MAX_MASK_SIZE = 3;
    if( E > 5'000 ) DST_MAX_MASK_SIZE = 2;

    int INIT_DST_MAX_MASK_SIZE = DST_MAX_MASK_SIZE;


    VVVI distances(1);
    int MMB = 0;
    VVI masks_bits(INIT_DST_MAX_MASK_SIZE+1);

    auto createDistances = [&](int v) { // creates distances for given subset of taken nodes in S
        DST_MAX_MASK_SIZE = INIT_DST_MAX_MASK_SIZE;
        if( DST_MAX_MASK_SIZE > 0 && S.size() >= 15 ) DST_MAX_MASK_SIZE--;
        if( DST_MAX_MASK_SIZE > 0 && S.size() >= 20 ) DST_MAX_MASK_SIZE--;

        int MAX_SNODES_IND = 0;
        for( int i=0; i<S_nodes.size(); i++ ) if(S_nodes[i] != -1) MAX_SNODES_IND = i+1;

        DistanceFinder2 df(N);
        if( (1 << MAX_SNODES_IND) > MMB ) {
            for( int mask = MMB; mask < (1 << MAX_SNODES_IND); mask++ ) {
                int b = popcount((unsigned)mask);
                if( b < masks_bits.size() ) masks_bits[b].push_back(mask);
            }
            MMB = (1 << MAX_SNODES_IND);
        }

        // for( int mask = 0; mask < (1 << MAX_SNODES_IND); mask++ ) {
        for( int mb = 0; mb <= DST_MAX_MASK_SIZE; mb++ ) {
            for( int mask : masks_bits[mb] ) {
                // if( popcount((unsigned)mask) > DST_MAX_MASK_SIZE ) continue;
                assert( popcount((unsigned)mask) == mb );

                bool can_continue = true;
                for( int i=0; i<S_nodes.size(); i++ ) if( (mask & (1<<i)) && S_nodes[i] == -1 ) can_continue = false;
                if(!can_continue) continue;

                while( distances.size() <= mask ) distances.emplace_back(); // for pathwidth <= 32 it will work :)
                distances[mask] = VVI(MAX_SNODES_IND);

                S.insert(v);
                for(int d : S) was[d] = false;
                for( int i=0; i<32; i++ ) if( mask & (1<<i) ) was[ S_nodes[i] ] = true;
                for(int d : S) {
                    int a = ind_in_S[d];
                    distances[mask][a] = VI(MAX_SNODES_IND,inf);

                    // if( !was[d] ) {
                    auto temp = was[d];
                    was[d] = false;
                    // auto dst = Solver2::findDistancesInclusive(g, d, was).first;
                    df.findDistancesInclusive(g, d, was);
                    for( int i=0; i<S_nodes.size(); i++ ) {
                        // if( S_nodes[i] != -1 ) distances[mask][a][i] = dst[ S_nodes[i] ];
                        if( S_nodes[i] != -1 ) distances[mask][a][i] = df.getDst(S_nodes[i]);
                    }

                    // last element will always have distance to vt if vt is not in S
                    // if( ind_in_S[vt] == -1 ) distances[mask][a].push_back( dst[vt] );
                    if( ind_in_S[vt] == -1 ) distances[mask][a].push_back( df.getDst(vt) );
                    was[d] = temp;
                    // }
                }
                for(int d : S) was[d] = true;
                was[v] = true;
                S.erase(v);
            }
        }
    };


    VB is_state_end(N,false);
    auto getLooseNodes = [&]( FBSS3D & st ) {
        VI loose;
        is_state_end[st.vs_node] = true;
        for(auto [a,b] : st.ends) is_state_end[a] = is_state_end[b] = true;
        for( int d : S ) if( !is_state_end[d] ) loose.push_back(d);
        is_state_end[st.vs_node] = false;
        for(auto [a,b] : st.ends) is_state_end[a] = is_state_end[b] = false;
        return loose;
    };

    constexpr bool use_small_neighborhood_trimming = true;
    constexpr bool trim_biconnected = false; // time consuming!
    constexpr bool use_additional_tm = true;

    VB _was_(N);
    VB _was2_(N);
    vector<unsigned short> neigh_in_S(N);
    VVB ends(N, VB(N,false));
    auto isEndsNeighborhoodLargeEnough = [&]( FBSS3D & st ) {
        if constexpr (!use_small_neighborhood_trimming) return true;

        if( st.ends.empty()  ) return true;
        if( st.vs_node == vt && !st.ends.empty() ) return false; // this is another rule, but can be done here :)

        _was_ = was;
        for( int i=0; i<S_nodes.size(); i++ ) if( S_nodes[i] != -1 ) _was_[ S_nodes[i] ] = true;

        {
            int cnt = 0;
            for( int d : underV[st.vs_node] ) if( !_was_[d] ) cnt++;
            if( st.vs_node != vt && cnt == 0 ) return false;
        }

        for( auto [a,b] : st.ends ) {
            int cnt = 0;
            for( int d : underV[a] ) if( !_was_[d] ) cnt++;
            if( a != vt && cnt == 0 ) return false;

            cnt = 0;
            for( int d : underV[b] ) if( !_was_[d] ) cnt++;
            if( b != vt && cnt == 0 ) return false;
        }

        int neigh_size = 0;
        auto checkNode = [&](int d) { if( !_was_[d] ){ neigh_size++; _was_[d] = true; } };

        fill(ALL(_was2_),false);

        constexpr bool exclude_uniquely_assignable_nodes = false;
        if constexpr (exclude_uniquely_assignable_nodes){
            bool changes_done = true;

            while(changes_done) {
                changes_done = false;

                for( auto [a,b] : st.ends ) {
                    for( int d : underV[a] ) neigh_in_S[d] = 0;
                    for( int d : underV[b] ) neigh_in_S[d] = 0;
                }
                for( auto [a,b] : st.ends ) {
                    if( a != vt && !_was2_[a] ) for( int d : underV[a] ) neigh_in_S[d]++;
                    if( b != vt && !_was2_[b] ) for( int d : underV[b] ) neigh_in_S[d]++;
                }
                for( auto [a,b] : st.ends ) {
                    if(a != vt) for( int d : underV[a] ) {
                        if( !_was_[d] && !_was2_[a] && neigh_in_S[d] == 1 ) {
                            _was2_[a] = changes_done = true;
                            neigh_size++;
                        }
                    }
                    if(b != vt) for( int d : underV[b] ) {
                        if( !_was_[d] && !_was2_[b] && neigh_in_S[d] == 1 ) {
                            _was2_[b] = changes_done = true;
                            neigh_size++;
                        }
                    }
                }
            }
        }

        int valid_ends = st.ends.size();
        for( auto [a,b] : st.ends ) {
            if( a != vt && !_was2_[a] ) for( auto d : underV[a] ) checkNode(d);
            if( b != vt && !_was2_[b] ) for( auto d : underV[b] ) checkNode(d);
        }

        bool res = true;
        if( st.vs_node != vt && st.containsNode(vt) ) {
            if constexpr (debug) clog << "CASE1 " << endl;
            res = (neigh_size >= valid_ends);
        }
        else res = (neigh_size >= valid_ends+1);

        if constexpr (debug) if(!res) {
            clog << "State " << st << " has too small ends-neighborhood" << endl;
            DEBUG(neigh_size);
            DEBUG(valid_ends);
        }
        return res;
    };


    auto getAdditionalTm = [&](FBSS3D & st, PII m, int v)-> int {
        _was_ = was;
        for( int i=0; i<S_nodes.size(); i++ ) if( S_nodes[i] != -1 ) _was_[ S_nodes[i] ] = true;
        _was_[v] = false;

        m.first |= ( 1 << ind_in_S[st.vs_node] );
        for( auto [a,b] : st.ends ) {
            m.first |= ( 1 << ind_in_S[a] );
            m.first |= ( 1 << ind_in_S[b] );
        }

        int neigh_size = 0;
        auto checkNode = [&](int d) { if( !_was2_[d] ){ neigh_size++; _was2_[d] = true; } };

        bool changes_done = true;
        int tm = 0;
        while( changes_done ) {
            changes_done = false;
            for( int i=0; i<32; i++ ) {
                if( S_nodes[i] != -1 && S_nodes[i] != vt && ( m.first & (1<<i) ) == 0 ) {
                    neigh_size = 0;
                    _was2_ = _was_;

                    int a = S_nodes[i];

                    for( int d : underV[a] ) checkNode(d);
                    if( neigh_size <= 1 ) {
                        tm |= (1 << i);
                        m.first |= (1 << i);
                        changes_done = true;
                    }
                }
            }
        }

        return tm;
    };

    BiconnectedFinder bf(N);
    DistanceFinder2 df(N);
    // returns pair where first boolean value tells whether state can be trimmed and second one - if state cannot be
    // trimmed - tells additional tm value to mark some loose nodes as taken
    auto trimBiconnectedOrFindDispensibleLooseNodes = [&]( FBSS3 & st, int v ) -> pair<bool, int> {
        if constexpr (!trim_biconnected) return {false,0};
        _was_ = was;
        for( int i=0; i<32; i++ ) if( S_nodes[i] != -1 ) _was_[ S_nodes[i] ] = false;


        fill(ALL(_was2_),false);
        for( auto [a,b] : st.ends ) _was2_[a] = _was2_[b] = true;

        for( auto [a,b] : st.ends ) {
            underV[a].push_back(b);
            underV[b].push_back(a);
            ends[a][b] = ends[b][a] = true;
        }

        auto clear = [&]() {
            for( auto [a,b] : st.ends ) {
                underV[a].pop_back();
                underV[b].pop_back();
                ends[a][b] = ends[b][a] = false;
            }
        };

        Mask mask(N);
        for( int i=0; i<N; i++ ) mask.set( i, _was_[i] );

        df.findDistancesFBS(underV, st.vs_node, mask, ind_in_S, v, ends);
        if( df.getDst(vt) == inf ){ clear(); return {true, 0}; }
        VI path;
        for( int u=vt; u != -1 ; u = df.getPar(u) ) path.push_back(u);
        reverse(ALL(path));
        Mask biconnected_trimmed = bf.getNodesAfterBiconnectedReduction2FBS(underV, mask, st.vs_node, vt,
                                    path, ind_in_S, v, ends );
        clear();

        int tm = 0;
        for( int i=0; i<N; i++ ) {
            if(biconnected_trimmed[i]) continue;
            if( _was2_[i] ){ return {true,0}; }
            if( ind_in_S[i] != -1 && i != st.vs_node && !_was2_[i] ) tm |= ( 1 << ind_in_S[i] );
        }


        return {false, tm};
    };

    auto trimDispensibleLooseNodes = [&](FBSS3D & st, VI & loose, int v) {
        if constexpr ( !use_additional_tm ) return;

        auto tm2 = getAdditionalTm(st, {0,0}, v);
        for( int i=(int) loose.size()-1; i>=0; i--) {
            // if(st.vs_node == vs) continue;
            int a = loose[i];
            if( a == vt ) continue;

            if( tm2 & (1 << ind_in_S[a]) ) {

                // ENDL(2);
                // clog << "Adding " << v << " as start node for state " << st << endl;
                // clog << "Removing node " << a << " from loose nodes " << loose << endl;
                swap(loose[i], loose.back());
                loose.pop_back();
            }
        }
    };


    constexpr int BUCKETS_BITS = 10;
    constexpr int BUCKETS = (1<<BUCKETS_BITS);
    constexpr int BH = BUCKETS-1;

    auto getHashForTakenMask = [&](int m) {
        HashType h;
        for( int i=0; i<32; i++ ) if(m & (1<<i)) h ^= node_hashes[ S_nodes[i] ];
        if constexpr (include_iter_cnt_in_hash) h.hash &= ( ULLONG_MAX >> expansion_iteration_shift_bits );
        return h;
    };

    vector<vector<FBSS3D>> states(BUCKETS);
    vector<vector<FBSS3D>> new_states(BUCKETS);

    auto countAllStates = [&](auto & sts) {
        return accumulate(ALL(sts), 0, [](int a, auto & vec){ return a + vec.size(); });
    };

    vector<HashType> S_hashes(N, 0);
    VI tm_shifter(N,-1);
    VB is_vsnode_or_end(N);

    constexpr bool apply_shifters = true; // works very well for dense parts of graphs

    auto applyShifters = [&](int &m, FBSS3D & st ) {
        is_vsnode_or_end[st.vs_node] = true;
        for( auto [a,b] : st.ends ) is_vsnode_or_end[a] = is_vsnode_or_end[b] = true;

        for( int i=0; i<32; i++ ) {
            if( m & (1 << i) ) {
                int u = S_nodes[i];
                if( tm_shifter[u] == -1 ){ m ^= (1<<i); continue; }

                int prev = i;
                while( tm_shifter[u] != u ) {
                    if( !is_vsnode_or_end[u] && ( m & (1 << ind_in_S[u]) ) == 0 ) {
                        // clog << "Shifting node " << u << " to " << tm_shifter[u] << endl;
                        m ^= ( 1 << prev );
                        m ^= ( 1 << ind_in_S[u] );
                        prev = ind_in_S[u];
                    }
                    u = tm_shifter[u];
                }
            }
        }

        is_vsnode_or_end[st.vs_node] = false;
        for( auto [a,b] : st.ends ) is_vsnode_or_end[a] = is_vsnode_or_end[b] = false;
    };

    auto placeIntoNewBucket = [&]( FBSS3D & st ) {
        if constexpr (use_additional_tm || apply_shifters) st.makeUnique(st.taken_masks);

        if constexpr (debug) {
            if(!st.isCorrectState(ind_in_S)) {
                clog << endl << "!!!!" << endl << "State is incorrect!" << endl;
                DEBUG(st);
                assert( st.isCorrectState(ind_in_S) );
            };
        }
        if( st.taken_masks.empty() ) return;

        new_states[st.hash.hash & BH].push_back( st );
        if constexpr (debug) clog << "\t\tCreated and added new_state: " << st << endl;
    };



    auto considerTransition = [&]( PII m, int tm, HashType st_hash, int minl_addition, FBSS3D & new_st,
        int a, int b, int c = -1, int d = -1 ) {
        auto h1 = st_hash ^ getHashForTakenMask(m.first);

        m.first |= tm;
        if constexpr (apply_shifters) applyShifters(m.first, new_st);

        auto h2 = new_st.hash ^ getHashForTakenMask(m.first);
        if(h1 != h2) addEdge( h1,h2, (a == -1 ? -1 : getArcId(a,b)), (c == -1 ? -1 :getArcId(c,d))  );
        new_st.taken_masks.emplace_back(m.first, m.second + minl_addition);
    };

    auto addAsIgnoredNode = [&]( int v, const FBSS3D & st, VI & loose ) {
        if constexpr (debug) clog << "\tAdding as ignored node" << endl;
        auto new_st = st;
        VPII().swap(new_st.taken_masks);
        new_st.recalculateHash();
        new_st.hash ^= cur_vis_hash; // do not create 'new state' for
        if constexpr (include_iter_cnt_in_hash) updateHashWithIterationCnt(new_st.hash);
        int tm = 0;

        if( !isEndsNeighborhoodLargeEnough( new_st ) ) return;
        if constexpr (use_additional_tm) {
            int tm2 = getAdditionalTm( new_st, {0,0}, v );
            tm |= tm2;
        }

        for( PII m : st.taken_masks ) {
            if constexpr (debug) clog << "\tconsidering m: " << m << endl;
            considerTransition( m, tm, st.hash, 0, new_st, -1, -1 );
        }

        placeIntoNewBucket(new_st);
    };

    auto addAsVsNode = [&](int v, const FBSS3D & st, VI & loose) {
        if constexpr (debug) clog << "\tAdding as vs_node" << endl;
        if( !in_neigh[st.vs_node] ) return;

        auto new_st = st;
        new_st.vs_node = v;
        VPII().swap(new_st.taken_masks);
        new_st.minl += g.getArcMinl(st.vs_node,v);
        new_st.recalculateHash();
        new_st.hash ^= cur_vis_hash;
        if constexpr (include_iter_cnt_in_hash) updateHashWithIterationCnt(new_st.hash);
        int tm = (1 << ind_in_S[st.vs_node]);
        int minl_addition = g.getArcMinl(st.vs_node,v);

        if( !isEndsNeighborhoodLargeEnough( new_st ) ) return;
        if constexpr (use_additional_tm) {
            int tm2 = getAdditionalTm( new_st, {0,0}, v );
            tm |= tm2;
        }

        for( PII m : st.taken_masks ) {
            assert( (m.first & (1 <<  ind_in_S[st.vs_node] )) == 0 ); // st.vs_node cannot be taken in this case...
            if constexpr (debug) clog << "\tconsidering m: " << m << endl;
            considerTransition( m, tm, st.hash, minl_addition, new_st, st.vs_node, v );
        }

        placeIntoNewBucket(new_st);
    };

    auto addAsStartNode = [&](int v, FBSS3D & st, VI loose) {
        trimDispensibleLooseNodes( st,loose, v );

        { // adding as start node to an existing loose node
            if constexpr (debug) clog << "\tAdding as start node: v -> loose" << endl;

            for( int a : loose ) {
                if( !out_neigh[a] ) continue;

                auto new_st = st;
                VPII().swap(new_st.taken_masks);
                new_st.ends.emplace_back( v, a );
                new_st.minl += g.getArcMinl(v,a);
                new_st.recalculateHash();
                new_st.hash ^= cur_vis_hash;
                if constexpr (include_iter_cnt_in_hash) updateHashWithIterationCnt(new_st.hash);
                int minl_addition = g.getArcMinl(v,a);
                int tm = 0;

                if( !isEndsNeighborhoodLargeEnough( new_st ) ) continue;
                if constexpr (use_additional_tm) {
                    int tm2 = getAdditionalTm( new_st, {0,0}, v );
                    tm |= tm2;
                }

                for( PII m : st.taken_masks ) {
                    if( m.first & (1 << ind_in_S[a]) ) continue; // if a is taken, then cannot crete new state...
                    considerTransition( m, tm, st.hash, minl_addition, new_st, v,a );
                }

                placeIntoNewBucket(new_st);
            }
        }

        { // adding as start node to an existing path end
            if constexpr (debug) clog << "\tAdding as start node: v -> ends" << endl;
            for( int i=0; i<st.ends.size(); i++ ) {
                for( int swp = 0; swp < 2; swp++) {
                    int a = st.ends[i].first;
                    if( !out_neigh[a] ) {
                        swap( st.ends[i].first, st.ends[i].second );
                        continue;
                    }

                    auto new_st = st;
                    VPII().swap(new_st.taken_masks);
                    new_st.ends[i].first = v;
                    new_st.minl += g.getArcMinl(v,a);
                    new_st.recalculateHash();
                    new_st.hash ^= cur_vis_hash;
                    if constexpr (include_iter_cnt_in_hash) updateHashWithIterationCnt(new_st.hash);
                    int tm = ( 1 << ind_in_S[a] ); // these nodes will be added to taken...
                    int minl_addition =  g.getArcMinl(v,a);

                    if( !isEndsNeighborhoodLargeEnough( new_st ) ) {
                        swap( st.ends[i].first, st.ends[i].second );
                        continue;
                    }
                    if constexpr (use_additional_tm) {
                        int tm2 = getAdditionalTm( new_st, {0,0}, v );
                        tm |= tm2;
                    }

                    for( PII m : st.taken_masks ) {
                        assert( !(m.first & (1 << ind_in_S[a])) ); // since we add v to path end, then a cannot be in taken_masks
                        considerTransition( m, tm, st.hash, minl_addition, new_st, v,a );
                    }

                    placeIntoNewBucket(new_st);

                    swap( st.ends[i].first, st.ends[i].second );
                }
            }
        }

    };

    auto addAsMiddleNode = [&](int v, FBSS3D & st, VI loose) {
        trimDispensibleLooseNodes( st,loose, v );

        { // adding as middle node, vs_node -> v -> loose
            if constexpr (debug) clog << "\tAdding as middle node : vs_node -> v -> loose" << endl;
            for( int a : loose ) {
                if( !in_neigh[st.vs_node] ) continue;
                if( !out_neigh[a] ) continue;

                auto new_st = st;
                VPII().swap(new_st.taken_masks);
                new_st.vs_node = a;
                new_st.minl += g.getArcMinl(st.vs_node,v) + g.getArcMinl(v,a);
                new_st.recalculateHash();
                new_st.hash ^= cur_vis_hash;
                if constexpr (include_iter_cnt_in_hash) updateHashWithIterationCnt(new_st.hash);
                int tm = ( 1 << ind_in_S[st.vs_node] ) | ( 1 << ind_in_S[v] );
                int minl_addition = g.getArcMinl(st.vs_node,v) + g.getArcMinl(v,a);

                if( !isEndsNeighborhoodLargeEnough( new_st ) ) continue;
                if constexpr (use_additional_tm) {
                    int tm2 = getAdditionalTm( new_st, {0,0}, v );
                    tm |= tm2;
                }

                for( PII m : st.taken_masks ) {
                    if( m.first & (1 << ind_in_S[a]) ) continue;
                    considerTransition( m, tm, st.hash, minl_addition, new_st, st.vs_node, v, v, a );
                }

                placeIntoNewBucket(new_st);
            }

        }

        { // adding as middle node, vs_node -> v -> path end
            if constexpr (debug) clog << "\tAdding as middle node : vs_node -> v -> ends" << endl;
            for( auto & [a,b] : st.ends ) {
                if( !in_neigh[st.vs_node] ) continue;

                for(int swp = 0; swp < 2; swp++) {
                    if( !out_neigh[a] ) {
                        swap(a,b);
                        continue;
                    }
                    auto new_st = st;
                    VPII().swap(new_st.taken_masks);
                    new_st.vs_node = b;
                    new_st.removeEnds({a,b});
                    new_st.minl += g.getArcMinl(st.vs_node,v) + g.getArcMinl(v,a);
                    new_st.recalculateHash();
                    new_st.hash ^= cur_vis_hash;
                    if constexpr (include_iter_cnt_in_hash) updateHashWithIterationCnt(new_st.hash);
                    int tm = ( 1 << ind_in_S[st.vs_node] ) | ( 1 << ind_in_S[v] ) | ( 1 << ind_in_S[a] );
                    int minl_addition = g.getArcMinl(st.vs_node,v) + g.getArcMinl(v,a);

                    if( !isEndsNeighborhoodLargeEnough( new_st ) ) {
                        swap(a,b);
                        continue;
                    }
                    if constexpr (use_additional_tm) {
                        int tm2 = getAdditionalTm( new_st, {0,0}, v );
                        tm |= tm2;
                    }

                    for( PII m : st.taken_masks ) {
                        // all taken masks are passed, as no loose node is used...
                        considerTransition( m, tm, st.hash, minl_addition, new_st, st.vs_node, v, v, a );
                    }

                    placeIntoNewBucket(new_st);
                    swap(a,b);
                }
            }
        }

        { // adding as middle node, loose -> v -> loose
            if constexpr (debug) clog << "\tAdding as middle node : loose -> v -> loose" << endl;
            for( int a : loose ) {
                if( !in_neigh[a] ) continue;

                for( int b : loose ) {
                    if( a >= b ) continue;
                    if( !out_neigh[b] ) continue;

                    auto new_st = st;
                    VPII().swap( new_st.taken_masks );
                    new_st.ends.emplace_back(a,b);
                    new_st.minl += g.getArcMinl(a,v) + g.getArcMinl(v,b);
                    new_st.recalculateHash();
                    new_st.hash ^= cur_vis_hash;
                    if constexpr (include_iter_cnt_in_hash) updateHashWithIterationCnt(new_st.hash);
                    int tm = ( 1 << ind_in_S[v] );
                    int minl_addition =  g.getArcMinl(a,v) + g.getArcMinl(v,b);

                    if( !isEndsNeighborhoodLargeEnough( new_st ) ) continue;
                    if constexpr (use_additional_tm) {
                        int tm2 = getAdditionalTm( new_st, {0,0}, v );
                        tm |= tm2;
                    }

                    for( PII m : st.taken_masks ) {
                        if( m.first & ( 1 << ind_in_S[a] ) ) continue;
                        if( m.first & ( 1 << ind_in_S[b] ) ) continue;
                        considerTransition( m, tm, st.hash, minl_addition, new_st, a, v,v,b );
                    }

                    placeIntoNewBucket(new_st);
                }
            }
        }

        { // adding as middle node, loose -> v -> path end
            if constexpr (debug) clog << "\tAdding as middle node : loose -> v -> ends" << endl;
            for( int a : loose ) {
                if( !in_neigh[a] ) continue;

                for( auto & [b,c] : st.ends ) {
                    for( int swp = 0; swp < 2; swp++ ) {
                        if( !out_neigh[b] ) {
                            swap(b,c);
                            continue;
                        }

                        auto new_st = st;
                        VPII().swap( new_st.taken_masks );
                        new_st.removeEnds( {b,c} );
                        new_st.ends.emplace_back(a,c);
                        new_st.minl += g.getArcMinl(a,v) + g.getArcMinl(v,b);
                        new_st.recalculateHash();
                        new_st.hash ^= cur_vis_hash;
                        if constexpr (include_iter_cnt_in_hash) updateHashWithIterationCnt(new_st.hash);
                        int tm = ( 1 << ind_in_S[v] ) | ( 1 << ind_in_S[b] );
                        int minl_addition = g.getArcMinl(a,v) + g.getArcMinl(v,b);

                        if( !isEndsNeighborhoodLargeEnough( new_st ) ) {
                            swap(b,c);
                            continue;
                        }
                        if constexpr (use_additional_tm) {
                            int tm2 = getAdditionalTm( new_st, {0,0}, v );
                            tm |= tm2;
                        }

                        for( PII m : st.taken_masks ) {
                            if( m.first & ( 1 << ind_in_S[a] ) ) continue;
                            considerTransition( m, tm, st.hash, minl_addition, new_st, a,v,v,b );
                        }

                        placeIntoNewBucket(new_st);
                        swap(b,c);
                    }
                }
            }
        }

        { /* adding as middle node, path_end -> v -> loose */ }

        { // adding as middle node, path_end -> v -> path_end
            if constexpr (debug) clog << "\tAdding as middle node : ends -> v -> ends" << endl;
            for( int k1=0; k1 < st.ends.size(); k1++ ) {
                auto & [a,b] = st.ends[k1];
                for( int swp1= 0; swp1 < 2; swp1++ ){
                    if( !in_neigh[b] ) {
                        swap(a,b);
                        continue;
                    }

                    for( int k2=k1+1; k2 < st.ends.size(); k2++ ) {
                        auto & [c,d] = st.ends[k2];
                        for( int swp2 = 0; swp2 < 2; swp2++ ){
                            if( a == c || a == d || b == c || b == d || !out_neigh[c] ) {
                                swap(c,d);
                                continue;
                            }

                            auto new_st = st;
                            VPII().swap( new_st.taken_masks );
                            new_st.removeEnds({a,b});
                            new_st.removeEnds({c,d});
                            new_st.ends.emplace_back(a,d);
                            new_st.minl += g.getArcMinl(b,v) + g.getArcMinl(v,c);
                            new_st.recalculateHash();
                            new_st.hash ^= cur_vis_hash;
                            if constexpr (include_iter_cnt_in_hash) updateHashWithIterationCnt(new_st.hash);
                            int tm = ( 1 << ind_in_S[b] ) | ( 1 << ind_in_S[v] ) | ( 1 << ind_in_S[c] );
                            int minl_addition = g.getArcMinl(b,v) + g.getArcMinl(v,c);

                            if( !isEndsNeighborhoodLargeEnough( new_st ) ) {
                                swap(c,d);
                                continue;
                            }
                            if constexpr (use_additional_tm) {
                                int tm2 = getAdditionalTm( new_st, {0,0}, v );
                                tm |= tm2;
                            }

                            for(PII m :st.taken_masks) {
                                // all taken masks are passed, since no loose nodes are used
                                considerTransition( m, tm, st.hash, minl_addition, new_st, b,v,v,c );
                            }

                            placeIntoNewBucket(new_st);
                            swap(c,d);
                        }
                    }
                    swap(a,b);
                }
            }
        }
    };

    auto mergeFBSS3DStates = [&]( vector<FBSS3D> & sts ) {
        sort(ALL(sts));

        int p = 0, q = 0, r = 0;

        while( p < sts.size() ){
            q = p;
            while( q < sts.size() && (sts[p] == sts[q]) ) q++;
            for( int i=p+1; i<q; i++ ){
                if constexpr (debug) clog << "------>" << endl << "merging state " << sts[i] << " to " << sts[p] << endl;
                sts[p].taken_masks += sts[i].taken_masks;
            }
            if( p+1 < q ) FBSS3D::makeUnique(sts[p].taken_masks);
            swap( sts[r++], sts[p] );
            p = q;
        }

        while( sts.size() > r ) sts.pop_back();
    };

    auto mergeStatesSameNeighborhood = [&](auto & states) {
        return; // #TEST #CAUTION - this function does not work correctly yet... :(

        vector<HashType> S_hashes(N, -1);
        for( int i=0; i<32; i++ ) if( S_nodes[i] != -1 ) {
            int v = S_nodes[i];
            for( int d : underV[v] ) if( !was[d] ) {
                S_hashes[v] ^= node_hashes2[d] + all_arcs_hashes[ getArcId(v,d) ];
            }
        }

        UniformIntGenerator rnd(0, ULLONG_MAX);
        HashType H1 = rnd.rand(), H2 = rnd.rand();

        auto getHashForState = [&]( FBSS3D & st ) {
            HashType res = S_hashes[ st.vs_node ] + node_hashes2[st.vs_node];

            // belowe we take += instead of ^= to distinguish between 1,3,5,... pairs with the same hash of [a,b]
            HashType h = 0;
            for( auto [a,b] : st.ends ) h ^= S_hashes[a] + S_hashes[b];
            h += H1;
            res ^= h;

            return res;
        };


        auto mergeByEnds = [&]() {
            vector< pair< HashType, FBSS3D > > to_merge;
            to_merge.reserve( countAllStates(states) );

            for( auto & bucket : states ) {
                for( auto & st : bucket ) to_merge.emplace_back( getHashForState(st), st );
                bucket.clear();
            }

            sort( ALL(to_merge), []( auto & a, auto & b ) {
                if( a.first != b.first ) return a.first < b.first;
                // else return a.second.taken_masks.size() > b.second.taken_masks.size();
                // else return a.second.taken_masks.size() > b.second.taken_masks.size();
                return a.second.minl < b.second.minl;
            } );

            VB is_vs_or_end(N);

            int p=0, q = 0, r = 0;
            while( p < to_merge.size() ) {
                q = p;
                while( q < to_merge.size() && to_merge[p].first == to_merge[q].first ) {
                    // if hashes are identical, then states should have been merged in mergeFBSS3DStates
                    q++;
                }
                if( p+1 < q ) {
                    is_vs_or_end[to_merge[p].second.vs_node] = true;
                    for( auto [a,b] : to_merge[p].second.ends ) is_vs_or_end[a] = is_vs_or_end[b] = true;
                }
                for( int i=p+1; i<q; i++ ) {
                    assert( to_merge[p].second.hash != to_merge[i].second.hash );
                    assert( to_merge[p].second.vs_node == to_merge[i].second.vs_node );
                    // DEBUG(S_nodes);
                    // clog << endl << "MERGING BY ENDS, trying to merge state " << to_merge[i].second << " to state " << to_merge[p].second << endl;


                    auto & tms = to_merge[i].second.taken_masks;

                    for( int k=(int)tms.size()-1; k>=0; k-- ) {
                        PII m = tms[k];
                        bool is_disjoint = true;
                        for( int j=0; j<32; j++ ) if( m.first & (1 << j) && is_vs_or_end[ S_nodes[j] ] ) is_disjoint = false;
                        if( !is_disjoint ) {
                            // clog << "\tcannot merge mask " << FBSS3D::getSetBits(m.first) << endl;
                            continue;
                        }

                        // clog << "\tVALID end-merge found! transitioning mask " << FBSS3D::getSetBits(m.first) << endl;
                        //  {
                        //     clog << "Neighbors outside S:" << endl;
                        //     set<int> nodes;
                        //     nodes.insert(to_merge[p].second.vs_node);
                        //     nodes.insert(to_merge[i].second.vs_node);
                        //     for( auto [a,b] : to_merge[p].second.ends ){ nodes.insert(a); nodes.insert(b); }
                        //     for( auto [a,b] : to_merge[i].second.ends ){ nodes.insert(a); nodes.insert(b); }
                        //     for( int v : nodes ){
                        //         clog << "out-neigh[" << v << "]: ";
                        //         for( int d : underV[v] ) if(!was[d]) clog << d << " ";
                        //         clog << endl;
                        //     }
                        // }

                        considerTransition( m, 0, to_merge[i].second.hash, 0, to_merge[p].second, -1, -1 );

                        to_merge[p].second.minl = min( to_merge[p].second.minl, m.second ); // #TEST

                        swap( tms[k], tms.back() ); // remove mask m from tmss
                        tms.pop_back();
                    }
                }
                if( p+1 < q ) {
                    FBSS3D::makeUnique( to_merge[p].second.taken_masks );
                    is_vs_or_end[to_merge[p].second.vs_node] = false;
                    for( auto [a,b] : to_merge[p].second.ends ) is_vs_or_end[a] = is_vs_or_end[b] = false;
                }
                // assert(false && "This is wrong - we can only create transition if to_merge[p]'s vs_node ]" "and ends do not overlap with taken nodes of to_merge[i]");
                for( int i=p; i<q; i++ ) if( !to_merge[i].second.taken_masks.empty() ) swap( to_merge[r++], to_merge[i] );
                p = q;
            }
            while( to_merge.size() > r ) to_merge.pop_back();

            for( auto & [h,st] : to_merge ) states[st.hash.hash & BH].push_back( st );
        };

        mergeByEnds();

    };

    auto getMstEdges = [&](FBSS3D& st, int taken_mask1 = 0, int taken_mask2 = 0) {
        vector<Triple<int>> edges;
        edges.reserve( (S.size()+2) * (S.size()+2) / 4 );

        if constexpr (debug)
            clog << endl << "Finding MST edges for state " << st << endl;

        auto checkPair = [&](int a, int b) {
            if constexpr (debug)
                clog << "Considering pair (" << a << ", "  << b << ")" << endl;

            int aa = ind_in_S[a];
            assert( aa != -1 );
            int bb = ind_in_S[b];
            if( bb == -1 ) assert( b == vt );

            int d1 = ( (bb == -1) ? distances[taken_mask1][aa].back() : distances[taken_mask1][aa][bb] );
            int d2 = ( (bb == -1) ? distances[taken_mask2][aa].back() : distances[taken_mask2][aa][bb] );

            int _dst = max( d1,d2 );
            if( _dst != inf ) {
                edges.emplace_back( _dst, a, b );
            }
        };

        checkPair(st.vs_node, vt);

        for( auto [a,b] : st.ends ) {
            checkPair(st.vs_node,a);
            checkPair(st.vs_node,b);
            checkPair(a,vt);
            checkPair(b,vt);

            for( auto [c,d] : st.ends ) {
                if( a == c || a == d || b == c || b == d ) continue;
                checkPair(a,c);
                checkPair(a,d);
                checkPair(b,c);
                checkPair(b,d);
            }
        }
        if constexpr (debug) {
            clog << "edges: " << endl;
            for(auto [w,a,b] : edges) clog << "\t(" << w << "," << a << "," << b << ")" << endl;
        }

        FAU fau(N);
        for( auto [a,b] : st.ends ) fau.Union(a,b);
        sort(ALL(edges), []( auto & a, auto & b ){ return a.st < b.st; });
        vector<Triple<int>> res;
        for( auto & [w,a,b] : edges ) {
            if( fau.Find(a) != fau.Find(b) ) {
                if constexpr (debug) clog << "\tuniting nodes " << a << " and " << b << endl;
                res.emplace_back(w,a,b);
                fau.Union(a,b);
            }
        }

        bool is_connected_mst = ( fau.Find(st.vs_node) == fau.Find(vt) );
        for( auto [a,b] : st.ends ) if( fau.Find(a) != fau.Find(st.vs_node) ) is_connected_mst = false;

        return make_pair(is_connected_mst,res);
    };

    auto canTrimMST = [&](FBSS3D & st, int taken_mask1 = 0, int taken_mask2 = 0) {
        auto [is_connected,edges] = getMstEdges(st, taken_mask1, taken_mask2);

        if constexpr (debug) if( !is_connected ) {
            clog << "MST created for state st: " << st << " is not connected, trimming!" << endl;
            clog << "edges: " << endl;
            for(auto [w,a,b] : edges) clog << "\t(" << w << "," << a << "," << b << ")" << endl;
        }

        if( !is_connected ) return true;

        LL sumw = accumulate( ALL(edges), 0ll, [](LL a, auto & e){ return a + e.st; } );
        if( sumw >= inf || sumw + st.minl > L ) {
            if constexpr (debug) {
                clog << endl << "Length conditon violated for st: " << st << " and edges: " << endl;
                for(auto [w,a,b] : edges) clog << "\t(" << w << "," << a << "," << b << ")" << endl;
            }
            return true;
        }

        return false;
    };


    auto canTrimPermutation = [&]( FBSS3D & st, VPII & trimmed_masks ) {
        if constexpr (debug) clog << "canTrimPermutation, st: " << st << endl;

        VPII pairs(1, {-1, st.vs_node});
        pairs += st.ends;
        pairs.emplace_back( vt, -1 );

        sort( pairs.begin()+1, pairs.begin() + 1 + st.ends.size() );

        int taken_mask1, taken_mask2;
        VI taken_masks;
        constexpr bool use_multiple_tms = false;
        if constexpr (use_multiple_tms){
            int p = 1;
            unsigned mask = 0;
            while( p < pairs.size() ) {
                int a = pairs[p].first;
                if( a != -1 && ind_in_S[a] != -1 && popcount(mask) < DST_MAX_MASK_SIZE ) mask |= (1 << ind_in_S[a]);
                if( popcount(mask) == DST_MAX_MASK_SIZE ){ taken_masks.push_back(mask); mask = 0; }

                a = pairs[p].second;
                if( a != -1 && ind_in_S[a] != -1 && popcount(mask) < DST_MAX_MASK_SIZE ) mask |= (1 << ind_in_S[a]);
                if( popcount(mask) == DST_MAX_MASK_SIZE ){ taken_masks.push_back(mask); mask = 0; }

                p++;
            }
            if(mask) taken_masks.push_back(mask);
        }

        auto dst = [&](int a, int b) {
            if( a == -1 || b == -1 ) return inf;
            int aa = ind_in_S[a];
            int bb = ind_in_S[b];
            if constexpr (debug){ DEBUG(aa); DEBUG(bb); }

            if constexpr (use_multiple_tms) {
                int d = 0;
                for( auto mask : taken_masks ) {
                    d = max( d, ( (bb == -1) ? distances[mask][aa].back() : distances[mask][aa][bb] ) );
                }
                return d;
            }else {
                int d1 = ( (bb == -1) ? distances[taken_mask1][aa].back() : distances[taken_mask1][aa][bb] );
                int d2 = ( (bb == -1) ? distances[taken_mask2][aa].back() : distances[taken_mask2][aa][bb] );
                int d12 = max(d1,d2);

                return d12;
            }
        };

        vector<int[2]> dp(pairs.size());
        auto getDst = [&]() -> LL {
            LL l = 0;
            if constexpr (debug) { DEBUG(taken_mask1); DEBUG(taken_mask2); }

            constexpr bool use_dp = true;

            if constexpr(use_dp) {
                for(int i=1; i<pairs.size(); i++) dp[i][0] = dp[i][1] = inf;
                dp[0][0] = inf; dp[0][1] = 0;

                for(int i=1; i<pairs.size(); i++) {
                    auto [a,b] = pairs[i-1];
                    auto [c,d] = pairs[i];

                    dp[i][0] = min( dp[i-1][0] + dst(a,d),  // b -?> a -> d -?> c
                                    dp[i-1][1] + dst(b,d) ); // a -?> b -> d -?> c
                    dp[i][0] = min( dp[i][0],inf );

                    dp[i][1] = min( dp[i-1][0] + dst(a,c),  // b -?> a -> c -?> d
                                    dp[i-1][1] + dst(b,c) ); // a -?> b -> c -?> d
                    dp[i][1] = min( dp[i][1],inf );
                }

                return dp.back()[1];

            }else {
                for(int i=1; i<pairs.size(); i++) {
                    auto [a,b] = pairs[i-1];
                    auto [c,d] = pairs[i];
                    int e = dst(b,c);
                    if( a != -1 && d != -1 ) e = min(e, dst(a,d) );
                    if( a != -1 && c != -1 ) e = min(e, dst(a,c) );
                    if( b != -1 && d != -1 ) e = min(e, dst(b,d) );
                    if( e == inf ) return inf;
                    l += e;
                }

                return l;
            }

            return inf;
        };

        sort(st.taken_masks.rbegin(), st.taken_masks.rend());

        bool can_trim_zero_mask = false;
        int minl_among_tms = inf;
        for(auto [tm,ll] : st.taken_masks) minl_among_tms = min(minl_among_tms, ll);

        for( int i=(int)st.taken_masks.size()-1; i>=0; i-- ) {
            PII & m = st.taken_masks[i];
            tie(taken_mask1, taken_mask2) = FBSS3D::getTakenMasks( m.first, DST_MAX_MASK_SIZE );
            sort( pairs.begin()+1, pairs.begin() + 1 + st.ends.size() );

            if constexpr (use_multiple_tms) {
                taken_masks.push_back(taken_mask1);
                taken_masks.push_back(taken_mask2);
            }

            bool can_trim_m = true;

            do {
                LL l = getDst();
                if( l + m.second <= L ){ can_trim_m = false; break; }
            }while( next_permutation( pairs.begin()+1, pairs.begin() + 1 + st.ends.size() ) );

            if constexpr (use_multiple_tms) {
                taken_masks.pop_back();
                taken_masks.pop_back();
            }

            if( can_trim_m && m.first == 0 && m.second == minl_among_tms ) {
                assert( i == (int)st.taken_masks.size()-1 );
                trimmed_masks += st.taken_masks;
                st.taken_masks.clear();
                return true;
            }

            if( can_trim_m ) {
                trimmed_masks.push_back(st.taken_masks[i]);
                swap( st.taken_masks[i], st.taken_masks.back() );
                st.taken_masks.pop_back();
            }
        }

        return st.taken_masks.empty();
    };

    auto trimSmallEndsNeighborhood = [&]( FBSS3D & st ) {
         if constexpr (!use_small_neighborhood_trimming) return true;

        if( st.ends.empty()  ) return true;// this cannot happen, as it was checked before adding newly created state

        _was_ = was;
        for( int i=0; i<S_nodes.size(); i++ ) if( S_nodes[i] != -1 ) _was_[ S_nodes[i] ] = true;

        int neigh_size = 0;
        auto checkNode = [&](int d) { if( !_was_[d] ){ neigh_size++; _was_[d] = true; } };

        constexpr bool exclude_uniquely_assignable_nodes = true;
        fill(ALL(_was2_),false);
        if constexpr (exclude_uniquely_assignable_nodes){
            bool changes_done = true;
            while(changes_done) {
                changes_done = false;

                for( auto [a,b] : st.ends ) {
                    for( int d : underV[a] ) neigh_in_S[d] = 0;
                    for( int d : underV[b] ) neigh_in_S[d] = 0;
                }
                for( auto [a,b] : st.ends ) {
                    if( a != vt && !_was2_[a] ) for( int d : underV[a] ) neigh_in_S[d]++;
                    if( b != vt && !_was2_[b] ) for( int d : underV[b] ) neigh_in_S[d]++;
                }
                for( auto [a,b] : st.ends ) {
                    if(a != vt) for( int d : underV[a] ) {
                        if( !_was_[d] && !_was2_[a] && neigh_in_S[d] == 1 ) {
                            _was2_[a] = changes_done = true;
                            neigh_size++;
                        }
                    }
                    if(b != vt) for( int d : underV[b] ) {
                        if( !_was_[d] && !_was2_[b] && neigh_in_S[d] == 1 ) {
                            _was2_[b] = changes_done = true;
                            neigh_size++;
                        }
                    }
                }
            }
        }

        int valid_ends = st.ends.size();
        for( auto [a,b] : st.ends ) {
            if( a != vt && !_was2_[a] ) for( auto d : underV[a] ) checkNode(d);
            if( b != vt && !_was2_[b] ) for( auto d : underV[b] ) checkNode(d);
        }

        bool res = true;
        if( st.vs_node != vt && st.containsNode(vt) ) {
            if constexpr (debug) clog << "CASE1 " << endl;
            res = (neigh_size >= valid_ends);
        }
        else res = (neigh_size >= valid_ends+1);

        if constexpr (debug) if(!res) {
            clog << "State " << st << " has too small ends-neighborhood" << endl;
            DEBUG(neigh_size);
            DEBUG(valid_ends);
        }
        return res;
    };


    vector<vector<char>> cap(N+2, vector<char>(N+2,0));
    VVI flowV(N+2);

    auto trimSmallEndsNeighborhoodFlow = [&]( FBSS3D & st ) {

        if( st.ends.empty()  ) return true;// this cannot happen, as it was checked before adding newly created state
        if(st.vs_node == vt) return true;

        if constexpr (debug) clog << "Checking using flow method" << endl;

        int src = N;
        int trg = N+1;

        VI nodes(1,st.vs_node);
        for( auto [a,b] : st.ends ) {
            if( a != vt ) nodes.push_back(a);
            if( b != vt ) nodes.push_back(b);
        }

        VB vis(N,false);
        for( int a : nodes) {
            if( a == vt ) continue;

            cap[src][a] = 1;
            flowV[src].push_back(a);

            for( int d : underV[a] ) {
                if( !was[d] ) {
                    cap[a][d] = 1;
                    cap[d][trg] = 2;

                    flowV[a].push_back(d);
                    flowV[d].push_back(a);
                    if( !vis[d] ) flowV[d].push_back(trg);
                    vis[d] = true;
                }
            }
        }


        VB helper(N+2);
        VI pth;
        function<bool(int)> dfs = [&](int num) {
            if( num == trg ) {
                if constexpr (debug) {
                    pth.push_back(num);
                    clog << "Found augmenting path: " << pth << endl;
                    pth.pop_back();
                }
                return true;
            }
            helper[num] = true;
            if constexpr (debug) pth.push_back(num);

            for( int d : flowV[num] ) {
                if( !helper[d] && cap[num][d] > 0 ) {
                    if( dfs(d) ) {
                        cap[num][d]--;
                        cap[d][num]++;
                        if constexpr (debug) pth.pop_back();
                        return true;
                    }
                }
            }

            if constexpr (debug) pth.pop_back();
            return false;
        };

        int flow = 0;
        while( dfs(src) ) {
            flow++;
            helper = VB(N+2, false);
        }

        if constexpr (debug) DEBUG(S_nodes);

        int thr = nodes.size();
        if(st.containsNode(vt)) thr--;

        if constexpr (debug)
        if(flow < thr) {
            clog << "Trimming state " << st << " by flow rule!" << endl;
            DEBUG(vt);
            DEBUG(flow);
            DEBUG(nodes);
            for( auto d : nodes ) clog << "flowV[" << d << "]: " << flowV[d] << endl;
            // exit(1);
        }

        // now clearing
        for( int a : nodes) {
            cap[src][a] = cap[a][src] = 0;
            flowV[a].clear();
            for( int d : underV[a] ) if( !was[d] ) {
                flowV[d].clear();
                cap[a][d] = cap[d][a] = cap[d][trg] = cap[trg][d] = 0;
            }
        }
        flowV[src].clear();

        return flow >= thr;
    };


    auto trimFBSS3Dtates = [&](vector<FBSS3D> & sts) {
        for( int i=(int)sts.size()-1; i>=0; i-- ) {
            assert( !sts[i].taken_masks.empty() );

            bool can_trim = (sts[i].minl > L);
            if constexpr (debug) clog << "Trimming masks from state " << sts[i] << endl;

            if( !can_trim ) {
                if( sts[i].vs_node == vt && !sts[i].ends.empty() ) can_trim = true;
                // if(!can_trim) for( auto [a,b] : sts[i].ends ) if( a == vt ) can_trim = true;
            }

            if( !can_trim ) {
                auto check = [&](int a) {
                    for( int d : underV[a] ) if( ind_in_S[d] == -1 && !was[d] ) return false;
                    return true;
                };
                if( sts[i].vs_node != vt && check( sts[i].vs_node ) ) can_trim = true;
                for( auto [a,b] : sts[i].ends ) {
                    if( a != vt && check(a) ) can_trim = true;
                    if( b != vt && check(b) ) can_trim = true;
                }
            }

            VPII trimmed_masks;
            constexpr int PERMUTATION_THR = 6;
            if(!can_trim && sts[i].ends.size() <= PERMUTATION_THR ) can_trim = canTrimPermutation(sts[i], trimmed_masks);
            // if(!can_trim && sts[i].ends.size() > PERMUTATION_THR) can_trim = canTrimMST(sts[i]); // is there a point to use it?

            if( !can_trim ) can_trim = !trimSmallEndsNeighborhood(sts[i]);

            if( !can_trim ) can_trim = !trimSmallEndsNeighborhoodFlow(sts[i]);

            if constexpr (debug) clog << "\ttrimmed_masks: " << trimmed_masks << endl;

            if( can_trim ) {
                swap( sts[i], sts.back() );
                sts.pop_back();
            }
        }
    };

    auto addNodeForStates = [&](int v, vector<FBSS3D> & sts) {
        markNeighbors(v, true); // mark in- and out-neighbors of node v within boundary set S

        for( auto & st : sts ) {
            if constexpr (debug) clog << "Adding for state st: " << st << endl;

            VI loose = getLooseNodes(st);
            if constexpr (debug) clog << "loose nodes: " << loose << endl;

            addAsIgnoredNode(v,st, loose);
            addAsVsNode(v,st, loose);
            addAsStartNode(v,st, loose);
            addAsMiddleNode(v,st, loose);
            // addAsEndNode(v,st, loose); // this is done in AddAsStartNode
        }

        markNeighbors(v,false); // unmark marked nodes...
    };

    auto removeFBSS3Dtates = [&](int v, vector<FBSS3D> & sts, auto & states, int cur_st_id) {
        for( int i=(int)sts.size()-1; i>=0; i-- ) {

            if constexpr (debug) clog << "In removeFBSS3Dtates, st: " << sts[i] << endl;

            if( sts[i].containsNode(v) ) {
                if constexpr (debug) clog << "\t-x-x-x-x-x removing state " << sts[i] << endl;
                swap( sts[i], sts.back() );
                sts.pop_back();
                continue;
            }

            for( PII &m : sts[i].taken_masks ) {
                auto h1 = sts[i].hash ^ getHashForTakenMask(m.first);
                m.first &= ~( 1 << ind_in_S[v] );
                auto h2 = sts[i].hash ^ getHashForTakenMask(m.first);
                if( h1 != h2 ) addEdge(h1,h2,-1,-1);
            }

            if constexpr (debug) clog << "In removeFBSS3Dtates, after creating transitions, st: " << sts[i] << endl;
            FBSS3D::makeUnique(sts[i].taken_masks);
            if constexpr (debug) clog << "In removeFBSS3Dtates, after uniqu-ing tm's, st: " << sts[i] << endl;

            if( (sts[i].hash.hash & BH) != cur_st_id ) {
                if constexpr (debug) clog << "\tRemoving state " << sts[i] << " from bucket " << cur_st_id
                                          << " and putting into " << (sts[i].hash.hash & BH) << endl;
                states[ sts[i].hash.hash & BH ].push_back( sts[i] );
                swap( sts[i], sts.back() );
                sts.pop_back();
            }
        }
    };

    auto createSameNeighborhoodSameHashesAndTMShifter = [&]() {
        // clog << "Creating tm_shifter" << endl;
        // DEBUG(S_nodes);

        fill(ALL(S_hashes),0);
        iota(ALL(tm_shifter),0);
        map<HashType,VI> mapper;

        for( int i=0; i<32; i++ ) if( S_nodes[i] != -1 ) {
            int v = S_nodes[i];
            if( v == vt ) continue; // we do not shift vt
            for( int d : underV[v] ) if( !was[d] ) {
                S_hashes[v] ^= node_hashes2[d] + all_arcs_hashes[ getArcId(v,d) ];
            }
            if( S_hashes[v] == 0 ) tm_shifter[v] = -1;
            else mapper[ S_hashes[v] ].push_back(v);
        }

        for( auto & [h,v] : mapper ) {
            sort(ALL(v), [&]( int a, int b ){ return ind_in_S[a] < ind_in_S[b]; } );
            for( int i=1; i<v.size(); i++) tm_shifter[v[i]] = v[i-1];
            // DEBUG(v);
        }

        // clog << "tm_shifter:" << endl;
        // for(int i=0; i<N; i++) if(ind_in_S[i] != -1) {
        //     clog << "tm_shifter[" << i << "]: " << tm_shifter[i] << endl;
        // }
    };

    auto addNode = [&](int v) {
        for( int i=0; i<BUCKETS; i++ ) new_states[i].clear();

        createSameNeighborhoodSameHashesAndTMShifter();

        clog << "Old states: " << countAllStates(states) << endl;
        sw.start("addNodeForStates");
        for( int i=0; i<BUCKETS; i++ ) addNodeForStates( v, states[i] );
        sw.stop("addNodeForStates");

        // DEBUG(countAllStates(states));
        // DEBUG(countAllStates(new_states));
        clog << "New states before merging: " << countAllStates(new_states) << endl;

        sw.start("mergeFBSS3DStates");
        for( int i=0; i<BUCKETS; i++ ) mergeFBSS3DStates( new_states[i] );
        sw.stop("mergeFBSS3DStates");
        clog << "New states after merging: " << countAllStates(new_states) << endl;

        sw.start("trimFBSS3DStates");
        for( int i=0; i<BUCKETS; i++ ) trimFBSS3Dtates( new_states[i] );
        sw.stop("trimFBSS3DStates");
        clog << "New states after trimming: " << countAllStates(new_states) << endl;

        mergeStatesSameNeighborhood(new_states);
        clog << "New states after ends-merging: " << countAllStates(new_states) << endl;

        // DEBUG(countAllStates(new_states));
        swap(states, new_states);
    };



    auto removeNode = [&](int v) {
        for( int i=0; i<BUCKETS; i++ ) new_states[i].clear();

        sw.start("removeFBSS3Dtates");
        for( int i=0; i<BUCKETS; i++ ) removeFBSS3Dtates( v, states[i], new_states, i );
        for( int i=0; i<BUCKETS; i++ ) {
            states[i] += new_states[i];
            new_states[i].clear();
        }
        sw.stop("removeFBSS3Dtates");

        sw.start("mergeFBSS3DStates");
        for( int i=0; i<BUCKETS; i++ ) mergeFBSS3DStates( states[i] );
        sw.stop("mergeFBSS3DStates");

    };

    was[vs] = true;
    S.insert(vs);
    S_nodes[0] = vs;
    ind_in_S[vs] = 0;
    order.erase(order.begin());
    cur_vis_hash_immutable = node_hashes2[vs];
    cur_vis_hash = cur_vis_hash_immutable;

    FBSS3D initial_state({}, vs, 0 );
    initial_state.taken_masks.push_back({0,0});
    initial_state.recalculateHash();
    if constexpr ( include_iter_cnt_in_hash ) updateHashWithIterationCnt(initial_state.hash);
    states[ initial_state.hash.hash & BH ].push_back( initial_state );

    int iter_cnt = 0;
    for( int v : order ) {
        if( v == vt && was[v] ) continue;
        if(timer.tle("fbs")) return {};

        if( !was[v] ) {
            clog << "Adding node v: " << v << endl;
            S.insert(v);
        }else {
            clog << "Removing node v: " << v << endl;
            S.erase(v);
        }
        clog << "Current boundary nodes: " << S.size() << " -> " << S << endl;
        if( !was[v] ) S.erase(v); else S.insert(v);
        DEBUG(countAllStates(states));
        DEBUG(trg_edges_h.size());
        clog << "Seconds since start of FBS3D: " << timer.getUnstoppedTimeSinceLastStart("fbs") / 1000 << endl;


        if constexpr (debug) {
            ENDL(5); ENDLS(50, "*"); ENDL(5);
            clog << "Current states: " << endl;
            for(int i=0; i<BUCKETS; i++) {
                if( !states[i].empty() ) clog << "states[" << i << "]: ";
                for( auto & st : states[i] ) clog << st << endl;
            }
            ENDL(2);
        }

        if constexpr (debug) { DEBUG(S_nodes); DEBUG(ind_in_S); }

        for( auto & bucket : states ) for( auto & st : bucket ) st.recalculateMinl();

        if( was[v] ) {
            cur_vis_hash_immutable ^= node_hashes2[v];
            cur_vis_hash = cur_vis_hash_immutable;

            removeNode(v);

            S.erase(v);
            int ind = ind_in_S[v];
            S_nodes[ind] = -1;
            ind_in_S[v] = -1;
        }
        else{
            expansion_iteration_cnt++;

            cur_vis_hash_immutable ^= node_hashes2[v];
            cur_vis_hash = cur_vis_hash_immutable;

            was[v] = true;
            int ind = getFirstFreeSIndex();
            S_nodes[ind] = v;
            ind_in_S[v] = ind;

            sw.start("create distances");
            createDistances(v);
            sw.stop("create distances");

            if constexpr (debug) { DEBUG(S_nodes); DEBUG(ind_in_S); }

            addNode(v);
            S.insert(v);
        }


        { // some statistics...
            ENDL(1);
            VI ends_cnt(S.size(),0);
            for( int i=0; i<BUCKETS; i++ ) {
                for( auto & st : states[i] ) ends_cnt[st.ends.size()]++;
            }

            clog << "Number of states with given number of ends.size(): " << endl;
            for(int i=0; i < min(ends_cnt.size(), 2 + ends_cnt.size() / 2); i++) {
                clog << i << ": " << ends_cnt[i] << endl;
            }

            // int VAL = 6;
            // clog << "States with ends.size() == " << VAL << endl;
            // for( int i=0; i<BUCKETS; i++ ) {
            //     for( auto & st : states[i] ) if( st.ends.size() == VAL ) clog << st << endl;
            // }

            ENDL(1);
        }
    }


    if constexpr (debug) {
        vector<FBSS3D> all_states; // only for debugging...
        for(auto & vec : states) all_states += vec;
        clog << "All states: " << endl;
        for(auto & st : all_states) clog << st << endl;
        clog << "trg_edges: " << endl;
        for( auto e : trg_edges ) clog << "(" << e.a << "," << e.b << ")" << endl;
    }

    ULL all_states_cnt = 0;
    FBSS3D terminal_state;
    for(auto & vec : states) {
        all_states_cnt += vec.size();
        if(!vec.empty()) terminal_state = vec[0];
        vector<FBSS3D>().swap(vec); // releasing space
    }
    for(auto & vec : new_states) vector<FBSS3D>().swap(vec); // releasing space


    assert( all_states_cnt <= 1 ); // there should be only one state - the one with node vt
    if(all_states_cnt == 1) {
        assert( terminal_state.vs_node == vt );
        assert( terminal_state.taken_masks.size() == 1 );
    }

    if( all_states_cnt == 0 ) return VBIG(L+1,0);

    if constexpr (debug) DEBUG(initial_state.hash.hash);

    int init_state_id = -1;
    int last_state_id = -1;

    int trgN;

    // using countsort is considerably faster, but requires copy of arcs, so roughly doubles memory peak...
    bool use_countsort = ( trg_edges_h.size() < 3e8 );

    {
        clog << "Remapping trg_edges_h" << endl;
        sw.start("remapping trg_edges_h");
        vector<HashType> all_hashes; all_hashes.reserve( 2 * trg_edges_h.size() );

        int BUCKETS_BITS = max(1, min(18, (int)ceil( log2(trg_edges_h.size()) ) ) );

        for(auto &e : trg_edges_h) {
            all_hashes.push_back(e.a);
            if( e.b == terminal_state.hash ) all_hashes.push_back(e.b);
        }
        sw.start("remapping sort - all hashes");
        if (use_countsort) Solver2FBS::countSortTrgEdgesHash<HashType, 2>(BUCKETS_BITS, all_hashes);
        else sort( ALL(all_hashes) );
        // unsignedIntegerRadixSort(all_hashes);
        sw.stop("remapping sort - all hashes");
        all_hashes.resize( unique(ALL(all_hashes)) - all_hashes.begin() );

        for( int i=0; i<all_hashes.size(); i++ ) {
            if( all_hashes[i] == initial_state.hash ) init_state_id = i;
            if( all_hashes[i] == terminal_state.hash ) last_state_id = i;
        }

        sw.start("remapping sort");
        if (use_countsort) Solver2FBS::countSortTrgEdgesHash<EdgeH, 0>(BUCKETS_BITS, trg_edges_h); // sort by a
        else sort( ALL(trg_edges_h), [](auto &a, auto &b){ return a.a < b.a; } );
        sw.stop("remapping sort");

        int p = 0, q = 0;
        while( p < all_hashes.size() && q < trg_edges_h.size() ) {
            if( all_hashes[p] == trg_edges_h[q].a ) trg_edges_h[q++].a.hash = p;
            else if( all_hashes[p] < trg_edges_h[q].a ) p++;
            else q++;
        }

        sw.start("remapping sort");
        if (use_countsort) Solver2FBS::countSortTrgEdgesHash<EdgeH, 1>(BUCKETS_BITS, trg_edges_h); // sort by b
        else sort( ALL(trg_edges_h), [](auto &a, auto &b){ return a.b < b.b; } );
        sw.stop("remapping sort");

        p = 0, q = 0;
        while( p < all_hashes.size() && q < trg_edges_h.size() ) {
            if( all_hashes[p] == trg_edges_h[q].b ) trg_edges_h[q++].b.hash = p;
            else if( all_hashes[p] < trg_edges_h[q].b ) p++;
            else q++;
        }

        trg_edges.clear();
        trgN = 0;
        for( auto &e : trg_edges_h ) {
            if(e.a.hash > all_hashes.size() || e.b.hash > all_hashes.size()) continue;
            trgN = max( trgN, (int)max( e.a.hash, e.b.hash ) );
            trg_edges.emplace_back(e.a.hash, e.b.hash, e.e1, e.e2);
        }
        trgN++;
        vector<EdgeH>().swap(trg_edges_h);

        sw.stop("remapping trg_edges_h");
    }

    if constexpr (debug) {
        assert( trg_edges_h.size() == set(ALL(trg_edges_h)).size() );
    }


    if constexpr (debug) { DEBUG(trgN); DEBUG(init_state_id); DEBUG(last_state_id); }


    clog << "TRANSITION GRAPH ARCS CREATED" << endl;
    clog << "Before transition graph creation, it has " << trgN << " nodes and "
         << trg_edges.size() << " arcs" << endl;


    auto constructTransitionGraph = [&](int trgN, auto & trg_edges, bool trg_edges_sorted_by_a = false) {
        sw.start("TRG creation");
        if(!trg_edges_sorted_by_a) {
            if (use_countsort) Solver2FBS::countSortTrgEdges(trgN, trg_edges);
            else sort( ALL(trg_edges), [](auto &a, auto &b){ return a.a < b.a; } );
        }
        VI first_of(trgN+1, trg_edges.size());
        for( int i=0; i<trg_edges.size(); i++ ) {
            auto& e = trg_edges[i];
            first_of[e.a] = ( i <= first_of[e.a] ? i : first_of[e.a] );
        }
        for( int i=trgN-1; i>=0; i-- ) first_of[i] = (first_of[i] == trg_edges.size() ? first_of[i+1] : first_of[i]);
        sw.stop("TRG creation");

        return first_of;
    };


    clog << "Before trimming and remapping graph it has " << trgN << " nodes and "
         << trg_edges.size() << " arcs " << endl;

    auto trimTRG = [&constructTransitionGraph, &init_state_id, &last_state_id, &sw, &all_arcs_lengths](int trgN, vector<Edge> & trg_edges,
         bool trg_edges_sorted_by_b = false) {
        sw.start("trimTRG");
        vector<unsigned short> outdeg(trgN,0);
        for( auto &e : trg_edges ){ swap( e.a, e.b ); ++outdeg[e.b]; }
        auto first_of_r = constructTransitionGraph( trgN, trg_edges, trg_edges_sorted_by_b );

        VB valid(trgN,true);

        VI to_trim; to_trim.reserve(trgN/2);
        for( int i=0; i<trgN; i++ ) if( outdeg[i] == 0 && i != last_state_id ) to_trim.push_back(i);
        for( int j=0; j<to_trim.size(); j++ ) {
            int num = to_trim[j];
            valid[num] = false;
            for( int i=first_of_r[num]; i < first_of_r[num+1]; i++ ) {
                auto &e = trg_edges[i];
                if( --outdeg[e.b] == 0 ) to_trim.push_back(e.b);
            }
        }
        VI().swap(to_trim);
        vector<unsigned short>().swap(outdeg);
        for( auto &e : trg_edges ){ swap( e.a, e.b ); }

        assert( valid[init_state_id] );
        assert( valid[last_state_id] );

        VI mapper(trgN,-1);
        int ind = 0;
        for( int i=0; i<trgN; i++ ) if( valid[i] ) mapper[i] = ind++;
        int r = 0;
        for(int i=0; i<trg_edges.size(); i++) {
            auto &e = trg_edges[i];
            if( mapper[e.a] != -1 && mapper[e.b] != -1 ) {
                e.a = mapper[e.a];
                e.b = mapper[e.b];
                trg_edges[r++] = e;
            }
        }
        last_state_id = mapper[last_state_id];
        init_state_id = mapper[init_state_id];
        while(trg_edges.size() > r) trg_edges.pop_back();

        VI first_of = constructTransitionGraph( ind, trg_edges);



        sw.stop("trimTRG");
        return first_of;
    };

    VI first_of = trimTRG(trgN, trg_edges, true);
    trgN = first_of.size()-1;

    clog << "After trimming (without remapping) transition graph has " << trgN << " nodes and " << trg_edges.size()
         << " arcs " << endl;
    int path_compressable_nodes = 0;
    int fully_compressable_nodes1 = 0;
    int fully_compressable_nodes2 = 0;

    auto calculateCompressableNodes = [&path_compressable_nodes, &fully_compressable_nodes1,
        &fully_compressable_nodes2, &sw] (int trgN, auto & trg_edges) {
        return; // #TEST - this seems to be relatively useless - perhaps saves some space, but adds some time overhead...
        sw.start("calculateCompressableNodes");
        path_compressable_nodes = 0;
        fully_compressable_nodes1 = 0;
        fully_compressable_nodes2 = 0;
        VI indeg(trgN,0), outdeg(trgN,0);
        for(auto e : trg_edges){ outdeg[e.a]++; indeg[e.b]++; }
        for( int i=0; i<trgN; i++ ) {
            if( indeg[i] == 1 && outdeg[i] == 1 ) path_compressable_nodes++;
            if( indeg[i] >= 1 && outdeg[i] == 1 ) fully_compressable_nodes1++;
            if( indeg[i] == 1 && outdeg[i] >= 1 ) fully_compressable_nodes2++;
        }
        sw.stop("calculateCompressableNodes");
    };
    calculateCompressableNodes(trgN, trg_edges);



    auto compressTRG = [&](int trgN, VI & first_of, vector<Edge> & trg_edges) {
        sw.start("compressTRG");

        sw.start("forward_compression");
        vector<CompressedEdge> cmp_trg_edges;
        cmp_trg_edges.reserve( trgN - fully_compressable_nodes1 + 10 );

        VI cmp_trg_first_of;

        constexpr bool apply_backward_compression = false;

        {
            VB valid(trgN,true);

            vector<unsigned short> outdeg(trgN,0);
            for(auto e : trg_edges) ++outdeg[e.a];

            for( int i=0; i<trgN; i++ ) {
                for( int j=first_of[i]; j < first_of[i+1]; j++ ) {
                    Edge e = trg_edges[j];
                    if( e.a != init_state_id && outdeg[e.a] == 1  ) continue;

                    VI arcs;
                    if( e.e1 != -1 ) arcs.push_back(e.e1);
                    if( e.e2 != -1 ) arcs.push_back(e.e2);

                    while( e.b != last_state_id && outdeg[e.b] == 1 ) {
                        valid[e.b] = false;
                        auto ee = trg_edges[ first_of[e.b] ];
                        if( ee.e1 != -1 ) arcs.push_back(ee.e1);
                        if( ee.e2 != -1 ) arcs.push_back(ee.e2);
                        e.b = ee.b;
                    }

                    cmp_trg_edges.emplace_back( e.a, e.b );
                    swap(arcs, cmp_trg_edges.back().arcs);
                }
            }
            assert(valid[init_state_id]);
            assert(valid[last_state_id]);

            VI mapper(trgN,-1);
            int ind = 0;
            for( int i=0; i<trgN; i++ ) if( valid[i] ) mapper[i] = ind++;
            last_state_id = mapper[last_state_id];
            init_state_id = mapper[init_state_id];
            for( auto & e : cmp_trg_edges ) { e.a = mapper[e.a]; e.b = mapper[e.b]; }

            trgN = ind;

            if constexpr (!apply_backward_compression) {
                cmp_trg_first_of = VI(trgN+1, cmp_trg_edges.size());
                for( int i=0; i<cmp_trg_edges.size(); i++ ) {
                    auto& e = cmp_trg_edges[i];
                    cmp_trg_first_of[e.a] = ( i <= cmp_trg_first_of[e.a] ? i : cmp_trg_first_of[e.a] );
                }
                for( int i=trgN-1; i>=0; i-- ) {
                    cmp_trg_first_of[i] = (cmp_trg_first_of[i] == cmp_trg_edges.size() ? cmp_trg_first_of[i+1] : cmp_trg_first_of[i]);
                }
            }
        }
        sw.stop("forward_compression");

        clog << "After forward-compression graph has " << trgN << " nodes and " << cmp_trg_edges.size()
             << " arcs" << endl;
        calculateCompressableNodes(trgN, cmp_trg_edges);



        if constexpr (apply_backward_compression){
            sw.start("backward_compression");
            for(auto & e : cmp_trg_edges) swap(e.a, e.b);
            auto cmp_trg_first_of_r = constructTransitionGraph( trgN, cmp_trg_edges );

            { // doing the same as before, but for reversed graph
                VB valid(trgN,true);

                vector<CompressedEdge> new_cmp_trg_edges;
                new_cmp_trg_edges.reserve( trgN - fully_compressable_nodes1 + 10 );

                VI outdeg(trgN,0);
                for(auto e : cmp_trg_edges) ++outdeg[e.a];

                for( int i=0; i<trgN; i++ ) {
                    for( int j=cmp_trg_first_of_r[i]; j < cmp_trg_first_of_r[i+1]; j++ ) {
                        CompressedEdge e = cmp_trg_edges[j];
                        if( e.a != last_state_id && outdeg[e.a] == 1  ) continue;

                        VI arcs = e.arcs;

                        while( e.b != last_state_id && outdeg[e.b] == 1 ) {
                            valid[e.b] = false;
                            auto & ee = cmp_trg_edges[ cmp_trg_first_of_r[e.b] ];
                            arcs += ee.arcs;
                            e.b = ee.b;
                        }

                        new_cmp_trg_edges.emplace_back( e.a, e.b );
                        swap(arcs, new_cmp_trg_edges.back().arcs);
                    }
                }
                assert(valid[init_state_id]);
                assert(valid[last_state_id]);

                VI mapper(trgN,-1);
                int ind = 0;
                for( int i=0; i<trgN; i++ ) if( valid[i] ) mapper[i] = ind++;
                last_state_id = mapper[last_state_id];
                init_state_id = mapper[init_state_id];
                for( auto & e : new_cmp_trg_edges ) { e.a = mapper[e.a]; e.b = mapper[e.b]; }
                trgN = ind;
                swap(new_cmp_trg_edges, cmp_trg_edges);
            }

            for(auto & e : cmp_trg_edges) swap(e.a, e.b);
            cmp_trg_first_of = constructTransitionGraph( trgN, cmp_trg_edges );
            sw.stop("backward_compression");
        }


        sw.stop("compressTRG");

        return make_tuple(cmp_trg_first_of, cmp_trg_edges);
    };


    vector<CompressedEdge> cmp_trg_arcs;
    tie(first_of, cmp_trg_arcs) = compressTRG(trgN, first_of, trg_edges);
    trgN = first_of.size()-1;
    vector<Edge>().swap(trg_edges); // releasing space
    clog << "After compression transition graph has " << trgN << " nodes and " << cmp_trg_arcs.size() << " arcs " << endl;

    calculateCompressableNodes(trgN, cmp_trg_arcs);
    // DEBUG(path_compressable_nodes);
    // DEBUG(fully_compressable_nodes1);
    // DEBUG(fully_compressable_nodes2);



    bool use_large_bigint = false;




    VBIG res(L+1,0);
    // auto calculateResult = [&]<typename BI, typename LNG>( vector<LNG> & lengths, vector<LNG> & all_arcs_lengths,
    auto calculateResult = [&]<typename BI, typename LNG>( vector<LNG*> & lengths, vector<LNG> & all_arcs_lengths,
        BI _bigint_, BI thr_value ) {

        was = VB(trgN,false);
        sw.start("Toposort result calculation");
        clog << "Solving iteratively by toposort" << endl;

        sw.start("creating topological order");
        VI topo_order;
        vector<int> indeg(trgN);
        for( auto & e : cmp_trg_arcs ) ++indeg[e.b];

        function<void(int)> toposort = [&](int num) {
            topo_order.push_back(num);
            was[num] = true;
            for( int i=first_of[num]; i < first_of[num+1]; i++ ) {
                auto & e = cmp_trg_arcs[i];
                if( --indeg[e.b] == 0 ) toposort(e.b);
            }
        };

        for( int i=0; i<trgN; i++ ) if( !was[i] && indeg[i] == 0 ) toposort(i);
        sw.stop("creating topological order");
        // for( int i=0; i<trgN; i++ ) if( indeg[i] == 0 ) toposort(i);

        DEBUG(trgN);
        DEBUG(topo_order.size());
        assert(topo_order.size() == trgN);

        // lengths[last_state_id] = LNG( vector<BI>(1,MULTIPLIER),0 );
        lengths[last_state_id] = new LNG( vector<BI>(1,1),0 );

        VI first_pred(trgN,-1);
        for( int v : topo_order ) {
            for( int i=first_of[v]; i < first_of[v+1]; i++ ) {
                int b = cmp_trg_arcs[i].b;
                if( first_pred[b] == -1 ) first_pred[b] = v;
            }
        }

        VI src_dst(trgN,inf);
        src_dst[init_state_id] = 0;
        for( int v : topo_order ) {
            for( int i=first_of[v]; i < first_of[v+1]; i++ ) {
                auto & e = cmp_trg_arcs[i];
                int suml = 0;
                for( int eid : e.arcs ) suml += all_arcs_lengths[eid].minl();
                src_dst[e.b] = min( src_dst[e.b], src_dst[v] + suml );
            }
        }

        sw.start("calculating for topological order");
        reverse(ALL(topo_order));
        VI to_clear;
        int dispensible_arcs = 0;
        int valid_arcs = 0;

        // fill(ALL(indeg),0);
        // for( auto & e : cmp_trg_arcs ) ++indeg[e.b];
        VI().swap(indeg);


        for( int v : topo_order ) {
            to_clear.clear();

            if( v != last_state_id ) lengths[v] = new LNG(); // zero

            // if( indeg[v] == 1 ) continue;
            // for( int i=first_of[v]; i < first_of[v+1]; i++ ) {
            //     auto & e = cmp_trg_arcs[i];
            //     if( indeg[e.b] == 1 && e.arcs.empty() ) {
            //         swap( lengths[v], lengths[e.b] );
            //         vector<BI>().swap( lengths[e.b]->l ); // this will make these lengths invalid in next loop
            //         break;
            //     }
            // }

            for( int i=first_of[v]; i < first_of[v+1]; i++ ) {
                auto & e = cmp_trg_arcs[i];
                // if( !lengths[e.b].valid() ) continue;
                assert( lengths[e.b] != nullptr );
                if( !lengths[e.b]->valid() ) continue;

                int suml = 0;
                for( int eid : e.arcs ) suml += all_arcs_lengths[eid].minl();

                if( src_dst[e.a] + suml + lengths[e.b]->minl() > L ) {
                    dispensible_arcs++;
                    if( first_pred[e.b] == v ) to_clear.push_back(e.b);
                    continue;
                }

                LNG mul = LNG( vector<BI>(1,1), 0 ); // equivalent to SP2Arc::mul_one
                for( int eid : e.arcs ) {
                    mul.trimToMax(L - all_arcs_lengths[eid].minl());
                    mul *= all_arcs_lengths[eid];
                }

                mul.trimToMax(L - lengths[e.b]->minl());
                mul.minimize();
                LNG temp = *lengths[e.b];
                temp.trimToMax(L - mul.minl());

                if( mul.valid() && temp.valid() ) {
                    valid_arcs++;
                    temp *= mul;

                    *lengths[v] += temp;
                    lengths[v]->trimToMax(L);
                    lengths[v]->minimize();
                }

                if( first_pred[e.b] == v ) to_clear.push_back(e.b);
            }

            for( auto & bi : lengths[v]->l ) if( bi > thr_value ) use_large_bigint = true;

            // for(int b : to_clear) vector<BI>().swap( lengths[b].l );
            for(int b : to_clear) if( lengths[b] != nullptr ){ delete lengths[b]; lengths[b] = nullptr; }
            if( v == init_state_id ) break;

        }

        DEBUG(dispensible_arcs);
        DEBUG(valid_arcs);
    };



    auto toStringSmall = [&](small_bigint b) {
        string s = "";
        while( b > 0 ) {
            small_bigint rem = (b % 10);
            small_bigint temp = rem;
            for( int i=0; i<10; i++ ) {
                temp = i;
                if( temp == rem ) s += char( '0' + i );
            }
            b /= 10;
        }
        reverse(ALL(s));
        return s;
    };

    auto toStringLarge = [&](bigint b) {
        string s = "";
        while( b > 0 ) {
            bigint rem = (b % 10);
            bigint temp = rem;
            for( int i=0; i<10; i++ ) {
                temp = i;
                if( temp == rem ) s += char( '0' + i );
            }
            b /= 10;
        }
        reverse(ALL(s));
        return s;
    };

    auto convertLargeToSmall = [&](bigint large, small_bigint small) {
        auto s = toStringLarge(large);
        small = 0;
        for( auto c : s ) {
            small *= 10;
            small += (int)(c - '0');
        }
        return small;
    };

    auto convertSmallToLarge = [&](small_bigint small, bigint large) {
        auto s = toStringSmall(small);
        large = 0;
        for( auto c : s ) {
            large *= 10;
            large += (int)(c - '0');
        }
        return large;
    };


    bool use_small_bigint = true;

    if( use_small_bigint ) {
        clog << "Using small bigint" << endl;

        using BI = small_bigint;
        using LNG = LengthsT<BI>;
        vector<LNG*> lengths(trgN); // calculate using ULL
        vector<LNG> all_arcs;
        all_arcs.reserve( all_arcs_lengths.size() );
        for( auto & arc : all_arcs_lengths ) {
            vector<BI> t( arc.l.size() );
            for( int i=0; i<t.size(); i++ ) t[i] = convertLargeToSmall(arc[ arc.minl() + i], (BI)0);
            all_arcs.emplace_back(t, arc.minl());
        }

        small_bigint small_thr_value = 1;
        {
            int times;
#ifdef USE_BOOST_SMALL_BIGINT
            times = 60;
#else
            times = 32;
#endif
            for( int i=0; i<times; i++ ) small_thr_value *= 10;
        }
        calculateResult(lengths, all_arcs, (BI)0, small_thr_value);



        auto & lr = *lengths[init_state_id];
        if(lr.valid()) {
             for(int i=lr.minl(); i <= lr.maxl(); i++) res[i] = convertSmallToLarge( lr[i], (bigint)0 );
        }

        bigint thr_value = 1;
        int times;
#ifdef USE_BOOST_SMALL_BIGINT
        times = 60;
#else
        times = 32;
#endif
        for( int i=0; i<times; i++ ) thr_value *= 10;

        auto sum = accumulate( ALL(res), (bigint)0 );
        // auto sum = lr.sum();

        if( sum < 0 || sum > thr_value ) {
            clog << "result " << sum << " is < 0   or   > 10^" << times
                 << ", recalculating using arbitrary precision" << endl;
            res = VBIG(L+1,0);
            use_large_bigint = true;
        }

        for( auto ptr : lengths ) if( ptr != nullptr ){ delete ptr; ptr = nullptr; }
    }


    if(use_large_bigint){

#if defined(USE_BOOST_BIGINT)
        clog << "Using large bigint - BOOST" << endl;
#elif defined(USE_GMP_BIGINT)
        clog << "Using large bigint - GMP" << endl;
#else
        clog << "Using large bigint - InfInt" << endl;
#endif

        // vector<Lengths> lengths(trgN);
        vector<Lengths*> lengths(trgN);
        calculateResult(lengths, all_arcs_lengths, (bigint)0, (bigint)0); // calculate using InfInt or any other multiple-precision


        auto & lr = *lengths[init_state_id];
        if(lr.valid()) for(int i=lr.minl(); i <= lr.maxl(); i++) res[i] = lr[i];

        for( auto ptr : lengths ) if( ptr != nullptr ){ delete ptr; ptr = nullptr; }
    }


    sw.stop("calculating for topological order");



    sw.stop("Toposort result calculation");


    sw.start("solveFBS3DForNodeOrder");
    sw.writeAll();


    clog << "returning result" << endl;


    timer.stop("fbs");
    timer.writeAll();

    return res;
}

