//
// Created by sylwe on 04/09/2024.
//

#include "Solver2MIM.h"

#include <ConnectedComponents.h>
#include <GraphUtils.h>
#include <Stopwatch.h>

#include "BiconnectedFinder.h"
#include "Kernelizer.h"
#include "Solver.h"
#include "Solver2.h"
#include "CollectionOperators.h"
#include "DistanceFinder2.h"


Solver2MIM::Solver2MIM(GraphSP2 _g, int _L, int _vs, int _vt) : g(_g), L(_L), vs(_vs), vt(_vt) {
    N = g.size();
    revg = g.reverseGraph();
}





void Solver2MIM::trimStates(GraphSP2 &g, GraphSP2 &revg, int vs, int vt, vector<SP2> &sp2s) {
    clog << "Trimming states" << endl;
    sort(ALL(sp2s), [&]( auto & sp1, auto & sp2 ) { return sp1.mask < sp2.mask; } );

    VB indices_to_trim( sp2s.size() );
    int p = 0, q = 0;
    while( p < sp2s.size() ) {
        auto mask = sp2s[p].mask;
        q = p+1;
        while( q < sp2s.size() && sp2s[q].mask == mask) q++;
        auto [dsts, pars] = Solver2::findDistances( g, vs, mask );
        auto [dstt, part] = Solver2::findDistances( revg, vt, mask );

        while(p < q) {
            auto & sp = sp2s[p];
            int a = sp.beg;
            int b = sp.end;

            int dstsa = dsts[a];
            for( auto & arc : revg[a] ) dstsa = min( dstsa, 1 + dsts[arc.v] );
            int dsttb = dstt[b];
            for( auto & arc : g[b] ) dsttb = min( dsttb, 1 + dstt[arc.v] );

            if( dstsa + sp.lengths.minl() + dsttb > L )  indices_to_trim[p] = true;
            p++;
        }
    }

    for( int i=0; i<sp2s.size(); i++ ) {
        auto & sp = sp2s[i];
        if( sp.mask[vs] && vs != sp.beg ) indices_to_trim[i] = true;
        if( sp.mask[vt] && vt != sp.end ) indices_to_trim[i] = true;
    }

    for( int i=(int)sp2s.size()-1; i>=0; i-- ) {
        if( indices_to_trim[i] ) {
            swap( sp2s[i], sp2s.back() );
            sp2s.pop_back();
        }
    }
}


VBIG Solver2MIM::solveMIM() {
    constexpr bool debug = false;

    clog << "Solving using solver2.mim" << endl;
    if( N > MAXN ) {
        cerr << "CANNOT PROCESS GRAPHS WITH MORE THAN " << MAXN << " NODES USING MIM..." << endl;
        exit(1);
    }


    vector<bigint> res_for_length(L+1,0);
    bool has_vsvt_arc = false;
    for( auto & arc : g[vs] ) if( arc.v == vt ) has_vsvt_arc = true;
    if( has_vsvt_arc ) {
        for( auto & arc : g[vs] ) {
            if( arc.v == vt ) {
                for( int i=arc.minl(); i <= arc.maxl(); i++ ) res_for_length[i] += arc[i];
                break;
            }
        }
        Kernelizer2::removeArc(g, vs,vt);
        Kernelizer2::removeArc(g, vt,vs);
        Kernelizer2::removeArc(revg, vs,vt);
        Kernelizer2::removeArc(revg, vt,vs);
    }


    vector<vector<SP4>> dpvs1(N), dpvs2(N), dpvt1(N), dpvt2(N);

    dpvs1[vs].emplace_back( Mask(N), node_hashes[vs], Lengths( VBIG(1,1), 0 ) );
    dpvs1[vs].back().mask.set(vs,1);

    dpvt1[vt].emplace_back( Mask(N), node_hashes[vt], Lengths( VBIG(1,1), 0 ) );
    dpvt1[vt].back().mask.set(vt,1);

    auto mergeEntries = [&](auto & new_entries) {
        int p = 0, q = 0, r = 0;
        while( p < new_entries.size() ){
            q = p;
            auto sp1 = new_entries[p], sp2 = new_entries[q];
            while( q < new_entries.size() && (sp1 == sp2) ) if(++q <new_entries.size()) sp2 = new_entries[q];
            while( p+1 < q ) sp1.counts += new_entries[++p].counts;
            new_entries[r++] = sp1;
            p = q;
        }

        // new_entries.resize(r);
        while(new_entries.size() > r) new_entries.pop_back();
    };

    auto trimStates = [L=L]( GraphSP2 & V, int source, int v, auto & sps ) {
        sort(ALL(sps) );

        VB indices_to_trim( sps.size() );
        int p = 0;
        while( p < sps.size() ) {
            if( sps[p].counts.minl() > L || sps[p].counts.empty() ) indices_to_trim[p] = true;

            auto& mask = sps[p].mask;
            assert(mask[v]);
            mask.set(v,0);
            auto [dst, par] = Solver2::findDistances( V, source, mask );
            mask.set(v,1);
            if( mask.popcnt()-1 + dst[v] > L ) indices_to_trim[p] = true;
            p++;
        }

        for( int i=(int)sps.size()-1; i>=0; i-- ) {
            if( indices_to_trim[i] ) {
                if constexpr (debug) clog << "Trimming state: " << sps[i].toString() << endl;
                swap( sps[i], sps.back() );
                sps.pop_back();
            }
        }
    };

    int minl_vs = 0;
    int minl_vt = 0;
    auto updateMinls = [&]() {
        minl_vs = inf;
        minl_vt = inf;
        for(int i=0; i<N; i++) {
            for( auto & sp : dpvs1[i] ) minl_vs = min(minl_vs, sp.counts.minl());

            for( auto & sp : dpvt1[i] ) minl_vt = min(minl_vt, sp.counts.minl());
            for( auto & sp : dpvt2[i] ) minl_vt = min(minl_vt, sp.counts.minl());
        }
    };
    updateMinls();

    int subsets_size = 0;
    while( minl_vs + minl_vt + 2 <= L ) {
        subsets_size++;

        // if constexpr (debug)
            clog << endl << endl << "minl_vs: " << minl_vs << ", minl_vt: " << minl_vt << ", S: " << subsets_size << endl;

        // now increase path length by 1
        swap(dpvs1, dpvs2);
        swap( dpvt1, dpvt2 );

        for( int i=0; i<N; i++ ) {
            dpvs1[i].clear();
            dpvt1[i].clear();
        }

        for( int i=0; i<N; i++ ) { // create in dpvs1 and dpvt1 entries with new lengths
            for( auto & sp3 : dpvs2[i] ) {
                for( auto & arc : g[i] ) {
                    int d = arc.v;
                    if( !sp3.mask[d] && d != vs && d != vt ) {
                        dpvs1[d].emplace_back( sp3.mask, sp3.hash ^ node_hashes[d], sp3.counts * arc.lengths );
                        dpvs1[d].back().mask.set(d,1);
                        dpvs1[d].back().counts.trimToMax(L);
                    }
                }
            }

            for( auto & sp3 : dpvt2[i] ) {
                for(auto & arc : revg[i]) {
                    int d = arc.v;
                    if( !sp3.mask[d] && d != vs && d != vt ) {
                        dpvt1[d].emplace_back( sp3.mask, sp3.hash ^ node_hashes[d], sp3.counts * arc.lengths );
                        dpvt1[d].back().mask.set(d,1);
                        dpvt1[d].back().counts.trimToMax(L);
                    }
                }
            }
        }

        int total_sps = 0;
        for(int i=0; i<N; i++) total_sps += dpvs1[i].size() + dpvt1[i].size() + dpvt2[i].size();
        clog << "\tTotal sps: " << total_sps << endl;

        if constexpr (debug) {
            clog << "Created SP3s:" << endl;

            for(int i=0; i<N; i++) {
                if( dpvs1[i].empty() ) continue;
                clog << "dpvs1[" << i << "]: " << endl;
                for( auto & sp : dpvs1[i] ) clog << "\t" << sp.toString() << endl;
            }

            for(int i=0; i<N; i++) {
                if( dpvt1[i].empty() ) continue;
                clog << "dpvt1[" << i << "]: " << endl;
                for( auto & sp : dpvt1[i] ) clog << "\t" << sp.toString() << endl;
            }

        }

        if constexpr (debug) clog << "\tProceeding to merge entries and trim states" << endl;
        for( int i=0; i<N; i++ ) {
            sort(ALL(dpvs1[i]));
            mergeEntries( dpvs1[i] );
            trimStates( revg, vt, i, dpvs1[i] );

            sort(ALL(dpvt1[i]));
            mergeEntries( dpvt1[i] );
            trimStates( g, vs,i, dpvt1[i] );
        }
        updateMinls();
        if(minl_vs + minl_vt > L) break;

        total_sps = 0;
        for(int i=0; i<N; i++) total_sps += dpvs1[i].size() + dpvt1[i].size() + dpvt2[i].size();
        clog << "\tAfter merging and trimming, total sps: " << total_sps << endl;

        // finally, calculate results...
        if constexpr (debug) {
            clog << "\tProceeding to calculate results for lengths " << 2*subsets_size-1 << " and "
                 << 2*subsets_size << endl;
        }
        ULL total_comparisons = 0;
        for( int i=0; i<N; i++ ) total_comparisons += dpvs1[i].size() * ( dpvt1[i].size() + dpvt2[i].size() );
        clog << "\tThere needs to be done at most " << total_comparisons << " bitset mask comparisons" << endl;

        auto cmp = []( auto & sp1, auto & sp2 ) {
            return sp1.counts.minl() < sp2.counts.minl();
        };

        for( int i=0; i<N; i++ ) {

            sort(ALL(dpvs1[i]), cmp);
            sort(ALL(dpvt1[i]), cmp);
            sort(ALL(dpvt2[i]), cmp);

            for( auto & sp1 : dpvs1[i] ) {
                Lengths res( VBIG(1,0), 0 );

                if( 2*subsets_size <= L ) {
                    for( auto & sp2 : dpvt1[i] ) {
                        if( sp1.counts.minl() + sp2.counts.minl() > L ) break;
                        // if((sp1.mask & sp2.mask).popcnt() == 1) res += sp2.counts;
                        if(sp1.mask.singleNodeIntersection(sp2.mask)) res += sp2.counts;
                    }
                    res *= sp1.counts;
                    res.trimToMax(L);
                    for( int j=res.minl(); j <= res.maxl(); j++ ) res_for_length[j] += res[j];
                }

                if( 2*subsets_size-1 <= L ) {
                    res = Lengths(VBIG(1,0), 0);
                    for( auto & sp2 : dpvt2[i] ) {
                        if( sp1.counts.minl() + sp2.counts.minl() > L ) break;
                        if(sp1.mask.singleNodeIntersection(sp2.mask)) res += sp2.counts;
                    }
                    res *= sp1.counts;
                    res.trimToMax(L);
                    for( int j=res.minl(); j <= res.maxl(); j++ ) res_for_length[j] += res[j];
                }
            }
        }
    }

    clog << "solver2.mim finished!" << endl;

    while( res_for_length.size() > L+1 ) res_for_length.pop_back();
    return res_for_length;
}

Mask Solver2MIM::getNodesAfterBiconnectedReduction(VVI & underV, Mask & mask, int beg, int vt, VI pth) {
    return getNodesAfterBiconnectedReduction2(underV, mask, beg, vt, pth); // #TEST

    constexpr bool debug = false;

    int N = underV.size();
    VB was(N);

    if constexpr (debug) {
        clog << endl << endl << "updateSP4HashBiconnected, sp: " << mask.toString() << endl << "beg: " << beg << ", vt: " << vt << endl;
    }

    for( int j=0; j<N; j++ ) was[j] = !mask[j];
    was[beg] = true; // remove node beg from mask before finding bicomponents
     auto bicomps = ConnectedComponents::getBiconnectedComponents(underV, was);
    fill(ALL(was),false);

    auto path = Solver::findShortestPath(underV, beg, vt, mask);
    if constexpr (debug) DEBUG(bicomps);
    if constexpr (debug) DEBUG(path);

    if( path.empty() ) return Mask(N); // no nodes after reduction, so no path...
    for(int d : path) was[d] = true;

    VI steps_in_bicomp(bicomps.size(),0);

    VB to_remove(bicomps.size(),false);
    for( int i=0; i<bicomps.size(); i++ ) {
        for(int d : bicomps[i]) {
            if( was[d] ) steps_in_bicomp[i]++;
        }
    }

    for( int i=0; i<bicomps.size(); i++ ) {
        if constexpr (debug) clog << "Steps in bicomp: " << bicomps[i] << ": " << steps_in_bicomp[i] << endl;
        if( bicomps[i].size() > 1 && steps_in_bicomp[i] <= 1 ) to_remove[i] = true;
        else if( bicomps[i].size() == 1 && steps_in_bicomp[i] == 0 && bicomps[i][0] != beg && bicomps[i][0] != vt) {
            to_remove[i] = true;
        }
    }

    VB nodes_to_retain(N); // nodes that belong to a bicomponent with at lest two steps in it
    for( int i=0; i<bicomps.size(); i++ ) if(steps_in_bicomp[i] >= 2) for(auto d : bicomps[i]) nodes_to_retain[d] = true;


    Mask res(N);
    VB in_rest(N,false);
    for(int i=0; i<bicomps.size(); i++) {
        for(auto x : bicomps[i]) if( !in_rest[x] && x != beg ) {
            in_rest[x] = true;
            res.set(x,true);
            if constexpr (debug) clog << "\t+adding hash for node " << x << endl;
        }
    }

    for(int i=0; i<bicomps.size(); i++) {
        if( to_remove[i] ) {
            if constexpr (debug) clog << "From biconnected component " << bicomps[i] << endl;
            for(int v : bicomps[i]){
                if( !was[v] && !nodes_to_retain[v] && in_rest[v] ) {
                    in_rest[v] = false;
                    res.set(v,false);
                    if constexpr (debug) clog << "\t-clearing hash for node " << v << endl;
                }
            }
        }
    }

    return res;
}



Mask Solver2MIM::getNodesAfterBiconnectedReduction2(VVI & underV, Mask & mask, int beg, int vt, VI pth) {
    constexpr bool debug = false;

    int N = underV.size();
    VB helper(N);


    for( int j=0; j<N; j++ ) helper[j] = !mask[j];
    helper[beg] = true; // remove node beg from mask before finding bicomponents

    Mask res(N);

    VI path = pth;
    if( path.empty() ) path = Solver::findShortestPath(underV, beg, vt, mask);
    // DEBUG(path);
    if( path.empty() ) return Mask(N); // no nodes after reduction, so no path...
    VB was_path(N,false);
    for(int d : path) was_path[d] = true;

    VI tin(N,-1);
    VI low(N, 1e9);
    VB was(N,false);
    VB was2(N,false);

    VI bicomp; bicomp.reserve(N);

    function<void(int,int)> findBicomp = [&]( int u, int forbid ){
        bicomp.push_back(u);
        was2[u] = true;
        if(u != forbid) {
            for( int d : underV[u] ) if( helper[d] && low[d] < tin[u] && !was2[d] ) findBicomp(d,forbid);
        }
    };

    int time_stamp = -1;
    function<void(int,int)> dfs = [&]( int u, int par ){
        was[u] = true;
        time_stamp++;
        tin[u] = time_stamp;
        low[u] = tin[u];

        for( int d : underV[u] ){
            if( helper[d] && !was[d] ){
                dfs(d,u);
                if( low[d] == tin[u] ){
                    bicomp.clear();
                    findBicomp(d,u);
                    for(auto x : bicomp) was2[x] = false;

                    int steps = 0;
                    for( int x : bicomp ) if( was_path[x] ) steps++;
                    // DEBUG(bicomp);
                    if( steps >= 2 ) for( int x : bicomp ) res.set(x,true);
                    else if( steps == 1 ) {
                        for( int x : bicomp ) if( was_path[x] ) res.set(x,true);
                    }
                }
            }
        }
        for(int d : underV[u]) if( d != par && helper[d] ) low[u] = min(low[u], low[d]);
    };

    for( int i=0; i<N; i++ ) if( helper[i] && !was[i] ) dfs(i,i);

    for( int i=0; i<N; i++ ) if( was_path[i]) res.set(i,true);


    if( beg != vt ) res.set(beg,false);

    return res;
}


HashType Solver2MIM::getMergeStateHashBiconnected(VVI & underV, Mask & mask, int beg, int vt) {
    return getHashForMask( getNodesAfterBiconnectedReduction(underV, mask, beg, vt), underV.size() );
}

// Mask Solver2MIM::getPassableNodes(GraphSP2 & g, GraphSP2& revg, Mask mask, int minl, int beg, int vt, int L ) {
Mask Solver2MIM::getPassableNodes(GraphSP2 & g, GraphSP2& revg, DistanceFinder2 & df_beg, DistanceFinder2 & df_vt,
    Mask mask, int minl, int beg, int vt, int L ) {
    int N = g.size();
    Mask passable_nodes_mask(N);

    assert(mask[beg]);
    mask.set(beg,0);
    // auto [dst_beg, par_beg] = Solver2::findDistances( g, beg, mask );
    df_beg.findDistances( g, beg, mask );
    // auto [dst_beg, par_beg] = findDistancesInclusive( g, beg, mask );
    // auto [dst_beg, par_beg] = findDistancesInclusive( g, beg, Mask::zeros );
    // auto [dst_vt, par_vt] = Solver2::findDistances( revg, vt, mask );
    df_vt.findDistances( revg, vt, mask );
    // auto [dst_vt, par_vt] = findDistancesInclusive( revg, vt, mask );
    // auto [dst_vt, par_vt] = findDistancesInclusive( revg, vt, Mask::zeros );
    mask.set(beg,1);

    for( int i=0; i<N; i++ ) {
        // if( !mask[i] && minl + dst_beg[i] + dst_vt[i] <= L ) {
        if( !mask[i] && minl + df_beg.getDst(i) + df_vt.getDst(i) <= L ) {
            passable_nodes_mask.set(i,true);
        }
    }

    return passable_nodes_mask;
}

tuple<HashType, HashType> Solver2MIM::getAllMergeStateHashes(GraphSP2 & g, GraphSP2 &revg, VVI & underV,
    SP4 & sp, int beg, int vt, int L) {
    constexpr bool debug = false;

    int N = underV.size();

    auto & mask = sp.mask;

    auto merge_state_mask = getNodesAfterBiconnectedReduction(underV, mask, beg, vt);
    auto merge_state_hash = getHashForMask(merge_state_mask, N);

    return {merge_state_hash, merge_state_hash};
}

pair<Mask, Mask> Solver2MIM::createReachableNodeMasksVsVt() {
    Mask reachable_nodes_vs(N), reachable_nodes_vt(N);

    auto V = g.getStruture();
    auto [dst_vs, par_vs] = Solver::findDistances( V, vs );
    for( int i=0; i<N; i++ ) if( dst_vs[i] <= L/2 ) reachable_nodes_vs.set(i,true);

    auto revV = GraphUtils::reverseGraph(V);
    auto [dst_vt, par_vt] = Solver::findDistances( revV, vt );
    for( int i=0; i<N; i++ ) if( dst_vt[i] <= L/2 ) reachable_nodes_vt.set(i,true);

    return { reachable_nodes_vs, reachable_nodes_vt };
}


VBIG Solver2MIM::solveMIM3() {
    constexpr bool debug = false;
    using VSP4 = vector<SP4>;

    if( !timer.hasSetLimit("mim3") ) timer.setLimit("mim3", 1e9);
    timer.start("mim3");

    if constexpr (debug) DEBUG(g);

    clog << "Solving using solver2.mim3" << endl;
    clog << "Time limit set to " << timer.getLimit("mim3") / 1000 << " seconds" << endl;
    // if( N > MAXN ) {
    //     cerr << "CANNOT PROCESS GRAPHS WITH MORE THAN " << MAXN << " NODES USING MIM3..." << endl;
    //     exit(1);
    // }

    if constexpr (debug) {
        DEBUG(N);
        DEBUG(g);
        DEBUG(revg);
        DEBUG(L);
        DEBUG(vs); DEBUG(vt);
    }


    vector<bigint> res_for_length(L+1,0);
    bool has_vsvt_arc = false;
    for( auto & arc : g[vs] ) if( arc.v == vt ) has_vsvt_arc = true;
    if( has_vsvt_arc ) {
        for( auto & arc : g[vs] ) {
            arc.lengths.trimToMax(L);
            if( arc.v == vt ) {
                for( int i=arc.minl(); i <= arc.maxl(); i++ ) res_for_length[i] += arc[i];
                break;
            }
        }
        Kernelizer2::removeArc(g, vs,vt);
        Kernelizer2::removeArc(g, vt,vs);
        Kernelizer2::removeArc(revg, vs,vt);
        Kernelizer2::removeArc(revg, vt,vs);
    }

    // DEBUG(g);
    // DEBUG(revg);

    vector<ULL> sps_counts;

    constexpr int masking_size = 8;
    constexpr int MS = (1<<masking_size);

    constexpr bool use_most_frequent_nodes_for_masking = true; // in mergeTrimExclude
    constexpr bool merge_states_for_calculation = true; // before proper calculating results secrion
    constexpr bool use_intersection_masks = false; // in calculating results

    if constexpr (debug) clog << "Creating arrays dp?s" << endl;

    vector<vector<SP4>> dpvs1(N), dpvs2(N);
    vector<vector<vector<SP4>>> dpvt1(N, vector<vector<SP4>>(MS)), dpvt2(N, vector<vector<SP4>>(MS));

    dpvs1[vs].emplace_back( Mask(N), node_hashes[vs], Lengths( VBIG(1,1), 0 ) );
    dpvs1[vs].back().mask.set(vs,1);

    dpvt1[vt][0].emplace_back( Mask(N), node_hashes[vt], Lengths( VBIG(1,1), 0 ) );
    dpvt1[vt][0].back().mask.set(vt,1);

    if constexpr (debug) clog << "Creating mask_nodes" << endl;

    VVI mask_nodes(N); // this is a vector of nodes used for
    VB was(N);
    for( int i=0; i<N; i++ ) {
        VI A;
        for(auto & arc : g[i]) { if(!was[arc.v]) { A.push_back(arc.v); was[arc.v] = true; } }
        for(auto & arc : revg[i]) { if(!was[arc.v]) { A.push_back(arc.v); was[arc.v] = true; } }
        auto B = A;
        for(int a : B) for( auto & arc : g[a] ) { if(!was[arc.v] && arc.v != i) { A.push_back(arc.v); was[arc.v] = true; } }
        B = A;
        for(int a : B) for( auto & arc : revg[a] ) { if(!was[arc.v] && arc.v != i) { A.push_back(arc.v); was[arc.v] = true; } }
        for(int d : A) was[d] = false;

        constexpr int seed = 123814;
        StandardUtils::shuffle(A, seed);
        if( A.size() > masking_size ) A.resize(masking_size);
        mask_nodes[i] = A;
    }

    Stopwatch sw;


    auto createMaskNodeMostFrequent = [&](int beg) {
        VLL freqs_vs(N,0);
        for( auto & sp : dpvs1[beg] ) for( int i=0; i<N; i++ ) if( sp.mask[i] ) freqs_vs[i]++;
        VLL freqs_vt(N,0);
        for( auto & vec : dpvt1[beg] ) {
            for( auto & sp : vec ) for( int i=0; i<N; i++ ) if( sp.mask[i] ) freqs_vt[i]++;
        }
        for( auto & vec : dpvt2[beg] ) {
            for( auto & sp : vec ) for( int i=0; i<N; i++ ) if( sp.mask[i] ) freqs_vt[i]++;
        }

        VI nodes(N); iota(ALL(nodes),0);
        sort(ALL(nodes), [&](int a, int b){ return freqs_vs[a] * freqs_vt[a] > freqs_vs[b] * freqs_vt[b]; });
        mask_nodes[beg].clear();
        int p=0;
        while( p < nodes.size() && mask_nodes[beg].size() < masking_size ) {
            int a = nodes[p];
            if( a != vs && a != vt && a != beg ) mask_nodes[beg].push_back(a);
            p++;
        }
    };


    auto getNeighborhoodMask = [&](Mask & m, int v) {
        // sw.start("creating neighborhood mask");
        assert(m[v]);
        int mask = 0;
        for(int i=0; i<mask_nodes[v].size(); i++) {
            if( m[mask_nodes[v][i]] ) mask |= ( 1 << i );
        }
        // sw.stop("creating neighborhood mask");
        return mask;
    };


    if constexpr (debug) clog << "Created" << endl;

    assert(getNeighborhoodMask( dpvt1[vt][0].back().mask, vt ) == 0);

    VI minl_vs(N), minl_vt(N), maxl_vs(N), maxl_vt(N);

    auto updateMinls = [&]() {
        // sw.start("updating min_ls/max_ls");
        fill(ALL(minl_vs),inf);
        fill(ALL(minl_vt),inf);
        fill(ALL(maxl_vs),-inf);
        fill(ALL(maxl_vt),-inf);


        for(int i=0; i<N; i++) {
            for( auto & sp : dpvs1[i] ) {
                minl_vs[i] = min(minl_vs[i], sp.counts.minl());
                maxl_vs[i] = max( maxl_vs[i], sp.counts.maxl() );
            }

            for( int j=0; j<MS; j++ ) {
                for( auto & sp : dpvt1[i][j] ) {
                    minl_vt[i] = min(minl_vt[i], sp.counts.minl());
                    maxl_vt[i] = max(maxl_vt[i], sp.counts.maxl());
                }
                for( auto & sp : dpvt2[i][j] ) {
                    minl_vt[i] = min(minl_vt[i], sp.counts.minl());
                    maxl_vt[i] = max(maxl_vt[i], sp.counts.maxl());
                }
            }
        }
        // sw.stop("updating min_ls/max_ls");
    };
    updateMinls();

    auto isLastIteration = [&]() {
        auto V = g.getStruture();
        for( int i=0; i<N; i++ ) {
            auto dst = Solver2::findDistances(g, i).first;
            auto dstV = Solver::findDistances(V,i).first;
            for( int j=0; j<N; j++ ) if( minl_vs[i] + minl_vt[j] + max(2,dst[j]) <= L ) return false;
        }
        return true;
    };

    // constexpr bool merge_states_biconnected = true;
    bool merge_states_biconnected = cnf.use_mim_advanced_state_merging;

    auto V = g.getStruture();
    auto revV = GraphUtils::reverseGraph(V);

    VB helper(N, true);
    VVI underV(N);
    for( int i=0; i<N; i++ ) {
        for( auto & arc : g[i] ) {
            underV[i].push_back(arc.v);
            underV[arc.v].push_back(i);
        }
    }
    underV = GraphUtils::makeSimple(underV);

    // those two masks mark nodes that can be reached by any path starting from vs/vt during algorithm execution.
    // in fact, it marks all nodes that are within path SIZE (not length) <= L/2 from vs/vt
    // this might hugely reduce number of states, if L is not much grater than dst(vs,vt) and edge lengths are mostly 1
    auto [reachable_nodes_vs, reachable_nodes_vt] = createReachableNodeMasksVsVt();

    DEBUG(reachable_nodes_vs.getSetBits().size());
    DEBUG(reachable_nodes_vt.getSetBits().size());

    int subsets_size = 1;

    auto updateSP4HashBiconnected = [&](auto & g, auto & revg, auto & underV, auto & sp, int beg, int vt) {
        sp.hash = get<1>( getAllMergeStateHashes( g, revg, underV, sp, beg, vt, L ) );
    };

    auto mergeEntries = [&](auto & new_entries) {
        sw.start("\tmerging");
        // if constexpr (debug) clog << "Merging new entries" << endl;
        int p = 0, q = 0, r = 0;
        // while( p < new_entries.size() ){
        //     q = p;
        //     auto sp1 = new_entries[p], sp2 = new_entries[q];
        //     while( q < new_entries.size() && (sp1 == sp2) ) if(++q <new_entries.size()) sp2 = new_entries[q];
        //     while( p+1 < q ) {
        //         sp1.counts += new_entries[++p].counts;
        //         if constexpr (debug) clog << "------>" << endl << "merging sp4 " << new_entries[p] << " to " << sp1 << endl;
        //     }
        //     new_entries[r++] = sp1;
        //     p = q;
        // }

        while( p < new_entries.size() ){
            q = p;
            while( q < new_entries.size() && (new_entries[p] == new_entries[q]) ) q++;
            for( int i=p+1; i<q; i++ ){
                new_entries[p].counts += new_entries[i].counts;
                if constexpr (debug) clog << "------>" << endl << "merging sp4 " << new_entries[i] << " to " << new_entries[p] << endl;
            }
            swap( new_entries[r++], new_entries[p] );
            p = q;
        }

        // new_entries.resize(r);
        while(new_entries.size() > r) new_entries.pop_back();
        sw.stop("\tmerging");
    };

    auto canTrimState = [&]( // GraphSP2 & g, GraphSP2 & revg,
        int source,
        int beg, auto & sp,
        // VI & dst_source, VI & dst_beg,
        DistanceFinder2 &df_source, DistanceFinder2 & df_beg,
        Mask * biconnected_trimmed_mask = nullptr,
        const bool check_only_biconnected = false,
        const bool calculate_biconnected = false ) {

        auto& mask = sp.mask;
        assert(mask[beg]);

        // mask.set(beg,0);
        // auto [dst_source, par_source] = findDistances( g, source, mask );
        // mask.set(beg,1);
        // if( !check_only_biconnected && sp.counts.minl() + dst_source[beg] > L ) return true;
        if( !check_only_biconnected && sp.counts.minl() + df_source.getDst(beg) > L ) return true;

        if(!check_only_biconnected){
            // now we check if the connected component in graph V \ mask that contains source contains at
            // least subsets_size-2 vertices
            // if not, then this state can be trimmed, as it will no longer be used in result calculation
            // only nodes that can be reached from source are considered
            int cmp_size = 0;
            // for( int i=0; i<dst_source.size(); i++ ) if( sp.counts.minl() + dst_source[i] <= L ) cmp_size++;
            for( int i=0; i<N; i++ ) if( sp.counts.minl() + df_source.getDst(i) <= L ) cmp_size++;
            if( cmp_size < subsets_size-1 ) {
                if constexpr (debug) clog << "Setting state " << sp << " to trim, #2" << endl;
                return true;
            }
        }

        if(!check_only_biconnected){
            int cmp_size = 0;
            // for( int i=0; i<dst_source.size(); i++ ) if( sp.counts.minl() + dst_beg[i] + dst_source[i] <= L ) cmp_size++;
            for( int i=0; i<N; i++ ) if( sp.counts.minl() + df_beg.getDst(i) + df_source.getDst(i) <= L ) cmp_size++;
            if( cmp_size < subsets_size-1 ) {
                if constexpr (debug) clog << "Setting state " << sp << " to trim, #3" << endl;
                return true;
            }
        }

        if(biconnected_trimmed_mask != nullptr) {
            assert(mask[beg]);
            mask.set(beg,0);
            Mask mask_left = (*biconnected_trimmed_mask);
            if(calculate_biconnected) mask_left = getNodesAfterBiconnectedReduction(underV, mask, beg, source);
            // DEBUG( mask_left.useBits() );
            assert(mask_left[beg] == false);
            int max_nodes_in_paired_path = mask_left.popcnt()+1; // beg can also be, this is a shared node
            mask.set(beg,1);

            if( max_nodes_in_paired_path < subsets_size-1 ) return true;
        }

        return false;
    };

    auto trimStates = [L=L, &subsets_size, &underV, &sw]( GraphSP2 & g, GraphSP2 & revg, int source, int beg, auto & sps ) {
        sw.start("\ttrimming");
        sort(ALL(sps) );

        VB indices_to_trim( sps.size() );
        int p = 0;
        while( p < sps.size() ) {
            if( sps[p].counts.minl() > L || sps[p].counts.empty() || !sps[p].counts.valid() ) indices_to_trim[p] = true;


            if(!indices_to_trim[p]) {
                auto& mask = sps[p].mask;
                assert(mask[beg]);


                mask.set(beg,0);
                auto [dst, par] = Solver2::findDistances( g, source, mask );
                mask.set(beg,1);
                if( sps[p].counts.minl() + dst[beg] > L ) {
                    if constexpr (debug) {
                        clog << "Setting state " << sps[p] << " to trim, #1" << endl;
                        DEBUG(dst);  DEBUG(beg);  DEBUG(dst[beg]); DEBUG(source);
                    }
                    indices_to_trim[p] = true;
                }

                if( !indices_to_trim[p] ) {
                    // now we check if the connected component in graph V \ mask that contains source contains at
                    // least subsets_size-2 vertices
                    // if not, then this state can be trimmed, as it will no longer be used in result calculation
                    // only nodes that can be reached from source are considered
                    int cmp_size = 0;
                    for( int i=0; i<dst.size(); i++ ) if( sps[p].counts.minl() + dst[i] <= L ) cmp_size++;
                    if( cmp_size < subsets_size-1 ) {
                        if constexpr (debug) clog << "Setting state " << sps[p] << " to trim, #2" << endl;
                        indices_to_trim[p] = true;
                    }
                }

                if(!indices_to_trim[p]) {
                    mask.set(beg,0);
                    auto [dst2, par2] = Solver2::findDistances(revg, beg, mask);
                    mask.set(beg,1);

                    int cmp_size = 0;
                    for( int i=0; i<dst.size(); i++ ) if( sps[p].counts.minl() + dst2[i] + dst[i] <= L ) cmp_size++;
                    if( cmp_size < subsets_size-1 ) {
                        if constexpr (debug) clog << "Setting state " << sps[p] << " to trim, #3" << endl;
                        indices_to_trim[p] = true;
                    }
                }

                if(!indices_to_trim[p]) {
                    mask.set(beg,0);
                    auto mask_left = getNodesAfterBiconnectedReduction(underV, mask, beg, source);
                    assert(mask_left[beg] == false);
                    int max_nodes_in_paired_path = mask_left.popcnt()+1; // beg can also be, this is a shared node
                    mask.set(beg,1);

                    if( max_nodes_in_paired_path < subsets_size-1 ) indices_to_trim[p] = true;
                }
            }

            p++;
        }

        for( int i=(int)sps.size()-1; i>=0; i-- ) {
            if( indices_to_trim[i] ) {
                if constexpr (debug) clog << "Trimming state: " << sps[i].toString() << endl;
                swap( sps[i], sps.back() );
                sps.pop_back();
            }
        }

        sw.stop("\ttrimming");
    };

    // merges states, trims states and excludes from computation in current iteration
    auto mergeTrimExclude = [&]() {

        sw.start("merging and trimming");
        for( int i=0; i<N; i++ ) {
            // sw.start("\tsorting");
            sort(ALL(dpvs1[i]));
            // sw.stop("\tsorting");

            mergeEntries( dpvs1[i] );
            trimStates( revg, g, vt, i, dpvs1[i] );

            if(merge_states_biconnected) {
                sw.start("\tmerging biconnected");
                for( auto & sp : dpvs1[i] ) updateSP4HashBiconnected(g, revg, underV, sp,i, vt);

                // sw.start("\tsorting");
                sort(ALL(dpvs1[i]));
                // sw.stop("\tsorting");

                mergeEntries( dpvs1[i] );
                for( auto & sp : dpvs1[i] ) { // getting back original state hash
                    sp.hash = 0;
                    for( int j=0; j<N; j++ ) if(sp.mask[j]) sp.hash ^= node_hashes[j];
                }
                sw.stop("\tmerging biconnected");
            }
        }

        // clog << "Trimming now dpvt" << endl;
        for( int i=0; i<N; i++ ) {

            for(int j=0; j<MS; j++) {
                sort(ALL(dpvt1[i][j]));

                mergeEntries( dpvt1[i][j] );
                trimStates( g, revg, vs,i, dpvt1[i][j] );
            }

            if(merge_states_biconnected) {
                sw.start("\tmerging biconnected");
                vector<SP4> to_update;

                if constexpr (use_most_frequent_nodes_for_masking) createMaskNodeMostFrequent(i);

                for( int j=0; j<MS; j++ ) {
                    to_update += dpvt1[i][j];
                    dpvt1[i][j].clear();
                }
                for( auto & sp : to_update ) updateSP4HashBiconnected(revg, g, underV, sp,i, vs);
                sort(ALL(to_update));

                mergeEntries( to_update );
                for( auto & sp : to_update ) { // getting back original state hash
                    sp.hash = 0;
                    for( int j=0; j<N; j++ ) if(sp.mask[j]) sp.hash ^= node_hashes[j];
                }


                // now we need to put updated states back to buckets...
                for( auto & sp : to_update ) {
                    auto neigh_mask = getNeighborhoodMask( sp.mask, i );
                    dpvt1[i][neigh_mask].emplace_back( sp );
                }

                if constexpr (use_most_frequent_nodes_for_masking) {
                    // relocating dpvt2 - it is neccessary if createMaskNodeMostFrequent
                    to_update.clear();
                    for( int j=0; j<MS; j++ ) {
                        to_update += dpvt2[i][j];
                        dpvt2[i][j].clear();
                    }
                    for( auto & sp : to_update ) {
                        auto neigh_mask = getNeighborhoodMask( sp.mask, i );
                        dpvt2[i][neigh_mask].emplace_back( sp );
                    }
                }

                sw.stop("\tmerging biconnected");
            }
        }
        sw.stop("merging and trimming");
    };

    auto mergeTrimExcludeFast = [&]() {
        // mergeTrimExclude(); // #TEST
        // return;
        clog << "Using mergeTrimExcludeFast" << endl;

        int total_sps = 0;
        for(int i=0; i<N; i++) {
            total_sps += dpvs1[i].size();
            for( int j=0; j<MS; j++ ) total_sps += dpvt1[i][j].size() + dpvt2[i][j].size();
        }
        bool use_trimming = !isLastIteration();
        // bool use_trimming = true;

        BiconnectedFinder bf(N);
        DistanceFinder2 df_beg(N), df_vt(N);

        auto process = [N=N, L=L, &underV, &mergeEntries, &canTrimState, &use_trimming, &subsets_size, &bf,
            &df_beg, &df_vt]( auto & g, auto & revg,

            int source, int beg, auto & sps, auto & minls, Mask & reachable_source_nodes ) {
            sort(ALL(sps));
            mergeEntries( sps );

            if( !use_trimming ) return;

            VB indices_to_trim(sps.size());

            for( int j=0; j<sps.size(); j++ ) {
                auto & sp = sps[j];

                assert(sp.mask[beg]);
                sp.mask.set(beg,0);
                // auto [dst_vt, par_vt] = Solver2::findDistances( revg, source, sp.mask );
                // auto [dst_beg, par_beg] = Solver2::findDistances( g, beg, sp.mask );
                df_vt.findDistances(revg, source, sp.mask);
                df_beg.findDistances( g, beg, sp.mask );
                sp.mask.set(beg,1);

                // if( canTrimState( source, beg, sp, dst_vt, dst_beg, nullptr ) ) {
                if( canTrimState( source, beg, sp, df_vt, df_beg, nullptr ) ) {
                    indices_to_trim[j] = true;
                    continue;
                }

                // now checking, if it is possible, that a path with state sp will ever have length <= L
                // such a path must be sp -> i -> source, which is bounded from belowe by the following
                // sp.counts.minl() + dst_beg[i] + minls[i]
                int min_pathlength = inf;
                // for( int i=0; i<N; i++ ) min_pathlength = min( min_pathlength, sp.counts.minl() + dst_beg[i] + minls[i] );
                for( int i=0; i<N; i++ ) min_pathlength = min( min_pathlength, sp.counts.minl() + df_beg.getDst(i) + minls[i] );
                if( min_pathlength > L ) {
                    indices_to_trim[j] = true;
                    continue;
                }

                assert(sp.mask[beg]);
                // Mask biconnected_trimmed_mask = getNodesAfterBiconnectedReduction(underV, sp.mask, beg, source);
                VI path;
                // for( int u=beg; u != -1; u = par_vt[u] ) path.push_back(u);
                for( int u=beg; u != -1; u = df_vt.getPar(u) ) path.push_back(u);
                // Mask biconnected_trimmed_mask = getNodesAfterBiconnectedReduction2(underV, sp.mask, beg, source, path);
                Mask biconnected_trimmed_mask = bf.getNodesAfterBiconnectedReduction2(underV, sp.mask, beg, source, path);

                constexpr bool trim_unreachable_nodes = true;
                // this seems to be correct - at least passes stress tests
                if constexpr (trim_unreachable_nodes) {
                    for( int i=0; i<N; i++ ) {
                        // if( biconnected_trimmed_mask[i] && (dst_beg[i] == inf || dst_vt[i] == inf) ) {
                        if( biconnected_trimmed_mask[i] && (df_beg.getDst(i) == inf || df_vt.getDst(i) == inf) ) {
                            biconnected_trimmed_mask.set(i,false);
                        }
                    }
                }

                // if( canTrimState(source, beg, sp, dst_vt, dst_beg, &biconnected_trimmed_mask, true ) ) {
                if( canTrimState(source, beg, sp, df_vt, df_beg, &biconnected_trimmed_mask, true ) ) {
                    indices_to_trim[j] = true;
                    continue;
                }

                sp.hash = getHashForMask( biconnected_trimmed_mask, N );
            }


            for( int j=(int)sps.size()-1; j>=0; j-- ) { // trimming states
                if( indices_to_trim[j] ) {
                    swap( sps[j], sps.back() );
                    sps.pop_back();
                }
            }

            sort(ALL(sps));
            mergeEntries( sps );

            bool find_passable_nodes = ( subsets_size & 1 ); // #TEST - do it every second iteration
            if (find_passable_nodes) {
                indices_to_trim = VB(sps.size(), false);
                for( int j=0; j<sps.size(); j++ ) {
                    auto & sp = sps[j];
                    // Mask passable_nodes_mask = getPassableNodes( g, revg, sp.mask, sp.counts.minl(), beg, source, L );
                    Mask passable_nodes_mask = getPassableNodes( g, revg, df_beg, df_vt, sp.mask, sp.counts.minl(), beg, source, L );
                    sp.hash = getHashForMask(passable_nodes_mask, N);

                    if constexpr (debug) clog << "\tsp: " << sp << ", passable_nodes: " << passable_nodes_mask.getSetBits() << endl;
                }


                sort(ALL(sps));

                // here must be special merging - it needs to be checked if state sp2 can be merged to sp1
                // it can be done if the set of passable nodes does not changed when mask of sp2 is changed to the mask
                // of sp1.
                int p = 0, q = 0;
                VB merged(sps.size(),false);

                // check if can merge sp2 to sp1 or sp1 to sp2.
                // if can merge sp1 to sp2, then -1 is returned, if sp2 to sp1, then 1, otherwise 0
                auto canMerge = [&]( auto & sp1, auto & sp2 ) {
                    // auto sp2hash_with_sp1_mask = getHashForMask( getPassableNodes(g, revg, sp1.mask, sp2.counts.minl(), beg, source, L), N);
                    auto sp2hash_with_sp1_mask = getHashForMask( getPassableNodes(g, revg, df_beg, df_vt, sp1.mask, sp2.counts.minl(), beg, source, L), N);
                    if( sp2.hash == sp2hash_with_sp1_mask ) return 1;
                    if constexpr (debug) clog << "Cannot merge sp2: " << sp2 << " to sp1: " << sp1 << endl;

                    // auto sp1hash_with_sp2_mask = getHashForMask( getPassableNodes(g, revg, sp2.mask, sp1.counts.minl(), beg, source, L), N);
                    auto sp1hash_with_sp2_mask = getHashForMask( getPassableNodes(g, revg, df_beg, df_vt, sp2.mask, sp1.counts.minl(), beg, source, L), N);
                    if( sp1.hash == sp1hash_with_sp2_mask ) return -1;
                    if constexpr (debug)  clog << "Cannot merge sp1: " << sp1 << " to sp2: " << sp1 << endl;

                    return 0;
                };

                while( p < sps.size() ){
                    q = p;
                    // auto sp1 = sps[p], sp2 = sps[q];
                    // while( q < sps.size() && (sp1 == sp2) ) if(++q <sps.size()) sp2 = sps[q];
                    while( q < sps.size() && (sps[p] == sps[q]) ) q++;
                    for( int i=p; i<q; i++ ) {
                        if( merged[i] ) continue;
                        auto & sp1 = sps[i];

                        for( int j=i+1; j<q; j++ ) {
                            if( merged[j] ) continue;
                            auto & sp2 = sps[j];

                            int can_merge = canMerge(sp1,sp2);

                            if( can_merge == -1 ) {
                                if constexpr (debug) {
                                    clog << endl << endl << "passables-merging sp1: " << sp1 << " to " << endl
                                            << "sp2: " << sp2 << endl << endl;
                                }
                                sp2.counts += sp1.counts;
                                merged[i] = true;
                                break;
                            }else if( can_merge == 1 ) {
                                if constexpr (debug) {
                                    clog << endl << endl << "passables-merging sp2: " << sp2 << " to " << endl
                                        << "sp1: " << sp1 << endl << endl;
                                }
                                sp1.counts += sp2.counts;
                                merged[j] = true;
                                // we do not need to break here, sp2 for other j might still be merged to sp1
                            }
                        }
                    }

                    p = q;
                }

                int r = 0;
                for(int i=0; i<sps.size(); i++) if( !merged[i] ) swap( sps[r++], sps[i] );
                while(sps.size() > r) sps.pop_back();
            }

            for( auto & sp : sps ) sp.hash = getHashForMask(sp.mask, N);
        };

        sw.start("merging and trimming");
        for( int i=0; i<N; i++ ) {
            if( timer.tle("mim3") && !isLastIteration() ) return;
           process( g, revg,vt, i, dpvs1[i], minl_vt, reachable_nodes_vt );
        }

        for( int i=0; i<N; i++ ) {
            if( timer.tle("mim3") && !isLastIteration() ) return;
            vector<SP4> to_update;

            // // use here to get best mask to nodes for states before merging and trimming
            // if constexpr (use_most_frequent_nodes_for_masking) createMaskNodeMostFrequent(i);

            for( int j=0; j<MS; j++ ) {
                to_update += dpvt1[i][j];
                dpvt1[i][j].clear();
            }

            process( revg, g, vs, i, to_update, minl_vs, reachable_nodes_vs );

            // now we need to put updated states back to buckets...
            for( auto & sp : to_update ) {
                auto neigh_mask = getNeighborhoodMask( sp.mask, i );
                dpvt1[i][neigh_mask].emplace_back( sp );
            }

            if constexpr (use_most_frequent_nodes_for_masking) {
                // relocating dpvt2 - it is neccessary if createMaskNodeMostFrequent
                createMaskNodeMostFrequent(i);

                auto relocateSps = [&]( auto & sps ) {
                    to_update.clear();
                    for( int j=0; j<MS; j++ ) {
                        to_update += sps[j];
                        sps[j].clear();
                    }
                    for( auto & sp : to_update ) {
                        auto neigh_mask = getNeighborhoodMask( sp.mask, i );
                        sps[neigh_mask].emplace_back( sp );
                    }
                };

                relocateSps( dpvt1[i] );
                relocateSps( dpvt2[i] );
            }

        }
        sw.stop("merging and trimming");
    };


    auto continueIteration = [&]() {
        if constexpr (debug) {
            DEBUG(minl_vs);
            DEBUG(minl_vt);
            DEBUG(subsets_size);
        }
        if( 2*subsets_size-3 > N-1 ) return false;
        // int mvs = inf, mvt = inf;
        // for(int i=0; i<N; i++) {
        //     mvs = min(mvs, minl_vs[i]);
        //     mvt = min(mvt, minl_vt[i]);
        // }
        // return mvs + mvt <= L;

        for( int i=0; i<N; i++ ) {
            auto dst = Solver2::findDistances(g, i).first;
            for( int j=0; j<N; j++ ) if( minl_vs[i] + dst[j] + minl_vt[j] <= L ) return true;
        }
        return false;
    };

    auto calculateUsingBitmasksMS = [&getNeighborhoodMask, &subsets_size, L=L, &res_for_length,N=N]
        (auto& dpvs1, auto& dpvt1, auto& dpvt2, int beg) {

        ULL all_bitmask_checks = 0;
        ULL masked_bitmask_checks = 0;
        ULL real_masked_bitmask_checks = 0;
        ULL valid_bitmask_checks = 0;

        vector<Mask> inters1;
        vector<Mask> inters2;

        if constexpr (use_intersection_masks) {
            inters1 = vector<Mask>(MS, Mask::ones(N));
            inters2 = vector<Mask>(MS, Mask::ones(N));

            for( int j=0; j<MS; j++ ) {
                for(auto & sp : dpvt1[j]) inters1[j] &= sp.mask;
                inters1[j].set(beg,false);
            }
            for( int j=0; j<MS; j++ ) {
                for(auto & sp : dpvt2[j]) inters2[j] &= sp.mask;
                inters2[j].set(beg,false);
            }
        }

        for( auto & sp1 : dpvs1 ) {
            int neigh_mask_sp1 = getNeighborhoodMask( sp1.mask, beg );

            sp1.mask.set(beg,0);

            if( 2*subsets_size-2 <= L ) {
                Lengths res( VBIG(1,0), 0 );
                for( int j=0; j<MS; j++ ){
                    all_bitmask_checks += dpvt1[j].size();
                    if( neigh_mask_sp1 & j ) continue; // if those masks intersect, then no need to check...
                    if constexpr (use_intersection_masks) {
                        // if( !((sp1.mask & inters1[j]) == Mask::zeros) ) continue;
                        if( !sp1.mask.noIntersection(inters1[j]) ) continue; // #TEST
                    }
                    masked_bitmask_checks += dpvt1[j].size();

                    for( auto & sp2 : dpvt1[j] ) {
                        // if( sp1.counts.minl() + sp2.counts.minl() > L ) break; // #TEST
                        real_masked_bitmask_checks++;
                        // if((sp1.mask & sp2.mask).m == Mask::zeros(N)) {
                        if(sp1.mask.noIntersection(sp2.mask)) {
                            if constexpr (debug) {
                                clog << "\t\tFound valid path: " << sp1 << " -> " << beg << " -> " << sp2 << endl;
                            }
                            res += sp2.counts;
                            valid_bitmask_checks++;
                        }
                    }
                }
                res *= sp1.counts;
                res.trimToMax(L);
                for( int j=res.minl(); j <= res.maxl(); j++ ) res_for_length[j] += res[j];
            }

            if( 2*subsets_size-3 <= L ) {
                Lengths res = Lengths(VBIG(1,0), 0);
                for( int j=0; j<MS; j++ ) {
                    all_bitmask_checks += dpvt2[j].size();
                    if( neigh_mask_sp1 & j ) continue; // if those masks intersect, then no need to check...
                    if constexpr (use_intersection_masks) {
                        // if( !((sp1.mask & inters2[j]) == Mask::zeros) ) continue;
                        if( !sp1.mask.noIntersection(inters2[j]) ) continue; // #TEST
                    }
                    masked_bitmask_checks += dpvt2[j].size();

                    for( auto & sp2 : dpvt2[j] ) {
                        // if( sp1.counts.minl() + sp2.counts.minl() > L ) break; // #TEST
                        real_masked_bitmask_checks++;
                        // if((sp1.mask & sp2.mask).m == Mask::zeros(N)) {
                        if(sp1.mask.noIntersection(sp2.mask)) {
                            if constexpr (debug) {
                                clog << "\t\tFound valid path: " << sp1 << " -> " << beg << " -> " << sp2 << endl;
                            }
                            res += sp2.counts;
                            valid_bitmask_checks++;
                        }
                    }
                }
                res *= sp1.counts;
                res.trimToMax(L);
                for( int j=res.minl(); j <= res.maxl(); j++ ) res_for_length[j] += res[j];
            }

            sp1.mask.set(beg,1);
        }

        if(all_bitmask_checks) {
            clog << "\tThere are " << masked_bitmask_checks << " masked check, out of " << all_bitmask_checks <<
                " all bitmskask checks - that is " << 100.0 * masked_bitmask_checks / all_bitmask_checks << "%" << endl;
            clog << "\t\treal_masked_bitmask_checks: " << real_masked_bitmask_checks << ", that is "
                << 100.0 * real_masked_bitmask_checks / masked_bitmask_checks << "% of masked_checks" << endl;
            clog << "\t\tvalid ones: " << valid_bitmask_checks << ", that is "
                << 100.0 * valid_bitmask_checks / real_masked_bitmask_checks << "% of real checks done" << endl;
        }
    };



    auto calculateUsingIEP = [cnf=cnf, &calculateUsingBitmasksMS,
        N=N, &res_for_length, L=L, &vt = this->vt,
        &vs = this->vs, &g=this->g, &revg = this->revg
        ]
        (VSP4& dpvs1, vector<VSP4>& dpvt1, vector<VSP4>& dpvt2, int beg) {

        if constexpr (debug) {
            clog << endl << endl << "using IEP, beg: " << beg << endl;
            // DEBUG(dpvs1); DEBUG(dpvt1); DEBUG(dpvt2);
        }


        const int MS_THR = cnf.mim_iep_size_threshold;

        vector<vector<pair<Mask,SP4>>>  small_vt1(MS),  small_vt2(MS);
        vector<VSP4> large_vt1(MS), large_vt2(MS);
        vector<pair<Mask,SP4>> small_vs;
        VSP4 large_vs;

        Mask sum_dpvs = Mask::zeros(N);
        for(auto & sp : dpvs1) sum_dpvs |= sp.mask;
        Mask sum_dpvt = Mask::zeros(N);
        for(auto & vec : dpvt1) for(auto & sp : vec) sum_dpvt |= sp.mask;
        for(auto & vec : dpvt2) for(auto & sp : vec) sum_dpvt |= sp.mask;


        for(auto & sp : dpvs1) {
            auto mask = sp.mask;
            mask &= sum_dpvt;
            if(mask.popcnt() > MS_THR) large_vs.emplace_back(sp);
            else small_vs.emplace_back(mask,sp);
        }
        for(int j=0; j<MS; j++){
            for( auto & sp : dpvt1[j] ) {
                auto mask = sp.mask;
                mask &= sum_dpvs;
                if(mask.popcnt() > MS_THR) large_vt1[j].emplace_back(sp);
                else small_vt1[j].emplace_back(mask,sp);
            }
        }
        for(int j=0; j<MS; j++) {
            for( auto & sp : dpvt2[j] ) {
                auto mask = sp.mask;
                mask &= sum_dpvs;
                if(mask.popcnt() > MS_THR) large_vt2[j].emplace_back(sp);
                else small_vt2[j].emplace_back(mask,sp);
            }
        }

        if constexpr (debug) {
            DEBUG(large_vs);  DEBUG(small_vs);
            DEBUG(large_vt1); DEBUG(small_vt1);
            DEBUG(large_vt2); DEBUG(small_vt2);
        }

        ULL large_cnt = large_vs.size() + large_vt1.size() + large_vt2.size();
        ULL small_cnt = small_vs.size() + small_vt1.size() + small_vt2.size();
        if constexpr (debug) {
            DEBUG(large_cnt); DEBUG(small_cnt);
        }

        ULL large_small_pairs = large_vs.size() + (dpvt1.size() + dpvt2.size())
            + small_vs.size() * (large_vt1.size() + large_vt2.size());
        if constexpr (debug) {
            DEBUG(large_small_pairs);
        }

        {
            calculateUsingBitmasksMS( large_vs, dpvt1, dpvt2, beg );

            VSP4 small_vs_sps;
            small_vs_sps.reserve(small_vs.size());
            for( auto & [mask,sp] : small_vs ) small_vs_sps.push_back(sp);
            calculateUsingBitmasksMS( small_vs_sps, large_vt1, large_vt2, beg  );
        }


        // now calculate results contatining SPs with only small masks

        constexpr int hs_bits = 10;
        constexpr int HS = (1<<hs_bits); // HS buckets are created for IEP subsets hashes

        // inserts all entries for subsets of mask of sp to sub_map
        auto insertSubsets = [N=N, &beg](auto & mask, auto & counts, auto & sub_map, auto & odd_masks) {
            VI sub; sub.reserve(mask.popcnt());
            for( int i=0; i<N; i++ ) if(mask[i] && i != beg) sub.push_back(i);
            int n = sub.size();
            for( int i=1; i<(1<<n); i++ ) {
                HashType h = 0;
                for( int j=0; j<n; j++ ) if( i & (1<<j) ) h ^= node_hashes[sub[j]];
                sub_map[h.hash  & (HS-1)].emplace_back(h,counts);
                if( popcount(unsigned(i)) & 1 ) odd_masks.insert(h);
            }
        };


        vector<vector<pair<HashType,Lengths>>> sub_map_vs(HS), sub_map_vt(HS);

        unordered_set<HashType, HashTypeHash> odd_masks;
        ULL total_subsets = 0;
        ULL total_created_subsets = 0;
        Lengths sum_vs, sum_vt;

        for( auto& [mask,sp] : small_vs ) {
            total_subsets += (1ll << (sp.mask.popcnt()-1))-1;
            total_created_subsets += (1ll << (mask.popcnt()))-1;
            sum_vs += sp.counts;
            // insertSubsets(sp.mask, sp.counts, sub_map_vs, odd_masks);
            insertSubsets(mask, sp.counts, sub_map_vs, odd_masks);
        }
        for(int j=0; j<MS; j++) {
            for( auto& [mask,sp] : small_vt1[j] ) {
                total_subsets += (1ll << (sp.mask.popcnt()-1))-1;
                total_created_subsets += (1ll << (mask.popcnt()))-1;
                sum_vt += sp.counts;
                // insertSubsets(sp.mask, sp.counts, sub_map_vt, odd_masks);
                insertSubsets(mask, sp.counts, sub_map_vt, odd_masks);
            }
            for( auto& [mask,sp] : small_vt2[j] ) {
                total_subsets += (1ll << (sp.mask.popcnt()-1))-1;
                total_created_subsets += (1ll << (mask.popcnt()))-1;
                sum_vt += sp.counts;
                // insertSubsets(sp.mask, sp.counts, sub_map_vt, odd_masks);
                insertSubsets(mask, sp.counts, sub_map_vt, odd_masks);
            }
        }

        Lengths res = sum_vs * sum_vt;
        ULL distinct_subsets = 0;

        { // merge sub_maps - only if vector is used instaed of maps...
            auto mergeSubMap = []( auto & v ) {
                sort(ALL(v), [](auto & a, auto & b){ return a.first < b.first; });
                int p=0, q= 0, r = 0;
                while(p < v.size()) {
                    q = p;
                    while( q < v.size() && v[p].first == v[q].first ) q++;
                    for(int i=p+1; i<q; i++) v[p].second += v[i].second;
                    swap(v[r++], v[p]);
                    p = q;
                }
                while(v.size() > r) v.pop_back();
            };

            for(int i=0; i<HS; i++) {
                mergeSubMap(sub_map_vs[i]);
                mergeSubMap(sub_map_vt[i]);
                int p = 0, q = 0;
                auto & v1 = sub_map_vs[i];
                auto & v2 = sub_map_vt[i];
                while(p < v1.size() && q < v2.size()) {
                    if( v1[p].first == v2[q].first ) {
                        auto temp = (v1[p].second * v2[q].second);
                        if( odd_masks.contains(v1[p].first) ) res -= temp;
                        else res += temp;
                        p++;
                    }else if( v1[p].first < v2[q].first ) p++;
                    else q++;
                }

                distinct_subsets += v1.size() + v2.size();
            }
        }

        if(distinct_subsets)
        clog << "IEP: created_subsets: " << total_created_subsets << ", from which distinct: " << distinct_subsets
            << " distinct subsets for IEP, while total number of susbets is: "
        << total_subsets << ", checked only " << 100.0 * total_created_subsets / total_subsets
            << "%" << endl;

        if constexpr (debug) clog << "res: " << res << endl;
        clog << endl; // #TEST

        res.trimToMax(L);
        for( int j=res.minl(); j <= res.maxl(); j++ ) res_for_length[j] += res[j];
    };

    bool calculate_using_iep = cnf.use_mim_iep;

    auto calculateIterationResult = [&]() {
        if constexpr (debug) {
            clog << "\tProceeding to calculate results for lengths " << 2*subsets_size-1 << " and "
                 << 2*subsets_size << endl;
        }

        ULL total_comparisons = 0;
        for(int i=0; i<N; i++) {
            for( auto & sp1 : dpvs1[i] ) {
                int neigh_mask_sp1 = getNeighborhoodMask( sp1.mask, i );
                for(int j=0; j<MS; j++) {
                    if( (neigh_mask_sp1 & j) == 0 ) total_comparisons += dpvt1[i][j].size() + dpvt2[i][j].size();
                }
            }
        }
        clog << "\tThere needs to be done at most " << total_comparisons << " bitset mask comparisons" << endl;


        auto cmp = []( auto & sp1, auto & sp2 ) { return sp1.counts.minl() < sp2.counts.minl(); };

        for( int i=0; i<N; i++ ) {
            if( timer.tle("mim3") && !isLastIteration() ) return;

            sort(ALL(dpvs1[i]), cmp);
            for( int j=0; j<MS; j++ ) {
                sort(ALL(dpvt1[i][j]), cmp);
                sort(ALL(dpvt2[i][j]), cmp);
            }

            bool any_pairs = true;
            if( dpvs1[i].empty() ) any_pairs = false;
            else {
                ULL r = 0;
                for(int j=0; j<MS; j++) r += dpvt1[i][j].size() + dpvt2[i][j].size();
                any_pairs = (r > 0);
            }
            if( !any_pairs ) continue;


            auto _dpvs = dpvs1[i];
            auto _dpvt1 = dpvt1[i];
            auto _dpvt2 = dpvt2[i];

            if constexpr (merge_states_for_calculation) {
                sw.start("merge_states_for_calculation");
                { // merge entries from dpvs1
                    for( int t = (int)_dpvs.size()-1; t>=0; t-- ) {
                        auto & sp = _dpvs[t];
                        Mask sum_vt(N);

                        auto neigh_mask = getNeighborhoodMask(sp.mask,i);
                        for(int j=0; j<MS; j++) {
                            if(neigh_mask & j) continue;
                            // for(auto & sp2 : dpvt1[i][j]) sum_vt |= sp2.mask;
                            for(auto & sp2 : _dpvt1[j]) sum_vt |= sp2.mask;
                            // for(auto & sp2 : dpvt2[i][j]) sum_vt |= sp2.mask;
                            for(auto & sp2 : _dpvt2[j]) sum_vt |= sp2.mask;
                        }

                        if( sum_vt.popcnt() == 0 ) {
                            // we can exclude sp from checks, as it brings absolutely nothing...
                            swap(_dpvs[t], _dpvs.back()); // can i do that ? exclude from checks...
                            _dpvs.pop_back();
                            continue;
                        }

                        sp.hash = 0;
                        for( int k=0; k<N; k++ ) if( sp.mask[k] && sum_vt[k] ) sp.hash ^= node_hashes[k];
                        for(int j=0; j<mask_nodes[i].size(); j++) {
                            int d = mask_nodes[i][j];
                            if( neigh_mask & (1<<j) )  sp.hash ^= node_hashes2[d];
                        }
                    }

                    sort( ALL(_dpvs) );
                    mergeEntries(_dpvs);

                    for( auto & sp : _dpvs ) { // bring hashes back to original values
                        sp.hash = 0;
                        for( int k=0; k<N; k++ ) if( sp.mask[k] ) sp.hash ^= node_hashes[k];
                    }

                    sort( ALL(_dpvs) );
                }

                // { // this seems to work, but it can be done better...
                //     // merge entries from dpvt1 and dpvt2 - this is expected to have slighly worse
                //     // effect than merging dpvs1, because here all sp in dpvs1 are taken for mask union, instead of just
                //     // those with nonintersecting masking nodes
                //
                //     Mask sum_vs;
                //     for(auto & sp : _dpvs) sum_vs |= sp.mask;
                //
                //     auto process = [&sum_vs, N=N, &mergeEntries, &i, &getNeighborhoodMask](auto & _dpvt) {
                //         vector<SP4> sps;
                //         sps.reserve( accumulate(ALL(_dpvt),0,[](int a, auto & b){return a + b.size();}) );
                //         for( auto & v : _dpvt ) {
                //             sps += v;
                //             v.clear();
                //         }
                //
                //         for( auto & sp : sps ) sp.hash = getHashForMask( sp.mask & sum_vs, N );
                //
                //         sort( ALL(sps) );
                //         mergeEntries(sps);
                //
                //         for( auto & sp : sps ) {
                //             sp.hash = getHashForMask( sp.mask, N );
                //             auto neigh_mask = getNeighborhoodMask( sp.mask, i );
                //             _dpvt[neigh_mask].push_back(sp);
                //         }
                //     };
                //
                //     if(sum_vs.popcnt() == 0) continue; // no need to consider even calculation...
                //
                //     process(_dpvt1);
                //     process(_dpvt2);
                // }

                { // merge entries from dpvt1 and dpvt2
                    vector<Mask> sum_vs(MS, Mask(N));
                    for(auto & sp : _dpvs) {
                        auto neigh_mask = getNeighborhoodMask(sp.mask,i);
                        for(int j=0; j<MS; j++) {
                            if( (neigh_mask & j) == 0) sum_vs[j] |= sp.mask;
                        }
                    }

                    auto process = [&sum_vs, N=N, &mergeEntries, &i, &getNeighborhoodMask, &MS, &mask_nodes]
                        (auto & _dpvt) {

                        vector<SP4> sps;
                        sps.reserve( accumulate(ALL(_dpvt),0,[](int a, auto & b){return a + b.size();}) );

                        for(int j=0; j<MS; j++) {
                            HashType node2_hash = 0;
                            for(int k=0; k<mask_nodes[i].size(); k++) {
                                int d = mask_nodes[i][k];
                                if( j & (1<<k) )  node2_hash ^= node_hashes2[d];
                            }

                            for( auto & sp : _dpvt[j] ) {
                                auto m = sp.mask;
                                m &= sum_vs[j];
                                sp.hash = getHashForMask( m, N );
                                sp.hash ^= node2_hash;
                                sps.push_back(sp);
                            }
                            _dpvt[j].clear();
                        }

                        sort( ALL(sps) );
                        mergeEntries(sps);

                        for( auto & sp : sps ) {
                            sp.hash = getHashForMask( sp.mask, N );
                            auto neigh_mask = getNeighborhoodMask( sp.mask, i );
                            _dpvt[neigh_mask].push_back(sp);
                        }
                    };

                    process(_dpvt1);
                    process(_dpvt2);
                }

                sw.stop("merge_states_for_calculation");

                clog << "dpvs/vt1/vt2.size(), before merging: " << dpvs1[i].size() << ", "
                    <<  accumulate(ALL(dpvt1[i]), 0, [](int a, auto & b){ return a + b.size(); } ) << ", "
                    <<  accumulate(ALL(dpvt2[i]), 0, [](int a, auto & b){ return a + b.size(); } ) << endl
                    << "after merging: " << _dpvs.size() << ", "
                    << accumulate(ALL(_dpvt1), 0, [](int a, auto & b){ return a + b.size(); } ) << ", "
                    << accumulate(ALL(_dpvt2), 0, [](int a, auto & b){ return a + b.size(); } ) << endl;
            }

            if(calculate_using_iep) {
                // calculateUsingIEP(_dpvs, dpvt1[i], dpvt2[i], i);
                calculateUsingIEP(_dpvs, _dpvt1, _dpvt2, i);
            }else{
                // calculateUsingBitmasksMS(_dpvs, dpvt1[i], dpvt2[i], i);
                calculateUsingBitmasksMS(_dpvs, _dpvt1, _dpvt2, i);
            }
        }
    };

    auto isLastIteartion = [&]() {
        for( int i=0; i<N; i++ ) {
            auto dst = Solver2::findDistances(g, i).first;
            for( int j=0; j<N; j++ ) if( minl_vs[i] + dst[j] + minl_vt[j] <= L-2 ) return false;
        }
        return true;
    };


    while( continueIteration() ) {
        if( timer.tle("mim3") && !isLastIteartion() ) return VBIG();
        subsets_size++;

        clog << endl << endl << endl << endl << "S: " << subsets_size << endl;

        // now increase path length by 1
        swap(dpvs1, dpvs2);
        swap( dpvt1, dpvt2 );

        sw.start("creating new states");
        for( int i=0; i<N; i++ ) {
            dpvs1[i].clear();
            for( int j=0; j<MS; j++ ) dpvt1[i][j].clear();
        }

        for( int i=0; i<N; i++ ) { // create in dpvs1 and dpvt1 entries with new lengths

            for( auto & sp3 : dpvs2[i] ) {
                for( auto & arc : g[i] ) {
                    int d = arc.v;
                    if( !sp3.mask[d] && d != vs && d != vt ) {
                        dpvs1[d].emplace_back( sp3.mask, sp3.hash ^ node_hashes[d], sp3.counts * arc.lengths );
                        dpvs1[d].back().mask.set(d,1);
                        dpvs1[d].back().counts.trimToMax(L);
                    }
                }
            }

            for(int j=0; j<MS; j++) {
                for( auto & sp3 : dpvt2[i][j] ) {
                    for(auto & arc : revg[i]) {
                        int d = arc.v;
                        if( !sp3.mask[d] && d != vs && d != vt ) {
                            auto newsp = SP4(sp3.mask, sp3.hash ^ node_hashes[d], sp3.counts * arc.lengths);
                            newsp.mask.set(d,1);
                            newsp.counts.trimToMax(L);

                            auto neigh_mask = getNeighborhoodMask( newsp.mask, d );
                            dpvt1[d][neigh_mask].emplace_back( newsp );
                        }
                    }
                }
            }
        }
        sw.stop("creating new states");

        if constexpr (debug) {
            clog << "Created SP4s:" << endl;

            for(int i=0; i<N; i++) {
                if( dpvs1[i].empty() ) continue;
                clog << "dpvs1[" << i << "]: " << endl;
                for( auto & sp : dpvs1[i] ) clog << "\t" << sp.toString() << endl;
            }

            for(int i=0; i<N; i++) {
                if( dpvt1[i].empty() ) continue;
                clog << "dpvt1[" << i << "]: " << endl;
                for(int j=0; j<MS; j++)
                for( auto & sp : dpvt1[i][j] ) clog << "\t" << sp.toString() << endl;
            }
        }

        // if constexpr (debug)
            clog << "\tProceeding to merge entries and trim states" << endl;

        updateMinls();
        if( timer.tle("mim3") && !isLastIteration() ) return VBIG();

        if( !continueIteration() ) {
            clog << "MIM iterations are no longer required, finishing" << endl;
            break;
        }

        ULL total_sps = 0;
        for(int i=0; i<N; i++) {
            total_sps += dpvs1[i].size();
            for( int j=0; j<MS; j++ ) total_sps += dpvt1[i][j].size() + dpvt2[i][j].size();
        }
        clog << "\tBefore merging and trimming, total sps: " << total_sps << endl;

        mergeTrimExcludeFast();

        total_sps = 0;
        for(int i=0; i<N; i++) {
            total_sps += dpvs1[i].size();
            for( int j=0; j<MS; j++ ) total_sps += dpvt1[i][j].size() + dpvt2[i][j].size();
        }
        clog << "\tAfter merging and trimming, total sps: " << total_sps << endl;
        sps_counts.push_back(total_sps);


        sw.start("calculating result");
        calculateIterationResult();
        sw.stop("calculating result");
    }

    clog << "solver2.mim3 finished!" << endl;
    sw.writeAll();

    DEBUG(res_for_length);
    DEBUG(sps_counts);

    while( res_for_length.size() > L+1 ) res_for_length.pop_back();
    return res_for_length;
}

HashType Solver2MIM::getHashForMask(Mask m, int N) {
    HashType res = 0;
    for(int i=0; i<N; i++) if( m[i] ) res ^= node_hashes[i];
    return res;
}

bool Solver2MIM::isDAG(GraphSP2 &G, const VB &was) {
    int N = G.size();
    VB on_path(N);
    VB helper(N);

    function<bool(int)> hasCycle = [&](int num) {
        if( on_path[num] ) return true;
        on_path[num] = helper[num] = true;

        for( auto & arc : G[num] ) {
            int v = arc.v;
            if( !was[v] && !helper[v]) {
                if( hasCycle(v) ) return true;
            }
        }

        on_path[num] = false;
        return false;
    };

    for( int i=0; i<N; i++ ) if( !helper[i] ) if( hasCycle(i) ) return false;
    return true;
}

