//
// Created by sylwester on 4/15/23.
//

#ifndef ALGORITHMSPROJECT_GVVISUALIZER_H
#define ALGORITHMSPROJECT_GVVISUALIZER_H

#include "Makros.h"


class GVConfig{
public:

    string graph_name = "";

    /**
     * Color of node border and fill.
     */
    string node_color = "black";
    string node_fillcolor = "white";
    string node_shape = "ellipse"; // e.g. ellipse, box, diamond, square, Mdiamond, Msquare
    string node_style = "rounded"; // e.g. filled, diagonals, rounded

    /**
     * Color of edges.
     */
    string edge_color = "blue";
    string edge_style = "bold"; // e.g. solid, dashed, bold, dotted

    /**
     * If true, then a directed graph will be visualized.
     * Value of [edge_dir] will be set to "forward".
     */
    bool directed = true;

    /**
     * Direction of the dge.
     * forward - arrow forward
     * back - arrow back
     * both - arrow forward and back
     * none - no arrows
     */
    string edge_dir = "none";

};

/**
 * Class used to visualize graphs.
 *
 * Usually map<string,string> is passed to functions to define what is specified for nodes or edges.
 * What can be set, is due to specification of GraphViz DOT format - see GVConfig for some concepts.
 *
 * Files are opened using    "google-chrome file &" command
 */
class GVVisualizer{
public:

    using _format = map<string,string>;

    enum OutputFormat{
        PDF = 0,
        PNG = 1
    };

    /**
     * Extension of the file to create.
     * If PDF, then a .pdf with the graph will be created.
     * If PNG, then a .png file will be created.
     */
    OutputFormat out_format = PDF;


    /**
     * Adds a node with a specification given in [args].
     */
    void addNode( string name, _format specs );
    void addNode( int a, _format specs ){ addNode(to_string(a), specs); }

    /**
     * Adds an edge (name1,name2) to the graph, with specifications given in [specs].
     */
    void addEdge( string name1, string name2, _format specs );
    void addEdge( int a, int b, _format specs ){ addEdge( to_string(a), to_string(b), specs ); }

    /**
     * Marks all nodes in [nodes] to belong to the same subgraph.
     */
    void markSubgraph( vector<string> nodes, _format specs );
    void markSubgraph( VI nodes, _format specs );

    vector<pair<string,_format>> nodes;
    vector< tuple<string,string,_format> > edges;
    vector< pair<vector<string>, _format> > subgraphs;


    static _format default_node_specs;
    static _format default_edge_specs;
    static _format default_subgraph_specs;


    /**
     * Prepares all data to visualize a subgraph of [V] induced by nodes in [nds].
     * A fragment of [V] containing all edges incident to some node in [nodes] is created.
     * Also a fragment containing nodes remapped to a new graph is visualized besides.
     */
    void preprocessInducedGraph( VVI & V, VI nds, bool directed, bool add_induced_visualization = true,
                                _format node_specs = default_node_specs,
                                _format edge_specs = default_edge_specs,
                                _format subgraph_specs = default_subgraph_specs);

    /**
     * Preprocesses the graph to be fully visualized.
     */
    void preprocessFullGraph( VVI & V, bool directed, _format node_specs = default_node_specs,
                              _format edge_specs = default_edge_specs,
                              _format subgraph_specs = default_subgraph_specs);




    /**
     * Visualizes a graph given in [V].
     * Uses GraphViz to draw the graph.
     * Creates a .pdf with the graph visualized.
     *
     * @param V graph to visualize
     * @param cnf configuration object
     * @param file_name name of the file containing the graph. There will be a .gv and .pdf file
     * @param open_when_built if true, then the pdf with the graph is opened after creation
     */
    static void visualizeBasic( VVI & V, GVConfig cnf,
                           string file_name = "temp_graph",
                           bool open_when_built = true );




    /**
     * Visualizes graph with node specs given in [nodes], edges and their specs given in [edges] and
     * marked subgraphs given in [subgraphs].
     * #CAUTION! The size of the graph is set to be limited by 7x10 inches (to fit into A4).
     */
    void visualize( string file_name, bool open_when_built = true );

    /**
    * Function is used to modify specs in [nodes].
    * [fun] must accept two parameters (string, _format) representing (node_name, specs).
    * Typically names are integers given as a string.
    */
    template<class F1>
    void modifyNodeSpecs(F1 fun){ for( auto & [name,specs] : nodes ) fun( name,specs ); }

    /**
     * Function is used to modify specs in [edges].
     * [fun] must accept two parameters (string, string, _format) representing (name1, name2, specs).
     * Typically names are integers given as a string.
     */
    template<class F1>
    void modifyEdgeSpecs(F1 fun){ for( auto & [name1, name2, specs] : edges ) fun( name1, name2, specs ); }

    /**
     * Function is used to modify specs in [subgraphs].
     * [fun] must accept two parameters (vector<string>,_format) representing (nodes_in_subgraph, specs).
     * Typically node names are integers given as strings.
     */
    template<class F1>
    void modifySubgraphSpecs(F1 fun){ for( auto & [nds,specs] : subgraphs ) fun(nds, specs); }

private:



};

#endif //ALGORITHMSPROJECT_GVVISUALIZER_H
