//
// Created by sylwester on 8/8/19.
//

#include "Makros.h"
#include <GraphUtils.h>
#include <ConnectedComponents.h>

// #include "combinatorics/CombinatoricUtils.h"
#include <StandardUtils.h>
#include "BridgesAndArtPoints.h"

VVI GraphUtils::getUndirectedStructure( VVI & V, VI S ) {
    int N = V.size();
    VB was(N);
    for(int d : S) was[d] = true;
    auto revV = reverseGraph(V);

    VVI res(N);
    for(int i=0; i<N; i++) {
        if( was[i] ) continue;
        for(int d : V[i]) if(!was[d]){ res[i].push_back(d); res[d].push_back(i); }
    }

    for(int i=0; i<N; i++) StandardUtils::makeUnique(res[i]);

    return res;
}

// checks if graph G \ S, where set S is marked by set bits in [was] is a DAG
bool GraphUtils::isDAG(VVI & V, const VB &was) {
    int N = V.size();
    VB on_path(N);
    VB helper(N);

    function<bool(int)> hasCycle = [&](int num) {
        if( on_path[num] ) return true;
        on_path[num] = helper[num] = true;

        for( int v : V[num] ) {
            if( !was[v] && !helper[v]) {
                if( hasCycle(v) ) return true;
            }
        }

        on_path[num] = false;
        return false;
    };

    for( int i=0; i<N; i++ ) if( !helper[i] ) if( hasCycle(i) ) return false;
    return true;
}

void GraphUtils::writeInstanceToFile(ofstream & str, VVI & V, int vs, int vt, int L, bool directed) {
    auto arcs = getGraphEdges(V,directed);
    str << "p edge " << V.size() << " " << arcs.size() << "\n";
    for(auto [a,b] : arcs) str << (directed ? "a" : "e") << " " << a+1 << " " << b+1 << "\n";
    str << "l " << L << "\n";
    str << "t " << vs+1 << " " << vt+1 << "\n";
    str << flush;
}


VPII GraphUtils::getGraphEdges( VVI & V, bool directed ){
    if(directed) return getDirectedGraphEdges(V);

    VPII res;
    res.reserve( countEdges(V) );
    for( int i=0; i<V.size(); i++ ){
        for( int d : V[i] ){
            if(d>i) res.push_back( {i,d} );
        }
    }
    return res;
}

VPII GraphUtils::getDirectedGraphEdges( VVI & V ){
    VPII res;
    res.reserve( countEdges(V) );
    for( int i=0; i<V.size(); i++ ){
        for( int d : V[i] ){
            res.push_back( {i,d} );
        }
    }
    return res;
}

VI GraphUtils::getNeighborhood(VVI &V, VI &S, bool useSet) {
    VI res;
    if( useSet ){
        unordered_set<int> zb;
        for( int p : S ){
            for(int d : V[p]){
                zb.insert(d);
            }
        }
        res = VI(ALL(zb));
    }else{
        VB was(V.size(),false);
        for( int p : S ){
            for(int d : V[p]){
                if( !was[d] ){
                    was[d] = true;
                    res.push_back(d);
                }
            }
        }
    }
    return res;
}

VI GraphUtils::getNeighborhoodOfAInB(VVI &V, VI &A, VI &B, bool useSet) {
    unordered_set<int> inBSet;
    VB inB;
    if( useSet ){
        inBSet.insert( ALL(B) );
    }else{
        inB = VB(V.size(),false);
        for(int d : B) inB[d] = true;
    }

    function< bool (int d) > inBFunction = [&inBSet, &inB, &useSet]( int d ) -> bool {
        if( useSet ) return (inBSet.count(d) > 0);
        else return inB[d];
    };

    VI res;
    if( useSet ){
        unordered_set<int> zb;
        for( int p : A ){
            for(int d : V[p]){
                if( inBFunction(d) ) zb.insert(d);
            }
        }
        res = VI(ALL(zb));
    }else{
        VB was(V.size(),false);
        for( int p : A ){
            for(int d : V[p]){
                if( inBFunction(d) && !was[d] ){
                    was[d] = true;
                    res.push_back(d);
                }
            }
        }
    }
    return res;
}


VVI GraphUtils::transposeGraph(VVI &v) {
    VVI g( SIZE(v) );
    REP( i,SIZE(v) )	REP( k,SIZE(v[i]) )	g[ v[i][k] ].PB(i);
    return g;
}


void GraphUtils::addEdge(VVI &V, int a, int b, bool directed) {
    V[a].push_back(b);
    if( !directed ) V[b].push_back(a);
}

bool GraphUtils::addEdgeSafe(VVI &V, int a, int b, bool directed) {
    auto exists = StandardUtils::find(V[a],b);
    if(!exists) V[a].push_back(b);

    bool x = false;
    if( !directed ) x = addEdgeSafe( V, b,a,true );

    return (!exists || x);
}

void GraphUtils::removeEdge(VVI &V, int a, int b, bool directed) {
    auto rem = [ &V ](int a, int b){
        for( int i=(int)V[a].size()-1; i>=0; i-- ){
            if( V[a][i] == b ){
                swap(V[a][i], V[a].back() );
                V[a].pop_back();
                break;
            }
        }
    };


    rem(a,b);
    if(!directed) rem(b,a);

}

void GraphUtils::removeNeighborsOfNode(VVI &V, int a, unordered_set<int> neighbors) {
    for( int i=(int)V[a].size()-1; i>=0; i-- ){
        if( neighbors.count( V[a][i] ) > 0 ){
            swap(V[a][i], V[a].back() );
            V[a].pop_back();
        }
    }
}

void GraphUtils::removeNodeFromGraph(VVI &V, int a) {
    for( int d : V[a] ) removeEdge( V,d,a,true );
    V[a].clear();
}

int GraphUtils::countNodesWithDegree(VVI &V, int d, int D) {
    int res = 0;
    for( auto & v : V ){
        if(v.size() >= d && v.size() <= D) res++;
    }
    return res;
}

bool GraphUtils::containsEdge(VVI &V, int a, int b) {
    return find( ALL(V[a]),b ) != V[a].end();
}

int GraphUtils::countEdges(VVI &V, const bool directed) {
    int res = 0;
    for(auto& v : V) res += v.size();
    if(!directed) return res >> 1;
    else return res;
}

VVI GraphUtils::sortNodeNeighborhoods( VVI & V ){
    VVI V2(V.size());
    for(int i=0; i<V.size(); i++) V2[i].reserve(V[i].size());
    for( int i=0; i<V.size(); i++ ){
        for( int d : V[i] ) V2[d].push_back(i);
    }
    return V2;
}


// VVI GraphUtils::getComplimentaryGraph(VVI &V) {
//     VVI V2(V.size());
//     for( int i=0; i<V.size(); i++ ){
//         VI tmp = V[i];
//         tmp.push_back(i);
//         sort(ALL(tmp));
//         V2[i] = getComplimentaryNodes( V,tmp );
//     }
//
//     bool correct = true;
//     for( int i=0; i<V.size(); i++ ){
//         VB was(V.size(),false);
//         for( int d : V[i] ) was[d] = true;
//         for(int d : V2[i]) was[d] = true;
//         if( V[i].size() + V2[i].size() != V.size() -1 ) correct = false;
//         for(int k=0; k<V.size(); k++) if( k != i && !was[k]) correct = false;
//     }
//
//     assert(correct);
//
//     return V2;
//
//
// //    VVI V2 = V;
// //    return makeCompliment(V2, CombinatoricUtils::getRandomPermutation(V.size()) );
// }

void GraphUtils::removeNodesFromGraph(VVI &V, VI nodes) {

    unordered_set<int> nd(ALL(nodes));
    VI neigh = getNeighborhood( V,nodes,true );
    for( int t : neigh ){
        if( nd.count(t) ) continue;

        for( int i = (int)V[t].size()-1; i>=0; i-- ){
            if( nd.count(V[t][i]) ){
                swap( V[t][i], V[t].back() );
                V[t].pop_back();
            }
        }
    }

    for( int d : nodes ) V[d].clear();
}

void GraphUtils::removeNodesFromGraph(VVI &V, VVI &W, VI nodes) {
    unordered_set<int> nd(ALL(nodes));
    VI neigh = getNeighborhood( V,nodes,true );
    for( int t : neigh ){
        if( nd.count(t) ) continue;

        for( int i = (int)V[t].size()-1; i>=0; i-- ){
            if( nd.count(V[t][i]) ){
                swap( V[t][i], V[t].back() );
                V[t].pop_back();

                swap( W[t][i], W[t].back() );
                W[t].pop_back();
            }
        }
    }

    for( int d : nodes ){
        V[d].clear();
        W[d].clear();
    }
}

void GraphUtils::writeGraphHumanReadable(VVI &V) {
    clog << "****" << endl;
    for( int i=0; i<V.size(); i++ ){
        clog << i << ": ";
        VI neigh = V[i];
        sort(ALL(neigh));
        for(int d : neigh) clog << d << "  ";
        clog << endl;
    }
}

void GraphUtils::mergeNodeToNode(VVI &V, int a, int b) {
//    assert( containsEdge( V,a,a ) == false );
//    assert( containsEdge( V,b,b ) == false );

    VI Na = V[a];
    Na.push_back(a);

    VI Nb = V[b];
    Nb.push_back(b);


    sort(ALL(Na));
    sort(ALL(Nb));

//    DEBUG(VI({a,b}));
//    DEBUG(Na);
//    DEBUG(Nb);

    VI diff;
    set_difference( ALL(Na), ALL(Nb), back_inserter(diff) ); // diff is Na \ Nb

//    DEBUG(diff);

    removeNodeFromGraph(V,a);
    for( int d : diff ){

//        assert( containsEdge( V,d,b ) == false );
        V[d].push_back(b);

//        assert( containsEdge( V,b,d ) == false );
        V[b].push_back(d);
    }

//    cerr << "GraphUtils::mergeNodeToNode not tested yet" << endl;
//    exit(1);
}

void GraphUtils::contractEdge(VVI &V, int a, int b) {
    mergeNodeToNode(V,a,b);
}

#ifndef QUICK_BUILD
void GraphUtils::writeBasicGraphStatistics(VVI &V) {
    clog << "V.size(): " << V.size() << endl;
    clog << "|E(V)|: " << countEdges(V) << endl;
    clog << "average degree: " << ( (double)2 * countEdges(V) ) / V.size() << endl;
    clog << "largest degree: " << max_element( ALL(V), []( VI& v, VI& w ){ return v.size() < w.size(); } )->size() << endl;
    int m = min( (int)V.size()-1,3 );

    VVI comps = ConnectedComponents::getConnectedComponents(V);
    clog << "There are " << comps.size() << " connected components" << endl;
    sort( ALL(comps), [](VI& v1, VI& v2){ return v1.size() > v2.size(); } );
    m = min( (int)comps.size(),7 );
    clog << "first " << m << " largest component sizes: "; for( int i=0; i<m; i++ ) clog << comps[i].size() << " "; clog << endl;
    clog << "first " << m << " smallest component sizes: "; for( int i=(int)comps.size()-m; i< comps.size(); i++ ) clog << comps[i].size() << " "; clog << endl;

    m = 10;
    VI degs(V.size(),0);
    for( VI& v : V ) degs[v.size()]++;
    for(int i=0; i<=m; i++) clog << "nodes with deg " << i << ": " << degs[i] << endl;
    clog << endl;

//    cerr << "isTree: " << Tree::isTree(V) << endl;

    auto artsAndBridges = BridgesAndArtPoints::getBridgesAndArtPoints(V);
    clog << "There are " << artsAndBridges.first.size() << " articulation points and " << artsAndBridges.second.size() << " bridges" << endl;
    clog << endl;


//    { // clique size by greedy extension
//        ENDL(1);
//        if(V.size() < 3'000){
//            VVI compV = getComplimentaryGraph(V);
//            VI vc = Utils::getUpperBoundByVCOnSuperPIGraph(compV, 3000);
//            VI compVmis = CombinatoricUtils::getFullSetDifference( V.size(), vc );
//            VI clq = compVmis;
//            clog << "maximal clique found using FastVC on complimentary graph: " << clq.size() << endl;
//        }else{
//            VI clq = CliqueExtension::findMaximalNodeCliqueExtension(V, true);
//            clog << "maximal clique found via sparse greedy extension: " << clq.size() << endl;
//        }
//    }

//    { // VC and MIS sizes
//        ENDL(1);
//        VI vc = Utils::getUpperBoundByVCOnSuperPIGraph(V, 3000);
//        clog << "Estimated VC size (run FastVC for 3s): " << vc.size() << "  ->  mis: " << V.size() - vc.size() << endl;
//    }

//    { // treewidth
//        TREEWIDTH trw;
//        int max_cnt = 8;
//        volatile sig_atomic_t tle = 0;
//        TreewidthDecomposition decomp = trw.main( V, max_cnt, tle );
//        ENDL(1);
//        clog << "Found treewidth decomposition of size: " << decomp.getWidth() << endl;
//    }
//
//    { // separator
//        ENDL(5);
////
//        VVI comps = ConnectedComponents::getConnectedComponents(V);
//        VI cmp = *max_element(ALL(comps), [](auto & a, auto & b){ return a.size() < b.size(); });
//        InducedGraph g = GraphInducer::induce(V,cmp);
//        DEBUG(g.V.size());
//        DEBUG(GraphUtils::countEdges(g.V));
//
//        FlowCutter fc( SeparatorEvaluators::sepEvalToUse );
//        fc.setBalanceParameter(5);
//        vector<Separator> separators = fc.createSeparators(g.V, 5);
//        clog << "Separators found using 5 iterations of FlowCutter:" << endl;
//        for(auto sep : separators){
//            DEBUG(sep.stats);
//            InducedGraph g = GraphInducer::induce(V, sep.nodes);
//            clog << "\tGraph induced by sep has " << countEdges(g.V) << " edges" << endl;
//        }
//
////        ComponentExpansionSeparatorCreator exp(SeparatorEvaluators::sepEvalToUse);
////        vector<Separator> separators = exp.createSeparators(g.V,5);
////        clog << "Separators found using 5 iterations of ComponentExpansionSeparatorCreator:" << endl;
////        for(auto sep : separators) DEBUG(sep.stats);
//    }

}
#endif // QUICKBUILD

void GraphUtils::removeEdges(VVI &V, VPII &edges, bool directed) {

//    cerr << "removeEdges not tested yet" << endl;

    if( directed == false ){
        int E = edges.size();
        for( int i=0; i<E; i++ ) edges.emplace_back( edges[i].second, edges[i].first ); // adding reverse edges to remove
    }

    sort( ALL(edges) );

    for( int i=0; i<edges.size(); i++ ){
        int p = i;
        unordered_set<int> toRemove;
        while( p < edges.size() && edges[p].first == edges[i].first ){
            toRemove.insert( edges[p].second );
            p++;
        }

        int t = edges[i].first;
        for( int k=(int)V[t].size()-1; k>=0; k-- ){
            int d = V[t][k];
            if( toRemove.count(d) ){
//                cerr << "Removing edge " << t << " -> " << V[t][k] << endl;
                swap( V[t][k], V[t].back() );
                V[t].pop_back();
            }
        }

        i = p-1;
    }

}

bool GraphUtils::isConnected(VVI &V) {
//    cerr << "isConnected not tested yet!" << endl; exit(1);
    VB was(V.size(),false);
    int cnt = 0;
    function< void(int) > dfs = [&V,&was, &dfs, &cnt](int num){
        was[num] = true;
        cnt++;
        for( int d : V[num] ) if( !was[d] ) dfs(d);
    };
    dfs(0);
    return (cnt == V.size());
}

VI GraphUtils::getNeighborhoodExclusive(VVI &V, VI &A, VB &helper) {
    VI res;
    for( int a : A ) helper[a] = true;
    for( int a : A ){
        for( int w : V[a] ){
            if(!helper[w]){
                res.push_back(w);
                helper[w] = true;
            }
        }
    }

    for( int a : A ) helper[a] = false;
    for(int d : res) helper[d] = false;
    return res;
}

int GraphUtils::regularity(VVI V) {
    int k = V[0].size();
    for(int i=1; i<V.size(); i++) if( V[i].size() != k ) return -1;
    return k;
}

VVI GraphUtils::makeSimple(VVI V) {
    for( int i=0; i<V.size(); i++ ){
        sort(ALL(V[i]));
        V[i].resize(unique(ALL(V[i])) - V[i].begin()); // removing parallel edges

        for( int j=(int)V[i].size()-1; j>=0; j--){ // removing loops
            if( V[i][j] == i ){
                swap(V[i][j], V[i].back());
                V[i].pop_back();
            }
        }
    }
    return V;
}

VVI GraphUtils::getGraphForEdges(VPII edges, bool directed) {
    int N = 0;
    for(auto & [a,b] : edges) N = max(N, max(a,b));
    VVI V(N+1);
    for( auto & [a,b] : edges ) addEdge(V,a,b,directed);
    return V;
}

VVI GraphUtils::remapGraph(VVI V, VI perm) {
    int N = V.size();
    VVI H(N);
    for( int i=0; i<N; i++ ){
        H[perm[i]].reserve(V[i].size());
        for(int d : V[i]) H[perm[i]].push_back(perm[d]);
    }

    return H;
}

bool GraphUtils::isDirected(VVI V) {
    set<PII> zb;
    for( int i=0; i<V.size(); i++ ) for( int d : V[i] ) zb.insert( {i,d} );
    for( int i=0; i<V.size(); i++ ) for( int d : V[i] ) if( !zb.contains({d,i}) ) return true;
    return false;
}

bool GraphUtils::isSimple(VVI V) {
    int N = V.size();
     VB helper(N,false);

     for( int i=0; i<N; i++ ){
         for( int d : V[i] ){
             if( d == i || helper[d] ) return false;
             helper[d] = true;
         }

         for(int d : V[i]) helper[d] = false;
     }

     return true;
}

double GraphUtils::density(VVI &V, bool directed) {
    LL denom = V.size();
    denom *= denom-1;
    denom >>= 1;

    double density = countEdges(V, directed);
    return density / denom;
}


