//
// Created by sylwe on 12.07.2024.
//


#include "Kernelizer.h"

#include <BridgesAndArtPoints.h>
#include <ConnectedComponents.h>
#include <GraphUtils.h>
#include <GVVisualizer.h>
#include <StandardUtils.h>

#include "Solver.h"
#include "../external/GraphInducer.h"
#include "../external/StronglyConnectedComponents.h"
#include "../external/GraphUtils.h"
#include "CollectionOperators.h"
#include "Solver2.h"

ostream& operator<<(ostream & str, Core & c) {
    str << "V: " << c.V << ", vs: " << c.vs << ", vt: " << c.vt << ", L:" << c.L;
    return str;
}

ostream& operator<<(ostream & str, Core2 & c) {
    str << "G: " << c.G << ", vs: " << c.vs << ", vt: " << c.vt << ", L:" << c.L;
    return str;
}


Kernelizer::Kernelizer(VVI &_V, VVI &_revV, int _L, int _vs, int _vt) : V(_V), revV(_revV), L(_L), N(V.size()),
vs(_vs), vt(_vt){
    helper.resize(N);
    was.resize(N);
}

Core Kernelizer::createCoreGraph() {

    bool changes_done = true;
    int iters_cnt = 0;

    auto getCore = [&]() {
        VI nodes;
        for(int i=0; i<N; i++) {
            if( V[i].empty() && revV[i].empty() ) continue;
            nodes.push_back(i);
        }

        // clog << "Inducing by nodes: " << nodes << endl;

        InducedGraph g = GraphInducer::induceDirected(V, revV, nodes);
        Core core;
        core.V = g.V;
        core.revV.resize(g.V.size());
        for( int i=0; i<g.V.size(); i++ ) for(int d : g.V[i]) core.revV[d].push_back(i);
        core.vs = g.perm[vs];
        core.vt = g.perm[vt];
        core.L = L;

        return core;
    };

    auto applyCore = [&](auto & core) {
        V = core.V;
        revV = core.revV;
        vs = core.vs;
        vt = core.vt;
        N = V.size();
        L = min(core.L, N-1);
    };

    constexpr bool apply_core_in_each_iteration = true;


    bool is_init_symmetric = !GraphUtils::isDirected(V);

    while(changes_done) {
        // clog << "Starting reduction iteration #" << ++iters_cnt << endl;
        changes_done = false;
        changes_done |= trimLeaves();
        changes_done |= removeDistantNodes();
        changes_done |= removeDispensibleSCCs();
        changes_done |= removeDispensibleBiconnectedComponents();

        if constexpr (apply_core_in_each_iteration) {
            auto core = getCore(); // this can be used to reudce size of the graph before time-consuming dominated-nodes reduction
            applyCore(core);
            if(V.empty()) return core;
        }

        if( N < 20'000 ) { // reductions with complexity O(N*M)
            changes_done |= removeDistantNodesV2Dominators();
            changes_done |= removeDominatedNodes(); // this is probably useless if removeDistantNodesV2Dominators is used...
        }

        if constexpr (apply_core_in_each_iteration) {
            auto core = getCore(); // this can be used to reudce size of the graph before time-consuming dominated-nodes reduction
            applyCore(core);
            if(V.empty()) return core;
        }

        if( GraphUtils::countEdges( V, true ) < 50'000 ) {
            if( !is_init_symmetric || cnf.use_all_reduction_rules_even_for_undirected_graphs ) {
                changes_done |= removeDistantArcsV2Dominators();
            }
        }

        iters_cnt++;
    }

    if( is_init_symmetric && !cnf.use_all_reduction_rules_even_for_undirected_graphs ) {
        assert( !GraphUtils::isDirected(V) );
    }


    auto core = getCore();

    constexpr bool write_kernel = true;
    if constexpr (write_kernel){
        ofstream f( "kernel1.col" );
        GraphUtils::writeInstanceToFile( f, core.V, core.vs, core.vt, core.L, directed_track );
        f.close();
    }

    return core;
}

bool Kernelizer::trimLeaves() {
    if( GraphUtils::countEdges(V, true) == 0 ) return false;

    VI nodes(N);
    iota(ALL(nodes),0);

    int leaves_trimmed = 0;

    for( int i=0; i<nodes.size(); i++ ) {
        int v = nodes[i];
        if( v == vs || v == vt ) continue;

        int cnt = 0;
        for(int d : V[v]) {
            if(!was[d]) {
                was[d] = true;
                cnt++;
            }
        }
        for(int d : revV[v]) {
            if(!was[d]) {
                was[d] = true;
                cnt++;
            }
        }


        if( cnt == 1 ) {
            for(int d : V[v]) was[d] = false;
            for(int d : revV[v]) was[d] = false;

            for(int d : V[v]) nodes.push_back(d);
            for(int d : revV[v]) nodes.push_back(d);
            // clog << "\tTrimming leaf " << v << endl;

            leaves_trimmed++;
            clearNode(V, revV, v);
        }

        for(int d : V[v]) was[d] = false;
        for(int d : revV[v]) was[d] = false;

    }

    clog << "Leaves trimmed: " << leaves_trimmed << endl;

    return leaves_trimmed > 0;
}

bool Kernelizer::removeDispensibleSCCs() {
    constexpr bool debug = false && kernelizer_debug;

    if( GraphUtils::countEdges(V, true) == 0 ) return false;

    StronglyConnectedComponents scc_creator(V);
    scc_creator.createStronglyConnectedComponents();
    auto sccs = scc_creator.getComponents();
    auto g = scc_creator.getCondensationGraph();

    // DEBUG(sccs);
    // DEBUG(g);

    int vs_scc = -1, vt_scc = -1;
    for( int i=0; i<sccs.size(); i++ ) {
        for( int d : sccs[i] ) {
            if( d == vs ) vs_scc = i;
            if( d == vt ) vt_scc = i;
        }
    }

    VVI revG = GraphUtils::reverseGraph(g);

    auto [dst_vs, par_vs] = Solver::findDistances(g,vs_scc);
    auto [dst_vt, par_vt] = Solver::findDistances(revG,vt_scc);

    // DEBUG(dst_vs);
    // DEBUG(dst_vt);

    VB to_remove(g.size(), false);
    for( int i=0; i<g.size(); i++ ) if(dst_vs[i] == inf || dst_vt[i] == inf) to_remove[i] = true;


    int nodes_removed = 0;
    for(int i=0; i<g.size(); i++) {
        if( to_remove[i] ) {
            for(int v : sccs[i]) {
                if(!( V[v].empty() && revV[v].empty() )) {
                    nodes_removed++;
                    clearNode(V, revV, v);
                    if constexpr (debug) clog << "From SCC: " << sccs[i] <<" removing node " << v << endl;
                }
            }
        }
    }

    clog << "Removed " << nodes_removed << " nodes using removeDispensibleSCCs reduction" << endl;

    return nodes_removed > 0;
}

bool Kernelizer::removeDistantNodesV2Dominators() {
    constexpr bool debug = false && kernelizer_debug;
    if( GraphUtils::countEdges(V,true) == 0 ) return false;

    VI to_remove;
    for( int v=0; v<N; v++ ) {
        if( v == vs || v == vt ) continue;
        if( V[v].empty() && revV[v].empty() ) continue;

        auto B = findDominators(revV, vt,v);
        for(int b : B) was[b] = true;
        auto [dst_vs, par_vs] = Solver::findDistances(V, vs, was);
        for(int b : B) was[b] = false;

        auto A = findDominators(V, vs,v);
        for(int a : A) was[a] = true;
        auto [dst_vt, par_vt] = Solver::findDistances(revV, vt, was);
        for(int a : A) was[a] = false;

        if( dst_vs[v] + dst_vt[v] > L ) {
            to_remove.push_back(v);
            if constexpr (debug) {
                clog << "Removing distant node (dominator version) v: " << v << endl;
                DEBUG(A);
                DEBUG(B);
                clog << "dst_vs: " << dst_vs[v] << endl;
                clog << "dst_vt: " << dst_vt[v] << endl;
                clog << endl;
            }
        }
    }

    for(int v : to_remove) clearNode(V, revV, v);
    clog << "Distant nodes (dominator version) removed: " << to_remove.size() << endl;

    return !to_remove.empty();
}

bool Kernelizer::removeDistantArcsV2Dominators() {
    constexpr bool debug = false;
    if( GraphUtils::countEdges(V,true) == 0 ) return false;

    VPII to_remove;

    for( int v=0; v<N; v++ ) {
        if( V[v].empty() && revV[v].empty() ) continue;

        for( int d : V[v] ) {

            helper[v] = true; // excluding begin of the arc (v,d) from dominator search
            auto B = findDominators(revV, helper,vt,d);
            fill(ALL(helper),false);
            for(int b : B) was[b] = true;
            was[d] = true; // node d cannot be on path vs->v
            auto [dst_vs, par_vs] = Solver::findDistances(V, vs, was);
            for(int b : B) was[b] = false;
            was[d] = false;

            helper[d] = true; // excluding end of the arc (v,d) from dominator search
            auto A = findDominators(V, helper, vs,v);
            fill(ALL(helper),false);
            for(int a : A) was[a] = true;
            was[v] = true; // node v cannot be on path d->vt (that is vt->d in revV)
            auto [dst_vt, par_vt] = Solver::findDistances(revV, vt, was);
            for(int a : A) was[a] = false;
            was[v] = false;

            if( dst_vs[v] + dst_vt[d] > L ) {
                to_remove.emplace_back( v,d );
                if constexpr (debug) {
                    clog << "Removing distant arc (dominator version) " << v << " -> " << d << endl;
                    DEBUG(A);  DEBUG(B);
                    clog << "dst_vs: " << dst_vs[v] << endl;
                    clog << "dst_vt: " << dst_vt[d] << endl;
                    clog << endl;

                    // if( v == 30 && d == 74 ) {
                    //     DEBUG(V[30]);
                    //     DEBUG(revV[30]);
                    //     DEBUG(V[74]);
                    //     DEBUG(revV[74]);
                    // }
                }
            }
        }
    }

    for(auto [v,d] : to_remove) removeArc( V, revV, v,d );
    clog << "Distant arcs (dominator version) removed: " << to_remove.size() << endl;

    return !to_remove.empty();
}


bool Kernelizer::removeDistantNodes() {
    constexpr bool debug = false; // && kernelizer_debug;
    if( GraphUtils::countEdges(V,true) == 0 ) return false;

    auto [dst_vs, par_vs] = Solver::findDistances(V, vs);
    auto [dst_vt, par_vt] = Solver::findDistances(revV, vt);

    VI to_remove;
    for(int i=0; i<N; i++) {
        if( ( !V[i].empty() || !revV[i].empty() ) && dst_vs[i] + dst_vt[i] > L ) {
            to_remove.push_back(i);
            if constexpr (debug) clog << "Removing distant node: " << i << endl;
        }
    }

    for(int v : to_remove) clearNode(V, revV, v);

    clog << "Distant nodes removed: " << to_remove.size() << endl;

    return !to_remove.empty();
}

bool Kernelizer::removeDispensibleBiconnectedComponents() {
    constexpr bool debug = false; // && kernelizer_debug;
    if( GraphUtils::countEdges(V,true) == 0 ) return false;

    VVI underV = V;
    for( int i=0; i<N; i++ ) {
        underV[i] = V[i] + revV[i];
        sort(ALL(underV[i]));
        underV[i].resize( unique(ALL(underV[i])) - underV[i].begin() );
    }

    if constexpr (debug) DEBUG(underV);

    auto bicomps = ConnectedComponents::getBiconnectedComponents(underV);
    auto path = Solver::findShortestPath(underV, vs, vt, Mask::zeros(N));
    if constexpr (debug) DEBUG(bicomps);
    if constexpr (debug) DEBUG(path);

    if( path.empty() ) return false;

    // auto [art_points,bridges] = BridgesAndArtPoints::getBridgesAndArtPoints(V);
    // auto is_art_point = StandardUtils::toVB(N, art_points);


    for(int d : path) was[d] = true;

    VI steps_in_bicomp(bicomps.size(),0);

    VB to_remove(bicomps.size(),false);
    for( int i=0; i<bicomps.size(); i++ ) {
        for(int d : bicomps[i]) {
            if( was[d] ) steps_in_bicomp[i]++;
        }
    }

    VB nodes_to_retain(N);
    for( int i=0; i<bicomps.size(); i++ ) if(steps_in_bicomp[i] >= 2) for(auto d : bicomps[i]) nodes_to_retain[d] = true;

    for( int i=0; i<bicomps.size(); i++ ) {
        if constexpr (debug) clog << "Steps in bicomp: " << bicomps[i] << ": " << steps_in_bicomp[i] << endl;
        if( bicomps[i].size() > 1 && steps_in_bicomp[i] <= 1 ) to_remove[i] = true;
        else if( bicomps[i].size() == 1 && steps_in_bicomp[i] == 0 && bicomps[i][0] != vs && bicomps[i][0] != vt) to_remove[i] = true;
    }

    int nodes_removed = 0;
    for(int i=0; i<bicomps.size(); i++) {
        if( to_remove[i] ) {
            if constexpr (debug) clog << "From biconnected component " << bicomps[i] << endl;
            for(int v : bicomps[i]){
                if( bicomps[i].size() == 1 && underV[v].empty() ) continue;
                // if( !was[v] && !is_art_point[v] ) {
                if( !was[v] && !nodes_to_retain[v] ) {

                    clearNode(V, revV, v);
                    nodes_removed++;
                    if constexpr (debug) clog << "\tremoving node " << v << endl;
                }
            }
        }
    }

    fill(ALL(was),false);

    clog << "Removed " << nodes_removed << " nodes using removeDispensibleBiconnectedComponents reduction" << endl;

    return nodes_removed > 0;
}

bool Kernelizer::removeDominatedNodes() {
    if( GraphUtils::countEdges(V,true) == 0 ) return false;

    constexpr bool debug = false && kernelizer_debug;

    VI to_remove;
    for( int i=0; i<N; i++ ) {
        if( i == vs || i == vt || (V[i].empty() && revV[i].empty()) ) continue;
        // if(debug) clog << endl << endl << "Considering node " << i << endl;
        VI dom = findDominators(V, vs,i);
        VI dom_rev = findDominators(V, i,vt);
        // if(debug) DEBUG(dom);
        // if(debug) DEBUG(dom_rev);

        bool has_common_dominator = false;
        for(int d : dom) was[d] = true;
        for(int d : dom_rev) if(was[d]) has_common_dominator = true;
        for(int d : dom) was[d] = false;
        if(has_common_dominator) to_remove.push_back(i);
    }

    if(debug) clog << "removeDominatedNodes, to_remove: " << to_remove << endl;
    for(int d : to_remove) clearNode(V, revV, d);

    clog << "Dominated nodes removed: " << to_remove.size() << endl;

    return to_remove.size() > 0;
}


void Kernelizer::removeArc(VVI & V, VVI & revV, int a, int b) {
    for( int i=(int)V[a].size()-1; i >=0; i-- ) {
        if( V[a][i] == b ) {
            swap( V[a][i], V[a].back() );
            V[a].pop_back();
        }
    }

    for( int i=(int)revV[b].size()-1; i >=0; i-- ) {
        if( revV[b][i] == a ) {
            swap( revV[b][i], revV[b].back() );
            revV[b].pop_back();
        }
    }
}

void Kernelizer::removeArc(VVI &V, int a, int b) {
    for( int i=(int)V[a].size()-1; i>=0; i-- ) {
        if( V[a][i] == b ) {
            swap(V[a][i], V[a].back() );
            V[a].pop_back();
        }
    }
}

void Kernelizer::clearNode(VVI &V, VVI &revV, int v) {
    VPII to_remove;
    for( int b : V[v] ) {
        to_remove.emplace_back(v,b);
        to_remove.emplace_back(b,v);
    }
    for( int b : revV[v] ) {
        to_remove.emplace_back(v,b);
        to_remove.emplace_back(b,v);
    }

    for( auto [a,b] : to_remove ) removeArc(V,revV, a,b);
}

VI Kernelizer::findDominators(VVI &V, int vs, int vt) {
    VB was(V.size());
    return findDominators(V,was,vs,vt);

    // constexpr bool debug = false && kernelizer_debug;
    // int N = V.size();
    //
    // if(debug) clog << "Seeking dominators, beg: " << vs << ", end: " << vt << endl;
    // // DEBUG(V);
    //
    // VB was(N);
    // auto [dst,par] = Solver::findDistances(V, vs);
    //
    // auto path = Solver::findShortestPath( V, vs, vt, Mask::zeros );
    // if( path.empty() ) return {}; // no path, so no dominators...
    //
    // if(debug) DEBUG(path);
    //
    //
    // VI ind_on_path(N,-1);
    // for(int i=0; i<path.size(); i++) ind_on_path[ path[i] ] = i;
    // for( int d : path ) was[d] = true;
    //
    // VI reach(N,-1);
    // function<int(int)> dfs = [&](int num) {
    //     was[num] = true;
    //     int res = -1;
    //     if(debug) clog << "\tnum: " << num << endl;
    //     for( int d : V[num] ) {
    //         res = max(res, ind_on_path[d]);
    //         if(!was[d]) res = max(res, dfs(d));
    //     }
    //
    //     reach[num] = max(reach[num],res);
    //     return reach[num];
    // };
    //
    // VI dominators;
    // for( int i=0; i<(int)path.size()-1; i++ ) {
    //     int v = path[i];
    //     was[v] = false;
    //     if( i > 0 ) reach[v] = max( reach[v], reach[path[i-1]] );
    //     if(debug) clog << "Cumulative reach[" << v << "]: " << reach[v] << endl;
    //     if( reach[v] == i && i > 0 ) {
    //         dominators.push_back(v);
    //         if(debug) clog << "Adding node " << v << " to dominator set" << endl;
    //     }
    //
    //     reach[v] = max(reach[v],i);
    //     if(debug) clog << "Starting dfs from node " << v << endl;
    //     dfs(v);
    //     if(debug) clog << "after dfs, reach[" << v << "]: " << reach[v] << endl;
    // }
    //
    // return dominators;
}

VI Kernelizer::findDominators(VVI &V, VB was, int vs, int vt) {
    constexpr bool debug = false && kernelizer_debug;
    int N = V.size();

    if(debug) clog << "Seeking dominators, beg: " << vs << ", end: " << vt << endl;
    // DEBUG(V);

    // auto [dst,par] = Solver::findDistances(V,vs, was);

    auto path = Solver::findShortestPath( V, vs, vt, was );
    if( path.empty() ) return {}; // no path, so no dominators...

    if(debug) DEBUG(path);


    VI ind_on_path(N,-1);
    for(int i=0; i<path.size(); i++) ind_on_path[ path[i] ] = i;
    for( int d : path ) was[d] = true;

    VI reach(N,-1);
    function<int(int)> dfs = [&](int num) {
        was[num] = true;
        int res = -1;
        if(debug) clog << "\tnum: " << num << endl;
        for( int d : V[num] ) {
            res = max(res, ind_on_path[d]);
            if(!was[d]) res = max(res, dfs(d));
        }

        reach[num] = max(reach[num],res);
        return reach[num];
    };

    VI dominators;
    for( int i=0; i<(int)path.size()-1; i++ ) {
        int v = path[i];
        was[v] = false;
        if( i > 0 ) reach[v] = max( reach[v], reach[path[i-1]] );
        if(debug) clog << "Cumulative reach[" << v << "]: " << reach[v] << endl;
        if( reach[v] == i && i > 0 ) {
            dominators.push_back(v);
            if(debug) clog << "Adding node " << v << " to dominator set" << endl;
        }

        reach[v] = max(reach[v],i);
        if(debug) clog << "Starting dfs from node " << v << endl;
        dfs(v);
        if(debug) clog << "after dfs, reach[" << v << "]: " << reach[v] << endl;
    }

    return dominators;
}

Kernelizer2::Kernelizer2(VVI &_V, int _L, int _vs, int _vt) :
G(_V), /*V(_V), revV(_revV),*/ L(_L), N(_V.size()), vs(_vs), vt(_vt){
    helper.resize(N);
    was.resize(N);
}

Kernelizer2::Kernelizer2(GraphSP2 &_G, int _L, int _vs, int _vt) : G(_G), L(_L), vs(_vs), vt(_vt), N(_G.size()) {
    helper.resize(N);
    was.resize(N);
}

Kernelizer2::InducedGraphSP2 Kernelizer2::InducedGraphSP2::induce(GraphSP2 &g, VI nodes) {
    // clog << "Inducing by nodes: " << nodes << endl;
    auto gv = g.getStruture();
    VVI rev_gv = GraphUtils::reverseGraph(gv);

    InducedGraph indg = GraphInducer::induceDirected(gv, rev_gv, nodes);

    InducedGraphSP2 res;
    int n = indg.V.size();
    res.G.V.resize(n);
    swap( res.nodes, indg.nodes );
    swap(res.perm, indg.perm);

    for( int i=0; i<n; i++ ) {
        int a = res.nodes[i];
        for( auto arc : g[a] ) {
            if( res.perm.contains( arc.v ) ) {
                arc.v = res.perm[arc.v];
                res[i].push_back( arc );
            }
        }
    }

    return res;
}

Kernelizer2::InducedGraphSP2 Kernelizer2::InducedGraphSP2::induceByNonisolatedNodes(GraphSP2 &g) {
    int N = g.size();
    VB was(N);
    for( int i=0; i<N; i++ ) for( auto & arc : g[i] ) was[arc.v] = was[i] = true;
    VI nodes;
    for(int i=0; i<N; i++) if(was[i]) nodes.push_back(i);
    return induce(g,nodes);
}

Core2 Kernelizer2::createCore2Graph() {

    assert( GraphUtils::isSimple( G.getStruture() ) );
    assert( GraphUtils::isSimple( G.reverseGraph().getStruture() ) );


    bool changes_done = true;
    constexpr bool apply_core_in_each_iteration = true;
    constexpr bool write_size = false;

    bool is_init_symmetric = G.isSymmetric();
    // auto initG = G;

    auto applyCore = [&]() {
        auto ind_g = InducedGraphSP2::induceByNonisolatedNodes( G );
        G = ind_g.G;
        N = G.size();
        was.resize(N);
        helper.resize(N);
        vs = ind_g.perm[vs];
        vt = ind_g.perm[vt];

        // if( is_init_symmetric && !cnf.use_all_reduction_rules_even_for_undirected_graphs ) {
        //     DEBUG(G);
        //     DEBUG(initG);
        //     assert( G.isSymmetric() );
        // }
    };

    while(changes_done) {
        changes_done = false;

        if constexpr (write_size) { DEBUG(G.size()); if(G.size() <= 5) DEBUG(G); }
        clog << "Contracting maximal paths..." << endl;
        changes_done |= contractMaximalPaths(G);
        if constexpr (apply_core_in_each_iteration) applyCore();
        assert( G.isValid() );

        if constexpr (write_size) { DEBUG(G.size()); if(G.size() <= 5) DEBUG(G); }
        clog << "Removing dispensible biconnected components..." << endl;
        changes_done |= removeDispensibleBiconnectedComponents();
        if constexpr (apply_core_in_each_iteration) applyCore();
        assert( G.isValid() );

        if constexpr (write_size) { DEBUG(G.size()); if(G.size() <= 5) DEBUG(G); }
        clog << "Removing dispensible SCCs..." << endl;
        changes_done |= removeDispensibleSCCs();
        if constexpr (apply_core_in_each_iteration) applyCore();
        assert( G.isValid() );

        if constexpr (write_size) { DEBUG(G.size()); if(G.size() <= 5) DEBUG(G); }
        clog << "Removing dominated nodes..." << endl;
        changes_done |= removeDominatedNodes();
        if constexpr (apply_core_in_each_iteration) applyCore();
        assert( G.isValid() );

        if constexpr (write_size) { DEBUG(G.size()); if(G.size() <= 5) DEBUG(G); }
        clog << "Removing distant nodes V2 (dominator version) ..." << endl;
        changes_done |= removeDistantNodesV2Dominators();
        if constexpr (apply_core_in_each_iteration) applyCore();
        assert( G.isValid() );

        if( G.countArcs() < 50'000 ) {
            if( !G.isSymmetric() || cnf.use_all_reduction_rules_even_for_undirected_graphs ) {
                changes_done |= removeDistantArcsV2Dominators();
                if constexpr (apply_core_in_each_iteration) applyCore();
            }
        }

        if( !G.isSymmetric() || cnf.use_all_reduction_rules_even_for_undirected_graphs ) {
            if constexpr (write_size) { DEBUG(G.size()); if(G.size() <= 5) DEBUG(G); }
            clog << "Proceeding to find dominated components..." << endl;
            changes_done |= dominatedComponents(G, L, vs, vt);
            if constexpr (apply_core_in_each_iteration) applyCore();
            assert( G.isValid() );
        }

        if constexpr (write_size) { DEBUG(G.size()); if(G.size() <= 5) DEBUG(G); }
        clog << "Proceeding to contract components..." << endl;
        changes_done |= contractComponents( G );
        if constexpr (apply_core_in_each_iteration) applyCore();
        assert( G.isValid() );
    }

    if( is_init_symmetric && !cnf.use_all_reduction_rules_even_for_undirected_graphs ) {
        if(!G.isSymmetric()) {
            DEBUG(G);
            // DEBUG(initG);
        }
        assert( G.isSymmetric() );
    }

    assert( G.isValid() );
    auto ind_g = InducedGraphSP2::induceByNonisolatedNodes( G );
    assert( ind_g.G.isValid() );
    auto core2 = Core2( ind_g.G, ind_g.perm[vs], ind_g.perm[vt], L );
    // DEBUG(core2);

    // clog << "Graph induced by nonisolate nodes: " << endl << ind_g.G << endl;

    constexpr bool write_kernel = true;
    if constexpr (write_kernel){
        ofstream f( "kernel2.col" );
        auto gv = core2.G.getStruture();
        GraphUtils::writeInstanceToFile( f, gv, core2.vs, core2.vt, core2.L, directed_track );
        f.close();
    }

    for( int i=0; i<core2.G.size(); i++ ) for( auto & arc : core2.G[i] ) arc.lengths.trimToMax(L);
    core2.G.removeInvalidArcs();

    if(!core2.G.isValid()) {
        DEBUG(core2.G);
        assert( core2.G.isValid() );
    }

    return core2;
}



inline void Kernelizer2::removeArc(GraphSP2 &G, int a, int b) {
    for( int i=(int)G[a].size()-1; i>=0; i-- ) {
        if( G[a][i].v == b ) {
            swap( G[a][i], G[a].back() );
            G[a].pop_back();
        }
    }
}

void Kernelizer2::clearNode(GraphSP2 &g, GraphSP2 &revg, int v) {
    VPII to_remove;
    for( auto & arc : g[v] ) {
        int b = arc.v;
        to_remove.emplace_back(v,b);
        to_remove.emplace_back(b,v);
    }
    for( auto & arc : revg[v] ) {
        int b = arc.v;
        to_remove.emplace_back(v,b);
        to_remove.emplace_back(b,v);
    }

    for( auto [a,b] : to_remove ) {
        removeArc(g, a,b);
        removeArc(revg, a,b);
    }
}

bool Kernelizer2::contractComponents(GraphSP2 & g) {
    constexpr bool debug = false; // && kernelizer_debug;

    if( g.countArcs() == 0)  return false;
    // fill(ALL(was),false);
    // fill(ALL(helper),false);

    if constexpr (debug) {
        clog << "Contracting components in Kernelizer 2" << endl << "g: " << endl << g << endl;
        DEBUG(vs);
        DEBUG(vt);
    }

    auto revg = g.reverseGraph();

    // bool changes_done = false;

    bool is_g_symmetric = g.isSymmetric();
    // #CAUTION! Symmetric feature can be used if we take as L for indg max(L1,L2), where L1 and L2 are path lengtsh
    // for induced graphs with start and end points a->b and b->a. If we use smaller of the two lengths, then we might
    // get WA...
    // bool is_g_symmetric = false;

    vector<tuple<int,int,Lengths>> arcs_to_add;
    arcs_to_add.reserve(10);
    VI nodes_to_clear;

    auto considerNodePair = [&](int a, int b) {
        // Mask mask;
        VB mask(N,false);

        auto gv = g.getStruture();
        auto underV = GraphUtils::getUndirectedStructure( gv, {} );
        // undirecred structure is needed only to create [unreachables]...
        // mask.set( a,1);  mask.set( b,1);
        mask[a] = mask[b] = true;
        auto [dst_vs,par_vs] = Solver::findDistances( underV, vs, mask );
        auto [dst_vt,par_vt] = Solver::findDistances( underV, vt, mask );
        // mask.set(a,0);  mask.set(b,0);
        mask[a] = mask[b] = false;

        // if(a == 0 && b == 4) {
        //     DEBUG(dst_vs);
        //     DEBUG(dst_vt);
        //
        //     int trg = 96;
        //     clog << "Path from " << vs << " to " << trg << ": " << endl;
        //     VI path;
        //     for(int u=trg; u != -1; u = par_vs[u]) path.push_back(u);
        //     reverse(ALL(path));
        //     DEBUG(path);
        // }

        VI unreachables({a,b});
        for(int i=0; i<N; i++) {
            if( !( g[i].empty() && revg[i].empty() ) && dst_vs[i] == inf && dst_vt[i] == inf ) {
                unreachables.push_back(i);
            }
        }
        StandardUtils::makeUnique(unreachables);

        // distances in g and revg are needed to reduce as much as possible value of L in induced subgraph.
        // here could perhaps also be taken into account dominators, which might reduce those distance even more...
        // mask.set( a,1);  mask.set( b,1);
        mask[a] = mask[b] = true;
        tie(dst_vs,par_vs) = Solver2::findDistances( g, vs, mask );
        tie(dst_vt,par_vt) = Solver2::findDistances( revg, vt, mask );
        // mask.set(a,0);  mask.set(b,0);
        mask[a] = mask[b] = false;

        int dst_vsa = dst_vs[a];
        for( auto & arc : revg[a] ) {
            if( dst_vs[arc.v] != inf ) dst_vsa = min( dst_vsa, dst_vs[arc.v] + arc.lengths.minl() );
        }
        int dst_vtb = dst_vt[b];
        for( auto & arc : g[b] ) {
            if( dst_vt[arc.v] != inf ) dst_vtb = min( dst_vtb, dst_vt[arc.v] + arc.lengths.minl() );
        }


        if constexpr (debug) {
            clog << "Considering node pair (" << a << "," << b << ")" << endl;
            DEBUG(unreachables);
            DEBUG(g[a]);
            DEBUG(g[b]);
            DEBUG(revg[a]);
            DEBUG(revg[b]);
            DEBUG(dst_vsa);
            DEBUG(dst_vtb);
            // DEBUG(g);
        }



        if( L <= dst_vsa + dst_vtb ) {
            if constexpr (debug) {
                DEBUG(L);
                DEBUG(dst_vs[a]);
                DEBUG(dst_vt[b]);
                DEBUG(dst_vs[a] + dst_vt[b]);
                DEBUG(dst_vs[b] + dst_vt[a]);
                clog << "\t, there is certainly no path " << vs << " ->" << a << " -> " << b << " -> " << vt << endl;
            }
            return;
        }

        InducedGraphSP2 indg = InducedGraphSP2::induce(g, unreachables);
        int indg_vs = indg.perm[a];
        int indg_vt = indg.perm[b];
        int indg_L = L - dst_vsa - dst_vtb;
        removeArc( indg.G, indg_vs, indg_vt );
        removeArc( indg.G, indg_vt, indg_vs );


        auto getSolver2 = [&]() { auto s2 = Solver2 ( indg.G, indg_L, indg_vs, indg_vt ); s2.cnf = cnf; return s2; };

        assert( unreachables.size() >= 2 );
        if( unreachables.size() == 2 ) return; // if only a and b are unreachable, then there is nothing to do...

        auto contract = [&, indg_vs]( auto & gg, int _vs, int _vt ) {
            if( Solver2::existsPath( gg.G, _vs, _vt, mask ) ) {
                // there is a path from a to b using nodes in [unreachables], so there will be arc a->b added...
                if constexpr (debug) {
                    clog << "There exists a path from " << _vs << " to " << _vt << ", contracting components..." << endl;
                    DEBUG(unreachables);
                    DEBUG( gg.G );
                    DEBUG( gg.nodes );
                    DEBUG( gg.perm );
                    DEBUG(indg_L);

                    DEBUG(L);
                    DEBUG(dst_vs[a]);
                    DEBUG(dst_vt[b]);
                    DEBUG(dst_vs[a] + dst_vt[b]);
                    DEBUG(dst_vs[b] + dst_vt[a]);
                }
                auto solver2 = getSolver2();
                clog << endl << endl << "Solver2 created, proceeding to find result for new component, indg_L: "
                     << indg_L << endl << endl;
                auto res_tab = solver2.solve();

                if constexpr (debug)
                    clog << "res_tab: " << res_tab << endl;
                bigint _sum = 0;
                for( auto & _x_ : res_tab ) _sum += _x_;
                // if( _sum > 0 ) {
                    auto lengths = Lengths::createFromVBIG(res_tab);
                    arcs_to_add.emplace_back( indg.nodes[_vs],indg.nodes[_vt],lengths );
                    assert(indg.nodes[_vs] == a);
                    assert(indg.nodes[_vt] == b);
                    // if( is_g_symmetric ) arcs_to_add.emplace_back( indg.nodes[_vt],indg.nodes[_vs],lengths );

                    if constexpr (debug) clog << "\tAdding arc (" << indg.nodes[_vs] << "," << indg.nodes[_vt] << "," << lengths << ")" << endl;
                    return true;
                // }
            }
            return false;
        };

        auto component_contracted = contract( indg, indg_vs, indg_vt );

        if( component_contracted ) {
            for( int d : unreachables ) {
                if( d != a && d != b ) {
                    if constexpr (debug) clog << "Clearing unreachable node " << d << endl;
                    // clearNode(g, revg, d);
                    nodes_to_clear.push_back(d);
                }
            }
        }
    };

    int nodes_cleared = 0;
    int components_contracted = 0;

    auto V = g.getStruture();
    auto revV = GraphUtils::reverseGraph(V);
    auto underV = V;
    for(int i=0; i<N; i++) {
        underV[i] += revV[i];
        StandardUtils::makeUnique(underV[i]);
    }

    for( int a=0; a<N; a++ ) {
        if( a == vs || a == vt ) continue;
        if( g[a].empty() && revg[a].empty() ) continue;

        VB helper(N,true);
        for( int i=0; i<N; i++ ) if( underV[i].empty() ) helper[i] = false;
        helper[a] = false;
        auto [art_points, bridges] = BridgesAndArtPoints::getBridgesAndArtPoints(underV, helper);

        // if( a == 0 ) DEBUG(art_points);

        // for( int b = a+1; b < N; b++ ) {
        for( int b : art_points ) {
            if( b == vs || b == vt ) continue;
            if( g[b].empty() && revg[b].empty() ) continue;

            clog << "\ra: " << a << ", b: " << b << flush;
            // clog << "a: " << a << ", b: " << b << endl;

            arcs_to_add.clear();
            nodes_to_clear.clear();

            considerNodePair(a,b);
            considerNodePair(b,a);

            // assert( arcs_to_add.size() <= 2 );

            if( is_g_symmetric ) {
                Lengths *l1 = nullptr, *l2 = nullptr;
                for( auto & [x,y,lengths] : arcs_to_add ) {
                    // if( !lengths.valid() ) continue;
                    if( x == a ) l1 = &lengths;
                    if( x == b ) l2 = &lengths;
                }

                // if(l1 != nullptr) DEBUG(*l1);
                // if(l2 != nullptr) DEBUG(*l2);

                if( l1 != nullptr || l2 != nullptr ) {
                    if( l1 == nullptr ) { // desperate struggle to fix a bug...
                        arcs_to_add.emplace_back( a,b,*l2 );
                        l1 = &get<2>(arcs_to_add.back());
                    }
                    if( l2 == nullptr ) { // desperate struggle to fix a bug...
                        arcs_to_add.emplace_back( b,a,*l1 );
                        l2 = &get<2>(arcs_to_add.back());
                    }
                    // DEBUG(*l1);
                    // DEBUG(*l2);
                    assert( l1->minl() == l2->minl() || !l1->valid() || !l2->valid() );
                    if( l1->maxl() > l2->maxl() || !l2->valid() ) *l2 = *l1;
                    else if( l1->maxl() < l2->maxl() || !l1->valid() ) *l1 = *l2;
                    else assert( *l1 == *l2 );
                    // DEBUG(*l1);
                    // DEBUG(*l2);
                }
            }


            for( auto & [x,y,lengths] : arcs_to_add ) {
                if( !lengths.valid() ) continue;

                g[x].emplace_back(y,lengths);
                g.mergeParallellEdges(x);
                revg[y].emplace_back(x,lengths);
                revg.mergeParallellEdges(y);

                if( !StandardUtils::find(underV[x], y) ) underV[x].push_back(y);
                if( !StandardUtils::find(underV[y], x) ) underV[y].push_back(x);
            }
            StandardUtils::makeUnique(nodes_to_clear);
            if( !nodes_to_clear.empty() || !arcs_to_add.empty() ) {
                if constexpr (debug) {
                    DEBUG(nodes_to_clear);
                    DEBUG(arcs_to_add);
                    DEBUG(g);
                }
                assert( !nodes_to_clear.empty() );
                assert( !arcs_to_add.empty() );
                nodes_cleared += (int)nodes_to_clear.size();
                components_contracted++;

                // clog << "Before clearing nodes " << nodes_to_clear << ", g:" << endl << g << endl;
                for(int d : nodes_to_clear) {
                    clearNode( g, revg, d );
                    Kernelizer::clearNode(underV, underV,d);
                }
                // clog << "After clearing nodes " << nodes_to_clear << ", g:" << endl << g << endl;
            }

            // DEBUG(g);
            // ENDL(5);

        }
    }

    clog << "In contractComponents, components contracted: " << components_contracted
         << ", total nodes cleared: " << nodes_cleared << endl;

    if constexpr (debug) {
        clog << "After contracting components in Kernelizer 2" << endl << "g: " << endl << g << endl;
        DEBUG(vs);
        DEBUG(vt);
    }

    assert( GraphUtils::isSimple( G.getStruture() ) );
    // assert( G.isSymmetric() );
    for( int i=0; i<N; i++ ) {
        for( auto & arc : g[i] ) {
            if(!arc.lengths.valid()) {
                DEBUG(g);
                DEBUG(g[i]);
                DEBUG(arc.lengths);
                assert( arc.lengths.valid() );
            }
        }
    }

    return components_contracted > 0;
}

bool Kernelizer2::dominatedComponents(GraphSP2 &G, int L, int vs, int vt) {
    constexpr bool debug = false;

    if( G.countArcs() == 0)  return false;

    if constexpr (debug) clog << "Entering dominatedComponents()" << endl;

    const int N = G.size();
    VB was(N);
    bool changes_done = false;
    auto revG = G.reverseGraph();

    auto processForX = [L=L, &changes_done, cnf=cnf](GraphSP2& G, GraphSP2& revG, int x, VI X, VI & neighX, int vs, int vt ) {
        X.push_back(x);
        changes_done = true;

        auto [dst_vs, par_vs] = Solver2::findDistances( G, vs );
        auto [dst_vt, par_vt] = Solver2::findDistances( revG, vt );

        if constexpr (debug)
            clog << "Processing for set X: " << X << ", x: " << x << ", neighX: " << neighX
                 << endl << "G: " << endl << G << endl;

        for( int y : neighX ) {
            X.push_back(y);

            if constexpr (debug) clog << "\tFinding paths from " << x << " to " << y << endl;

            InducedGraphSP2 indg = InducedGraphSP2::induce( G,X );
            int indg_vs = indg.perm[x];
            int indg_vt = indg.perm[y];
            removeArc(indg.G, indg_vs, indg_vt ); // we need to remove possible arc between x and y
            removeArc(indg.G, indg_vt, indg_vs );

            int newL = L - dst_vs[x] - dst_vt[y];
            newL = max(0,newL);

            // { // uncomment this to visualize what is done on the fly
            //     GVVisualizer vis;
            //     auto V = G.getStruture();
            //     vis.preprocessFullGraph(V, true );
            //     vis.modifyNodeSpecs( [&]( string &name, GVVisualizer::_format& specs ) {
            //         int v = stoi(name);
            //         if( find(ALL(X),v) != X.end() ) specs["color"] = "orange";
            //         if( v == vs ) specs["color"] = "green";
            //         if( v == vt ) specs["color"] = "red";
            //     } );
            //     auto filename = "test_file_dominated_components_" + to_string(x) + "_" + to_string(y) + "_"
            //     + to_string(vs) + "_" + to_string(vt) + "_" + to_string(L);
            //     vis.visualize( filename );
            //
            //     ofstream str("test_file_dominated_components.col");
            //     GraphUtils::writeInstanceToFile(str, V, vs, vt, L );
            // }

            Solver2 solver( indg.G, newL, indg_vs, indg_vt );
            solver.cnf = cnf;
            auto indg_res = solver.solve();

            if constexpr (debug) clog << "\tindg_res: " << indg_res << endl;

            Lengths len = Lengths::createFromVBIG(indg_res);

            if( len.valid() ) {
                if constexpr (debug) clog << "\tadding arc " << x << " -> " << y << ": " << len << endl;
                G[x].emplace_back( y, len );
                revG[y].emplace_back(x,len);

                G.mergeParallellEdges(x);
                revG.mergeParallellEdges(y);
            }

            X.pop_back();
        }
        X.pop_back();

        if constexpr (debug) clog << "Clearing set X: " << X << endl;

        for( int d : X ) clearNode(G, revG, d);
    };

    auto processForGraph = [N=N, &was, &processForX, L=L]( GraphSP2 & G, GraphSP2 & revG, int vs, int vt ) {

        auto [dst_vs_orig, par_vs_orig] = Solver2::findDistances(G, vs);

        for(int x = 0; x < N; x++) {
            if(x == vs || x == vt) continue;
            was[x] = true;

            auto [dstvs, par_vs] = Solver2::findDistances(G,vs, was);

            VI X;
            for(int i=0; i<N; i++) {
                if( i != vs && i != vt && i != x && dstvs[i] == inf && dst_vs_orig[i] != inf ) {
                    X.push_back(i);
                    was[i] = true;
                }
            }

            set<int> temp;
            for( int d : X ) for( auto & arc : G[d] ) if(!was[arc.v]) temp.insert(arc.v);
            assert(!temp.contains(x));
            VI neighX(ALL(temp));

            if constexpr (debug) {
                clog << "\tx: " << x << ", X: " << X << " neighX: " << neighX << endl;
            }

            if( !neighX.empty() ) {
                processForX(G, revG, x, X, neighX, vs, vt);

                tie(dstvs, par_vs) = Solver2::findDistances(G,vs, was);
                tie(dst_vs_orig, par_vs_orig) = Solver2::findDistances(G, vs);
            }

            for(int d : X) was[d] = false;
            was[x] = false;
        }
    };

    if constexpr (debug) clog << "Processing for graph G, L: " << L << ", vs: "
        << vs << ", vt: " << vt << endl;
    processForGraph( G, revG, vs, vt );

    if constexpr (debug) clog << "Processing for reversed graph (revG), L: " << L << ", vs: "
        << vs << ", vt: " << vt << endl;
    processForGraph( revG, G, vt, vs );

    if constexpr (debug) clog << "\tLeaving dominatedComponents()" << endl;

    // if(changes_done) {
    //     clog << "exiting in dominatedComponent()" << endl; exit(1);
    // }

    assert( GraphUtils::isSimple( G.getStruture() ) );
    for( int i=0; i<N; i++ ) {
        for( auto & arc : G[i] ) {
            if(!arc.lengths.valid()) {
                DEBUG(G);
                DEBUG(G[i]);
                DEBUG(arc.lengths);
                assert( arc.lengths.valid() );
            }
        }
    }

    return changes_done;
}

void Kernelizer2::contractNode( int v, GraphSP2 & G, VVI & revG, int L ) {
    constexpr bool debug = false;

    auto pred = revG[v];
    auto succ = G[v];

    vector<pair<int,SP2Arc>> to_add;

    if constexpr (debug) clog << "v: " << v << endl << endl;

    for( int a : pred ) {
        if constexpr (debug) clog << "a: " << a << endl;

        for(auto arc_vb : succ ) {
            int b = arc_vb.v;
            if( a == b ) continue;
            if constexpr (debug) clog << "\tb: " << b << endl;

            for( auto & arc_av : G[a] ) {
                if( arc_av.v != v ) continue;

                // we can now replace arc_av by joint arc...
                // this way we simultaneously 'remove' old arc and add new arc...
                // SP2Arc new_arc(b,1,1,1);
                SP2Arc new_arc(b,1,1,0);
                new_arc.lengths = Lengths::multiplyLengths( arc_av.lengths, arc_vb.lengths );
                new_arc.lengths.trimToMax(L);
                if( new_arc.lengths.valid() ) to_add.emplace_back(a,new_arc);
                if constexpr (debug) {
                    DEBUG(arc_av);
                    DEBUG(arc_vb);
                }
                if constexpr (debug) clog << "adding new arc: (" << a << "," << new_arc << ")" << endl;
            }
        }
    }

    for( int a : pred ) {
        for(auto arc_vb : succ ) {
            int b = arc_vb.v;
            removeArc(G,a,v);
            removeArc(G,v,a);
            removeArc(G,v,b);
            removeArc(G,b,v);

            Kernelizer::removeArc( revG, a,v );
            Kernelizer::removeArc( revG, v,a );
            Kernelizer::removeArc( revG, b,v );
            Kernelizer::removeArc( revG, v,b );
        }
    }

    for( auto & [a,arc] : to_add ) {
        G[a].push_back(arc);
        if( !StandardUtils::find(revG[arc.v],a) ) {
            revG[arc.v].push_back(a);
        }
    }

    for( auto & [a,arc] : to_add ) G.mergeParallellEdges(a);
    // for( int a : pred ) G.mergeParallellEdges(a);
    // for( auto & arc : succ ) G.mergeParallellEdges(arc.v);
};

bool Kernelizer2::contractMaximalPaths(GraphSP2 &G) {
    constexpr bool debug = false;
    if( G.countArcs() == 0 ) return false;

    // DEBUG(G);
    // DEBUG(PII(vs,vt));

    int N = G.size();

    VVI revG(N);
    for( int i=0; i<N; i++ ) for( auto & e : G[i] ) revG[e.v].push_back(i);

    bool changes_done = true;
    int nodes_contracted = 0;

    while(changes_done) {
        changes_done = false;

        VB was(N);
        for( int i=0; i<N; i++ ) {
            if( i == vs || i == vt || G[i].empty() || revG[i].empty() ) continue;

            int cnt = 0;
            int only_in_cnt = 0;
            int only_out_cnt = 0;

            for( int d : revG[i] ) was[d] = true;
            for( auto & e : G[i] ) if(!was[e.v]){ only_out_cnt++; }
            for( int d : revG[i] ) was[d] = false;

            for( auto & e : G[i] ) if(!was[e.v]){ was[e.v] = true; cnt++; }
            for( int d : revG[i] ) if(!was[d]){ was[d] = true; cnt++; only_in_cnt++; }

            bool can_contract = (cnt <= 2 || G[i].size() == 1 || revG[i].size() == 1
                // || ( cnt <= 3 && (only_in_cnt >= 2 || only_out_cnt >= 2 ) )
                || ( only_in_cnt >= cnt-1 || only_out_cnt >= cnt-1 )
                );

            // clear [was] before contracting node i
            for( auto & e : G[i] ) was[e.v] = false;
            for( int d : revG[i] ) was[d] = false;

            // if( cnt <= 2 ) { // contracting node with at most 2 neihgbors
            if( can_contract ) { // contracting node with at most 2 neihgbors
                assert( cnt >= 1 );

                if constexpr (debug) clog << "Contracting node " << i << endl;
                // if( cnt >= 3 ) {
                //     clog << "Contracting node " << i << " with " << cnt << " total neighbors, "
                //          << only_in_cnt << " only-in neighbors and " << only_out_cnt << " only-out neighbors" << endl;
                // }
                contractNode(i, G, revG, L);
                changes_done = true;
                nodes_contracted++;
            }
        }
    }

    clog << "nodes contracted: " << nodes_contracted << endl;

    assert(GraphUtils::isSimple(revG));
    assert( GraphUtils::isSimple( G.getStruture() ) );

    assert( G.isValid() );

    return nodes_contracted > 0;
}

bool Kernelizer2::removeDistantNodesV2Dominators() {
    constexpr bool debug = false;

    if( G.countArcs() == 0 ) return false;

    auto revG = G.reverseGraph();

    auto V = G.getStruture();
    auto revV = GraphUtils::reverseGraph(V);

    VI to_remove;
    for( int v=0; v<N; v++ ) {
        if( v == vs || v == vt ) continue;
        if( G[v].empty() && revG[v].empty() ) continue;

        auto B = Kernelizer::findDominators(revV, vt,v);
        for(int b : B) was[b] = true;
        auto [dst_vs, par_vs] = Solver2::findDistances(G, vs, was);
        for(int b : B) was[b] = false;

        auto A = Kernelizer::findDominators(V, vs,v);
        for(int a : A) was[a] = true;
        auto [dst_vt, par_vt] = Solver2::findDistances(revG, vt, was);
        for(int a : A) was[a] = false;

        if( dst_vs[v] + dst_vt[v] > L ) {
            to_remove.push_back(v);
            if constexpr (debug) {
                clog << "Removing distant node (dominator version) v: " << v << endl;
                DEBUG(A);
                DEBUG(B);
                clog << "dst_vs: " << dst_vs[v] << endl;
                clog << "dst_vt: " << dst_vt[v] << endl;
                clog << endl;
            }
        }
    }

    for(int v : to_remove) clearNode(G, revG, v);
    clog << "Distant nodes (dominator version) removed: " << to_remove.size() << endl;

    assert( G.isValid() );

    return !to_remove.empty();
}

bool Kernelizer2::removeDistantArcsV2Dominators() {
     constexpr bool debug = false;
    if( G.countArcs() == 0 ) return false;

    VPII to_remove;
    VB was(N,false), helper(N,false);

    auto revG = G.reverseGraph();

    for( int v=0; v<N; v++ ) {
        if( G[v].empty() && revG[v].empty() ) continue;

        auto V = G.getStruture();
        auto revV = GraphUtils::reverseGraph(V);

        for( auto & arc : G[v] ) {
            int d = arc.v;

            helper[v] = true; // excluding begin of the arc (v,d) from dominator search
            auto B = Kernelizer::findDominators(revV, helper,vt,d);
            fill(ALL(helper),false);
            for(int b : B) was[b] = true;
            was[d] = true; // node d cannot be on path vs->v
            auto [dst_vs, par_vs] = Solver2::findDistances(G, vs, was);
            for(int b : B) was[b] = false;
            was[d] = false;

            helper[d] = true; // excluding end of the arc (v,d) from dominator search
            auto A = Kernelizer::findDominators(V, helper, vs,v);
            fill(ALL(helper),false);
            for(int a : A) was[a] = true;
            was[v] = true; // node v cannot be on path d->vt (that is vt->d in revV)
            auto [dst_vt, par_vt] = Solver2::findDistances(revG, vt, was);
            for(int a : A) was[a] = false;
            was[v] = false;

            if( dst_vs[v] + dst_vt[d] > L ) {
                to_remove.emplace_back( v,d );
                if constexpr (debug) {
                    clog << "Removing distant arc (dominator version) " << v << " -> " << d << endl;
                    DEBUG(A);  DEBUG(B);
                    clog << "dst_vs: " << dst_vs[v] << endl;
                    clog << "dst_vt: " << dst_vt[d] << endl;
                    clog << endl;
                }
            }
        }
    }

    for(auto [v,d] : to_remove) {
        removeArc( G, v,d );
        removeArc( revG, d,v );
    }
    clog << "Distant arcs (dominator version) removed: " << to_remove.size() << endl;

    assert( G.isValid() );

    return !to_remove.empty();
}

bool Kernelizer2::removeDispensibleSCCs() {
    if( G.countArcs() == 0)  return false;

    auto V = G.getStruture();
    auto revV = GraphUtils::reverseGraph(V);
    Kernelizer kern(V, revV, 1, vs, vt);
    if( !kern.removeDispensibleSCCs() ) return false;

    VI nodes_cleared;
    for( int i=0; i<N; i++ ) {
        if( (V[i].size() != kern.V[i].size() || revV[i].size() != kern.revV[i].size() )
            && kern.V[i].empty() && kern.revV[i].empty() ) {
            nodes_cleared.push_back(i);
        }
    }

    auto revG = G.reverseGraph();
    for(auto d : nodes_cleared) clearNode(G, revG, d);

    assert(nodes_cleared.size() > 0);
    assert( G.isValid() );
    return true;
}

bool Kernelizer2::removeDominatedNodes() {
    if( G.countArcs() == 0)  return false;

    auto V = G.getStruture();
    auto revV = GraphUtils::reverseGraph(V);
    Kernelizer kern(V, revV, 1, vs, vt);
    if( !kern.removeDominatedNodes() ) return false;

    VI nodes_cleared;
    for( int i=0; i<N; i++ ) {
        if( (V[i].size() != kern.V[i].size() || revV[i].size() != kern.revV[i].size() )
            && kern.V[i].empty() && kern.revV[i].empty() ) {
            nodes_cleared.push_back(i);
        }
    }

    auto revG = G.reverseGraph();
    for(auto d : nodes_cleared) clearNode(G, revG, d);

    assert(nodes_cleared.size() > 0);
    assert( G.isValid() );
    return true;
}


bool Kernelizer2::removeDispensibleBiconnectedComponents() {
    if( G.countArcs() == 0)  return false;

    auto V = G.getStruture();
    auto revV = GraphUtils::reverseGraph(V);
    Kernelizer kern(V, revV, 1, vs, vt);
    if( !kern.removeDispensibleBiconnectedComponents() ) return false;

    VI nodes_cleared;
    for( int i=0; i<N; i++ ) {
        if( (V[i].size() != kern.V[i].size() || revV[i].size() != kern.revV[i].size() )
            && kern.V[i].empty() && kern.revV[i].empty() ) {
            nodes_cleared.push_back(i);
        }
    }

    auto revG = G.reverseGraph();
    for(auto d : nodes_cleared) clearNode(G, revG, d);

    assert(nodes_cleared.size() > 0);
    assert( G.isValid() );
    return true;
}
