//
// Created by sylwe on 14.07.2024.
//

#include "Structures.h"

Lengths SP2Arc::mul_one( VBIG(1,1),0 );

ostream& operator<<( ostream & str, HashType & h ) {
    str << h.hash;
    return str;
}

GraphSP2::GraphSP2(VVI &V) {
    createFromVVI(V);
}

void GraphSP2::createFromVVI(VVI &G) {
    int N = G.size();
    V.resize(N);
    for( int i=0; i<N; i++ ) {
        for( int d : G[i] ) {
            SP2Arc arc(d,1,1,1);
            arc[1] = 1;
            V[i].push_back(arc);
        }
    }
}

int GraphSP2::getArcMinl(int a, int b) {
    for( auto & arc : V[a] ) {
        if(arc.v == b) return arc.minl();
    }

    assert(false && "arc a->b must exist in getArcMinl");
    return inf;
}

int GraphSP2::countArcs() {
    int res = 0;
    for(auto & v : V) res += v.size();
    return res;
}

GraphSP2 GraphSP2::reverseGraph() {
    GraphSP2 res;
    res.V.resize(size());
    for( int i=0; i<size(); i++ ) for( auto & arc : (*this)[i] ) res[arc.v].emplace_back( i, arc.lengths );
    return res;
}

bool GraphSP2::isSymmetric() {
    int N = V.size();
    for( int i=0; i<N; i++ ) {
        for( auto & arc : V[i] ) {
            if( !hasArc(arc.v,i) ) return false;
            auto & l = getArc(arc.v,i).lengths;
            if( arc.lengths != l ) return false;
        }
    }
    return true;
}


VVI GraphSP2::getStruture() {
    int N = V.size();
    VVI res(N);
    for(int i=0; i<N; i++) {
        for(auto & arc : V[i]) res[i].push_back(arc.v);
    }
    return res;
}

void GraphSP2::mergeParallellEdges(int v) {
    sort( ALL(V[v]), []( auto & sp1, auto & sp2 ) {
        return sp1.v < sp2.v;
    } );

    // clog << "Merging parallel arcs: " << V[v] << endl;

    auto & vec = V[v];
    int p = 0, q = 0, r = 0;

    while( p < V[v].size() ) {
        q = p;
        while( q < V[v].size() && V[v][p].v == V[v][q].v ) q++;
        for( int i=p+1; i<q; i++ ) {
            // clog << "\t\tv: " << v << "merging arc " << vec[i] << " to arc " << vec[p] << endl;
            if( vec[i].size() > vec[p].size() ) vec[p].resize(vec[i].size());
            Lengths lengths( min( vec[p].lengths.minl(), vec[i].lengths.minl() ),
                max( vec[p].lengths.maxl(), vec[i].lengths.maxl() ) );
            for( int j=vec[p].lengths.minl(); j <= vec[p].lengths.maxl(); j++ ) lengths[j] += vec[p][j];
            for( int j=vec[i].lengths.minl(); j <= vec[i].lengths.maxl(); j++ ) lengths[j] += vec[i][j];
            vec[p].lengths = lengths;

            // for( int l=0; l<vec[i].size(); l++ ) vec[p][l] += vec[i][l];
            // clog << "\t\t\tresult: " << vec[p] << endl;
        }

        vec[r++] = vec[p];
        p = q;
    }


    while( vec.size() > r ) vec.pop_back();

    // clog << "After merging parallel arcs: " << V[v] << endl;
}

void GraphSP2::mergeParallellEdges() {
    for(int i=0; i<V.size(); i++) mergeParallellEdges(i);
}

ostream& operator<<( ostream & str, GraphSP2 & g ) {
    str << "{ " << endl;
    for( int i=0; i<g.size(); i++ ) {
        str << i << ": {";
        for( auto & arc : g[i] ) {
            str << " (" << int(arc.v) << " | minl: " << arc.lengths.minl() << " | ";
            for( int j=arc.lengths.minl(); j<=arc.lengths.maxl(); j++ ) str << arc[j] << " ";
            str << ")";
        }
        str << " } " << endl;
    }
    str << "}";
    return str;
}


ostream& operator<<(ostream& str, SP2Arc & arc) {
    str << "(" << int(arc.v) << ", " << arc.lengths << ")";
    return str;
}

ostream& operator<<( ostream& str, Lengths & l ) {
    str << "( minl: " << l.minl() << ", " << l.l << ")";
    return str;
}




ostream& operator<<( ostream& str, SubsetPaths4 & sp4 ) {
    str << sp4.toString();
    return str;
}


FBSState::FBSState(Lengths _counts, VPII _ends, VI _loose, int _vs_node) {
    counts = _counts;
    ends = _ends;
    loose = _loose;
    vs_node = _vs_node;
    hash = getHash();
}


bool FBSState::isCorrectState() {
    unordered_set<int> zb;
    zb.insert(vs_node);
    for( int d : loose ) {
        if(zb.contains(d)) return false;
        zb.insert(d);
    }
    for(auto [a,b] : ends) {
        if(zb.contains(a)) return false;
        zb.insert(a);
        if(zb.contains(b)) return false;
        zb.insert(b);
    }
    return true;
}

HashType FBSState::getHash() {
    HashType hash = (node_hashes[vs_node] ^ node_hashes2[vs_node]) + (node_hashes2[vs_node] ^ node_hashes[vs_node]);
    for( int d : loose ) hash ^= node_hashes[d];
    for( auto [a,b] : ends ) hash ^= node_hashes[a] + node_hashes2[b];
    return hash;
}

ostream& operator<<( ostream & str, FBSState & st ) {
    str << "{";
    str << "vs_node: " << st.vs_node << ", loose: " << st.loose << ", ends: " << st.ends << ", hash: " << st.hash;
    str << ", counts: " << st.counts;
    str << "}";
    return str;
}



FBSState2::FBSState2(Lengths _counts, VPII _ends, int _vs_node) {
    counts = _counts;
    ends = _ends;
    vs_node = _vs_node;
    hash = getHash();
}

void FBSState2::removeTaken(int v) {
    StandardUtils::removeAll(taken,v);
}

bool FBSState2::isCorrectState() {
    unordered_set<int> zb;
    zb.insert(vs_node);
    for( int a : taken) {
        if(zb.contains(a)) return false;
        zb.insert(a);
    }
    for(auto [a,b] : ends) {
        if(zb.contains(a)) return false;
        zb.insert(a);
        if(zb.contains(b)) return false;
        zb.insert(b);
    }
    return true;
}

HashType FBSState2::getHash() {
    HashType hash = (node_hashes[vs_node] ^ node_hashes2[vs_node]) + (node_hashes2[vs_node] ^ node_hashes[vs_node]);
    for( int d : taken ) hash ^= node_hashes[d];
    for( auto [a,b] : ends ) hash ^= node_hashes[a] + node_hashes2[b];
    return hash;
}


ostream& operator<<( ostream & str, FBSS2 & st ) {
    str << "{";
    str << "vs_node: " << st.vs_node << ", ends: " << st.ends << ", hash: " << st.hash
        << ", taken: " << st.taken;
    str << ", counts: " << st.counts;
    str << "}";
    return str;
}









FBSState3::FBSState3(VPII _ends, int _vs_node, int _minl, HashType bndhash) {
// FBSState3::FBSState3(VPII _ends, int _vs_node, int _minl) {
    ends = _ends;
    vs_node = _vs_node;
    minl = _minl;
    // bnd_hash = bndhash;
    hash = getHash();
}

void FBSState3::removeTaken(int v, VI & ind_in_S) {
    int pos = ind_in_S[v];
    assert(pos != -1);
    for( auto & m : taken_masks ) m.first &= ~( 1 << pos );
    StandardUtils::makeUnique(taken_masks);
}

bool FBSState3::isCorrectState(VI & ind_in_S) {
    unordered_set<int> zb;
    zb.insert(vs_node);
    // for( int a : taken) {
    //     if(zb.contains(a)) return false;
    //     zb.insert(a);
    // }
    for(auto [a,b] : ends) {
        if(zb.contains(a)) return false;
        zb.insert(a);
        if(zb.contains(b)) return false;
        zb.insert(b);
    }

    for( auto [m,ml] : taken_masks ) {
        for( int a : zb ) {
            assert( ind_in_S[a] != -1 );
            // DEBUG(ind_in_S[a]);
            if( m & ( 1 << ind_in_S[a] ) ){
                auto bits = getSetBits(m);
                clog << endl << "ERROR, node " << a << " is in taken mask: m: " << m << " -> bits: " << bits << endl;
                return false;
            }
        }
    }

    return true;
}

void FBSState3::makeUnique(VPII &tm) {
    sort(ALL(tm));
    int p = 0, q = 0, r = 0;

    while( p < tm.size() ){
        q = p;
        while( q < tm.size() && (tm[p].first == tm[q].first) ) q++;
        // tm[p] has the smallest minl value among all tm[p], tm[p+1], ..., tm[q-1]
        swap( tm[r++], tm[p] );
        p = q;
    }

    while( tm.size() > r ) tm.pop_back();
}

PII FBSState3::getTakenMasks(int m, int max_bits) {
    int tm1 = 0, tm2 = 0;
    int cnt = 0;
    for(int i=0; i<32 && cnt < max_bits; i++) {
        if( m & (1<<i) ) {
            tm1 |= (1<<i);
            cnt++;
        }
    }

    cnt = 0;
    for(int i=31; i>=0 && cnt < max_bits; i--) {
        if( m & (1<<i) ) {
            tm2 |= (1<<i);
            cnt++;
        }
    }

    return {tm1, tm2};
}

HashType FBSState3::getHash()  {
    HashType hash = (node_hashes[vs_node] ^ node_hashes2[vs_node]) + (node_hashes2[vs_node] ^ node_hashes[vs_node]);
    // hash ^= bnd_hash;
    for( auto [a,b] : ends ) hash ^= node_hashes[a] + node_hashes2[b];
    return hash;
}


ostream& operator<<( ostream & str, FBSState3 & st ){
    str << "{";
    str << "vs_node: " << st.vs_node << ", ends: " << st.ends << ", hash: " << (st.hash.hash % 1000)
        << ", taken_masks: ";
    int ind = 0;
    for( auto [m,ml] : st.taken_masks ) str << (ind++ ? ", " : "") << m << ":"  << st.getSetBits(m);
    str << ", minl: " << st.minl;
    str << "}";
    return str;
}










FBSStateD::FBSStateD(Lengths _counts, VPII _ends, int _vs_node) {
    counts = _counts;
    ends = _ends;
    vs_node = _vs_node;
    hash = getHash();
}

void FBSStateD::removeTaken(int v) {
    StandardUtils::removeAll(taken,v);
}

bool FBSStateD::isCorrectState() {
    unordered_set<int> zb;
    zb.insert(vs_node);
    for( int a : taken) {
        if(zb.contains(a)) return false;
        zb.insert(a);
    }
    for(auto [a,b] : ends) {
        if( a > b ) return false;
        if(zb.contains(a)) return false;
        zb.insert(a);
        if(zb.contains(b)) return false;
        zb.insert(b);
    }
    return true;
}

HashType FBSStateD::getHash() {
    HashType hash = (node_hashes[vs_node] ^ node_hashes2[vs_node]) + (node_hashes2[vs_node] ^ node_hashes[vs_node]);
    for( int d : taken ) hash ^= node_hashes[d];
    for( auto [a,b] : ends ) {
        // assert(a < b);
        hash ^= node_hashes[a] + node_hashes2[b];
    }
    return hash;
}


ostream& operator<<( ostream & str, FBSStateD & st ) {
    str << "{";
    str << "vs_node: " << st.vs_node << ", ends: " << st.ends << ", hash: " << st.hash
        << ", taken: " << st.taken;
    str << ", counts: " << st.counts;
    str << "}";
    return str;
}






FBSState3D::FBSState3D(VPII _ends, int _vs_node, int _minl) {
    ends = _ends;
    vs_node = _vs_node;
    minl = _minl;
    hash = getHash();
}

bool FBSState3D::isCorrectState(VI & ind_in_S) {
    unordered_set<int> zb;
    zb.insert(vs_node);
    // for( int a : taken) {
    //     if(zb.contains(a)) return false;
    //     zb.insert(a);
    // }
    for(auto [a,b] : ends) {
        if(zb.contains(a)) return false;
        zb.insert(a);
        if(zb.contains(b)) return false;
        zb.insert(b);
    }

    for( auto [m,ml] : taken_masks ) {
        for( int a : zb ) {
            assert( ind_in_S[a] != -1 );
            // DEBUG(ind_in_S[a]);
            if( m & ( 1 << ind_in_S[a] ) ){
                auto bits = getSetBits(m);
                clog << endl << "ERROR, node " << a << " is in taken mask: m: " << m << " -> bits: " << bits << endl;
                return false;
            }
        }
    }

    zb.clear();
    for( auto [m,ml] : taken_masks ) {
        if( zb.contains(m) ) {
            clog << "ERROR! Repeated taken mask " << m << endl;
            return false;
        }
        zb.insert(m);
    }

    return true;
}

void FBSState3D::makeUnique(VPII &tm) {
    sort(ALL(tm));
    int p = 0, q = 0, r = 0;

    while( p < tm.size() ){
        q = p;
        while( q < tm.size() && (tm[p].first == tm[q].first) ) q++;
        // tm[p] has the smallest minl value among all tm[p], tm[p+1], ..., tm[q-1]
        swap( tm[r++], tm[p] );
        p = q;
    }

    while( tm.size() > r ) tm.pop_back();
}

PII FBSState3D::getTakenMasks(int m, int max_bits) {
    int tm1 = 0, tm2 = 0;
    int cnt = 0;
    for(int i=0; i<32 && cnt < max_bits; i++) {
        if( m & (1<<i) ) {
            tm1 |= (1<<i);
            cnt++;
        }
    }

    cnt = 0;
    for(int i=31; i>=0 && cnt < max_bits; i--) {
        if( m & (1<<i) ) {
            tm2 |= (1<<i);
            cnt++;
        }
    }

    return {tm1, tm2};
}

HashType FBSState3D::getHash()  {
    HashType hash = (node_hashes[vs_node] ^ node_hashes2[vs_node]) + (node_hashes2[vs_node] ^ node_hashes[vs_node]);
    for( auto [a,b] : ends ) hash ^= node_hashes2[a] + node_hashes2[b]; // must be symmetric
    return hash;
}


ostream& operator<<( ostream & str, FBSState3D & st ){
    str << "{";
    str << "vs_node: " << st.vs_node << ", ends: " << st.ends << ", hash: " << (st.hash.hash % 1000)
        << ", taken_masks: ";
    int ind = 0;
    for( auto [m,ml] : st.taken_masks ) str << (ind++ ? ", " : "") << m << ":"  << st.getSetBits(m);
    str << ", minl: " << st.minl;
    str << "}";
    return str;
}


