// #pragma GCC target("avx2")

#include <GraphInducer.h>
#include <GraphUtils.h>
#include <RandomNumberGenerators.h>
#include <StandardUtils.h>
#include <Stopwatch.h>
#include <StronglyConnectedComponents.h>

#include "FlowCutter.h"
#include "HPP.h"
#include "Makros.h"
#include "Solver.h"
#include "Solver2.h"
#include "Solver2BT.h"
#include "Solver2FBS.h"
#include "Solver2FBSD.h"
#include "Solver2MIM.h"

using namespace std;


std::ostream& operator<<(std::ostream & str, v8si v) {
    str << "[";
    for( int i=0; i<8; i++ ) str << v[i] << " ";
    str << "]";
    return str;
}

std::ostream& operator<<( std::ostream& dest, __int128_t value ) // this is to be able to print __int128 to stdout
{
    std::ostream::sentry s( dest );
    if ( s ) {
        __uint128_t tmp = value < 0 ? -value : value;
        char buffer[ 128 ];
        char* d = std::end( buffer );
        do
        {
            -- d;
            *d = "0123456789"[ tmp % 10 ];
            tmp /= 10;
        } while ( tmp != 0 );
        if ( value < 0 ) {
            -- d;
            *d = '-';
        }
        int len = std::end( buffer ) - d;
        if ( dest.rdbuf()->sputn( d, len ) != len ) {
            dest.setstate( std::ios_base::badbit );
        }
    }
    return dest;
}



void testOnUndirectedGraphs() {
    int seed = 1;

    int TESTS = 1e5;
    int T = 0;

    VVI V;
    int L,vs,vt;

    while(T++ < TESTS){
        clog << endl << endl << "Considering test #" << T << " / " << TESTS << endl;

        UniformIntGenerator rnd(0,100, seed++);

        int N = 20 + rnd.nextInt(15);
        node_hashes = createNodeHashes(N);
        node_hashes2 = createNodeHashes(N);

        int deg = 3;
        if( T&1 ) deg++;
        V = VVI(N);
        vs = 0; vt = 1;
        for( int i=0; i<N; i++ ) {
            int r = rnd.nextInt(deg);
            for(int j=0; j<r; j++) {
                int a = rnd.nextInt(N);
                if( i == vs && a == vt ) continue;
                if( i == vt && a == vs ) continue;
                if(a != i) {
                    V[i].push_back(a);
                    V[a].push_back(i);
                }
            }
        }
        for(int i=0; i<N; i++) StandardUtils::makeUnique(V[i]);
        L = 1 + rnd.nextInt(N-1);

        {
            auto indg = GraphInducer::induceByNonisolatedNodes(V);
            V = indg.V;
            vs = indg.perm[vs];
            vt = indg.perm[vt];
            N = V.size();
            if( vs == vt ) {
                assert(vs == 0);
                vt = 1;
            }
            // GraphUtils::writeGraphHumanReadable(V);
            if(!GraphUtils::isConnected(V)) continue;
            assert( !GraphUtils::isDirected(V) );
            assert( GraphUtils::isConnected(V) );
            assert( GraphUtils::isSimple(V) );
        }

        ofstream str( "test.col" );
        GraphUtils::writeInstanceToFile(str, V, vs,vt,L);
        str.close();


        Solver2BT s(V,L, vs,vt);
        auto res1 = s.solveBTStateMerging();

        Solver2FBS s0( GraphSP2(V),L,vs,vt );
        auto [order,pathwidth] = s0.createFBSNodeOrder();

        assert( order.size() == 2*N );

        Solver2FBSD s2( GraphSP2(V), L, vs, vt );
        auto res2 = s2.solveFBS3DForNodeOrder(order);

        auto res3 = s0.solveFBS();

        Solver ss(V,L,vs,vt);
        auto res4 = ss.solveSmart();

        auto sum1 = accumulate(ALL(res1),(bigint)0);
        auto sum2 = accumulate(ALL(res2),(bigint)0);
        auto sum3 = accumulate(ALL(res3),(bigint)0);
        auto sum4 = accumulate(ALL(res4),(bigint)0);

        if( sum1 != sum2 || sum1 != sum3 || sum1 != sum4 ) {
            DEBUG(res1); DEBUG(sum1);
            DEBUG(res2); DEBUG(sum2);
            DEBUG(res3); DEBUG(sum3);
            DEBUG(res4); DEBUG(sum4);

            assert( sum1 == sum2 );
            assert( sum1 == sum3 );
            assert( sum1 == sum4 );
        }

        DEBUG(res3);
        DEBUG(sum3);
    }
}

int main() {
    ios_base::sync_with_stdio(0);

    DEBUG(sizeof(bigint));
    DEBUG(sizeof(Mask1));
    DEBUG(sizeof(Mask2));
    DEBUG(sizeof(Mask));
    DEBUG(sizeof(Lengths));
    DEBUG(sizeof(SP4));
    DEBUG(sizeof(SP2Arc));
    DEBUG(sizeof(FBSState2));
    DEBUG(sizeof(FBSState3));
    DEBUG(sizeof(FBSState3D));


    // testOnUndirectedGraphs();


    Stopwatch sw;
    sw.start("main");


    Solver solver( {}, -1, -1, -1 );
    solver.use_graphillion_for_all_paths = false;

    solver.readGraph(directed_track);
    solver.V = GraphUtils::makeSimple(solver.V);
    solver.revV = GraphUtils::reverseGraph(solver.V);

    auto V = solver.V;
    auto vs = solver.vs, vt = solver.vt;
    auto L = solver.L;

    node_hashes = createNodeHashes(V.size());
    node_hashes2 = createNodeHashes(V.size());


    constexpr bool make_directed = false; // #TEST #CAUTION
    if constexpr (make_directed) {
        clog << "MAKING GRAPH UNDIRECTED!!! CAUTION!" << endl << endl << "!!!!!!!!!!----" << endl << endl;
        auto V2 = V;
        for(int i=0; i<V.size(); i++){ for(int d : V[i]) V2[d].push_back(i);}
        V = V2;
        for(int i=0; i<V.size(); i++) StandardUtils::makeUnique(V[i]);
    }
    DEBUG(PII(vs,vt));
    DEBUG(L);
    DEBUG(GraphUtils::isDirected(V));
    DEBUG(GraphUtils::isConnected(V));


    { // check if graph is simple...
        auto & V = solver.V;
        int N = V.size();
        VB helper(N,false);

        for( int i=0; i<N; i++ ){
            for( int d : V[i] ){
                if( d == i || helper[d] ) {
                    clog << "loop " << i+1 << " -> " << i+1 << " or arc " << i+1 << " -> " << d+1 << " occurs twice" << endl;
                }
                helper[d] = true;
            }

            for(int d : V[i]) helper[d] = false;
        }
    }


    string option = "solver1.solveSmart";
    solver = Solver(V,L,vs,vt);
    sw.start( option );
    clog << "Starting " << option << endl;

    auto res_tab = solver.solveSmart();

    auto solver1_smart_res = accumulate(ALL(res_tab), (bigint)0 );
    clog << endl << option << " res: " << solver1_smart_res << endl;
    cout << solver1_smart_res << endl;
    sw.stop( option );

    clog << "Finished " << option << endl;


    sw.stop("main");
    sw.writeAll();


    return 0;
}
