//
// Created by sylwe on 04/09/2024.
//

#ifndef SOLVER2FBS_H
#define SOLVER2FBS_H

#include <Stopwatch.h>

#include "Config.h"
#include "Structures.h"

class Solver2FBS {
public:

    struct Edge {
        Edge( unsigned aa, unsigned bb, int ee1, int ee2 ) : a(aa), b(bb), e1(ee1), e2(ee2) {}
        unsigned int a,b;
        int e1,e2; // ids of arcs that need to be taken to make the transition
        // bool operator<(Edge & oth ) const{ return make_pair(a,b) < make_pair(oth.a, oth.b); }
        bool operator<(Edge & oth ) const { return ( a != oth.a ? a < oth.a : b < oth.b ); }
    };

    struct CompressedEdge {
        CompressedEdge( unsigned aa, unsigned bb ) : a(aa), b(bb) {}
        unsigned int a,b;
        VI arcs;
        // bool operator<(CompressedEdge & oth ) const{ return make_pair(a,b) < make_pair(oth.a, oth.b); }
        bool operator<(CompressedEdge & oth ) const { return ( a != oth.a ? a < oth.a : b < oth.b ); }
    };

    struct EdgeH {
        EdgeH( HashType aa, HashType bb, int ee1, int ee2 ) : a(aa), b(bb), e1(ee1), e2(ee2) {}
        HashType a,b;
        int e1,e2; // ids of arcs that need to be taken to make the transition
        // bool operator<(EdgeH & oth ) const{ return make_pair(a,b) < make_pair(oth.a, oth.b); }
        bool operator<(EdgeH & oth ) const { return ( a != oth.a ? a < oth.a : b < oth.b ); }
    };

    Solver2FBS( GraphSP2 _g, int L, int vs, int vt );

    int N; // size of the input graph
    int L; // max length of paths to consider
    int vs, vt;
    GraphSP2 g, revg;
    Config cnf;

    Stopwatch timer;

    // uses dp on implicit pathwidth/treewidth decomposition.
    VBIG solveFBS();
    pair<VI,int> createFBSNodeOrder();
    VI optimizeOrder(VI order);


    /**
     * Calculates and returns pathwidt for order on N nodes (component expansion method) and creates order of
     * 2N nodes in res_order that can be used for FBS
     */
    int calculatePathwidth(VI order, VI & res_order);

    VBIG solveFBSForNodeOrder( VI order );
    VBIG solveFBS2ForNodeOrder( VI order );

    /**
     * In this version we construct the transition graph and then calculate the result for the graph.
     * It may be much more memory-consuming than the previous two version, but should also be much faster
     * due to applied optimization techniques...
     */
    VBIG solveFBS3ForNodeOrder( VI order );


    // for directed symetric (lengths for arc a->b must be the same as for b->a) graphs
    VBIG solveFBSDForNodeOrder( VI order );


    template<class T>
    static void countSortTrgEdges( int trgN, vector<T> & trg_edges ) {
        VI cnt(trgN,0);
        for( auto &e : trg_edges ) cnt[e.a]++;
        for( int i=1; i<cnt.size(); i++ ) cnt[i] += cnt[i-1];
        auto cp = trg_edges;
        for( auto &e : trg_edges ) cp[ --cnt[e.a] ] = e;
        swap( trg_edges, cp );
        for( int i=1; i<trgN; i++ ) sort( trg_edges.begin() + cnt[i-1], trg_edges.begin() + cnt[i] );
        sort( trg_edges.begin() + cnt.back(), trg_edges.end() );
    }

    template<class T, int sort_type>
    static void countSortTrgEdgesHash( const int BITS, vector<T> & trg_edges ) {
        const int SHIFT = 64 - BITS;
        int trgN = (1 << BITS);
        VI cnt(trgN,0);

        if constexpr (sort_type == 0) for( auto &e : trg_edges ) cnt[e.a.hash >> SHIFT]++;
        else if constexpr (sort_type == 1) for( auto &e : trg_edges ) cnt[e.b.hash >> SHIFT]++;
        else for( auto &e : trg_edges ) cnt[e.hash >> SHIFT]++;

        for( int i=1; i<cnt.size(); i++ ) cnt[i] += cnt[i-1];
        auto cp = trg_edges;

        if constexpr (sort_type == 0) for( auto &e : trg_edges ) cp[ --cnt[e.a.hash >> SHIFT] ] = e;
        else if constexpr (sort_type == 1) for( auto &e : trg_edges ) cp[ --cnt[e.b.hash >> SHIFT] ] = e;
        else for( auto &e : trg_edges ) cp[ --cnt[e.hash >> SHIFT] ] = e;


        for( int i=1; i<trgN; i++ ) {
            if constexpr ( sort_type == 0 ) {
                sort( cp.begin() + cnt[i-1], cp.begin() + cnt[i], [](auto &a, auto &b){ return a.a < b.a; } );
            }
            else if constexpr ( sort_type == 1 ) { // sort by b
                sort( cp.begin() + cnt[i-1], cp.begin() + cnt[i], [](auto &a, auto &b){ return a.b < b.b; } );
            }
            else sort( cp.begin() + cnt[i-1], cp.begin() + cnt[i] ); // sort by default order given by <
        }
        if constexpr ( sort_type == 0 ) {
            sort( cp.begin() + cnt.back(), cp.end(), [](auto a, auto b){ return a.a < b.a; } );
        }
        else if constexpr ( sort_type == 1 ) {
            sort( cp.begin() + cnt.back(), cp.end(), [](auto a, auto b){ return a.b < b.b; } );
        }
        else sort( cp.begin() + cnt.back(), cp.end() );

        swap( trg_edges, cp );
    }

    static void unsignedIntegerRadixSort(vector<HashType> & A){
        alignas(64) unsigned cnt[256];
        const int N = A.size();
        auto B = A;

        for(int offset=0; offset < 8; offset++ ){
            memset( &cnt[0], 0, 256 * 4 );
            for( auto a : A ) cnt[ (a.hash >> (offset<<3) ) & 0xff ]++;
            for( int i=1; i<256; i++ ) cnt[i] += cnt[i-1];
            for( int i= N-1; i>=0; i-- ) B[ --cnt[(A[i].hash >> (offset<<3) ) & 0xff] ] = A[i];
            swap(A,B);
        }
    }

};


#endif //SOLVER2FBS_H
