//
// Created by sylwe on 04/09/2024.
//

#include "Solver2FBS.h"

#include <ConnectedComponents.h>
#include <FAU.h>
#include <GraphUtils.h>
#include <Stopwatch.h>

#include "BiconnectedFinder.h"
#include "Solver2.h"
#include "CollectionOperators.h"
#include "DistanceFinder2.h"
#include "FlowCutter.h"
#include "Kernelizer.h"
#include "Solver.h"
#include "Solver2FBSD.h"

Solver2FBS::Solver2FBS(GraphSP2 _g, int _L, int _vs, int _vt) : g(_g), L(_L), vs(_vs), vt(_vt) {
    N = g.size();
    revg = g.reverseGraph();
}


VBIG Solver2FBS::solveFBS() {
    constexpr bool debug = true;

    constexpr bool inflate_graph = false;
    if constexpr (inflate_graph) {
        GraphSP2 newg = g;
        for( int i=0; i<N; i++ ) newg[i].clear();
        int cnt = 0;
        constexpr int freq = 1; // transforms only 1 / freq part of all arcs

        map<PII,int> mapper;

        for( int i=0; i<N; i++ ) {
            for( auto & arc : g[i] ) {
                if( cnt++ % freq > 0 ) {
                    newg[i].push_back(arc);
                    continue;
                }

                int id;
                if( mapper.contains({arc.v,i}) ) id = mapper[{arc.v,i}];
                else {
                    newg.V.push_back( vector<SP2Arc>() );
                    id = newg.V.size()-1;
                    mapper[ {i,arc.v} ] = id;
                }

                auto newarc = arc;
                newarc.v = id;
                newg[i].push_back( newarc );

                newarc.lengths = SP2Arc::mul_one;
                newarc.v = arc.v;
                newg[id].push_back(newarc);
            }
        }

        swap(g, newg);
        revg = g.reverseGraph();
        N = g.size();
        if( node_hashes.size() < N ) {
            node_hashes = createNodeHashes(N);
            node_hashes2 = createNodeHashes(N);
        }
    }


    auto [order,pathwidth] = createFBSNodeOrder();
    // auto res = solveFBSForNodeOrder(order); // FBS
    // auto res = solveFBS2ForNodeOrder(order); // FBS2 - faster than FBS
    // auto res = solveFBSDForNodeOrder(order); // for directed, symmetric graphs only

    VBIG res;
    if( g.isSymmetric() ) {
        clog << endl << "Solving using FBS3D" << endl;
        Solver2FBSD s2fbsd(g,L,vs,vt);
        s2fbsd.timer.setLimit("fbs", timer.getLimit("fbs"));
        res = s2fbsd.solveFBS3DForNodeOrder(order);
    }else {
        clog << endl << "Solving using FBS3" << endl;
        timer.setLimit("fbs", timer.getLimit("fbs"));
        res = solveFBS3ForNodeOrder(order);
    }

    clog << "FBS result calculated and returned" << endl;
    return res;
}

pair<VI,int> Solver2FBS::createFBSNodeOrder() {
    constexpr bool debug = false;

    VVI V = g.getStruture();
    VVI revV = GraphUtils::reverseGraph(V);
    VVI underV = V;
    for(int i=0; i<N; i++) {
        underV[i] += revV[i];
        StandardUtils::makeUnique(underV[i]);
    }


    VI order;
    VI S;
    VI deg_out(N,0);
    VI deg_in_S(N,0);
    VI neigh_S_deg1(N,0);
    VB in_S(N);

    int pathwidth = 0;

    for( int i=0; i<N; i++ ) deg_out[i] = underV[i].size();

    // auto [dstvs, parvs] = Solver::findDistances(underV, vs);
    auto [dstvs, parvs] = Solver2::findDistances(g, vs);
    // auto [dstvt, parvt] = Solver::findDistances(underV, vt);
    auto [dstvt, parvt] = Solver2::findDistances(revg, vt);

    VB is_outside(N,true);

    auto cmp = [&](int a, int b) {
        if( neigh_S_deg1[a] != neigh_S_deg1[b] ) return neigh_S_deg1[a] > neigh_S_deg1[b];
        // else if( deg_in_S[a] - deg_out[a] != deg_in_S[b] - deg_out[b] ) {
        //     return deg_in_S[a] - deg_out[a] > deg_in_S[b] - deg_out[b];
        // }
        else if( deg_in_S[a] != deg_in_S[b] ) return deg_in_S[a] > deg_in_S[b];
        else if( deg_out[a] != deg_out[b] ) return deg_out[a] < deg_out[b];
        else if( dstvt[a] - dstvs[a] != dstvt[b] - dstvs[b] ) return dstvt[a] - dstvs[a] > dstvt[b] - dstvs[b];
        // else if( dstvt[a] != dstvt[b] ) return dstvt[a] > dstvt[b];
        // else if( dstvs[a] != dstvs[b] ) return dstvs[a] < dstvs[b];
        return a < b;
    };
    set<int, decltype(cmp)> left(cmp);
    for(int i=0; i<N; i++) left.insert(i);

    auto getNextNode = [&]() {
        if( order.empty() ) return vs;
        return *left.begin();
    };

    auto addNode = [&](int v) {
        if constexpr (debug) {
            clog << "Adding node " << v << endl;
            DEBUG(S);
        }
        S.push_back(v);
        pathwidth = max( pathwidth, (int)S.size() );
        in_S[v] = true;
        is_outside[v] = false;

        for( int d : underV[v] ) {
            if constexpr (debug) clog << "\td: " << d << endl;

            if( left.contains(d) ) {
                if constexpr (debug) clog << "\tupdating node in left, d: " << d << endl;
                left.erase(d);
                deg_out[d]--;
                deg_in_S[d]++;
                left.insert(d);
                assert(left.contains(d));
            }else if( in_S[d] ) {
                if constexpr (debug) clog << "\tupdating node in S, d: " << d << endl;
                deg_out[d]--;
                if( deg_out[d] == 1 ) {
                    for( int a : underV[d] ) {
                        if( left.contains(a) ) {
                            if constexpr (debug) clog << "\tupdating node a: " << a << endl;
                            left.erase(a);
                            neigh_S_deg1[a]++;
                            left.insert(a);
                            break;
                        }
                    }
                }
            }
            else {
                DEBUG(left.contains(d));
                for(auto x : left) DEBUG(x);
                assert(false && "node d must be either unvisited (in set left) or in S");
            }
        }


    };

    while( !left.empty() ) {
        if constexpr (debug) {
            for(int i=0; i<N; i++) {
                clog << i << " -> deg_in_S: " << deg_in_S[i] << " -> deg_out: " << deg_out[i] << " -> deg1_in_S: "
                    << neigh_S_deg1[i] << endl;
            }
        }
        int v = getNextNode();
        left.erase(v);
        order.push_back(v);

        addNode(v);


        for( int i=(int)S.size()-1; i>=0; i-- ) {
            int a = S[i];
            if( deg_out[a] == 0 ) {
                if constexpr (debug) clog << "Removing node " << a << " from S" << endl;
                swap(S[i], S.back());
                S.pop_back();
                in_S[a] = false;
                order.push_back(a);
            }
        }
    }

    assert( order.size() == 2*N );
    assert(order[0] == vs);

    clog << "Pathwidth for found order: " << pathwidth << endl;

    constexpr bool optimize_order = true;

    vector< pair<VI,int> > candidates = { {order, pathwidth} };

    if constexpr (optimize_order) {

        auto optimizeUsingFlowCutter = [&](VI order, int opt) {
            VVI expV = FlowCutter::getExpansionGraph(underV);

            FlowCutter fc;
            VI exp_ord = fc.getFlowCutterExpansionOrder(expV, {2*vs}, {2*vt});
            // DEBUG(exp_ord);
            auto borders = fc.getSourceTargetBorders();
            // clog << "borders: " << endl;
            // for( auto v : borders ) clog << v << endl;

            for( auto & v : borders ) {
                for( auto & a : v ) a >>= 1;
                StandardUtils::makeUnique(v);
            }

            VVI newb( 1, borders[0] );
            for( int i=1; i < borders.size(); i++ ) if( borders[i] != borders[i-1] ) newb.push_back( borders[i] );
            swap(borders, newb);

            if( opt == 0 ){} // do nothign
            else if( opt == 1 ) { // take only the largest border
                sort(ALL(borders), []( auto & v1, auto & v2 ){ return v1.size() < v2.size(); });
                borders = { borders.back() };
            }else if( opt == 2 ) { // take the middle border
                borders = { borders[ borders.size()/2 ] };
            }

            clog << "updated borders: " << endl;
            for( auto v : borders ) clog << v << endl;

            auto rearrange = [&]( auto & border ) {
                VB was(N);
                int cnt = 0;
                for( int d : border ) { was[d] = true; cnt++; }
                if(was[vs]) return;

                auto [dst,par] = Solver::findDistances(underV, vs, was );
                for( int i=0; i<N; i++ ) if( dst[i] < inf ){ was[i] = true; cnt++; }
                // clog << "There are " << cnt << " nodes reachable from vs to border " << border << endl;

                VI new_order; new_order.reserve(order.size());
                for( int b : order ) if( was[b] ) new_order.push_back(b);
                for( int b : order ) if( !was[b] ) new_order.push_back(b);

                // DEBUG(order);
                // DEBUG(new_order);
                // ENDL(1);

                swap(order, new_order);
            };

            for( auto v : borders ) rearrange(v);

            return order;
        };

        { // removing second occurrences of nodes in order
            VI new_order;
            VB vis(N,false);
            for( int d : order ) {
                if(!vis[d]) {
                    vis[d] = true;
                    new_order.push_back(d);
                }
            }
            swap(order, new_order);
            // DEBUG(order);
            assert(pathwidth == calculatePathwidth(order, new_order));
            // DEBUG(order);
        }

        // DEBUG(order.size());
        VI optimized_order = optimizeOrder(order);
        // DEBUG(optimized_order);
        // DEBUG(optimized_order.size());
        pathwidth = calculatePathwidth(optimized_order, order);
        if( order.size() != 2*N ) {
            DEBUG(optimized_order.size());
            DEBUG(N);
            DEBUG(order.size());
            DEBUG(order);
            DEBUG( set(ALL(order)) );
            assert(order.size() == 2*N);
        }
        clog << "Pathwidth for CE-optimized order: " << pathwidth << endl;
        candidates.emplace_back( order, pathwidth );

        constexpr bool use_flowcutter = false;

        if constexpr (use_flowcutter) {
            if( g.size() > 20 ) {
                for( int opt = 0; opt<3; opt++ ) {
                    auto opt_ord = optimized_order;
                    auto order2 = optimizeUsingFlowCutter(opt_ord, opt);
                    opt_ord = optimizeOrder(order2);
                    pathwidth = calculatePathwidth(opt_ord,order2);
                    clog << "Pathwidth for CE+FC+CE-optimized (" << opt << ") order: " << pathwidth << endl;
                    DEBUG(opt_ord.size());
                    candidates.emplace_back( order2, pathwidth );
                }
            }
        }

        // optimized_order = optimizeOrder(order);
        // pathwidth = calculatePathwidth(optimized_order, order);
        // clog << "Pathwidth for CE+FC+CE-optimized order: " << pathwidth << endl;
        // candidates.emplace_back( order, pathwidth );

    }

    for( auto & [v,pw] : candidates ) assert( v.size() == 2*N );

    clog << "Decomposition candidates' pathwidths: ";
    for( auto & [v,pw] : candidates ) clog << "(" << v.size() << ", " << pw << ") "; clog << endl;

    // use the following for FlowCutter version
    // sort( ALL(candidates), [](auto & c1, auto & c2){ return c1.second < c2.second; } );
    // clog << "Best pathwidth: " << candidates[0].second << endl;
    // return candidates[0];

    // return candidates[1];

    if( candidates[0].second < candidates[1].second ) return candidates[0]; // return first optimized order...
    else return candidates[1]; // if pathwidths are equal, then return optimized order
}

VI Solver2FBS::optimizeOrder(VI order) {
    bool debug = false;

    VVI V = g.getStruture();
    V = GraphUtils::getUndirectedStructure(V, VI());
    int N = V.size();

    VI new_order;
    {
        VB vis(N,false);
        for( int d : order ) {
            if(!vis[d]) {
                vis[d] = true;
                new_order.push_back(d);
            }
        }
        swap(order, new_order);
    }

    int firstInOrder = order[0];
    assert( order.size() == N );

    VVI traverseGraph(N);
    VI edgesInComp(N,0);
    VI firstInComp(N,0); // first in comp[ fau.Find(i) ] is the first node in @{order} in component that contains i, in given phase of the algorithm
    VI inOrder(N,0);
    for( int i=0; i<order.size(); i++ ) inOrder[ order[i] ] = i;

    FAU fau(N);

    if(debug)DEBUG(order);

    while( !order.empty() ){
        int p = order.back();
        order.pop_back();

        if(debug)cerr << "Adding " << p << endl;

        set<int> neigh;
        int edges = 0;
        for( int d : V[p] ){
            if( inOrder[d] > inOrder[p] ){
                int t = fau.Find(d);
                neigh.insert( t );
                edges++;
            }
        }

        if(debug) {cerr << "neigh: "; for(int t : neigh) cerr << t << " "; cerr << endl;}
        for( int d : neigh ) edges += edgesInComp[d];
        if(debug)DEBUG(edges);

        VI neighOrder(ALL(neigh));
        sort( ALL(neighOrder), [&edgesInComp]( int a, int b ){ return edgesInComp[a] < edgesInComp[b]; } );

        if(debug){
            DEBUG(neighOrder);
            cerr << "first in orders: "; for(int t : neighOrder) cerr << firstInComp[t] << " "; cerr << endl;
        }

        for( int d : neighOrder ) traverseGraph[p].push_back( firstInComp[d] );
        for( int d : neighOrder ) fau.Union(p,d);
        firstInComp[ fau.Find(p) ] = p;
        edgesInComp[ fau.Find(p) ] = edges;

        if(debug)ENDL(1);
    }

    if(debug) DEBUG(traverseGraph);

    VI newOrder;

    function< void(int) > restoreNewOrder = [&traverseGraph, &newOrder, &restoreNewOrder](int v){
        for( int d : traverseGraph[v] ){
            newOrder.push_back(d);
            restoreNewOrder(d);
        }
    };

    newOrder.push_back( firstInOrder );
    restoreNewOrder( firstInOrder );

    if(debug) {
        DEBUG(newOrder);
        // exit(1);
    }

    return newOrder;
}

int Solver2FBS::calculatePathwidth(VI order, VI & res_order) { // this is almost a copy-paste from createFBSNodeOrder...
    constexpr bool debug = false;

    res_order.clear();
    if constexpr (debug) {
        clog << "CAlculating pathwidth for order: " << order << endl;
    }

    VVI V = g.getStruture();
    VVI revV = GraphUtils::reverseGraph(V);
    VVI underV = V;
    for(int i=0; i<N; i++) {
        underV[i] += revV[i];
        StandardUtils::makeUnique(underV[i]);
    }


    VI S;
    VI deg_out(N,0);
    VI deg_in_S(N,0);
    VI neigh_S_deg1(N,0);
    VB in_S(N);

    int pathwidth = 0;

    for( int i=0; i<N; i++ ) deg_out[i] = underV[i].size();

    // auto [dstvs, parvs] = Solver::findDistances(underV, vs);
    auto [dstvs, parvs] = Solver2::findDistances(g, vs);
    // auto [dstvt, parvt] = Solver::findDistances(underV, vt);
    auto [dstvt, parvt] = Solver2::findDistances(revg, vt);

    VB is_outside(N,true);

    auto cmp = [&](int a, int b) {
        if( neigh_S_deg1[a] != neigh_S_deg1[b] ) return neigh_S_deg1[a] > neigh_S_deg1[b];
        else if( deg_in_S[a] != deg_in_S[b] ) return deg_in_S[a] > deg_in_S[b];
        else if( deg_out[a] != deg_out[b] ) return deg_out[a] < deg_out[b];
        else if( dstvt[a] - dstvs[a] != dstvt[b] - dstvs[b] ) return dstvt[a] - dstvs[a] > dstvt[b] - dstvs[b];
        return a < b;
    };
    set<int, decltype(cmp)> left(cmp);
    for(int i=0; i<N; i++) left.insert(i);

    int next_ind = 0;
    auto getNextNode = [&]() {
        return order[next_ind++];
    };

    auto addNode = [&](int v) {
        if constexpr (debug) {
            clog << "Adding node " << v << endl;
            DEBUG(S);
        }
        S.push_back(v);
        pathwidth = max( pathwidth, (int)S.size() );
        in_S[v] = true;
        is_outside[v] = false;

        for( int d : underV[v] ) {
            if constexpr (debug) clog << "\td: " << d << endl;

            if( left.contains(d) ) {
                if constexpr (debug) clog << "\tupdating node in left, d: " << d << endl;
                left.erase(d);
                deg_out[d]--;
                deg_in_S[d]++;
                left.insert(d);
                assert(left.contains(d));
            }else if( in_S[d] ) {
                if constexpr (debug) clog << "\tupdating node in S, d: " << d << endl;
                deg_out[d]--;
                if( deg_out[d] == 1 ) {
                    for( int a : underV[d] ) {
                        if( left.contains(a) ) {
                            if constexpr (debug) clog << "\tupdating node a: " << a << endl;
                            left.erase(a);
                            neigh_S_deg1[a]++;
                            left.insert(a);
                            break;
                        }
                    }
                }
            }
            else {
                DEBUG(left.contains(d));
                for(auto x : left) DEBUG(x);
                assert(false && "node d must be either unvisited (in set left) or in S");
            }
        }
    };

    VB vis(N,false);

    while( next_ind < order.size() ) {
        if constexpr (debug) {
            for(int i=0; i<N; i++) {
                clog << i << " -> deg_in_S: " << deg_in_S[i] << " -> deg_out: " << deg_out[i] << " -> deg1_in_S: "
                    << neigh_S_deg1[i] << endl;
            }
        }
        int v = getNextNode();
        // while( next_ind < order.size() && vis[v] ) {
        //     // res_order.push_back(v);
        //     v = getNextNode();
        //     assert(false);
        // }
        // if( next_ind == order.size() ) break;

        res_order.push_back(v);
        vis[v] = true;
        left.erase(v);

        if constexpr (debug) {
            clog << endl << "left nodes: ";
            for(int d : left) clog << d << " ";
            clog << endl;
        }

        addNode(v);

        for( int i=(int)S.size()-1; i>=0; i-- ) {
            int a = S[i];
            if( deg_out[a] == 0 ) {
                if constexpr (debug) clog << "Removing node " << a << " from S" << endl;
                swap(S[i], S.back());
                S.pop_back();
                in_S[a] = false;

                res_order.push_back(a);
            }
        }
    }

    return pathwidth;
}

VBIG Solver2FBS::solveFBSForNodeOrder(VI order) {
    constexpr bool debug = false;

    if constexpr (debug) {
        clog << "Entering solveFBSForNodeOrder" << endl;
        DEBUG(g);
        DEBUG(order);
    }

    VB was(N);
    VB in_neigh(N), out_neigh(N);
    set<int> S; // nodes on boundary - needed only for debugging

    auto markNeighbors = [&](int v, const bool val) {
        for( auto & arc : g[v] ) if(was[arc.v] && S.contains(arc.v)) out_neigh[arc.v] = val;
        for( auto & arc : revg[v] ) if(was[arc.v] && S.contains(arc.v)) in_neigh[arc.v] = val;
    };


    VVI distances(N);
    auto createDistances = [&](int v) {
        for(int i=0; i<N; i++) VI().swap(distances[i]);
        S.insert(v);
        for( int d : S ) {
            was[d] = false;
            distances[d] = Solver2::findDistancesInclusive(g, d, was).first;
            was[d] = true;
        }
        was[v] = false;
        S.erase(v);
    };


    constexpr int BUCKETS = (1<<10);
    constexpr int BH = BUCKETS-1;

    vector<vector<FBSS>> states(BUCKETS);
    vector<vector<FBSS>> new_states(BUCKETS);

    auto countAllStates = [&]() {
        return accumulate(ALL(states), 0, [](int a, auto & vec){ return a + vec.size(); });
    };

    bool use_fbs_for_directed_graph = false; // currently does not work for [true]
    // DEBUG(is_directed_graph);

    auto placeIntoNewBucket = [&]( FBSS & st ) {
        // if(!st.isCorrectState()) { // #TEST #CAUTION!! This is time-consuming as hell here! remove that after tests...
        //     clog << endl << "!!!!" << endl << "State is incorrect!" << endl;
        //     DEBUG(st);
        //     assert( st.isCorrectState() );
        // };
        if( use_fbs_for_directed_graph ) {
            for( auto & [a,b] : st.ends ) if( a > b ) swap(a,b);
        }

        st.recalculateHash();
        st.counts.trimToMax(L);
        new_states[st.hash.hash & BH].push_back( st );
        if constexpr (debug) clog << "\t\tCreated and added new_state: " << st << endl;
    };

    auto addAsIgnoredNode = [&]( int v, const FBSS & st ) {
        if constexpr (debug) clog << "\tAdding as ignored node" << endl;
        auto new_st = st;
        placeIntoNewBucket(new_st);
    };

    auto addAsLooseNode = [&](int v, const FBSS & st) {
        if constexpr (debug) clog << "\tAdding as loose node" << endl;
        auto new_st = st;
        new_st.loose.push_back(v);
        placeIntoNewBucket(new_st);
    };

    auto addAsVsNode = [&](int v, const FBSS & st) {
        if constexpr (debug) clog << "\tAdding as vs_node" << endl;
        if( !in_neigh[st.vs_node] ) return;

        auto new_st = st;
        new_st.vs_node = v;
        new_st.counts *= g.getArc(st.vs_node, v).lengths;
        placeIntoNewBucket(new_st);
    };

    auto addAsStartNode = [&](int v, const FBSS & st) {
        { // adding as start node to an existing loose node
            if constexpr (debug) clog << "\tAdding as start node: v -> loose" << endl;
            for( int a : st.loose ) {
                if( !out_neigh[a] ) continue;

                auto new_st = st;
                new_st.removeLooseNode(a);
                new_st.ends.emplace_back( v, a );
                new_st.counts *= g.getArc( v, a ).lengths;
                placeIntoNewBucket(new_st);
            }
        }

        { // adding as start node to an existing path end
            if constexpr (debug) clog << "\tAdding as start node: v -> ends" << endl;
            for( int i=0; i<st.ends.size(); i++ ) {
                int a = st.ends[i].first;
                if( !out_neigh[a] ) continue;

                auto new_st = st;
                new_st.ends[i].first = v;
                new_st.counts *= g.getArc( v,a ).lengths;
                placeIntoNewBucket(new_st);
            }
        }

    };

    auto addAsMiddleNode = [&](int v, const FBSS & st) {
        { // adding as middle node, vs_node -> v -> loose
            if constexpr (debug) clog << "\tAdding as middle node : vs_node -> v -> loose" << endl;
            for( int a : st.loose ) {
                if( !in_neigh[st.vs_node] ) continue;
                if( !out_neigh[a] ) continue;

                auto new_st = st;
                new_st.vs_node = a;
                new_st.removeLooseNode(a);
                new_st.counts *= g.getArc(st.vs_node, v).lengths;
                new_st.counts *= g.getArc(v, a).lengths;
                placeIntoNewBucket(new_st);
            }

        }

        { // adding as middle node, vs_node -> v -> path end
            if constexpr (debug) clog << "\tAdding as middle node : vs_node -> v -> ends" << endl;
            for( auto [a,b] : st.ends ) {
                if( !in_neigh[st.vs_node] ) continue;
                if( !out_neigh[a] ) continue;
                auto new_st = st;
                new_st.vs_node = b;
                new_st.removeEnds({a,b});
                new_st.counts *= g.getArc(st.vs_node, v).lengths;
                new_st.counts *= g.getArc(v, a).lengths;
                placeIntoNewBucket(new_st);
            }
        }

        { // adding as middle node, loose -> v -> loose
            if constexpr (debug) clog << "\tAdding as middle node : loose -> v -> loose" << endl;
            for( int a : st.loose ) {
                if( !in_neigh[a] ) continue;

                for( int b : st.loose ) {
                    if( a == b ) continue;
                    if( !out_neigh[b] ) continue;

                    auto new_st = st;
                    new_st.removeLooseNode(a);
                    new_st.removeLooseNode(b);
                    new_st.ends.emplace_back(a,b);
                    new_st.counts *= g.getArc(a, v).lengths;
                    new_st.counts *= g.getArc(v, b).lengths;
                    placeIntoNewBucket(new_st);
                }
            }
        }

        { // adding as middle node, loose -> v -> path end
            if constexpr (debug) clog << "\tAdding as middle node : loose -> v -> ends" << endl;
            for( int a : st.loose ) {
                if( !in_neigh[a] ) continue;

                for( auto [b,c] : st.ends ) {
                    if( !out_neigh[b] ) continue;

                    auto new_st = st;
                    new_st.removeLooseNode(a);
                    new_st.removeEnds( {b,c} );
                    new_st.ends.emplace_back(a,c);
                    new_st.counts *= g.getArc(a, v).lengths;
                    new_st.counts *= g.getArc(v, b).lengths;
                    placeIntoNewBucket(new_st);
                }
            }
        }

        { // adding as middle node, path_end -> v -> loose
            if constexpr (debug) clog << "\tAdding as middle node : ends -> v -> loose" << endl;
            for( auto [a,b] : st.ends ) {
                if( !in_neigh[b] ) continue;

                for( int c : st.loose ) {
                    if( !out_neigh[c] ) continue;

                    auto new_st = st;
                    new_st.removeLooseNode(c);
                    new_st.removeEnds({a,b});
                    new_st.ends.emplace_back(a,c);
                    new_st.counts *= g.getArc(b, v).lengths;
                    new_st.counts *= g.getArc(v, c).lengths;
                    placeIntoNewBucket(new_st);
                }
            }
        }

        { // adding as middle node, path_end -> v -> path_end
            if constexpr (debug) clog << "\tAdding as middle node : ends -> v -> ends" << endl;
            for( auto [a,b] : st.ends ) {
                if( !in_neigh[b] ) continue;

                for( auto & [c,d] : st.ends ) {
                    if( a == c || b == d ) continue;
                    if( !out_neigh[c] ) continue;

                    auto new_st = st;
                    new_st.removeEnds({a,b});
                    new_st.removeEnds({c,d});
                    new_st.ends.emplace_back(a,d);
                    new_st.counts *= g.getArc(b, v).lengths;
                    new_st.counts *= g.getArc(v, c).lengths;
                    placeIntoNewBucket(new_st);
                }
            }
        }
    };

    auto addAsEndNode = [&](int v, const FBSS & st) {
        { // adding as end node to an existing loose node
            if constexpr (debug) clog << "\tAdding as end node : loose -> v" << endl;
            for( int a : st.loose ) {
                if( !in_neigh[a] ) continue;
                auto new_st = st;
                new_st.removeLooseNode(a);
                new_st.ends.emplace_back( a,v );
                new_st.counts *= g.getArc( a,v ).lengths;
                placeIntoNewBucket(new_st);
            }
        }

        { // adding as end node to an existing path end
            if constexpr (debug) clog << "\tAdding as end node : ends -> v" << endl;
            for( int i=0; i<st.ends.size(); i++ ) {
                int b = st.ends[i].second;
                if( !in_neigh[b] ) continue;

                auto new_st = st;
                new_st.ends[i].second = v;
                new_st.counts *= g.getArc( b,v ).lengths;
                placeIntoNewBucket(new_st);
            }
        }
    };


    auto mergeFBSStates = [&]( vector<FBSS> & sts ) {
        sort(ALL(sts));

        int p = 0, q = 0, r = 0;
        while( p < sts.size() ){
            q = p;
            auto st1 = sts[p], st2 = sts[q];
            while( q < sts.size() && (st1 == st2) ) if(++q <sts.size()) st2 = sts[q];
            while( p+1 < q ) {
                st1.counts += sts[++p].counts;
                if constexpr (debug) clog << "------>" << endl << "merging state " << sts[p] << " to " << st1 << endl;
            }
            sts[r++] = st1;
            p = q;
        }

        while( sts.size() > r ) sts.pop_back();
    };

    auto getMstEdges = [&](FBSS& st) {
        vector<Triple<int>> edges;
        edges.reserve( (S.size()+2) * (S.size()+2) );

        if constexpr (debug)
            clog << endl << "Finding MST edges for state " << st << endl;

        auto checkPair = [&](int a, int b) {
            // if(use_fb_for_directed_graph && a > b) return;
            if constexpr (debug)
                clog << "Considering pair (" << a << ", "  << b << ")" << endl;
            if( distances[a][b] != inf ) {
                edges.emplace_back( distances[a][b], a, b );
            }
        };

        checkPair(st.vs_node, vt);

        for( int a : st.loose ) {
            checkPair(st.vs_node, a);
            checkPair(a,vt);

            for( int b : st.loose ) {
                if(a == b) continue;
                checkPair(a,b);
            }

            for( auto [c,d] : st.ends ) {
                checkPair(a,c);
                checkPair(d,a);
            }
        }

        for( auto [a,b] : st.ends ) {
            checkPair(st.vs_node,a);
            checkPair(b,vt);

            for( auto [c,d] : st.ends ) {
                if( a == c || a == d || b == c || b == d ) continue;
                checkPair(b,c);
            }
        }
        if constexpr (debug) {
            clog << "edges: " << endl;
            for(auto [w,a,b] : edges) clog << "\t(" << w << "," << a << "," << b << ")" << endl;
            DEBUG(distances[st.vs_node]);
        }

        // now find mst
        FAU fau(N);
        for( auto [a,b] : st.ends ) fau.Union(a,b);
        sort(ALL(edges), []( auto & a, auto & b ){ return a.st < b.st; });
        vector<Triple<int>> res;
        for( auto & [w,a,b] : edges ) {
            if( fau.Find(a) != fau.Find(b) ) {
                if constexpr (debug)
                    clog << "\tuniting nodes " << a << " and " << b << endl;
                res.emplace_back(w,a,b);
                fau.Union(a,b);
            }
        }

        bool is_connected_mst = ( fau.Find(st.vs_node) == fau.Find(vt) );
        for( int a : st.loose ) {
            if( fau.Find(a) != fau.Find(st.vs_node) ) is_connected_mst = false;
        }
        for( auto [a,b] : st.ends ) if( fau.Find(a) != fau.Find(st.vs_node) ) is_connected_mst = false;

        return make_pair(is_connected_mst,res);
    };

    auto canTrimMST = [&](FBSS & st) {
        auto [is_connected,edges] = getMstEdges(st);

        if constexpr (debug) if( !is_connected ) {
            clog << "MST created for state st: " << st << " is not connected, trimming!" << endl;
            clog << "edges: " << endl;
            for(auto [w,a,b] : edges) clog << "\t(" << w << "," << a << "," << b << ")" << endl;
        }

        if( !is_connected ) return true;

        int sumw = accumulate( ALL(edges), 0, [](int a, auto & e){ return a + e.st; } );
        if( st.counts.minl() + sumw > L ) {
            if constexpr (debug) {
                clog << endl << "Length conditon violated for st: " << st << " and edges: " << endl;
                for(auto [w,a,b] : edges) clog << "\t(" << w << "," << a << "," << b << ")" << endl;
            }
            return true;
        }
        return false;
    };

    auto trimFBSStates = [&](vector<FBSS> & sts) {
        for( int i=(int)sts.size()-1; i>=0; i-- ) {
            bool can_trim = (sts[i].counts.minl() > L);
            if(!can_trim) can_trim = canTrimMST(sts[i]);

            if( can_trim ) {
                swap( sts[i], sts.back() );
                sts.pop_back();
            }
        }
    };

    auto addNodeForStates = [&](int v, vector<FBSS> & sts) {
        markNeighbors(v, true); // mark in- and out-neighbors of node v within boundary set S

        for( auto & st : sts ) {
            addAsIgnoredNode(v,st);
            addAsLooseNode(v,st);
            addAsVsNode(v,st);
            addAsStartNode(v,st);
            addAsMiddleNode(v,st);
            addAsEndNode(v,st);
        }

        markNeighbors(v,false); // unmark marked nodes...
    };

    auto removeFBSStates = [&](int v, vector<FBSS> & sts) {
        for( int i=(int)sts.size()-1; i>=0; i-- ) {
            if( sts[i].containsNode(v) ) {
                if constexpr (debug) clog << "\t-x-x-x-x-x removing state " << sts[i] << endl;
                swap( sts[i], sts.back() );
                sts.pop_back();
            }
        }
    };

    auto addNode = [&](int v) {
        // if constexpr (debug)
            clog << "Adding node v: " << v << endl;
        for( int i=0; i<BUCKETS; i++ ) new_states[i].clear();
        for( int i=0; i<BUCKETS; i++ ) addNodeForStates( v, states[i] );
        for( int i=0; i<BUCKETS; i++ ) {
            mergeFBSStates( new_states[i] );
            trimFBSStates( new_states[i] );
        }
        swap(states, new_states);
    };



    auto removeNode = [&](int v) {
        // if constexpr (debug)
            clog << "Removing node v: " << v << endl;
        for( int i=0; i<BUCKETS; i++ ) removeFBSStates( v, states[i] );
    };

    FBSS initial_state( Lengths(VBIG(1,1),0), {}, {}, vs );
    states[0].push_back( initial_state );
    was[vs] = true;
    S.insert(vs);
    order.erase(order.begin());

    for( int v : order ) {
        if( v == vt && was[v] ) continue;

        clog << "Current boundary nodes: " << S.size() << " -> " << S << endl;
        DEBUG(countAllStates());

        if constexpr (debug) {
            ENDL(5); ENDLS(50, "*"); ENDL(5);
            clog << "Current states: " << endl;
            for(auto & vec : states) for( auto & st : vec ) clog << st << endl;
            ENDL(2);
        }

        if( was[v] ) {
            removeNode(v);
            S.erase(v);
        }
        else{
            createDistances(v);
            addNode(v);
            was[v] = true;
            S.insert(v);
        }
    }

    vector<FBSS> all_states;
    for(auto & vec : states) all_states += vec;

    if constexpr (debug) {
        clog << "All states: " << endl;
        for(auto & st : all_states) clog << st << endl;
    }

    assert( all_states.size() == 1 ); // there should be only one state - the one with node vt
    assert( all_states[0].vs_node == vt );
    assert( all_states[0].loose.empty());
    assert( all_states[0].ends.empty());

    VBIG res(L+1);
    auto & lengths = all_states[0].counts;
    for( int i=lengths.minl(); i <= min( L, lengths.maxl() ); i++ ) res[i] += lengths[i];
    return res;
}


VBIG Solver2FBS::solveFBS2ForNodeOrder(VI order) {
    constexpr bool debug = false;

    if constexpr (debug) {
        clog << "Entering solveFBSForNodeOrder" << endl;
        DEBUG(g);
        DEBUG(order);
    }

    Stopwatch sw;
    sw.start("solveFBS2ForNodeOrder");

    VB was(N);
    VB in_neigh(N), out_neigh(N);
    set<int> S; // nodes on boundary - needed only for debugging
    VI S_nodes;
    VB is_state_end(N,false);

    auto markNeighbors = [&](int v, const bool val) {
        for( auto & arc : g[v] ) if(was[arc.v] && S.contains(arc.v)) out_neigh[arc.v] = val;
        for( auto & arc : revg[v] ) if(was[arc.v] && S.contains(arc.v)) in_neigh[arc.v] = val;
    };


    VVI distances(N);
    auto createDistances = [&](int v) {
        for(int i=0; i<N; i++) VI().swap(distances[i]);
        S.insert(v);

        // for( int d : S ) {
        //     was[d] = false;
        //     distances[d] = findDistancesInclusive(g, d, was).first;
        //     was[d] = true;
        // }
        // was[v] = false;

        for(int d : S) was[d] = false;
        for(int d : S) distances[d] = Solver2::findDistancesInclusive(g, d, was).first;
        for(int d : S) was[d] = true;
        was[v] = false;

        S.erase(v);
    };

    auto getLooseNodes = [&]( FBSS2 & st ) {
        VI loose;
        is_state_end[st.vs_node] = true;
        for(auto [a,b] : st.ends) is_state_end[a] = is_state_end[b] = true;
        for(int a : st.taken) is_state_end[a] = true;
        for( int d : S_nodes ) if( !is_state_end[d] ) loose.push_back(d);
        is_state_end[st.vs_node] = false;
        for(auto [a,b] : st.ends) is_state_end[a] = is_state_end[b] = false;
        for(int a : st.taken) is_state_end[a] = false;
        return loose;
    };


    constexpr int BUCKETS = (1<<10);
    constexpr int BH = BUCKETS-1;

    vector<vector<FBSS2>> states(BUCKETS);
    vector<vector<FBSS2>> new_states(BUCKETS);

    auto countAllStates = [&](auto & states) {
        return accumulate(ALL(states), 0, [](int a, auto & vec){ return a + vec.size(); });
    };

    constexpr bool use_fbs_for_directed_graph = false; // currently does not work for [true]


    auto placeIntoNewBucket = [&]( FBSS2 & st ) {
        // if(!st.isCorrectState()) { // #TEST #CAUTION!! This is time-consuming as hell here! remove that after tests...
        //     clog << endl << "!!!!" << endl << "State is incorrect!" << endl;
        //     DEBUG(st);
        //     assert( st.isCorrectState() );
        // };
        if( use_fbs_for_directed_graph ) {
            for( auto & [a,b] : st.ends ) if( a > b ) swap(a,b);
        }

        st.recalculateHash();
        st.counts.trimToMax(L);
        new_states[st.hash.hash & BH].push_back( st );
        if constexpr (debug) clog << "\t\tCreated and added new_state: " << st << endl;
    };

    auto addAsIgnoredNode = [&]( int v, const FBSS2 & st, VI & loose ) {
        if constexpr (debug) clog << "\tAdding as ignored node" << endl;
        auto new_st = st;
        placeIntoNewBucket(new_st);
    };

    auto addAsVsNode = [&](int v, const FBSS2 & st, VI & loose) {
        if constexpr (debug) clog << "\tAdding as vs_node" << endl;
        if( !in_neigh[st.vs_node] ) return;

        auto new_st = st;
        new_st.vs_node = v;
        new_st.taken.push_back(st.vs_node);
        new_st.counts *= g.getArc(st.vs_node, v).lengths;
        placeIntoNewBucket(new_st);
    };

    auto addAsStartNode = [&](int v, const FBSS2 & st, VI & loose) {
        { // adding as start node to an existing loose node
            if constexpr (debug) clog << "\tAdding as start node: v -> loose" << endl;
            for( int a : loose ) {
                if( !out_neigh[a] ) continue;

                auto new_st = st;
                new_st.ends.emplace_back( v, a );
                new_st.counts *= g.getArc( v, a ).lengths;
                placeIntoNewBucket(new_st);
            }
        }

        { // adding as start node to an existing path end
            if constexpr (debug) clog << "\tAdding as start node: v -> ends" << endl;
            for( int i=0; i<st.ends.size(); i++ ) {
                int a = st.ends[i].first;
                if( !out_neigh[a] ) continue;

                auto new_st = st;
                new_st.ends[i].first = v;
                new_st.taken.push_back(a);
                new_st.counts *= g.getArc( v,a ).lengths;
                placeIntoNewBucket(new_st);
            }
        }

    };

    auto addAsMiddleNode = [&](int v, const FBSS2 & st, VI & loose) {
        { // adding as middle node, vs_node -> v -> loose
            if constexpr (debug) clog << "\tAdding as middle node : vs_node -> v -> loose" << endl;
            for( int a : loose ) {
                if( !in_neigh[st.vs_node] ) continue;
                if( !out_neigh[a] ) continue;

                auto new_st = st;
                new_st.vs_node = a;
                new_st.taken.push_back(st.vs_node);
                new_st.taken.push_back(v);
                new_st.counts *= g.getArc(st.vs_node, v).lengths;
                new_st.counts *= g.getArc(v, a).lengths;
                placeIntoNewBucket(new_st);
            }

        }

        { // adding as middle node, vs_node -> v -> path end
            if constexpr (debug) clog << "\tAdding as middle node : vs_node -> v -> ends" << endl;
            for( auto [a,b] : st.ends ) {
                if( !in_neigh[st.vs_node] ) continue;
                if( !out_neigh[a] ) continue;
                auto new_st = st;
                new_st.vs_node = b;
                new_st.removeEnds({a,b});
                new_st.taken.push_back(st.vs_node);
                new_st.taken.push_back(v);
                new_st.taken.push_back(a);
                new_st.counts *= g.getArc(st.vs_node, v).lengths;
                new_st.counts *= g.getArc(v, a).lengths;
                placeIntoNewBucket(new_st);
            }
        }

        { // adding as middle node, loose -> v -> loose
            if constexpr (debug) clog << "\tAdding as middle node : loose -> v -> loose" << endl;
            for( int a : loose ) {
                if( !in_neigh[a] ) continue;

                for( int b : loose ) {
                    if( a == b ) continue;
                    if( !out_neigh[b] ) continue;

                    auto new_st = st;
                    new_st.ends.emplace_back(a,b);
                    new_st.taken.push_back(v);
                    new_st.counts *= g.getArc(a, v).lengths;
                    new_st.counts *= g.getArc(v, b).lengths;
                    placeIntoNewBucket(new_st);
                }
            }
        }

        { // adding as middle node, loose -> v -> path end
            if constexpr (debug) clog << "\tAdding as middle node : loose -> v -> ends" << endl;
            for( int a : loose ) {
                if( !in_neigh[a] ) continue;

                for( auto [b,c] : st.ends ) {
                    if( !out_neigh[b] ) continue;

                    auto new_st = st;
                    new_st.removeEnds( {b,c} );
                    new_st.ends.emplace_back(a,c);
                    new_st.taken.push_back(v);
                    new_st.taken.push_back(b);
                    new_st.counts *= g.getArc(a, v).lengths;
                    new_st.counts *= g.getArc(v, b).lengths;
                    placeIntoNewBucket(new_st);
                }
            }
        }

        { // adding as middle node, path_end -> v -> loose
            if constexpr (debug) clog << "\tAdding as middle node : ends -> v -> loose" << endl;
            for( auto [a,b] : st.ends ) {
                if( !in_neigh[b] ) continue;

                for( int c : loose ) {
                    if( !out_neigh[c] ) continue;

                    auto new_st = st;
                    new_st.removeEnds({a,b});
                    new_st.ends.emplace_back(a,c);
                    new_st.taken.push_back(b);
                    new_st.taken.push_back(v);
                    new_st.counts *= g.getArc(b, v).lengths;
                    new_st.counts *= g.getArc(v, c).lengths;
                    placeIntoNewBucket(new_st);
                }
            }
        }

        { // adding as middle node, path_end -> v -> path_end
            if constexpr (debug) clog << "\tAdding as middle node : ends -> v -> ends" << endl;
            for( auto [a,b] : st.ends ) {
                if( !in_neigh[b] ) continue;

                for( auto & [c,d] : st.ends ) {
                    if( a == c || b == d ) continue;
                    if( !out_neigh[c] ) continue;

                    auto new_st = st;
                    new_st.removeEnds({a,b});
                    new_st.removeEnds({c,d});
                    new_st.ends.emplace_back(a,d);
                    new_st.taken.push_back(b);
                    new_st.taken.push_back(v);
                    new_st.taken.push_back(c);
                    new_st.counts *= g.getArc(b, v).lengths;
                    new_st.counts *= g.getArc(v, c).lengths;
                    placeIntoNewBucket(new_st);
                }
            }
        }
    };

    auto addAsEndNode = [&](int v, const FBSS2 & st, VI & loose) {
        { // adding as end node to an existing loose node
            if constexpr (debug) clog << "\tAdding as end node : loose -> v" << endl;
            for( int a : loose ) {
                if( !in_neigh[a] ) continue;
                auto new_st = st;
                new_st.ends.emplace_back( a,v );
                new_st.counts *= g.getArc( a,v ).lengths;
                placeIntoNewBucket(new_st);
            }
        }

        { // adding as end node to an existing path end
            if constexpr (debug) clog << "\tAdding as end node : ends -> v" << endl;
            for( int i=0; i<st.ends.size(); i++ ) {
                int b = st.ends[i].second;
                if( !in_neigh[b] ) continue;

                auto new_st = st;
                new_st.ends[i].second = v;
                new_st.taken.push_back(b);
                new_st.counts *= g.getArc( b,v ).lengths;
                placeIntoNewBucket(new_st);
            }
        }
    };


    auto mergeFBSS2States = [&]( vector<FBSS2> & sts ) {
        sort(ALL(sts));

        int p = 0, q = 0, r = 0;

        while( p < sts.size() ){
            q = p;
            while( q < sts.size() && (sts[p] == sts[q]) ) q++;
            for( int i=p+1; i<q; i++ ){
                sts[p].counts += sts[i].counts;
                if constexpr (debug) clog << "------>" << endl << "merging state " << sts[i] << " to " << sts[p] << endl;
            }
            swap( sts[r++], sts[p] );
            p = q;
        }

        while( sts.size() > r ) sts.pop_back();
    };

    auto getMstEdges = [&](FBSS2& st) {
        vector<Triple<int>> edges;
        edges.reserve( (S.size()+2) * (S.size()+2) / 4 );

        if constexpr (debug)
            clog << endl << "Finding MST edges for state " << st << endl;

        auto checkPair = [&](int a, int b) {
            if constexpr (debug)
                clog << "Considering pair (" << a << ", "  << b << ")" << endl;
            if( distances[a][b] != inf ) {
                edges.emplace_back( distances[a][b], a, b );
            }
        };

        checkPair(st.vs_node, vt);

        for( auto [a,b] : st.ends ) {
            checkPair(st.vs_node,a);
            checkPair(b,vt);

            for( auto [c,d] : st.ends ) {
                if( a == c || a == d || b == c || b == d ) continue;
                checkPair(b,c);
            }
        }
        if constexpr (debug) {
            clog << "edges: " << endl;
            for(auto [w,a,b] : edges) clog << "\t(" << w << "," << a << "," << b << ")" << endl;
            DEBUG(distances[st.vs_node]);
        }

        // now find mst
        FAU fau(N);
        for( auto [a,b] : st.ends ) fau.Union(a,b);
        sort(ALL(edges), []( auto & a, auto & b ){ return a.st < b.st; });
        vector<Triple<int>> res;
        for( auto & [w,a,b] : edges ) {
            if( fau.Find(a) != fau.Find(b) ) {
                if constexpr (debug)
                    clog << "\tuniting nodes " << a << " and " << b << endl;
                res.emplace_back(w,a,b);
                fau.Union(a,b);
            }
        }

        bool is_connected_mst = ( fau.Find(st.vs_node) == fau.Find(vt) );
        for( auto [a,b] : st.ends ) if( fau.Find(a) != fau.Find(st.vs_node) ) is_connected_mst = false;

        return make_pair(is_connected_mst,res);
    };

    auto canTrimMST = [&](FBSS2 & st) {
        auto [is_connected,edges] = getMstEdges(st);

        if constexpr (debug) if( !is_connected ) {
            clog << "MST created for state st: " << st << " is not connected, trimming!" << endl;
            clog << "edges: " << endl;
            for(auto [w,a,b] : edges) clog << "\t(" << w << "," << a << "," << b << ")" << endl;
        }

        if( !is_connected ) return true;

        LL sumw = accumulate( ALL(edges), 0ll, [](LL a, auto & e){ return a + e.st; } );
        if( sumw >= inf || st.counts.minl() + sumw > L ) {
            if constexpr (debug) {
                clog << endl << "Length conditon violated for st: " << st << " and edges: " << endl;
                for(auto [w,a,b] : edges) clog << "\t(" << w << "," << a << "," << b << ")" << endl;
            }
            return true;
        }
        return false;
    };

    auto canTrimPermutation = [&]( FBSS2 & st ) {
        VPII pairs(1, {-1, st.vs_node});
        pairs += st.ends;
        pairs.emplace_back( vt, -1 );

        sort( pairs.begin()+1, pairs.begin() + 1 + st.ends.size() );

        auto getDst = [&]() {
            int l = 0;
            for(int i=1; i<pairs.size(); i++) {
                int d = distances[ pairs[i-1].second ][ pairs[i].first ];
                if( d == inf ) return inf;
                l += d;
            }
            return l;
        };

        do {
            int l = getDst();
            if( st.counts.minl() + l <= L ) return false;
        }while( next_permutation( pairs.begin()+1, pairs.begin() + 1 + st.ends.size() ) );

        return true;
    };

    auto trimFBSS2tates = [&](vector<FBSS2> & sts) {
        for( int i=(int)sts.size()-1; i>=0; i-- ) {
            bool can_trim = (sts[i].counts.minl() > L);
            if(!can_trim && sts[i].ends.size() <= 4 ) can_trim = canTrimPermutation(sts[i]);
            if(!can_trim) can_trim = canTrimMST(sts[i]);

            if( can_trim ) {
                swap( sts[i], sts.back() );
                sts.pop_back();
            }
        }
    };

    auto addNodeForStates = [&](int v, vector<FBSS2> & sts) {
        markNeighbors(v, true); // mark in- and out-neighbors of node v within boundary set S

        for( auto & st : sts ) {
            VI loose = getLooseNodes(st);

            addAsIgnoredNode(v,st, loose);
            addAsVsNode(v,st, loose);
            addAsStartNode(v,st, loose);
            addAsMiddleNode(v,st, loose);
            addAsEndNode(v,st, loose);
        }

        markNeighbors(v,false); // unmark marked nodes...
    };

    auto removeFBSS2tates = [&](int v, vector<FBSS2> & sts, auto & states, int cur_st_id) {
        for( int i=(int)sts.size()-1; i>=0; i-- ) {
            sts[i].removeTaken(v);
            sts[i].recalculateHash();

            if( sts[i].containsNode(v) ) {
                if constexpr (debug) clog << "\t-x-x-x-x-x removing state " << sts[i] << endl;
                swap( sts[i], sts.back() );
                sts.pop_back();
            }else if( (sts[i].hash.hash & BH) != cur_st_id ) {
                if constexpr (debug) clog << "\tRemoving state " << sts[i] << " from bucket " << cur_st_id << " and putting into " << (sts[i].hash.hash & BH) << endl;
                states[  sts[i].hash.hash & BH ].push_back( sts[i] );
                swap( sts[i], sts.back() );
                sts.pop_back();
            }
        }
    };

    auto addNode = [&](int v) {
        // if constexpr (debug)
            clog << "Adding node v: " << v << endl;
        for( int i=0; i<BUCKETS; i++ ) new_states[i].clear();

        sw.start("addNodeForStates");
        for( int i=0; i<BUCKETS; i++ ) addNodeForStates( v, states[i] );
        sw.stop("addNodeForStates");

        DEBUG(countAllStates(states));
        DEBUG(countAllStates(new_states));

        // sw.start("mergeAndTrim");
        // for( int i=0; i<BUCKETS; i++ ) {
        //     mergeFBSS2tates( new_states[i] );
        //     trimFBSS2tates( new_states[i] );
        // }
        // sw.stop("mergeAndTrim");

        sw.start("mergeFBSS2tates");
        for( int i=0; i<BUCKETS; i++ ) mergeFBSS2States( new_states[i] );
        sw.stop("mergeFBSS2tates");

        sw.start("trimFBSS2tates");
        for( int i=0; i<BUCKETS; i++ ) trimFBSS2tates( new_states[i] );
        sw.stop("trimFBSS2tates");


        DEBUG(countAllStates(new_states));
        swap(states, new_states);
    };



    auto removeNode = [&](int v) {
        // if constexpr (debug)
            clog << "Removing node v: " << v << endl;
        for( int i=0; i<BUCKETS; i++ ) new_states[i].clear();

        sw.start("removeFBSS2tates");
        for( int i=0; i<BUCKETS; i++ ) removeFBSS2tates( v, states[i], new_states, i );
        for( int i=0; i<BUCKETS; i++ ) {
            states[i] += new_states[i];
            new_states[i].clear();
        }
        sw.stop("removeFBSS2tates");

        sw.start("mergeFBSS2States");
        for( int i=0; i<BUCKETS; i++ ) mergeFBSS2States( states[i] );
        sw.stop("mergeFBSS2States");

    };

    FBSS2 initial_state( Lengths(VBIG(1,1),0), {}, vs );
    states[0].push_back( initial_state );
    was[vs] = true;
    S.insert(vs);
    order.erase(order.begin());

    for( int v : order ) {
        if( v == vt && was[v] ) continue;

        clog << "Current boundary nodes: " << S.size() << " -> " << S << endl;
        DEBUG(countAllStates(states));

        if constexpr (debug) {
            ENDL(5); ENDLS(50, "*"); ENDL(5);
            clog << "Current states: " << endl;
            for(int i=0; i<BUCKETS; i++) {
                if( !states[i].empty() ) clog << "states[" << i << "]: ";
                for( auto & st : states[i] ) clog << st << endl;
            }
            ENDL(2);
        }

        if( was[v] ) {
            removeNode(v);
            S.erase(v);
        }
        else{
            sw.start("createDistances");
            createDistances(v);
            sw.stop("createDistances");

            S_nodes = VI(ALL(S));
            addNode(v);
            was[v] = true;
            S.insert(v);
        }
    }

    vector<FBSS2> all_states;
    for(auto & vec : states) all_states += vec;

    if constexpr (debug) {
        clog << "All states: " << endl;
        for(auto & st : all_states) clog << st << endl;
    }

    assert( all_states.size() <= 1 ); // there should be only one state - the one with node vt
    if(!all_states.empty()) {
        assert( all_states[0].vs_node == vt );
        assert( all_states[0].ends.empty());
    }

    sw.start("solveFBS2ForNodeOrder");
    sw.writeAll();

    VBIG res(L+1);
    if( !all_states.empty() ) {
        auto & lengths = all_states[0].counts;
        for( int i=lengths.minl(); i <= min( L, lengths.maxl() ); i++ ) res[i] += lengths[i];
    }

    return res;
}


VBIG Solver2FBS::solveFBSDForNodeOrder(VI order) {
    constexpr bool debug = false;

    if constexpr (debug) {
        clog << "Entering solveFBSDForNodeOrder" << endl;
        DEBUG(g);
        DEBUG(order);
    }

    Stopwatch sw;
    sw.start("solveFBSDForNodeOrder");

    VB was(N);
    VB in_neigh(N), out_neigh(N);
    set<int> S; // nodes on boundary - needed only for debugging
    VI S_nodes;
    VB is_state_end(N,false);

    auto markNeighbors = [&](int v, const bool val) {
        for( auto & arc : g[v] ) if(was[arc.v] && S.contains(arc.v)) out_neigh[arc.v] = val;
        for( auto & arc : revg[v] ) if(was[arc.v] && S.contains(arc.v)) in_neigh[arc.v] = val;
    };


    VVI distances(N);
    auto createDistances = [&](int v) {
        for(int i=0; i<N; i++) VI().swap(distances[i]);
        S.insert(v);

        for(int d : S) was[d] = false;
        for(int d : S) distances[d] = Solver2::findDistancesInclusive(g, d, was).first;
        for(int d : S) was[d] = true;
        was[v] = false;

        S.erase(v);
    };

    auto getLooseNodes = [&]( FBSSD & st ) {
        VI loose;
        is_state_end[st.vs_node] = true;
        for(auto [a,b] : st.ends) is_state_end[a] = is_state_end[b] = true;
        for(int a : st.taken) is_state_end[a] = true;
        for( int d : S_nodes ) if( !is_state_end[d] ) loose.push_back(d);
        is_state_end[st.vs_node] = false;
        for(auto [a,b] : st.ends) is_state_end[a] = is_state_end[b] = false;
        for(int a : st.taken) is_state_end[a] = false;
        return loose;
    };


    constexpr int BUCKETS = (1<<10);
    constexpr int BH = BUCKETS-1;

    vector<vector<FBSSD>> states(BUCKETS);
    vector<vector<FBSSD>> new_states(BUCKETS);

    auto countAllStates = [&](auto & states) {
        return accumulate(ALL(states), 0, [](int a, auto & vec){ return a + vec.size(); });
    };

    auto placeIntoNewBucket = [&]( FBSSD & st ) {
        for( auto & [a,b] : st.ends ) if( a > b ) swap(a,b);

        // if(!st.isCorrectState()) { // #TEST #CAUTION!! This is time-consuming as hell here! remove that after tests...
        //     clog << endl << "!!!!" << endl << "State is incorrect!" << endl;
        //     DEBUG(st);
        //     assert( st.isCorrectState() );
        // };


        st.recalculateHash();
        st.counts.trimToMax(L);
        new_states[st.hash.hash & BH].push_back( st );
        if constexpr (debug) clog << "\t\tCreated and added new_state: " << st << endl;
    };

    auto addAsIgnoredNode = [&]( int v, const FBSSD & st, VI & loose ) {
        if constexpr (debug) clog << "\tAdding as ignored node" << endl;
        auto new_st = st;
        placeIntoNewBucket(new_st);
    };

    auto addAsVsNode = [&](int v, const FBSSD & st, VI & loose) {
        if constexpr (debug) clog << "\tAdding as vs_node" << endl;
        if( !in_neigh[st.vs_node] ) return;

        auto new_st = st;
        new_st.vs_node = v;
        new_st.taken.push_back(st.vs_node);
        new_st.counts *= g.getArc(st.vs_node, v).lengths;
        placeIntoNewBucket(new_st);
    };

    auto addAsStartNode = [&](int v, const FBSSD & st, VI & loose) {
        { // adding as start node to an existing loose node
            if constexpr (debug) clog << "\tAdding as start node: v -> loose" << endl;
            for( int a : loose ) {
                if( !out_neigh[a] ) continue;

                auto new_st = st;
                new_st.ends.emplace_back( v, a );
                new_st.counts *= g.getArc( v, a ).lengths;
                placeIntoNewBucket(new_st);
            }
        }

        { // adding as start node to an existing path end
            if constexpr (debug) clog << "\tAdding as start node: v -> ends" << endl;
            for( auto [a,b] : st.ends ) {
                for(int i=0; i<2; i++) {
                    if( out_neigh[a] ) {
                        auto new_st = st;
                        new_st.removeEnds({a,b});
                        new_st.ends.emplace_back(v,b);
                        new_st.taken.push_back(a);
                        new_st.counts *= g.getArc( v,a ).lengths;
                        placeIntoNewBucket(new_st);
                    }
                    swap(a,b);
                }
            }
        }

    };

    auto addAsMiddleNode = [&](int v, const FBSSD & st, VI & loose) {
        { // adding as middle node, vs_node -> v -> loose
            if constexpr (debug) clog << "\tAdding as middle node : vs_node -> v -> loose" << endl;
            for( int a : loose ) {
                if( !in_neigh[st.vs_node] ) continue;
                if( out_neigh[a] ) {
                    auto new_st = st;
                    new_st.vs_node = a;
                    new_st.taken.push_back(st.vs_node);
                    new_st.taken.push_back(v);
                    new_st.counts *= g.getArc(st.vs_node, v).lengths;
                    new_st.counts *= g.getArc(v, a).lengths;
                    placeIntoNewBucket(new_st);
                }
            }

        }

        { // adding as middle node, vs_node -> v -> path end
            if constexpr (debug) clog << "\tAdding as middle node : vs_node -> v -> ends" << endl;
            for( auto [a,b] : st.ends ) {
                if( !in_neigh[st.vs_node] ) continue;

                for(int i=0; i<2; i++) {
                    if( out_neigh[a] ) {
                        auto new_st = st;
                        new_st.vs_node = b;
                        new_st.removeEnds({a,b});
                        new_st.taken.push_back(st.vs_node);
                        new_st.taken.push_back(v);
                        new_st.taken.push_back(a);
                        new_st.counts *= g.getArc(st.vs_node, v).lengths;
                        new_st.counts *= g.getArc(v, a).lengths;
                        placeIntoNewBucket(new_st);
                    }
                    swap(a,b);
                }
            }
        }

        { // adding as middle node, loose -> v -> loose
            if constexpr (debug) clog << "\tAdding as middle node : loose -> v -> loose" << endl;
            for( int a : loose ) {
                if( !in_neigh[a] ) continue;

                for( int b : loose ) {
                    if( a >= b ) continue;
                    if( out_neigh[b] ) {
                        auto new_st = st;
                        new_st.ends.emplace_back(a,b);
                        new_st.taken.push_back(v);
                        new_st.counts *= g.getArc(a, v).lengths;
                        new_st.counts *= g.getArc(v, b).lengths;
                        placeIntoNewBucket(new_st);
                    }
                }
            }
        }

        { // adding as middle node, loose -> v -> path end
            if constexpr (debug) clog << "\tAdding as middle node : loose -> v -> ends" << endl;
            for( int a : loose ) {
                if( !in_neigh[a] ) continue;

                for( auto [b,c] : st.ends ) {
                    for(int i=0; i<2; i++) {
                        if( out_neigh[b] ) {
                            auto new_st = st;
                            new_st.removeEnds( {b,c} );
                            new_st.ends.emplace_back(a,c);
                            new_st.taken.push_back(v);
                            new_st.taken.push_back(b);
                            new_st.counts *= g.getArc(a, v).lengths;
                            new_st.counts *= g.getArc(v, b).lengths;
                            placeIntoNewBucket(new_st);
                        }
                        swap(b,c);
                    }
                }
            }
        }

        // { // adding as middle node, path_end -> v -> loose
        //     if constexpr (debug) clog << "\tAdding as middle node : ends -> v -> loose" << endl;
        //     for( auto [a,b] : st.ends ) {
        //         for(int i=0; i<2; i++) {
        //             if( in_neigh[b] ) {
        //                 for( int c : loose ) {
        //                     if( !out_neigh[c] ) continue;
        //
        //                     auto new_st = st;
        //                     new_st.removeEnds({a,b});
        //                     new_st.ends.emplace_back(a,c);
        //                     new_st.taken.push_back(b);
        //                     new_st.taken.push_back(v);
        //                     new_st.counts *= g.getArc(b, v).lengths;
        //                     new_st.counts *= g.getArc(v, c).lengths;
        //                     placeIntoNewBucket(new_st);
        //                 }
        //             }
        //             swap(a,b);
        //         }
        //     }
        // }

        { // adding as middle node, path_end -> v -> path_end
            if constexpr (debug) clog << "\tAdding as middle node : ends -> v -> ends" << endl;
            // for( auto [a,b] : st.ends ) {
            for( int k1 = 0; k1 < st.ends.size(); k1++ ) {
                auto [a,b] = st.ends[k1];

                for( int i=0; i<2; i++ ){
                    if( in_neigh[b] ) {
                        // for( auto [c,d] : st.ends ) {
                        for( int k2 = k1+1; k2 < st.ends.size(); k2++ ) {
                            auto [c,d] = st.ends[k2];
                            // if( min(a,b) >= min(c,d) ) continue;

                            for(int j=0; j<2; j++) {
                                if( out_neigh[c] ) {
                                    auto new_st = st;
                                    new_st.removeEnds({a,b});
                                    new_st.removeEnds({c,d});
                                    new_st.ends.emplace_back(a,d);
                                    new_st.taken.push_back(b);
                                    new_st.taken.push_back(v);
                                    new_st.taken.push_back(c);
                                    new_st.counts *= g.getArc(b, v).lengths;
                                    new_st.counts *= g.getArc(v, c).lengths;
                                    placeIntoNewBucket(new_st);
                                }
                                swap(c,d);
                            }
                        }
                    }
                    swap(a,b);
                }
            }
        }
    };

    auto mergeFBSSDStates = [&]( vector<FBSSD> & sts ) {
        sort(ALL(sts));

        int p = 0, q = 0, r = 0;

        while( p < sts.size() ){
            q = p;
            while( q < sts.size() && (sts[p] == sts[q]) ) q++;
            for( int i=p+1; i<q; i++ ){
                sts[p].counts += sts[i].counts;
                if constexpr (debug) clog << "------>" << endl << "merging state " << sts[i] << " to " << sts[p] << endl;
            }
            swap( sts[r++], sts[p] );
            p = q;
        }

        while( sts.size() > r ) sts.pop_back();
    };

    auto getMstEdges = [&](FBSSD& st) {
        vector<Triple<int>> edges;
        edges.reserve( (S.size()+2) * (S.size()+2) / 4 );

        if constexpr (debug)
            clog << endl << "Finding MST edges for state " << st << endl;

        auto checkPair = [&](int a, int b) {
            if constexpr (debug)
                clog << "Considering pair (" << a << ", "  << b << ")" << endl;
            if( distances[a][b] != inf ) {
                edges.emplace_back( distances[a][b], a, b );
            }
        };

        checkPair(st.vs_node, vt);

        for( auto [a,b] : st.ends ) {
            checkPair(st.vs_node,a);
            checkPair(st.vs_node,b);
            checkPair(a,vt);
            checkPair(b,vt);

            for( auto [c,d] : st.ends ) {
                if( min(a,b) >= min(c,d) ) continue;
                checkPair(a,c);
                checkPair(a,d);
                checkPair(b,c);
                checkPair(b,d);
            }
        }
        if constexpr (debug) {
            clog << "edges: " << endl;
            for(auto [w,a,b] : edges) clog << "\t(" << w << "," << a << "," << b << ")" << endl;
            DEBUG(distances[st.vs_node]);
        }

        // now find mst
        FAU fau(N);
        for( auto [a,b] : st.ends ) fau.Union(a,b);
        sort(ALL(edges), []( auto & a, auto & b ){ return a.st < b.st; });
        vector<Triple<int>> res;
        for( auto & [w,a,b] : edges ) {
            if( fau.Find(a) != fau.Find(b) ) {
                if constexpr (debug)
                    clog << "\tuniting nodes " << a << " and " << b << endl;
                res.emplace_back(w,a,b);
                fau.Union(a,b);
            }
        }

        bool is_connected_mst = ( fau.Find(st.vs_node) == fau.Find(vt) );
        for( auto [a,b] : st.ends ) if( fau.Find(a) != fau.Find(st.vs_node) ) is_connected_mst = false;

        return make_pair(is_connected_mst,res);
    };

    auto canTrimMST = [&](FBSSD & st) {
        auto [is_connected,edges] = getMstEdges(st);

        if constexpr (debug) if( !is_connected ) {
            clog << "MST created for state st: " << st << " is not connected, trimming!" << endl;
            clog << "edges: " << endl;
            for(auto [w,a,b] : edges) clog << "\t(" << w << "," << a << "," << b << ")" << endl;
        }

        if( !is_connected ) return true;

        LL sumw = accumulate( ALL(edges), 0ll, [](LL a, auto & e){ return a + e.st; } );
        if( sumw >= inf || st.counts.minl() + sumw > L ) {
            if constexpr (debug) {
                clog << endl << "Length conditon violated for st: " << st << " and edges: " << endl;
                for(auto [w,a,b] : edges) clog << "\t(" << w << "," << a << "," << b << ")" << endl;
            }
            return true;
        }
        return false;
    };

    auto canTrimPermutation = [&]( FBSSD & st ) {
        VPII pairs(1, {-1, st.vs_node});
        pairs += st.ends;
        pairs.emplace_back( vt, -1 );

        sort( pairs.begin()+1, pairs.begin() + 1 + st.ends.size() );

        auto getDst = [&]() {
            int l = 0;
            for(int i=1; i<pairs.size(); i++) {
                auto [a,b] = pairs[i-1];
                auto [c,d] = pairs[i];
                int e = distances[b][c];
                if( a != -1 && d != -1 ) e = min(e, distances[a][d] );
                if( a != -1 && c != -1 ) e = min(e, distances[a][c] );
                if( b != -1 && d != -1 ) e = min(e, distances[b][d] );
                if( e == inf ) return inf;
                l += e;
            }
            return l;
        };

        do {
            int l = getDst();
            if( st.counts.minl() + l <= L ) return false;
        }while( next_permutation( pairs.begin()+1, pairs.begin() + 1 + st.ends.size() ) );

        return true;
    };

    auto trimFBSSDStates = [&](vector<FBSSD> & sts) {
        for( int i=(int)sts.size()-1; i>=0; i-- ) {
            bool can_trim = (sts[i].counts.minl() > L);
            if(!can_trim && sts[i].ends.size() <= 6 ) can_trim = canTrimPermutation(sts[i]);
            if(!can_trim) can_trim = canTrimMST(sts[i]);

            if( can_trim ) {
                swap( sts[i], sts.back() );
                sts.pop_back();
            }
        }
    };

    auto addNodeForStates = [&](int v, vector<FBSSD> & sts) {
        markNeighbors(v, true); // mark in- and out-neighbors of node v within boundary set S

        for( auto & st : sts ) {
            if constexpr (debug) clog << "Adding for state st: " << st << endl;

            VI loose = getLooseNodes(st);

            addAsIgnoredNode(v,st, loose);
            addAsVsNode(v,st, loose);
            addAsStartNode(v,st, loose);
            addAsMiddleNode(v,st, loose);
            // addAsEndNode(v,st, loose);
        }

        markNeighbors(v,false); // unmark marked nodes...
    };

    auto removeFBSSDtates = [&](int v, vector<FBSSD> & sts, auto & states, int cur_st_id) {
        for( int i=(int)sts.size()-1; i>=0; i-- ) {
            sts[i].removeTaken(v);
            sts[i].recalculateHash();

            if( sts[i].containsNode(v) ) {
                if constexpr (debug) clog << "\t-x-x-x-x-x removing state " << sts[i] << endl;
                swap( sts[i], sts.back() );
                sts.pop_back();
            }else if( (sts[i].hash.hash & BH) != cur_st_id ) {
                if constexpr (debug) clog << "\tRemoving state " << sts[i] << " from bucket " << cur_st_id << " and putting into " << (sts[i].hash.hash & BH) << endl;
                states[  sts[i].hash.hash & BH ].push_back( sts[i] );
                swap( sts[i], sts.back() );
                sts.pop_back();
            }
        }
    };

    auto addNode = [&](int v) {
        // if constexpr (debug)
            clog << "Adding node v: " << v << endl;
        for( int i=0; i<BUCKETS; i++ ) new_states[i].clear();

        sw.start("addNodeForStates");
        for( int i=0; i<BUCKETS; i++ ) addNodeForStates( v, states[i] );
        sw.stop("addNodeForStates");

        DEBUG(countAllStates(states));
        DEBUG(countAllStates(new_states));

        // sw.start("mergeAndTrim");
        // for( int i=0; i<BUCKETS; i++ ) {
        //     mergeFBSS2tates( new_states[i] );
        //     trimFBSS2tates( new_states[i] );
        // }
        // sw.stop("mergeAndTrim");

        sw.start("mergeFBSS2tates");
        for( int i=0; i<BUCKETS; i++ ) mergeFBSSDStates( new_states[i] );
        sw.stop("mergeFBSS2tates");

        sw.start("trimFBSS2tates");
        for( int i=0; i<BUCKETS; i++ ) trimFBSSDStates( new_states[i] );
        sw.stop("trimFBSS2tates");


        DEBUG(countAllStates(new_states));
        swap(states, new_states);
    };



    auto removeNode = [&](int v) {
        // if constexpr (debug)
            clog << "Removing node v: " << v << endl;
        for( int i=0; i<BUCKETS; i++ ) new_states[i].clear();

        sw.start("removeFBSS2tates");
        for( int i=0; i<BUCKETS; i++ ) removeFBSSDtates( v, states[i], new_states, i );
        for( int i=0; i<BUCKETS; i++ ) {
            states[i] += new_states[i];
            new_states[i].clear();
        }
        sw.stop("removeFBSS2tates");

        sw.start("mergeFBSS2States");
        for( int i=0; i<BUCKETS; i++ ) mergeFBSSDStates( states[i] );
        sw.stop("mergeFBSS2States");

    };

    FBSSD initial_state( Lengths(VBIG(1,1),0), {}, vs );
    states[0].push_back( initial_state );
    was[vs] = true;
    S.insert(vs);
    order.erase(order.begin());

    for( int v : order ) {
        if( v == vt && was[v] ) continue;

        clog << "Current boundary nodes: " << S.size() << " -> " << S << endl;
        DEBUG(countAllStates(states));

        if constexpr (debug) {
            ENDL(5); ENDLS(50, "*"); ENDL(5);
            clog << "Current states: " << endl;
            for(int i=0; i<BUCKETS; i++) {
                if( !states[i].empty() ) clog << "states[" << i << "]: ";
                for( auto & st : states[i] ) clog << st << endl;
            }
            ENDL(2);
        }

        if( was[v] ) {
            removeNode(v);
            S.erase(v);
        }
        else{
            sw.start("createDistances");
            createDistances(v);
            sw.stop("createDistances");

            S_nodes = VI(ALL(S));
            addNode(v);
            was[v] = true;
            S.insert(v);
        }
    }

    vector<FBSSD> all_states;
    for(auto & vec : states) all_states += vec;

    if constexpr (debug) {
        clog << "All states: " << endl;
        for(auto & st : all_states) clog << st << endl;
    }

    assert( all_states.size() <= 1 ); // there should be only one state - the one with node vt
    if(!all_states.empty()) {
        assert( all_states[0].vs_node == vt );
        assert( all_states[0].ends.empty());
    }

    sw.start("solveFBSDForNodeOrder");
    sw.writeAll();

    VBIG res(L+1);
    if( !all_states.empty() ) {
        auto & lengths = all_states[0].counts;
        for( int i=lengths.minl(); i <= min( L, lengths.maxl() ); i++ ) res[i] += lengths[i];
    }

    return res;
}




VBIG Solver2FBS::solveFBS3ForNodeOrder(VI order) {
    constexpr bool debug = false;

    {
        VB was(N);
        if( !Solver2::existsPath(g,vs,vt,was) ) return VBIG(L+1,0);
    }

    if constexpr (debug) {
        clog << "Entering solveFBS3ForNodeOrder" << endl;
        DEBUG(g);
        DEBUG(order);
    }

    Stopwatch sw;
    sw.start("solveFBS3ForNodeOrder");

    if( timer.getLimit("fbs") == 1e9 ) timer.setLimit("fbs",1e9);
    timer.start("fbs");

    VB was(N);
    VB in_neigh(N), out_neigh(N);
    set<int> S; // nodes on boundary - needed only for debugging


    VI S_nodes(32,-1);
    VI ind_in_S(N,-1);
    auto getFirstFreeSIndex = [&]() {
        int r = 0;
        while( S_nodes[r] != -1 ) r++;
        return r;
    };

    vector<unordered_map<int,int>> arc_id_mapper(N);
    VVI arc_id_mapper_tab;
    vector<Lengths> all_arcs_lengths;
    bool use_map_for_arc_id_mapper = g.countArcs() > 3'000;
    // bool use_map_for_arc_id_mapper = false;
    if(!use_map_for_arc_id_mapper) arc_id_mapper_tab.assign(N, VI(N,-1));
    for( int i=0; i<N; i++ ) {
        for( auto & arc : g[i] ) {
            if(use_map_for_arc_id_mapper) arc_id_mapper[i][ arc.v ] = all_arcs_lengths.size();
            else arc_id_mapper_tab[i][arc.v] = all_arcs_lengths.size();
            all_arcs_lengths.push_back(arc.lengths);
            all_arcs_lengths.back().minimize();
        }
    }
    auto getArcId = [&](int a, int b) {
        if( use_map_for_arc_id_mapper ) return arc_id_mapper[a][b];
        else return arc_id_mapper_tab[a][b];
    };

    vector<HashType> all_arcs_hashes; all_arcs_lengths.reserve( all_arcs_lengths.size() );
    for( auto & l : all_arcs_lengths ) all_arcs_hashes.push_back( l.hash() );

    HashType cur_vis_hash = 0;

    map<int,FBSState3> aux_st_mapper; // just for debugging purposes





    vector<Edge> trg_edges;
    vector<EdgeH> trg_edges_h;
    unordered_map<HashType,int, HashTypeHash> state_id_mapper;


    constexpr bool compress_on_the_fly = false;
    constexpr int PRED_CNT = 4;
    struct Pred {
        Pred(){ clear(); }
        void addPredecessor( int id ) {
            for(int i=0; i<PRED_CNT; i++) if( predecessors[i] == -1 ){ predecessors[i] = id; break; }
        }
        int countPredecessors() {
            int res = 0;
            for(int i=0; i<PRED_CNT; i++) if( predecessors[i] != -1 ) res++;
            return res;
        }
        void clear(){ fill( predecessors, predecessors + PRED_CNT, -1 ); }
        int predecessors[PRED_CNT];
    };

    vector<Pred> state_predecessors;
    VI state_outdeg;
    VI free_ids;
    vector<HashType> state_hash_remapper;
    VI first_valid_edge_index; // necessary to identify edges that were added earlier, but are no longer valid...
    auto getFreeNodeId = [&]() {
        if( free_ids.empty() ) {
            return (int)state_id_mapper.size();
            // if constexpr (compress_on_the_fly) return (int)state_id_mapper.size();
            // else return next_node_id_h++;
        }

        int t = free_ids.back(); free_ids.pop_back(); return t;
    };
    auto trimOnTheFly = [&](int x) { // iterative version
        VI to_trim(1,x);

        for( int j=0; j<to_trim.size(); j++ ) {
            int num = to_trim[j];

            auto h = state_hash_remapper[num];
            state_hash_remapper[num] = -1; // here can be anything, but it might be easier to debug it if it is -1
            state_id_mapper.erase(h);
            free_ids.push_back(num);
            first_valid_edge_index[num] = inf;

            for( int i=0; i<PRED_CNT; i++ ) {
                int pred = state_predecessors[num].predecessors[i];
                if( pred != -1 ) {
                    state_outdeg[pred]--;
                    if( state_outdeg[pred] == 0 ) to_trim.push_back(pred);
                }
            }

            state_predecessors[num].clear();
        }
    };
    ULL total_masks_trimmed = 0;


    ULL expansion_iteration_cnt = 0;
    constexpr int expansion_iteration_shift_bits = 6; // up to 256 nodes
    constexpr bool include_iter_cnt_in_hash = true && !compress_on_the_fly;
    auto updateHashWithIterationCnt = [&](HashType & h ) {
        if constexpr (include_iter_cnt_in_hash) {
            h.hash &= (ULLONG_MAX >> expansion_iteration_shift_bits);
            h.hash |= ( expansion_iteration_cnt << (64-expansion_iteration_shift_bits) );
        }
    };


    auto trackState = [&](HashType h) {
        if constexpr (compress_on_the_fly) {
            if( !state_id_mapper.contains(h) ) {
                auto s = getFreeNodeId();
                state_id_mapper[h] = s;
                if constexpr (debug) clog << "Adding new state to track: " << h << " -> " << s << endl;
                if( s >= state_predecessors.size() ) {
                    state_predecessors.emplace_back();
                    state_outdeg.push_back(0);
                    state_hash_remapper.emplace_back(0);
                    first_valid_edge_index.push_back(inf);
                }
                state_hash_remapper[s] = h;
                first_valid_edge_index[s] = trg_edges.size();
            }
            else if constexpr ( debug ) clog << "\tstate alredy tracked" << endl;
        }else {
            // we do not need to do anything here...
            // we just add edges as hashes to tgr_edges_h and then sort it and remap...
        }
    };
    auto addEdge = [&]( HashType h1, HashType h2, int e1, int e2 = -1  ) {
        if constexpr (debug && compress_on_the_fly) assert( state_id_mapper.contains(h1));
        if constexpr (debug && compress_on_the_fly) assert( state_id_mapper.contains(h2));
        assert(h1 != h2);
        if constexpr (compress_on_the_fly) {
            int a = state_id_mapper[h1];
            int b = state_id_mapper[h2];
            trg_edges.emplace_back( a, b, e1,e2 );

            state_predecessors[b].addPredecessor(a);
            state_outdeg[a]++;
        }else {

            if constexpr (debug) clog << "Adding to trg_edges_h arc " << h1 << " -> " << h2 << endl;
            trg_edges_h.emplace_back(h1,h2,e1,e2);
        }
    };


    auto markNeighbors = [&](int v, const bool val) {
        for( auto & arc : g[v] ) if(was[arc.v] && S.contains(arc.v)) out_neigh[arc.v] = val;
        for( auto & arc : revg[v] ) if(was[arc.v] && S.contains(arc.v)) in_neigh[arc.v] = val;
    };

    int DST_MAX_MASK_SIZE = 5;
    int E = g.countArcs();
    if( E > 100 ) DST_MAX_MASK_SIZE = 4;
    if( E > 1'000 ) DST_MAX_MASK_SIZE = 3;
    if( E > 5'000 ) DST_MAX_MASK_SIZE = 2;
    // DEBUG(DST_MAX_MASK_SIZE);
    int INIT_DST_MAX_MASK_SIZE = DST_MAX_MASK_SIZE;

    constexpr int DOM_MASK_SIZE = 64;
    using DOM_MASK = bitset<DOM_MASK_SIZE>;
    int DOM_MAX_MASK_SIZE = DST_MAX_MASK_SIZE-1;

    VVVI distances;
    vector<vector<vector<DOM_MASK>>> dominators;
    // bool use_dominator_trimming = g.size() <= DOM_MASK_SIZE;
    bool use_dominator_trimming = false;

    int MMB = 0;
    VVI masks_bits(INIT_DST_MAX_MASK_SIZE+1);


    auto createDistances = [&](int v) { // creates distances for given subset of taken nodes in S
        DST_MAX_MASK_SIZE = INIT_DST_MAX_MASK_SIZE;
        if( DST_MAX_MASK_SIZE > 0 && S.size() >= 15 ) DST_MAX_MASK_SIZE--;
        if( DST_MAX_MASK_SIZE > 0 && S.size() >= 20 ) DST_MAX_MASK_SIZE--;
        DOM_MAX_MASK_SIZE = max(0, DST_MAX_MASK_SIZE-1);

        int MAX_SNODES_IND = 0;
        for( int i=0; i<S_nodes.size(); i++ ) if(S_nodes[i] != -1) MAX_SNODES_IND = i+1;

        DistanceFinder2 df(N);
        if( (1 << MAX_SNODES_IND) > MMB ) {
            for( int mask = MMB; mask < (1 << MAX_SNODES_IND); mask++ ) {
                int b = popcount((unsigned)mask);
                if( b < masks_bits.size() ) masks_bits[b].push_back(mask);
            }
            MMB = (1 << MAX_SNODES_IND);
        }


        // for( int mask = 0; mask < (1 << MAX_SNODES_IND); mask++ ) {
        for( int mb = 0; mb <= DST_MAX_MASK_SIZE; mb++ ) {
            for( int mask : masks_bits[mb] ) {
                // if( popcount((unsigned)mask) > DST_MAX_MASK_SIZE ) continue;
                assert( popcount((unsigned)mask) == mb );

                bool can_continue = true;
                for( int i=0; i<S_nodes.size(); i++ ) if( (mask & (1<<i)) && S_nodes[i] == -1 ) can_continue = false;
                if(!can_continue) continue;

                while( distances.size() <= mask ) distances.emplace_back(); // for pathwidth <= 32 it will work :)
                distances[mask] = VVI(MAX_SNODES_IND);

                S.insert(v);
                for(int d : S) was[d] = false;
                for( int i=0; i<32; i++ ) if( mask & (1<<i) ) was[ S_nodes[i] ] = true;
                for(int d : S) {
                    int a = ind_in_S[d];
                    distances[mask][a] = VI(MAX_SNODES_IND,inf);

                    if( !was[d] ) {
                        // auto dst = Solver2::findDistancesInclusive(g, d, was).first;
                        df.findDistancesInclusive(g, d, was);
                        for( int i=0; i<S_nodes.size(); i++ ) {
                            // if( S_nodes[i] != -1 ) distances[mask][a][i] = dst[ S_nodes[i] ];
                            if( S_nodes[i] != -1 ) distances[mask][a][i] = df.getDst(S_nodes[i]);
                        }


                        // last element will always have distance to vt if vt is not in S
                        // if( ind_in_S[vt] == -1 ) distances[mask][a].push_back( dst[vt] );
                        if( ind_in_S[vt] == -1 ) distances[mask][a].push_back( df.getDst(vt) );
                    }
                }
                for(int d : S) was[d] = true;
                was[v] = true;
                S.erase(v);
            }
        }

        constexpr bool check_distances = debug; // this is jut for debugging
        if constexpr (check_distances) {
            ENDL(5);
            clog << "#CAUTION! Checking if distances are created properly" << endl;
            for( int mask = 1; mask < (1 << MAX_SNODES_IND); mask++ ) {
                if( popcount((unsigned)mask) > DST_MAX_MASK_SIZE ) continue;
                bool can_continue = true;
                for( int i=0; i<S_nodes.size(); i++ ) if( (mask & (1<<i)) && S_nodes[i] == -1 ) can_continue = false;
                if(!can_continue) continue;

                for( int a = 0; a < S_nodes.size(); a++ ) {
                    if( S_nodes[a] == -1 ) continue;
                    if( mask & (1<<a) ) continue;

                    for( int b = 0; b < S_nodes.size(); b++ ) {
                        if( S_nodes[b] == -1 ) continue;
                        if( mask & (1<<b) ) continue;

                        if( distances[mask][a][b] < distances[0][a][b] ) {
                            DEBUG(S_nodes);
                            DEBUG(FBSS3::getSetBits(mask));
                            DEBUG(a);
                            DEBUG(b);
                            DEBUG(distances[mask][a]);
                            DEBUG(distances[0][a]);

                            assert( distances[mask][a][b] >= distances[0][a][b] );
                        }

                        VB helper = was;
                        for(int d : S) helper[d] = false;
                        helper[v] = false;
                        if( distances[mask][a][b] < Solver2::findDistances( g, S_nodes[a], helper ).first[ S_nodes[b] ] ) {
                            DEBUG(S_nodes);
                            DEBUG(FBSS3::getSetBits(mask));
                            DEBUG(a);
                            DEBUG(b);
                            DEBUG(distances[mask][a]);
                            DEBUG(distances[0][a]);
                            DEBUG(Solver2::findDistances( g, S_nodes[a], helper ).first);

                            assert( distances[mask][a][b] >= Solver2::findDistances( g, S_nodes[a], helper ).first[ S_nodes[b] ] );
                        }
                    }
                }
            }
        }
    };

    auto createDominators = [&](int v) {
        int MAX_SNODES_IND = 0;
        for( int i=0; i<S_nodes.size(); i++ ) if(S_nodes[i] != -1) MAX_SNODES_IND = i+1;

        auto V = g.getStruture();
        for( int mask = 0; mask < (1 << MAX_SNODES_IND); mask++ ) {
            if( popcount((unsigned)mask) > DST_MAX_MASK_SIZE-1 ) continue;
            bool can_continue = true;
            for( int i=0; i<S_nodes.size(); i++ ) if( (mask & (1<<i)) && S_nodes[i] == -1 ) can_continue = false;
            if(!can_continue) continue;

            while( dominators.size() <= mask ) dominators.emplace_back();
            dominators[mask] = vector<vector<DOM_MASK>>(MAX_SNODES_IND);

            S.insert(v);
            for(int d : S) was[d] = false;
            for( int i=0; i<32; i++ ) if( mask & (1<<i) ) was[S_nodes[i]] = true;
            for(int d1 : S) {
                int a = ind_in_S[d1];
                assert(a != -1);
                dominators[mask][a] = vector<DOM_MASK>(MAX_SNODES_IND, DOM_MASK(0));
                for( auto & x : dominators[mask][a] ) x.reset();

                for( int d2 : S ) {
                    if(d1 == d2) continue;
                    int b = ind_in_S[d2];
                    assert(b != -1);

                    bool t1 = was[d1], t2 = was[d2];
                    was[d1] = was[d2] = false;
                    VI dom = Kernelizer::findDominators( V, was, d1,d2 );
                    was[d1] = t1; was[d2] = t2;

                    for( int x : dom ) dominators[mask][a][b].set(x,true);
                }

                if( ind_in_S[vt] == -1 ) {
                    dominators[mask][a].emplace_back( DOM_MASK(0) );
                    bool t1 = was[d1], t2 = was[vt];
                    was[d1] = was[vt] = false;
                    VI dom = Kernelizer::findDominators( V, was, d1,vt );
                    was[d1] = t1; was[vt] = t2;
                    for( int x : dom ) dominators[mask][a].back().set(x,true);
                }
            }
            for(int d : S) was[d] = true;
            was[v] = true;
            S.erase(v);
        }
    };

    VB is_state_end(N,false);
    auto getLooseNodes = [&]( FBSS3 & st ) {
        VI loose;
        is_state_end[st.vs_node] = true;
        for(auto [a,b] : st.ends) is_state_end[a] = is_state_end[b] = true;
        for( int d : S ) if( !is_state_end[d] ) loose.push_back(d);
        is_state_end[st.vs_node] = false;
        for(auto [a,b] : st.ends) is_state_end[a] = is_state_end[b] = false;
        return loose;
    };

    VVI underV = g.getStruture();
    {
        VVI revV = GraphUtils::reverseGraph(underV);
        for(int i=0; i<N; i++) {
            underV[i] += revV[i];
            StandardUtils::makeUnique(underV[i]);
        }
    }

    constexpr bool use_small_neighborhood_trimming = true;
    constexpr bool trim_biconnected = false; // time consuming!
    constexpr bool use_additional_tm = true;
    // constexpr bool use_additional_tm = true && !trim_biconnected; // biconnected finding is stronger, but slower

    VB _was_(N);
    VB _was2_(N);
    vector<unsigned short> neigh_in_S(N);
    VVB ends(N, VB(N,false));
    auto isEndsNeighborhoodLargeEnough = [&]( FBSS3 & st ) {
        if constexpr (!use_small_neighborhood_trimming) return true;

        if( st.ends.empty()  ) return true;
        if( st.vs_node == vt && !st.ends.empty() ) return false; // this is another rule, but can be done here :)
        for( auto [a,b] : st.ends ) if( a == vt ) return false; // this is another rule, but can be done here :)

        _was_ = was;
        for( int i=0; i<S_nodes.size(); i++ ) if( S_nodes[i] != -1 ) _was_[ S_nodes[i] ] = true;


        {
            int cnt = 0;
            for( auto & arc : g[st.vs_node] ) if( !_was_[arc.v] ) cnt++;
            if( st.vs_node != vt && cnt == 0 ) return false;
        }

        for( auto [a,b] : st.ends ) {
            int cnt = 0;
            for( auto & arc : revg[a] ) if( !_was_[arc.v] ) cnt++;
            if( a != vt && cnt == 0 ) return false;

            cnt = 0;
            for( auto & arc: g[b] ) if( !_was_[arc.v] ) cnt++;
            if( b != vt && cnt == 0 ) return false;
        }

        int neigh_size = 0;
        auto checkNode = [&](int d) { if( !_was_[d] ){ neigh_size++; _was_[d] = true; } };

        fill(ALL(_was2_),false);

        constexpr bool exclude_uniquely_assignable_nodes = false;
        if constexpr (exclude_uniquely_assignable_nodes){
            bool changes_done = true;

            while(changes_done) {
                changes_done = false;

                for( auto [a,b] : st.ends ) {
                    for( int d : underV[a] ) neigh_in_S[d] = 0;
                    for( int d : underV[b] ) neigh_in_S[d] = 0;
                }
                for( auto [a,b] : st.ends ) {
                    if( a != vt && !_was2_[a] ) for( int d : underV[a] ) neigh_in_S[d]++;
                    if( b != vt && !_was2_[b] ) for( int d : underV[b] ) neigh_in_S[d]++;
                }
                for( auto [a,b] : st.ends ) {
                    if(a != vt) for( auto & arc : revg[a] ) {
                        int d = arc.v;
                        if( !_was_[d] && !_was2_[a] && neigh_in_S[d] == 1 ) {
                            _was2_[a] = changes_done = true;
                            neigh_size++;
                        }
                    }
                    if(b != vt) for( auto & arc : g[b] ) {
                        int d = arc.v;
                        if( !_was_[d] && !_was2_[b] && neigh_in_S[d] == 1 ) {
                            _was2_[b] = changes_done = true;
                            neigh_size++;
                        }
                    }
                }
            }
        }

        int valid_ends = st.ends.size();
        for( auto [a,b] : st.ends ) {
            if( a != vt && !_was2_[a] ) for( auto & arc : revg[a] ) checkNode(arc.v);
            if( b != vt && !_was2_[b] ) for( auto & arc : g[b] ) checkNode(arc.v);
        }

        bool res = true;
        if( st.vs_node != vt && st.containsNode(vt) ) {
            if constexpr (debug) clog << "CASE1 " << endl;
            res = (neigh_size >= valid_ends);
        }
        else res = (neigh_size >= valid_ends+1);

        if constexpr (debug) if(!res) {
            clog << "State " << st << " has too small ends-neighborhood" << endl;
            DEBUG(neigh_size);
            DEBUG(valid_ends);
        }
        return res;
    };



    auto getAdditionalTm = [&](FBSS3 & st, PII m, int v)-> int {
        _was_ = was;
        for( int i=0; i<S_nodes.size(); i++ ) if( S_nodes[i] != -1 ) _was_[ S_nodes[i] ] = true;
        _was_[v] = false;

        m.first |= ( 1 << ind_in_S[st.vs_node] );
        for( auto [a,b] : st.ends ) {
            m.first |= ( 1 << ind_in_S[a] );
            m.first |= ( 1 << ind_in_S[b] );
        }

        int neigh_size = 0;
        auto checkNode = [&](int d) {
            if( !_was2_[d] ){ neigh_size++; _was2_[d] = true; return 1; }
            return 0;
        };

        bool changes_done = true;
        int tm = 0;
        while( changes_done ) {
            changes_done = false;
            for( int i=0; i<32; i++ ) {
                if( S_nodes[i] != -1 && S_nodes[i] != vt && ( m.first & (1<<i) ) == 0 ) {
                    neigh_size = 0;
                    _was2_ = _was_;

                    int a = S_nodes[i];
                    for( int d : underV[a] ) checkNode(d);
                    bool can_mark = (neigh_size <= 1);

                    {
                        if(!can_mark) {
                            neigh_size = 0;
                            _was2_ = _was_;
                            for( auto & arc : revg[a] ) neigh_size += checkNode(arc.v);
                            if( neigh_size == 0 ) can_mark = true;
                        }

                        if(!can_mark) {
                            neigh_size = 0;
                            _was2_ = _was_;
                            for( auto & arc : g[a] ) neigh_size += checkNode(arc.v);
                            if( neigh_size == 0 ) can_mark = true;
                        }

                        if(!can_mark) {
                            neigh_size = 0;
                            _was2_ = _was_;
                            for( int d : underV[a] ) checkNode(d);
                            can_mark = (neigh_size <= 1);
                        }
                    }

                    // if( neigh_size <= 1 ) {
                    if( can_mark ) {
                        tm |= (1 << i);
                        m.first |= (1 << i);
                        changes_done = true;
                    }
                }
            }
        }

        return tm;
    };

    BiconnectedFinder bf(N);
    DistanceFinder2 df(N);
    // returns pair where first boolean value tells whether state can be trimmed and second one - if state cannot be
    // trimmed - tells additional tm value to mark some loose nodes as taken
    auto trimBiconnectedOrFindDispensibleLooseNodes = [&]( FBSS3 & st, int v ) -> pair<bool, int> {
        if constexpr (!trim_biconnected) return {false,0};
        _was_ = was;
        for( int i=0; i<32; i++ ) if( S_nodes[i] != -1 ) _was_[ S_nodes[i] ] = false;


        fill(ALL(_was2_),false);
        for( auto [a,b] : st.ends ) _was2_[a] = _was2_[b] = true;

        for( auto [a,b] : st.ends ) {
            underV[a].push_back(b);
            underV[b].push_back(a);
            ends[a][b] = ends[b][a] = true;
        }

        auto clear = [&]() {
            for( auto [a,b] : st.ends ) {
                underV[a].pop_back();
                underV[b].pop_back();
                ends[a][b] = ends[b][a] = false;
            }
        };

        Mask mask(N);
        for( int i=0; i<N; i++ ) mask.set( i, _was_[i] );

        df.findDistancesFBS(underV, st.vs_node, mask, ind_in_S, v, ends);
        if( df.getDst(vt) == inf ){ clear(); return {true, 0}; }
        VI path;
        for( int u=vt; u != -1 ; u = df.getPar(u) ) path.push_back(u);
        reverse(ALL(path));
        Mask biconnected_trimmed = bf.getNodesAfterBiconnectedReduction2FBS(underV, mask, st.vs_node, vt,
                                    path, ind_in_S, v, ends );

        clear();

        int tm = 0;
        for( int i=0; i<N; i++ ) {
            if(biconnected_trimmed[i]) continue;
            if( _was2_[i] ){ return {true,0}; }
            if( ind_in_S[i] != -1 && i != st.vs_node && !_was2_[i] ) tm |= ( 1 << ind_in_S[i] );
        }


        return {false, tm};
    };


    auto trimDispensibleLooseNodes = [&](FBSS3 & st, VI & loose, int v) {
        if constexpr ( !trim_biconnected && !use_additional_tm ) return;

        auto tm2 = getAdditionalTm(st, {0,0}, v);
        for( int i=(int) loose.size()-1; i>=0; i--) {
            int a = loose[i];
            if( a == vt ) continue;

            if( tm2 & (1 << ind_in_S[a]) ) {
                swap(loose[i], loose.back());
                loose.pop_back();
            }
        }
    };



    constexpr int BUCKETS_BITS = 10;
    constexpr int BUCKETS = (1<<BUCKETS_BITS);
    constexpr int BH = BUCKETS-1;

    auto getHashForTakenMask = [&](int m) {
        HashType h;
        for( int i=0; i<32; i++ ) if(m & (1<<i)) h ^= node_hashes[ S_nodes[i] ];
        if constexpr (include_iter_cnt_in_hash) h.hash &= ( ULLONG_MAX >> expansion_iteration_shift_bits );
        return h;
    };

    vector<vector<FBSS3>> states(BUCKETS);
    vector<vector<FBSS3>> new_states(BUCKETS);

    auto countAllStates = [&](auto & sts) {
        return accumulate(ALL(sts), 0, [](int a, auto & vec){ return a + vec.size(); });
    };

    auto transformTakenMaskToNodes = [&]( int m ) {
        auto bits = FBSS3::getSetBits(m);
        VI temp;
        for(int d : bits) {
            assert( S_nodes[d] != -1 );
            temp.push_back( S_nodes[d] );
        }
        return temp;
    };

    vector<HashType> S_hashes(N, 0);
    VI tm_shifter(N,-1);
    VB is_vsnode_or_end(N);

    constexpr bool apply_shifters = true; // works very well for dense parts of graphs

    auto applyShifters = [&](int &m, FBSS3 & st ) {
        is_vsnode_or_end[st.vs_node] = true;
        for( auto [a,b] : st.ends ) is_vsnode_or_end[a] = is_vsnode_or_end[b] = true;

        for( int i=0; i<32; i++ ) {
            if( m & (1 << i) ) {
                int u = S_nodes[i];
                if( tm_shifter[u] == -1 ){ m ^= (1<<i); continue; }

                int prev = i;
                while( tm_shifter[u] != u ) {
                    if( !is_vsnode_or_end[u] && ( m & (1 << ind_in_S[u]) ) == 0 ) {
                        // clog << "Shifting node " << u << " to " << tm_shifter[u] << endl;
                        m ^= ( 1 << prev );
                        m ^= ( 1 << ind_in_S[u] );
                        prev = ind_in_S[u];
                    }
                    u = tm_shifter[u];
                }
            }
        }

        is_vsnode_or_end[st.vs_node] = false;
        for( auto [a,b] : st.ends ) is_vsnode_or_end[a] = is_vsnode_or_end[b] = false;
    };


    auto placeIntoNewBucket = [&]( FBSS3 & st ) {
        if constexpr (use_additional_tm || apply_shifters || trim_biconnected) st.makeUnique(st.taken_masks);

        if constexpr (debug) {
            if(!st.isCorrectState(ind_in_S)) {
                clog << endl << "!!!!" << endl << "State is incorrect!" << endl;
                DEBUG(st);
                assert( st.isCorrectState(ind_in_S) );
            };
        }
        if( st.taken_masks.empty() ) return;

        // st.recalculateHash();
        // StandardUtils::makeUnique(st.taken_masks);
        // FBSS3::makeUnique(st.taken_masks);
        new_states[st.hash.hash & BH].push_back( st );

        if constexpr (debug && compress_on_the_fly) {
            for( PII m : st.taken_masks ) {
                assert( state_id_mapper.contains( st.hash ^ getHashForTakenMask(m.first) ) );
                aux_st_mapper[ state_id_mapper[st.hash ^ getHashForTakenMask(m.first)] ] = st;
                VI newmasks = transformTakenMaskToNodes(m.first);
                VPII new_takenmasks(newmasks.size(), {-1,-1});
                for(int i=0; i<newmasks.size(); i++) new_takenmasks[i].first = newmasks[i];
                aux_st_mapper[ state_id_mapper[st.hash ^ getHashForTakenMask(m.first)] ].taken_masks = new_takenmasks;
            }
        }

        if constexpr (debug) clog << "\t\tCreated and added new_state: " << st << endl;
    };


    auto considerTransition = [&]( PII m, int tm, HashType st_hash, int minl_addition, FBSS3 & new_st, int a, int b, int c = -1, int d = -1 ) {
        auto h1 = st_hash ^ getHashForTakenMask(m.first);
        trackState(h1);

        m.first |= tm;
        if constexpr (apply_shifters) applyShifters(m.first, new_st);

        auto h2 = new_st.hash ^ getHashForTakenMask(m.first);
        trackState(h2);

        // if(h1 != h2) addEdge( h1,h2, (a == -1 ? -1 : arc_id_mapper[a][b]), (c == -1 ? -1 : arc_id_mapper[c][d])  );
        if(h1 != h2) addEdge( h1,h2, (a == -1 ? -1 : getArcId(a,b)), (c == -1 ? -1 :getArcId(c,d))  );
        new_st.taken_masks.emplace_back(m.first, m.second + minl_addition);
    };

    auto addAsIgnoredNode = [&]( int v, const FBSS3 & st, VI & loose ) {
        if constexpr (debug) clog << "\tAdding as ignored node" << endl;
        auto new_st = st;
        VPII().swap(new_st.taken_masks);
        // new_st.bnd_hash = cur_vis_hash; // do not create 'new state' for
        new_st.recalculateHash();
        new_st.hash ^= cur_vis_hash; // do not create 'new state' for
        if constexpr (include_iter_cnt_in_hash) updateHashWithIterationCnt(new_st.hash);
        int tm = 0;
        int minl_addition = 0;

        if( !isEndsNeighborhoodLargeEnough( new_st ) ) return;
        auto [can_trim, tm2] = trimBiconnectedOrFindDispensibleLooseNodes(new_st,v);
        if( can_trim ) return;
        else tm |= tm2;
        if constexpr (use_additional_tm) tm |= getAdditionalTm( new_st, {0,0}, v );

        for( PII m : st.taken_masks ) {
            if constexpr (debug) clog << "\tconsidering m: " << m << endl;
            considerTransition( m, tm, st.hash, minl_addition, new_st, -1, -1 );
        }

        placeIntoNewBucket(new_st);
    };

    auto addAsVsNode = [&](int v, const FBSS3 & st, VI & loose) {
        if constexpr (debug) clog << "\tAdding as vs_node" << endl;
        if( !in_neigh[st.vs_node] ) return;

        auto new_st = st;
        new_st.vs_node = v;
        VPII().swap(new_st.taken_masks);
        new_st.minl += g.getArcMinl(st.vs_node,v);
        // new_st.bnd_hash = cur_vis_hash;
        new_st.recalculateHash();
        new_st.hash ^= cur_vis_hash;
        if constexpr (include_iter_cnt_in_hash) updateHashWithIterationCnt(new_st.hash);
        int tm = (1 << ind_in_S[st.vs_node]);
        int minl_addition = g.getArcMinl(st.vs_node,v);

        if( !isEndsNeighborhoodLargeEnough( new_st ) ) return;
        auto [can_trim, tm2] = trimBiconnectedOrFindDispensibleLooseNodes(new_st,v);
        if( can_trim ) return;
        else tm |= tm2;
        if constexpr (use_additional_tm) tm |= getAdditionalTm( new_st, {0,0}, v );


        for( PII m : st.taken_masks ) {
            assert( (m.first & (1 <<  ind_in_S[st.vs_node] )) == 0 ); // st.vs_node cannot be taken in this case...
            if constexpr (debug) clog << "\tconsidering m: " << m << endl;
            considerTransition( m, tm, st.hash, minl_addition, new_st, st.vs_node, v );
        }

        placeIntoNewBucket(new_st);
    };

    auto addAsStartNode = [&](int v, FBSS3 & st, VI loose) {
        trimDispensibleLooseNodes( st, loose, v );

        { // adding as start node to an existing loose node
            if constexpr (debug) clog << "\tAdding as start node: v -> loose" << endl;
            for( int a : loose ) {
                if( !out_neigh[a] ) continue;

                auto new_st = st;
                // new_st.bnd_hash = cur_vis_hash;
                VPII().swap(new_st.taken_masks);
                new_st.ends.emplace_back( v, a );
                new_st.minl += g.getArcMinl(v,a);
                new_st.recalculateHash();
                new_st.hash ^= cur_vis_hash;
                if constexpr (include_iter_cnt_in_hash) updateHashWithIterationCnt(new_st.hash);
                int tm = 0;
                int minl_addition = g.getArcMinl(v,a);

                if( !isEndsNeighborhoodLargeEnough( new_st ) ) continue;
                auto [can_trim, tm2] = trimBiconnectedOrFindDispensibleLooseNodes(new_st,v);
                if( can_trim ) continue;
                else tm |= tm2;
                if constexpr (use_additional_tm) tm |= getAdditionalTm( new_st, {0,0}, v );

                for( PII m : st.taken_masks ) {
                    if( m.first & (1 << ind_in_S[a]) ) continue; // if a is taken, then cannot crete new state...
                    considerTransition( m, tm, st.hash, minl_addition, new_st, v,a );
                }

                placeIntoNewBucket(new_st);
            }
        }

        { // adding as start node to an existing path end
            if constexpr (debug) clog << "\tAdding as start node: v -> ends" << endl;
            for( int i=0; i<st.ends.size(); i++ ) {
                int a = st.ends[i].first;
                if( !out_neigh[a] ) continue;

                auto new_st = st;
                // new_st.bnd_hash = cur_vis_hash;
                VPII().swap(new_st.taken_masks);
                new_st.ends[i].first = v;
                new_st.minl += g.getArcMinl(v,a);
                new_st.recalculateHash();
                new_st.hash ^= cur_vis_hash;
                if constexpr (include_iter_cnt_in_hash) updateHashWithIterationCnt(new_st.hash);
                int tm = ( 1 << ind_in_S[a] ); // these nodes will be added to taken...
                int minl_addition =  g.getArcMinl(v,a);

                if( !isEndsNeighborhoodLargeEnough( new_st ) ) continue;
                auto [can_trim, tm2] = trimBiconnectedOrFindDispensibleLooseNodes(new_st,v);
                if( can_trim ) continue;
                else tm |= tm2;
                if constexpr (use_additional_tm) tm |=  getAdditionalTm( new_st, {0,0}, v );


                for( PII m : st.taken_masks ) {
                    assert( !(m.first & (1 << ind_in_S[a])) ); // since we add v to path end, then a cannot be in taken_masks
                    considerTransition( m, tm, st.hash, minl_addition, new_st, v,a );
                }

                placeIntoNewBucket(new_st);
            }
        }

    };

    auto addAsMiddleNode = [&](int v, FBSS3 & st, VI loose) {
        trimDispensibleLooseNodes(st, loose, v);

        { // adding as middle node, vs_node -> v -> loose
            if constexpr (debug) clog << "\tAdding as middle node : vs_node -> v -> loose" << endl;
            for( int a : loose ) {
                if( !in_neigh[st.vs_node] ) continue;
                if( !out_neigh[a] ) continue;

                auto new_st = st;
                // new_st.bnd_hash = cur_vis_hash;
                VPII().swap(new_st.taken_masks);
                new_st.vs_node = a;
                new_st.minl += g.getArcMinl(st.vs_node,v) + g.getArcMinl(v,a);
                new_st.recalculateHash();
                new_st.hash ^= cur_vis_hash;
                if constexpr (include_iter_cnt_in_hash) updateHashWithIterationCnt(new_st.hash);
                int tm = ( 1 << ind_in_S[st.vs_node] ) | ( 1 << ind_in_S[v] );
                int minl_addition = g.getArcMinl(st.vs_node,v) + g.getArcMinl(v,a);

                if( !isEndsNeighborhoodLargeEnough( new_st ) ) continue;
                auto [can_trim, tm2] = trimBiconnectedOrFindDispensibleLooseNodes(new_st,v);
                if( can_trim ) continue;
                else tm |= tm2;
                if constexpr (use_additional_tm) tm |= getAdditionalTm( new_st, {0,0}, v );

                for( PII m : st.taken_masks ) {
                    if( m.first & (1 << ind_in_S[a]) ) continue;
                    considerTransition( m, tm, st.hash, minl_addition, new_st, st.vs_node, v, v, a );
                }

                placeIntoNewBucket(new_st);
            }

        }

        { // adding as middle node, vs_node -> v -> path end
            if constexpr (debug) clog << "\tAdding as middle node : vs_node -> v -> ends" << endl;
            for( auto [a,b] : st.ends ) {
                if( !in_neigh[st.vs_node] ) continue;
                if( !out_neigh[a] ) continue;

                auto new_st = st;
                // new_st.bnd_hash = cur_vis_hash;
                VPII().swap(new_st.taken_masks);
                new_st.vs_node = b;
                new_st.removeEnds({a,b});
                new_st.minl += g.getArcMinl(st.vs_node,v) + g.getArcMinl(v,a);
                new_st.recalculateHash();
                new_st.hash ^= cur_vis_hash;
                if constexpr (include_iter_cnt_in_hash) updateHashWithIterationCnt(new_st.hash);
                int tm = ( 1 << ind_in_S[st.vs_node] ) | ( 1 << ind_in_S[v] ) | ( 1 << ind_in_S[a] );
                int minl_addition = g.getArcMinl(st.vs_node,v) + g.getArcMinl(v,a);

                if( !isEndsNeighborhoodLargeEnough( new_st ) ) continue;
                auto [can_trim, tm2] = trimBiconnectedOrFindDispensibleLooseNodes(new_st,v);
                if( can_trim ) continue;
                else tm |= tm2;
                if constexpr (use_additional_tm) tm |= getAdditionalTm( new_st, {0,0}, v );

                for( PII m : st.taken_masks ) {
                    // all taken masks are passed, as no loose node is used...
                    considerTransition( m, tm, st.hash, minl_addition, new_st, st.vs_node, v, v, a );
                }

                placeIntoNewBucket(new_st);
            }
        }

        { // adding as middle node, loose -> v -> loose
            if constexpr (debug) clog << "\tAdding as middle node : loose -> v -> loose" << endl;
            for( int a : loose ) {
                if( !in_neigh[a] ) continue;

                for( int b : loose ) {
                    if( a == b ) continue;
                    if( !out_neigh[b] ) continue;

                    auto new_st = st;
                    // new_st.bnd_hash = cur_vis_hash;
                    VPII().swap( new_st.taken_masks );
                    new_st.ends.emplace_back(a,b);
                    new_st.minl += g.getArcMinl(a,v) + g.getArcMinl(v,b);
                    new_st.recalculateHash();
                    new_st.hash ^= cur_vis_hash;
                    if constexpr (include_iter_cnt_in_hash) updateHashWithIterationCnt(new_st.hash);
                    int tm = ( 1 << ind_in_S[v] );
                    int minl_addition =  g.getArcMinl(a,v) + g.getArcMinl(v,b);

                    if( !isEndsNeighborhoodLargeEnough( new_st ) ) continue;
                    auto [can_trim, tm2] = trimBiconnectedOrFindDispensibleLooseNodes(new_st,v);
                    if( can_trim ) continue;
                    else tm |= tm2;
                    if constexpr (use_additional_tm) tm |= getAdditionalTm( new_st, {0,0}, v );

                    for( PII m : st.taken_masks ) {
                        if( m.first & ( 1 << ind_in_S[a] ) ) continue;
                        if( m.first & ( 1 << ind_in_S[b] ) ) continue;
                        considerTransition( m, tm, st.hash, minl_addition, new_st, a, v,v,b );
                    }

                    placeIntoNewBucket(new_st);
                }
            }
        }

        { // adding as middle node, loose -> v -> path end
            if constexpr (debug) clog << "\tAdding as middle node : loose -> v -> ends" << endl;
            for( int a : loose ) {
                if( !in_neigh[a] ) continue;

                for( auto [b,c] : st.ends ) {
                    if( !out_neigh[b] ) continue;

                    auto new_st = st;
                    // new_st.bnd_hash = cur_vis_hash;
                    VPII().swap( new_st.taken_masks );
                    new_st.removeEnds( {b,c} );
                    new_st.ends.emplace_back(a,c);
                    new_st.minl += g.getArcMinl(a,v) + g.getArcMinl(v,b);
                    new_st.recalculateHash();
                    new_st.hash ^= cur_vis_hash;
                    if constexpr (include_iter_cnt_in_hash) updateHashWithIterationCnt(new_st.hash);
                    int tm = ( 1 << ind_in_S[v] ) | ( 1 << ind_in_S[b] );
                    int minl_addition = g.getArcMinl(a,v) + g.getArcMinl(v,b);

                    if( !isEndsNeighborhoodLargeEnough( new_st ) ) continue;
                    auto [can_trim, tm2] = trimBiconnectedOrFindDispensibleLooseNodes(new_st,v);
                    if( can_trim ) continue;
                    else tm |= tm2;
                    if constexpr (use_additional_tm) tm |= getAdditionalTm( new_st, {0,0}, v );

                    for( PII m : st.taken_masks ) {
                        if( m.first & ( 1 << ind_in_S[a] ) ) continue;
                        considerTransition( m, tm, st.hash, minl_addition, new_st, a,v,v,b );
                    }

                    placeIntoNewBucket(new_st);
                }
            }
        }

        { // adding as middle node, path_end -> v -> loose
            if constexpr (debug) clog << "\tAdding as middle node : ends -> v -> loose" << endl;
            for( auto [a,b] : st.ends ) {
                if( !in_neigh[b] ) continue;

                for( int c : loose ) {
                    if( !out_neigh[c] ) continue;

                    auto new_st = st;
                    // new_st.bnd_hash = cur_vis_hash;
                    VPII().swap( new_st.taken_masks );
                    new_st.removeEnds({a,b});
                    new_st.ends.emplace_back(a,c);
                    new_st.minl += g.getArcMinl(b,v) + g.getArcMinl(v,c);
                    new_st.recalculateHash();
                    new_st.hash ^= cur_vis_hash;
                    if constexpr (include_iter_cnt_in_hash) updateHashWithIterationCnt(new_st.hash);
                    int tm = ( 1 << ind_in_S[b] ) | ( 1 << ind_in_S[v] );
                    int minl_addition = g.getArcMinl(b,v) + g.getArcMinl(v,c);

                    if( !isEndsNeighborhoodLargeEnough( new_st ) ) continue;
                    auto [can_trim, tm2] = trimBiconnectedOrFindDispensibleLooseNodes(new_st,v);
                    if( can_trim ) continue;
                    else tm |= tm2;
                    if constexpr (use_additional_tm) tm |= getAdditionalTm( new_st, {0,0}, v );

                    for( PII m : st.taken_masks ) {
                        if( m.first & ( 1 << ind_in_S[c] ) ) continue;
                        considerTransition( m, tm, st.hash, minl_addition, new_st, b,v,v,c );
                    }

                    placeIntoNewBucket(new_st);
                }
            }
        }

        { // adding as middle node, path_end -> v -> path_end
            if constexpr (debug) clog << "\tAdding as middle node : ends -> v -> ends" << endl;
            for( auto [a,b] : st.ends ) {
                if( !in_neigh[b] ) continue;

                for( auto & [c,d] : st.ends ) {
                    if( a == c || b == d ) continue;
                    if( !out_neigh[c] ) continue;

                    auto new_st = st;
                    // new_st.bnd_hash = cur_vis_hash;
                    VPII().swap( new_st.taken_masks );
                    new_st.removeEnds({a,b});
                    new_st.removeEnds({c,d});
                    new_st.ends.emplace_back(a,d);
                    new_st.minl += g.getArcMinl(b,v) + g.getArcMinl(v,c);
                    new_st.recalculateHash();
                    new_st.hash ^= cur_vis_hash;
                    if constexpr (include_iter_cnt_in_hash) updateHashWithIterationCnt(new_st.hash);
                    int tm = ( 1 << ind_in_S[b] ) | ( 1 << ind_in_S[v] ) | ( 1 << ind_in_S[c] );
                    int minl_addition = g.getArcMinl(b,v) + g.getArcMinl(v,c);

                    if( !isEndsNeighborhoodLargeEnough( new_st ) ) continue;
                    auto [can_trim, tm2] = trimBiconnectedOrFindDispensibleLooseNodes(new_st,v);
                    if( can_trim ) continue;
                    else tm |= tm2;
                    if constexpr (use_additional_tm) tm |= getAdditionalTm( new_st, {0,0}, v );

                    for(PII m :st.taken_masks) {
                        // all taken masks are passed, since no loose nodes are used
                        considerTransition( m, tm, st.hash, minl_addition, new_st, b,v,v,c );
                    }

                    placeIntoNewBucket(new_st);
                }
            }
        }
    };

    auto addAsEndNode = [&](int v, FBSS3 & st, VI loose) {
        trimDispensibleLooseNodes( st,loose, v );

        { // adding as end node to an existing loose node
            if constexpr (debug) clog << "\tAdding as end node : loose -> v" << endl;
            for( int a : loose ) {
                if( !in_neigh[a] ) continue;

                auto new_st = st;
                // new_st.bnd_hash = cur_vis_hash;
                VPII().swap(new_st.taken_masks);
                new_st.ends.emplace_back( a,v );
                new_st.minl += g.getArc( a,v ).minl();
                new_st.recalculateHash();
                new_st.hash ^= cur_vis_hash;
                if constexpr (include_iter_cnt_in_hash) updateHashWithIterationCnt(new_st.hash);
                int tm = 0;
                int minl_addition = g.getArc( a,v ).minl();

                if( !isEndsNeighborhoodLargeEnough( new_st ) ) continue;
                auto [can_trim, tm2] = trimBiconnectedOrFindDispensibleLooseNodes(new_st,v);
                if( can_trim ) continue;
                else tm |= tm2;
                if constexpr (use_additional_tm) tm |= getAdditionalTm( new_st, {0,0}, v );

                for( PII m : st.taken_masks ) {
                    if( m.first & ( 1 << ind_in_S[a] ) ) continue;
                    considerTransition( m, tm, st.hash, minl_addition, new_st, a,v );
                }

                placeIntoNewBucket(new_st);
            }
        }

        { // adding as end node to an existing path end
            if constexpr (debug) clog << "\tAdding as end node : ends -> v" << endl;
            for( int i=0; i<st.ends.size(); i++ ) {
                int b = st.ends[i].second;
                if( !in_neigh[b] ) continue;

                auto new_st = st;
                // new_st.bnd_hash = cur_vis_hash;
                VPII().swap( new_st.taken_masks );
                new_st.ends[i].second = v;
                new_st.minl += g.getArc( b,v ).minl();
                new_st.recalculateHash();
                new_st.hash ^= cur_vis_hash;
                if constexpr (include_iter_cnt_in_hash) updateHashWithIterationCnt(new_st.hash);
                int tm = ( 1 << ind_in_S[b] );
                int minl_addition = g.getArc( b,v ).minl();

                if( !isEndsNeighborhoodLargeEnough( new_st ) ) continue;
                auto [can_trim, tm2] = trimBiconnectedOrFindDispensibleLooseNodes(new_st,v);
                if( can_trim ) continue;
                else tm |= tm2;
                if constexpr (use_additional_tm) tm |= getAdditionalTm( new_st, {0,0}, v );

                for(PII m : st.taken_masks) {
                    // all taken masks are forwarded, since no loose nodes are used
                    considerTransition( m, tm, st.hash, minl_addition, new_st, b,v );
                }

                placeIntoNewBucket(new_st);
            }
        }
    };


    auto mergeFBSS3States = [&]( vector<FBSS3> & sts ) {
        sort(ALL(sts));

        int p = 0, q = 0, r = 0;

        while( p < sts.size() ){
            q = p;
            while( q < sts.size() && (sts[p] == sts[q]) ) q++;
            for( int i=p+1; i<q; i++ ){
                sts[p].taken_masks += sts[i].taken_masks;
                if constexpr (debug) clog << "------>" << endl << "merging state " << sts[i] << " to " << sts[p] << endl;
            }
            if( p+1 < q ) FBSS3::makeUnique(sts[p].taken_masks);
            swap( sts[r++], sts[p] );
            p = q;
        }

        while( sts.size() > r ) sts.pop_back();
    };

    auto getMstEdges = [&](FBSS3& st, int taken_mask1 = 0, int taken_mask2 = 0) {
        vector<Triple<int>> edges;
        edges.reserve( (S.size()+2) * (S.size()+2) / 4 );

        if constexpr (debug)
            clog << endl << "Finding MST edges for state " << st << endl;

        auto checkPair = [&](int a, int b) {
            if constexpr (debug)
                clog << "Considering pair (" << a << ", "  << b << ")" << endl;

            int aa = ind_in_S[a];
            assert( aa != -1 );
            int bb = ind_in_S[b];
            if( bb == -1 ) assert( b == vt );

            int d1 = ( (bb == -1) ? distances[taken_mask1][aa].back() : distances[taken_mask1][aa][bb] );
            int d2 = ( (bb == -1) ? distances[taken_mask2][aa].back() : distances[taken_mask2][aa][bb] );

            int _dst = max( d1,d2 );
            if( _dst != inf ) {
                edges.emplace_back( _dst, a, b );
            }
        };

        checkPair(st.vs_node, vt);

        for( auto [a,b] : st.ends ) {
            checkPair(st.vs_node,a);
            checkPair(b,vt);

            for( auto [c,d] : st.ends ) {
                if( a == c || a == d || b == c || b == d ) continue;
                checkPair(b,c);
            }
        }
        if constexpr (debug) {
            clog << "edges: " << endl;
            for(auto [w,a,b] : edges) clog << "\t(" << w << "," << a << "," << b << ")" << endl;
            // DEBUG(distances[st.vs_node]);
        }

        // now find mst
        FAU fau(N);
        for( auto [a,b] : st.ends ) fau.Union(a,b);
        sort(ALL(edges), []( auto & a, auto & b ){ return a.st < b.st; });
        vector<Triple<int>> res;
        for( auto & [w,a,b] : edges ) {
            if( fau.Find(a) != fau.Find(b) ) {
                if constexpr (debug)
                    clog << "\tuniting nodes " << a << " and " << b << endl;
                res.emplace_back(w,a,b);
                fau.Union(a,b);
            }
        }

        bool is_connected_mst = ( fau.Find(st.vs_node) == fau.Find(vt) );
        for( auto [a,b] : st.ends ) if( fau.Find(a) != fau.Find(st.vs_node) ) is_connected_mst = false;

        return make_pair(is_connected_mst,res);
    };

    auto canTrimMST = [&](FBSS3 & st, int taken_mask1 = 0, int taken_mask2 = 0) {
        auto [is_connected,edges] = getMstEdges(st, taken_mask1, taken_mask2);

        if constexpr (debug) if( !is_connected ) {
            clog << "MST created for state st: " << st << " is not connected, trimming!" << endl;
            clog << "edges: " << endl;
            for(auto [w,a,b] : edges) clog << "\t(" << w << "," << a << "," << b << ")" << endl;
        }

        if( !is_connected ) return true;

        LL sumw = accumulate( ALL(edges), 0ll, [](LL a, auto & e){ return a + e.st; } );
        if( sumw >= inf || sumw + st.minl > L ) {
            if constexpr (debug) {
                clog << endl << "Length conditon violated for st: " << st << " and edges: " << endl;
                for(auto [w,a,b] : edges) clog << "\t(" << w << "," << a << "," << b << ")" << endl;
            }
            return true;
        }

        return false;
    };

    auto canTrimUnreachable = [&](FBSS3 & st) {
        // checks if from st.vs_node one can reach vt in a graph, where taken nodes (marked by m.first) are already
        // taken
        VI SS;
        for( int i=0; i<S_nodes.size(); i++ ) if( S_nodes[i] != -1 ) SS.push_back(S_nodes[i]);
        for(int d : SS) assert(was[d]);
        VI neigh; neigh.reserve(N);
        VI paired_node(N,-1);
        for(auto [a,b] : st.ends) paired_node[a] = b;

        auto isConnected = [&](PII m) {
            for(int d : SS) was[d] = false;
            assert(!was[st.vs_node]);
            for( int i=0; i<32; i++ ) if( m.first & (1<<i) ) if( !was[S_nodes[i]] ) was[S_nodes[i]] = true;

            neigh.clear(); neigh.push_back(st.vs_node);
            VB vis(N,false);
            for( int i=0; i<neigh.size(); i++ ) {
                int v = neigh[i];
                vis[v] = true;
                for( auto & arc : g[v] ) {
                    int d = arc.v;
                    if( !vis[d] && !was[d] ) { // was marks all nodes visited so far, without boundary nodes except for taken nodes
                        vis[d] = true;
                        neigh.push_back(d);
                        int d2 = paired_node[d];
                        if( d2 != -1 && !vis[d2] && !was[d2] ) {
                            vis[d2] = true;
                            neigh.push_back(d2);
                        }
                    }
                    if( vis[vt] ) break;
                }
                if( vis[vt] ) break;
            }

            for(int d : SS) was[d] = true;
            return !vis[vt];
        };

        for( int i=(int)st.taken_masks.size()-1; i>=0; i-- ) {
            PII & m = st.taken_masks[i];
            bool can_trim_m = isConnected(m);
            if( can_trim_m ) {
                swap( st.taken_masks[i], st.taken_masks.back() );
                st.taken_masks.pop_back();
            }
        }

        return st.taken_masks.empty();
    };

    auto canTrimPermutation = [&]( FBSS3 & st, VPII & trimmed_masks ) {
        if constexpr (debug) clog << "canTrimPermutation, st: " << st << endl;

        VPII pairs(1, {-1, st.vs_node});
        pairs += st.ends;
        pairs.emplace_back( vt, -1 );

        sort( pairs.begin()+1, pairs.begin() + 1 + st.ends.size() );

        auto getTakenMasks34 = [&](VPII pairs, int taken_mask1, int taken_mask2, int max_size)
            -> PII{
            int _taken_mask3 = 0;
            int _taken_mask4 = 0;

            for( int j=0; j<max_size; j++ ) {
                int x = pairs[j % pairs.size()].second;
                int y = pairs[(j+1) % pairs.size()].first;
                if( x != -1 && popcount((unsigned)_taken_mask3) < max_size ) {
                    _taken_mask3 |= (1 << ind_in_S[x] );
                }
                if( y != -1 && y != vt && popcount((unsigned)_taken_mask3) < max_size ) {
                    _taken_mask3 |= (1 << ind_in_S[y] );
                }

                x = pairs[(j+pairs.size()/2) % pairs.size()].second;
                y = pairs[(j+1+pairs.size()/2) % pairs.size()].first;
                if( x != -1 && popcount((unsigned)_taken_mask4) < max_size ) {
                    _taken_mask4 |= (1 << ind_in_S[x] );
                }
                if( y != -1 && y != vt && popcount((unsigned)_taken_mask4) < max_size ) {
                    _taken_mask4 |= (1 << ind_in_S[y] );
                }
            }

            assert( (_taken_mask3 & taken_mask1) == 0 );
            assert( (_taken_mask4 & taken_mask2) == 0 );

            if( popcount((unsigned)(_taken_mask3 | taken_mask1)) <= max_size ) _taken_mask3 |= taken_mask1;
            if( popcount((unsigned)(_taken_mask4 | taken_mask2)) <= max_size ) _taken_mask4 |= taken_mask2;

            return {_taken_mask3, _taken_mask4};
        };



        auto getDst = [&](int taken_mask1 = 0, int taken_mask2 = 0, int curl) {
            LL l = 0;
            if constexpr (debug) { DEBUG(taken_mask1); DEBUG(taken_mask2); }
            // auto [_taken_mask3, _taken_mask4] = getTakenMasks34(pairs, taken_mask1, taken_mask2, DST_MAX_MASK_SIZE);

            for(int i=1; i<pairs.size(); i++) {
                int a = pairs[i-1].second;
                int b = pairs[i].first;
                if constexpr (debug) {DEBUG(a); DEBUG(b);}

                int aa = ind_in_S[a];
                // assert( aa != -1 );
                int bb = ind_in_S[b];
                if constexpr (debug){ DEBUG(aa); DEBUG(bb); }

                // assert( bb != -1 || b == vt );
                // if( ind_in_S[vt] != -1 ) assert( bb != -1 );

                // int taken_mask3 = ( _taken_mask3 & ~(1 << ind_in_S[a]));
                // int taken_mask4 = ( _taken_mask4 & ~(1 << ind_in_S[a]));
                // if(bb != -1) {
                //     taken_mask3 &= ~(1 << ind_in_S[b]);
                //     taken_mask4 &= ~(1 << ind_in_S[b]);
                // }

                // int taken_mask5 = 0;
                // auto check = [&](){ return popcount((unsigned)taken_mask5) < DST_MAX_MASK_SIZE; };
                // if( pairs[i].second != -1 && check() ) taken_mask5 |= (1 << ind_in_S[pairs[i].second] );
                // if( i+1 < pairs.size() && pairs[i+1].first != vt && check() ) taken_mask5 |= (1 << ind_in_S[pairs[i+1].first] );
                // if( i+1 < pairs.size() && pairs[i+1].second != -1 && check() ) taken_mask5 |= (1 << ind_in_S[pairs[i+1].second] );
                // if( i > 1 && check() ) taken_mask5 |= (1 << ind_in_S[pairs[i-2].second] );
                // if( i > 1 && pairs[i-2].first != -1 && check() ) taken_mask5 |= (1 << ind_in_S[pairs[i-2].first] );

                int d1 = ( (bb == -1) ? distances[taken_mask1][aa].back() : distances[taken_mask1][aa][bb] );
                int d2 = ( (bb == -1) ? distances[taken_mask2][aa].back() : distances[taken_mask2][aa][bb] );
                // int d3 = ( (bb == -1) ? distances[taken_mask3][aa].back() : distances[taken_mask3][aa][bb] );
                // int d4 = ( (bb == -1) ? distances[taken_mask4][aa].back() : distances[taken_mask4][aa][bb] );
                // int d5 = ( (bb == -1) ? distances[taken_mask5][aa].back() : distances[taken_mask5][aa][bb] );

                int d12 = max(d1,d2);
                // int d34 = max(d3,d4);

                if constexpr (debug) {DEBUG(d1); DEBUG(d2);}
                if constexpr (debug) DEBUG(distances[taken_mask1][aa]);
                if constexpr (debug) DEBUG(distances[taken_mask2][aa]);
                if constexpr (debug) if(bb != -1) DEBUG(distances[taken_mask1][bb]);
                if constexpr (debug) if(bb != -1) DEBUG(distances[taken_mask2][bb]);
                if constexpr (debug) DEBUG(distances[0][aa]);

                // l += max( d5, max(d12,d34) );
                // l += max(d12,d34);
                l += d12;
                if constexpr (debug) clog << "\tadding d12: " << d12 << endl;

                // if( l + curl > L ) return l;
            }

            return l;
        };

        auto checkDominators = [&](int taken_mask1, int taken_mask2) {
            if( dominators.empty() ) return false;

            if constexpr (debug) { DEBUG(taken_mask1); DEBUG(taken_mask2); }
            DOM_MASK all_dom(0);
            auto [_taken_mask3, _taken_mask4] =
                getTakenMasks34(pairs, taken_mask1, taken_mask2, DOM_MAX_MASK_SIZE);

            for(int i=1; i<pairs.size(); i++) {
                int a = pairs[i-1].second;
                int b = pairs[i].first;
                if constexpr (debug) {DEBUG(a); DEBUG(b);}

                int aa = ind_in_S[a];
                int bb = ind_in_S[b];
                if constexpr (debug){ DEBUG(aa); DEBUG(bb); }

                int taken_mask3 = ( _taken_mask3 & ~(1 << aa));
                if(bb != -1) taken_mask3 &= ~(1 << bb);
                int taken_mask4 = ( _taken_mask4 & ~(1 << aa));
                if(bb != -1) taken_mask4 &= ~(1 << bb);

                if( bb == -1 ) assert( b == vt );

                DOM_MASK cur_dom(0);
                cur_dom |= ((bb == -1) ? dominators[taken_mask1][aa].back() : dominators[taken_mask1][aa][bb]);
                cur_dom |= ((bb == -1) ? dominators[taken_mask2][aa].back() : dominators[taken_mask2][aa][bb]);
                cur_dom |= ((bb == -1) ? dominators[taken_mask3][aa].back() : dominators[taken_mask3][aa][bb]);
                cur_dom |= ((bb == -1) ? dominators[taken_mask4][aa].back() : dominators[taken_mask4][aa][bb]);

                if( (cur_dom & all_dom).count() > 0 ) return true;
                all_dom |= cur_dom;
            }
            return false;
        };

        sort(st.taken_masks.rbegin(), st.taken_masks.rend());

        bool can_trim_zero_mask = false;
        int minl_among_tms = inf;
        for(auto [tm,ll] : st.taken_masks) minl_among_tms = min(minl_among_tms, ll);

        for( int i=(int)st.taken_masks.size()-1; i>=0; i-- ) {
            PII & m = st.taken_masks[i];
            auto [taken_masks1, taken_masks2] = FBSS3::getTakenMasks( m.first, DST_MAX_MASK_SIZE );
            auto [taken_masks_dom1, taken_masks_dom2] = FBSS3::getTakenMasks( m.first, DOM_MAX_MASK_SIZE );

            bool can_trim_m = true;

            constexpr bool use_iterative_version = true;

            sort( pairs.begin()+1, pairs.begin() + 1 + st.ends.size() );

            if constexpr (use_iterative_version) {
                do {
                    if( use_dominator_trimming && checkDominators(taken_masks_dom1, taken_masks_dom2) ) continue;

                    if constexpr (debug) clog << endl << "pairs: " << pairs << endl;
                    LL l = getDst(taken_masks1, taken_masks2, m.second);
                    if constexpr (debug) DEBUG(PII(l, m.second));
                    if( l + m.second <= L ){ can_trim_m = false; break; }

                }while( next_permutation( pairs.begin()+1, pairs.begin() + 1 + st.ends.size() ) );

                // if( can_trim_zero_mask && !can_trim_m && m.first != 0) {
                //     clog << endl << "--------------------> ERROR  !! !! !!" << endl;
                //     clog << "Can trim state " << st << " for mask " << 0 << endl;
                //     clog << "But cannot trim it for mask " << m.first << endl;
                //     assert(false && "dst-trimming error!");
                // }

                if( can_trim_m && m.first == 0 && m.second == minl_among_tms ) {
                    assert( i == (int)st.taken_masks.size()-1 );
                    // clog << "Can trim state st: " << st << endl << "for mask m: " << m << endl;
                    // DEBUG(S_nodes); can_trim_zero_mask = true; exit(2);
                    trimmed_masks += st.taken_masks;
                    st.taken_masks.clear();
                    return true;
                }

            }
            else {
                auto [taken_masks3, taken_masks4] =
                    getTakenMasks34(pairs, taken_masks1, taken_masks2, DST_MAX_MASK_SIZE);
                auto [taken_masks_dom3, taken_masks_dom4] =
                    getTakenMasks34(pairs, taken_masks_dom1, taken_masks_dom2, DOM_MAX_MASK_SIZE);

                // DEBUG(pairs);
                // function<bool(int, DOM_MASK, int)> considerForPairs = [&]( int ind, DOM_MASK all_dom, int dst ) {
                function<bool(int, int)> considerForPairs = [&]( int ind, int dst ) {
                    if(dst > L) return true;
                    // here we solve in true O(n!) instead of O(n! * n) as in the iterative version...
                    // but there probably is a bug somewhere here... and it does not seem to be faster in practice...
                    if( ind == pairs.size() ) {
                        // DEBUG(dst);
                        if( dst <= L ) return false;
                        else return true;
                    }

                    for( int i=ind; i<pairs.size(); i++ ) {
                        bool swap_cnd = (i+1 < pairs.size());
                        // bool swap_cnd = (i > ind && i+1 < pairs.size());
                        if( swap_cnd ) swap( pairs[i], pairs[ind] );
                        // DEBUG(pairs[ind]);

                        int a = pairs[ind-1].second, b = pairs[ind].first;
                        int aa = ind_in_S[a], bb = ind_in_S[b];


                        // DOM_MASK cur_dom(0);
                        // if( use_dominator_trimming ) {
                        //     int tm1_dom = ( taken_masks_dom1 & ~(1 << ind_in_S[a])); if(bb != -1) tm1_dom &= ~(1 << ind_in_S[b]);
                        //     int tm2_dom = ( taken_masks_dom2 & ~(1 << ind_in_S[a])); if(bb != -1) tm2_dom &= ~(1 << ind_in_S[b]);
                        //     int tm3_dom = ( taken_masks_dom3 & ~(1 << ind_in_S[a])); if(bb != -1) tm3_dom &= ~(1 << ind_in_S[b]);
                        //     int tm4_dom = ( taken_masks_dom4 & ~(1 << ind_in_S[a])); if(bb != -1) tm4_dom &= ~(1 << ind_in_S[b]);
                        //
                        //     cur_dom |= ((bb == -1) ? dominators[tm1_dom][aa].back() : dominators[tm1_dom][aa][bb]);
                        //     cur_dom |= ((bb == -1) ? dominators[tm2_dom][aa].back() : dominators[tm2_dom][aa][bb]);
                        //     cur_dom |= ((bb == -1) ? dominators[tm3_dom][aa].back() : dominators[tm3_dom][aa][bb]);
                        //     cur_dom |= ((bb == -1) ? dominators[tm4_dom][aa].back() : dominators[tm4_dom][aa][bb]);
                        //     if( (all_dom & cur_dom).count() > 0 ) {
                        //         if( swap_cnd ) swap( pairs[i], pairs[ind] );
                        //         return true;
                        //     }
                        // }


                        int tm3 = ( taken_masks3 & ~(1 << ind_in_S[a])); if(bb != -1) tm3 &= ~(1 << ind_in_S[b]);
                        int tm4 = ( taken_masks4 & ~(1 << ind_in_S[a])); if(bb != -1) tm4 &= ~(1 << ind_in_S[b]);

                        int d1 = ( (bb == -1) ? distances[taken_masks1][aa].back() : distances[taken_masks1][aa][bb] );
                        int d2 = ( (bb == -1) ? distances[taken_masks2][aa].back() : distances[taken_masks2][aa][bb] );
                        int d3 = ( (bb == -1) ? distances[tm3][aa].back() : distances[tm3][aa][bb] );
                        int d4 = ( (bb == -1) ? distances[tm4][aa].back() : distances[tm4][aa][bb] );
                        int d12 = max(d1,d2);
                        int d34 = max(d3,d4);
                        int d1234 = max(d12,d34);

                        // if( considerForPairs(ind+1, all_dom | cur_dom, min(inf,dst+d1234)) == false ) {
                        if( considerForPairs(ind+1, min(inf,dst+d1234)) == false ) {
                            if( swap_cnd ) swap( pairs[i], pairs[ind] );
                            return false;
                        }

                        if( swap_cnd ) swap( pairs[i], pairs[ind] );
                    }

                    return true;
                };

                // can_trim_m = considerForPairs(1, DOM_MASK(0), m.second);
                can_trim_m = considerForPairs(1, m.second);
            }

            if( can_trim_m ) {
                trimmed_masks.push_back(st.taken_masks[i]);
                swap( st.taken_masks[i], st.taken_masks.back() );
                st.taken_masks.pop_back();
            }
        }

        return st.taken_masks.empty();
    };


    bool use_mcf_trimming = false;
    bool mcf_trim_all_masks = false;
    int MCF_ENDS_SIZE_THR = 6;
    VVI mcfCost;
    VVI mcfCapacity;
    VVI mcf_adj;
    VVB mcf_ends;
    struct MCFEdge { int from, to, capacity, cost; };
    vector<MCFEdge> all_mcf_g_arcs;
    if(use_mcf_trimming) {
        mcf_adj = VVI(2*N+2);
        mcfCost.assign(2*N+2, VI(2*N+2,0));
        mcfCapacity.assign(2*N+2,VI(2*N+2,0));
        for(int i=0; i<N; i++) {
            all_mcf_g_arcs.emplace_back(2*i, 2*i+1, 1, 0);
            for(auto & arc : g[i]) all_mcf_g_arcs.emplace_back(2*i+1, 2*arc.v, 1, arc.minl());
        }
        mcf_ends.assign(N, VB(N,0));
    }


    // finds min-cost max-flow to bound the minimal length of a path for current state
    auto canTrimFlow = [&]( FBSS3 & st, VPII & trimmed_masks ) {
        if( st.ends.empty() ) return false;

        VVI &cost = mcfCost, &capacity = mcfCapacity, &adj = mcf_adj;
        const int INF = 1e9;

        auto spfa = [&adj, &capacity, &cost, &INF](int n, int v0, vector<int>& d, vector<int>& p) {
            d.assign(n, INF);
            d[v0] = 0;
            vector<bool> inq(n, false);
            queue<int> q;
            q.push(v0);
            p.assign(n, -1);

            while (!q.empty()) {
                int u = q.front();
                q.pop();
                inq[u] = false;
                for (int v : adj[u]) {
                    if (capacity[u][v] > 0 && d[v] > d[u] + cost[u][v]) {
                        d[v] = d[u] + cost[u][v];
                        p[v] = u;
                        if (!inq[v]) {
                            inq[v] = true;
                            q.push(v);
                        }
                    }
                }
            }
        };

        auto minCostFlow = [&adj, &capacity, &cost, &spfa](int N, vector<MCFEdge> edges, int K, int s, int t)
            -> PII {
            for(int i=0; i<N; i++) adj[i].clear();

            for (auto e : edges) {
                adj[e.from].push_back(e.to);
                adj[e.to].push_back(e.from);
                cost[e.from][e.to] = e.cost;
                cost[e.to][e.from] = -e.cost;
                capacity[e.from][e.to] = e.capacity;
            }

            int flow = 0;
            int res_cost = 0;
            vector<int> d, p;
            while (flow < K) {
                spfa(N, s, d, p);
                if (d[t] == INF) break;

                // find max flow on that path
                int f = K - flow;
                int cur = t;
                while (cur != s) {
                    f = min(f, capacity[p[cur]][cur]);
                    cur = p[cur];
                }

                // apply flow
                flow += f;
                res_cost += f * d[t];
                cur = t;
                while (cur != s) {
                    capacity[p[cur]][cur] -= f;
                    capacity[cur][p[cur]] += f;
                    cur = p[cur];
                }
            }

            for (auto e : edges) { // clearing
                cost[e.from][e.to] = cost[e.to][e.from] = 0;
                capacity[e.from][e.to] = capacity[e.to][e.from] = 0;
            }


            if (flow < K) return {-1,flow};
            else return {res_cost,flow};
        };

        sort(ALL(st.taken_masks));
        reverse(ALL(st.taken_masks));
        auto masks = st.taken_masks;
        if (!mcf_trim_all_masks) masks = VPII(1, {0, st.minl});

        for( int k=(int)masks.size()-1; k>=0; k-- ) {
            PII m = masks[k];

            vector<MCFEdge> g_arcs; g_arcs.reserve(all_mcf_g_arcs.size());
            for(int i=0; i<S_nodes.size(); i++) if(S_nodes[i] != -1) was[S_nodes[i]] = false;
            for( auto [a,b] : st.ends ) mcf_ends[a][b] = mcf_ends[b][a] = true;
            for( int i=0; i<32; i++ ) if( m.first & (1<<i) ) {
                for( int j=0; j<N; j++ ) mcf_ends[S_nodes[i]][j] = mcf_ends[j][S_nodes[i]] = true;
            }

            for(auto e : all_mcf_g_arcs){
                if( e.from+1 == e.to ) g_arcs.push_back(e);
                else {
                    int a = e.from, b = e.to;
                    a -= a&1; b -= b&1;
                    a >>= 1; b >>= 1;
                    if( !was[a] && !was[b] && !mcf_ends[a][b] ) g_arcs.push_back(e);
                }
            }
            for( auto [a,b] : st.ends ) mcf_ends[a][b] = mcf_ends[b][a] = false;
            for( int i=0; i<32; i++ ) if( m.first & (1<<i) ) {
                for( int j=0; j<N; j++ ) mcf_ends[S_nodes[i]][j] = mcf_ends[j][S_nodes[i]] = false;
            }
            for(int i=0; i<S_nodes.size(); i++) if(S_nodes[i] != -1) was[S_nodes[i]] = true;

            // now add source and sink - from source to vs_node and second nodes from ends, from first nodes and vt to sink
            // source - 2*N, sink - 2*N+1
            g_arcs.emplace_back( 2*N, 2*st.vs_node+1, 1, 0 );
            bool has_vt_end = false;
            for( auto [a,b] : st.ends ) {
                if( b == vt ) has_vt_end = true;
                if( b != vt ) g_arcs.emplace_back( 2*N, 2*b+1, 1, 0 );
                g_arcs.emplace_back( 2*a, 2*N+1, 1, 0 );
            }
            if( !has_vt_end ) g_arcs.emplace_back( 2*vt+1, 2*N+1, 1, 0 );

            int K = st.ends.size() + 1;
            if( has_vt_end ) K--;

            auto [flow_cost,flow] = minCostFlow( 2*N+2, g_arcs, K, 2*N, 2*N+1 );
            if( flow_cost != -1 ) assert( flow == K );
            else assert( flow < K );

            if constexpr (debug) if(  flow_cost == -1 || st.minl + flow_cost > L ) {
                ENDL(2);
                string msg = ( flow_cost == -1 ? "invalid" : "EXCESSIVE" );
                clog << "Found an " << msg << " flow for state st: " << st << endl;
                DEBUG(flow);
                DEBUG(flow_cost);
                ENDL(2);
            }

            if( flow_cost == -1 || m.second + flow_cost > L ) {
                if( m.first == 0 ) return true;
                else if( mcf_trim_all_masks ){
                    trimmed_masks.push_back(st.taken_masks[k]);
                    swap( st.taken_masks[k], st.taken_masks.back() );
                    st.taken_masks.pop_back();
                }
            }
        }
        return st.taken_masks.empty();
    };

    auto trimSmallEndsNeighborhood = [&]( FBSS3 & st ) {
        if constexpr (!use_small_neighborhood_trimming) return true;

        if( st.ends.empty()  ) return true;
        // if( st.vs_node == vt && !st.ends.empty() ) return false; // this is another rule, but can be done here :)
        // for( auto [a,b] : st.ends ) if( a == vt ) return false; // this is another rule, but can be done here :)

        _was_ = was;
        for( int i=0; i<S_nodes.size(); i++ ) if( S_nodes[i] != -1 ) _was_[ S_nodes[i] ] = true;

        // for( auto [a,b] : st.ends ) {
        //     int cnt = 0;
        //     for( auto & arc : revg[a] ) if( !_was_[arc.v] ) cnt++;
        //     if( a != vt && cnt == 0 ) return false;
        //
        //     cnt = 0;
        //     for( auto & arc: g[b] ) if( !_was_[arc.v] ) cnt++;
        //     if( b != vt && cnt == 0 ) return false;
        // }

        int neigh_size = 0;
        auto checkNode = [&](int d) { if( !_was_[d] ){ neigh_size++; _was_[d] = true; } };

        fill(ALL(_was2_),false);

        constexpr bool exclude_uniquely_assignable_nodes = true;
        if constexpr (exclude_uniquely_assignable_nodes){
            bool changes_done = true;

            while(changes_done) {
                changes_done = false;

                for( auto [a,b] : st.ends ) {
                    for( int d : underV[a] ) neigh_in_S[d] = 0;
                    for( int d : underV[b] ) neigh_in_S[d] = 0;
                }
                for( auto [a,b] : st.ends ) {
                    if( a != vt && !_was2_[a] ) for( int d : underV[a] ) neigh_in_S[d]++;
                    if( b != vt && !_was2_[b] ) for( int d : underV[b] ) neigh_in_S[d]++;
                }
                for( auto [a,b] : st.ends ) {
                    if(a != vt) for( auto & arc : revg[a] ) {
                        int d = arc.v;
                        if( !_was_[d] && !_was2_[a] && neigh_in_S[d] == 1 ) {
                            _was2_[a] = changes_done = true;
                            neigh_size++;
                        }
                    }
                    if(b != vt) for( auto & arc : g[b] ) {
                        int d = arc.v;
                        if( !_was_[d] && !_was2_[b] && neigh_in_S[d] == 1 ) {
                            _was2_[b] = changes_done = true;
                            neigh_size++;
                        }
                    }
                }
            }
        }

        int valid_ends = st.ends.size();
        for( auto [a,b] : st.ends ) {
            if( a != vt && !_was2_[a] ) for( auto & arc : revg[a] ) checkNode(arc.v);
            if( b != vt && !_was2_[b] ) for( auto & arc : g[b] ) checkNode(arc.v);
        }

        bool res = true;
        if( st.vs_node != vt && st.containsNode(vt) ) {
            if constexpr (debug) clog << "CASE1 " << endl;
            res = (neigh_size >= valid_ends);
        }
        else res = (neigh_size >= valid_ends+1);

        if constexpr (debug) if(!res) {
            clog << "State " << st << " has too small ends-neighborhood" << endl;
            DEBUG(neigh_size);
            DEBUG(valid_ends);
        }
        return res;
    };

    vector<vector<char>> cap(N+2, vector<char>(N+2,0));
    VVI flowV(N+2);

    auto trimSmallEndsNeighborhoodFlow = [&]( FBSS3 & st ) {

        if( st.ends.empty()  ) return true;// this cannot happen, as it was checked before adding newly created state
        if(st.vs_node == vt) return true;

        if constexpr (debug) clog << "Checking using flow method" << endl;

        int src = N;
        int trg = N+1;

        VI nodes(1,st.vs_node);
        VB is_b(N); // nodes from which flow must be outgoing...
        is_b[st.vs_node] = true;

        for( auto [a,b] : st.ends ) {
            if( a != vt ) nodes.push_back(a);
            if( b != vt ) nodes.push_back(b);
            is_b[b] = true;
        }

        VB vis(N);
        for( int a : nodes) {
            if( a == vt ) continue;

            cap[src][a] = 1;
            flowV[src].push_back(a);

            GraphSP2 * graph;
            if(is_b[a]) graph = &g;
            else graph = &revg;

            for( auto & arc : (*graph)[a] ) {
                int d = arc.v;

                if( !was[d] ) {
                    cap[a][d] = 1;
                    cap[d][trg] = 2;

                    flowV[a].push_back(d);
                    flowV[d].push_back(a);
                    if( !vis[d] ) flowV[d].push_back(trg);
                    vis[d] = true;
                }
            }
        }


        VB helper(N+2);
        VI pth;
        function<bool(int)> dfs = [&](int num) {
            if( num == trg ) {
                if constexpr (debug) {
                    pth.push_back(num);
                    clog << "Found augmenting path: " << pth << endl;
                    pth.pop_back();
                }
                return true;
            }
            helper[num] = true;
            if constexpr (debug) pth.push_back(num);

            for( int d : flowV[num] ) {
                if( !helper[d] && cap[num][d] > 0 ) {
                    if( dfs(d) ) {
                        cap[num][d]--;
                        cap[d][num]++;
                        if constexpr (debug) pth.pop_back();
                        return true;
                    }
                }
            }

            if constexpr (debug) pth.pop_back();
            return false;
        };

        int flow = 0;
        while( dfs(src) ) {
            flow++;
            helper = VB(N+2, false);
        }

        if constexpr (debug) DEBUG(S_nodes);

        int thr = nodes.size();
        if(st.containsNode(vt)) thr--;

        if constexpr (debug)
        if(flow < thr) {
            clog << "Trimming state " << st << " by flow rule!" << endl;
            DEBUG(vt);
            DEBUG(flow);
            DEBUG(nodes);
            for( auto d : nodes ) clog << "flowV[" << d << "]: " << flowV[d] << endl;
            // exit(1);
        }

        // now clearing
        for( int a : nodes) {
            cap[src][a] = cap[a][src] = 0;
            flowV[a].clear();
            for( int d : underV[a] ) if( !was[d] ) {
                flowV[d].clear();
                cap[a][d] = cap[d][a] = cap[d][trg] = cap[trg][d] = 0;
            }
        }
        flowV[src].clear();

        return flow >= thr;
    };

    auto trimFBSS3tates = [&](vector<FBSS3> & sts, const bool use_only_reachability = false) {
        for( int i=(int)sts.size()-1; i>=0; i-- ) {
            assert( !sts[i].taken_masks.empty() );

            bool can_trim = (sts[i].minl > L);
            if constexpr (debug) clog << "Trimming masks from state " << sts[i] << endl;

            if( !can_trim ) {
                if( sts[i].vs_node == vt && !sts[i].ends.empty() ) {
                    // clog << endl << "Trimming state with vs_node == vt and nonempty ends" << endl << endl;
                    can_trim = true;
                }
                if(!can_trim) {
                    for( auto [a,b] : sts[i].ends ) if( a == vt ) can_trim = true;
                }
            }

            if( !can_trim ) {
                auto check = [&](int a) {
                    for( int d : underV[a] ) if( ind_in_S[d] == -1 && !was[d] ) return false;
                    return true;
                };
                if( sts[i].vs_node != vt && check( sts[i].vs_node ) ) can_trim = true;
                for( auto [a,b] : sts[i].ends ) {
                    if( a != vt && check(a) ) can_trim = true;
                    if( b != vt && check(b) ) can_trim = true;
                }
            }

            VPII trimmed_masks;
            if(!use_only_reachability) { // traditional removal
                constexpr int PERMUTATION_THR = 6;
                if(!can_trim && sts[i].ends.size() <= PERMUTATION_THR ) can_trim = canTrimPermutation(sts[i], trimmed_masks);
                if(sts[i].ends.size() > PERMUTATION_THR && !can_trim) can_trim = canTrimMST(sts[i]); // is there a point to use it?

                if( !can_trim ) {
                    // #TEST - this is a bit unneccessarily timconsuming, but might be in some cases beneficial
                    can_trim = !trimSmallEndsNeighborhood(sts[i]);
                }
                // if( !can_trim ) can_trim = !trimSmallEndsNeighborhoodFlow(sts[i]);

                // setting this to a large value, but this is so slowe, that is should probably be removed... :(
                int FLOW_THR = MCF_ENDS_SIZE_THR;
                if( use_mcf_trimming && !can_trim && sts[i].ends.size() >= FLOW_THR ) {
                    can_trim = canTrimFlow(sts[i], trimmed_masks);
                }
            }

            constexpr int G_SIZE_THR = 0; // #TEST - disabled checking as it is time consuming and brings almost none benefit...
            // if( use_only_reachability && !can_trim && g.size() <= G_SIZE_THR ) {
            if( !can_trim && g.size() <= G_SIZE_THR ) {
                // this might be very time-consuming, depending on grach characteristics
                // a BFS is run for each taken_mask to check reachability
                can_trim = canTrimUnreachable(sts[i]);
            }

             if constexpr (debug) clog << "\ttrimmed_masks: " << trimmed_masks << endl;


            if constexpr (compress_on_the_fly) {
                if(can_trim) trimmed_masks += sts[i].taken_masks;
                total_masks_trimmed += trimmed_masks.size();

                for(PII m : trimmed_masks) { // we iterate over all masks
                    auto h = sts[i].hash ^ getHashForTakenMask(m.first);
                    if constexpr (debug) assert( state_id_mapper.contains(h) );
                    int num = state_id_mapper[h];
                    if( state_outdeg[num] == 0 ) {
                        // clog << endl << "!!! Trimming on the fly!" << endl;
                        trimOnTheFly(num);
                    }
                }

                if(!can_trim) can_trim = sts[i].taken_masks.empty();
            }

            if( can_trim ) {
                swap( sts[i], sts.back() );
                sts.pop_back();
            }
        }
    };

    auto addNodeForStates = [&](int v, vector<FBSS3> & sts) {
        markNeighbors(v, true); // mark in- and out-neighbors of node v within boundary set S

        for( auto & st : sts ) {
            if constexpr (debug) clog << "Adding for state st: " << st << endl;

            VI loose = getLooseNodes(st);
            if constexpr (debug) clog << "loose nodes: " << loose << endl;

            addAsIgnoredNode(v,st, loose);
            addAsVsNode(v,st, loose);
            addAsStartNode(v,st, loose);
            addAsMiddleNode(v,st, loose);
            addAsEndNode(v,st, loose);
        }

        markNeighbors(v,false); // unmark marked nodes...
    };

    auto removeFBSS3tates = [&](int v, vector<FBSS3> & sts, auto & states, int cur_st_id) {
        for( int i=(int)sts.size()-1; i>=0; i-- ) {

            if( sts[i].containsNode(v) ) {
                if constexpr (debug) clog << "\t-x-x-x-x-x removing state " << sts[i] << endl;
                if constexpr (compress_on_the_fly) {
                    for(PII m : sts[i].taken_masks) {
                        auto h = sts[i].hash ^ getHashForTakenMask(m.first);
                        int num = state_id_mapper[h];
                        if( state_outdeg[num] == 0 ) {
                            // clog << endl << "!!! Trimming on the fly!" << endl;
                            trimOnTheFly(num);
                        }
                    }
                }
                swap( sts[i], sts.back() );
                sts.pop_back();
                continue;
            }

            for( PII &m : sts[i].taken_masks ) {
                int mm = m.first;

                auto hm = getHashForTakenMask(m.first);
                auto h1 = sts[i].hash ^ hm;

                if constexpr (debug && compress_on_the_fly) assert( state_id_mapper.contains(h1) );
                m.first &= ~( 1 << ind_in_S[v] );

                hm = getHashForTakenMask(m.first);
                auto h2 = sts[i].hash ^ hm;
                trackState(h2);

                if constexpr (debug && compress_on_the_fly) {
                    aux_st_mapper[ state_id_mapper[h2] ] = sts[i];
                    VI newmasks = transformTakenMaskToNodes(m.first);
                    VPII new_takenmasks(newmasks.size(), {-1,-1});
                    for(int j=0; j<newmasks.size(); j++) new_takenmasks[j].first = newmasks[j];
                    aux_st_mapper[ state_id_mapper[h2] ].taken_masks = new_takenmasks;
                }

                if( h1 != h2 ) addEdge(h1,h2,-1,-1); // this is just a transition from states after removing v from taken masks
            }

            FBSS3::makeUnique(sts[i].taken_masks);

            if( (sts[i].hash.hash & BH) != cur_st_id ) {
                if constexpr (debug) clog << "\tRemoving state " << sts[i] << " from bucket " << cur_st_id
                                          << " and putting into " << (sts[i].hash.hash & BH) << endl;
                states[  sts[i].hash.hash & BH ].push_back( sts[i] );
                swap( sts[i], sts.back() );
                sts.pop_back();
            }
        }
    };

    auto createSameNeighborhoodSameHashesAndTMShifter = [&]() {
        fill(ALL(S_hashes),0);
        iota(ALL(tm_shifter),0);
        map<HashType,VI> mapper;

        for( int i=0; i<32; i++ ) if( S_nodes[i] != -1 ) {
            int v = S_nodes[i];
            if( v == vt ) continue; // we do not shift vt
            for( int d : underV[v] ) if( !was[d] ) {
                S_hashes[v] ^= node_hashes2[d] + all_arcs_hashes[ getArcId(v,d) ];
            }
            if( S_hashes[v] == 0 ) tm_shifter[v] = -1;
            else mapper[ S_hashes[v] ].push_back(v);
        }

        for( auto & [h,v] : mapper ) {
            sort(ALL(v), [&]( int a, int b ){ return ind_in_S[a] < ind_in_S[b]; } );
            for( int i=1; i<v.size(); i++) tm_shifter[v[i]] = v[i-1];
        }
    };

    auto addNode = [&](int v) {
        // if constexpr (debug)
            clog << "Adding node v: " << v << endl;
        for( int i=0; i<BUCKETS; i++ ) new_states[i].clear();

        createSameNeighborhoodSameHashesAndTMShifter();

        clog << "Old states: " << countAllStates(states) << endl;
        sw.start("addNodeForStates");
        for( int i=0; i<BUCKETS; i++ ) addNodeForStates( v, states[i] );
        sw.stop("addNodeForStates");

        // DEBUG(countAllStates(states));
        // DEBUG(countAllStates(new_states));
        clog << "New states before merging: " << countAllStates(new_states) << endl;

        sw.start("mergeFBSS3States");
        for( int i=0; i<BUCKETS; i++ ) mergeFBSS3States( new_states[i] );
        sw.stop("mergeFBSS3States");
        clog << "New states after merging: " << countAllStates(new_states) << endl;

        sw.start("trimFBSS3States");
        for( int i=0; i<BUCKETS; i++ ) trimFBSS3tates( new_states[i] );
        sw.stop("trimFBSS3States");
        clog << "New states after trimming: " << countAllStates(new_states) << endl;


        // DEBUG(countAllStates(new_states));
        swap(states, new_states);
    };



    auto removeNode = [&](int v) {
        // if constexpr (debug)
            clog << "Removing node v: " << v << endl;
        for( int i=0; i<BUCKETS; i++ ) new_states[i].clear();

        sw.start("removeFBSS3tates");
        for( int i=0; i<BUCKETS; i++ ) removeFBSS3tates( v, states[i], new_states, i );
        for( int i=0; i<BUCKETS; i++ ) {
            states[i] += new_states[i];
            new_states[i].clear();
        }
        sw.stop("removeFBSS3tates");

        sw.start("mergeFBSS3States");
        for( int i=0; i<BUCKETS; i++ ) mergeFBSS3States( states[i] );
        sw.stop("mergeFBSS3States");

    };

    was[vs] = true;
    S.insert(vs);
    S_nodes[0] = vs;
    ind_in_S[vs] = 0;
    order.erase(order.begin());
    cur_vis_hash = node_hashes2[vs];

    FBSS3 initial_state({}, vs, 0, cur_vis_hash );
    // FBSS3 initial_state({}, vs, 0 );
    initial_state.taken_masks.push_back({0,0});
    initial_state.recalculateHash();
    trackState(initial_state.hash);
    if constexpr ( include_iter_cnt_in_hash ) updateHashWithIterationCnt(initial_state.hash);

    if constexpr (compress_on_the_fly) assert(state_id_mapper.size() == 1);
    if constexpr (debug) aux_st_mapper[0] = initial_state;

    states[ initial_state.hash.hash & BH ].push_back( initial_state );

    for( int v : order ) {
        if( v == vt && was[v] ) continue;

        if(timer.tle("fbs")) return {};

        clog << "Current boundary nodes: " << S.size() << " -> " << S << endl;
        DEBUG(countAllStates(states));
        if constexpr (compress_on_the_fly){ DEBUG( state_id_mapper.size() ); }
        else DEBUG(trg_edges_h.size());
        clog << "Time since start of FBS3: " << (int)(timer.getUnstoppedTimeSinceLastStart("fbs") / 1000) << "s" << endl;


        if constexpr (debug) {
            ENDL(5); ENDLS(50, "*"); ENDL(5);
            clog << "Current states: " << endl;
            for(int i=0; i<BUCKETS; i++) {
                if( !states[i].empty() ) clog << "states[" << i << "]: ";
                for( auto & st : states[i] ) clog << st << endl;
            }
            ENDL(2);
        }

        if constexpr (debug) {
            DEBUG(S_nodes);
            DEBUG(ind_in_S);
        }

        for( auto & bucket : states ) for( auto & st : bucket ) st.recalculateMinl();

        if( was[v] ) {
            cur_vis_hash ^= node_hashes2[v];
            removeNode(v);

            S.erase(v);
            int ind = ind_in_S[v];
            S_nodes[ind] = -1;
            ind_in_S[v] = -1;
        }
        else{
            expansion_iteration_cnt++;

            cur_vis_hash ^= node_hashes2[v];
            was[v] = true;
            int ind = getFirstFreeSIndex();
            S_nodes[ind] = v;
            ind_in_S[v] = ind;

            sw.start("create distances");
            createDistances(v);
            sw.stop("create distances");

            sw.start("create dominators");
            if( use_dominator_trimming ) createDominators(v);
            sw.stop("create dominators");

            if constexpr (debug) {
                DEBUG(S_nodes);
                DEBUG(ind_in_S);
            }

            addNode(v);

            S.insert(v);
        }


        { // some statistics...
            ENDL(1);
            VI ends_cnt(S.size(),0);
            for( int i=0; i<BUCKETS; i++ ) {
                for( auto & st : states[i] ) ends_cnt[st.ends.size()]++;
            }

            clog << "Number of states with given number of ends.size(): " << endl;
            for(int i=0; i < min(ends_cnt.size(), 2 + ends_cnt.size() / 2); i++) {
                clog << i << ": " << ends_cnt[i] << endl;
            }

            // int VAL = 6;
            // clog << "States with ends.size() == " << VAL << endl;
            // for( int i=0; i<BUCKETS; i++ ) {
            //     for( auto & st : states[i] ) if( st.ends.size() == VAL ) clog << st << endl;
            // }

            ENDL(1);
        }
    }

    vector<FBSS3> all_states; // only for debugging...
    if constexpr (debug) {
        for(auto & vec : states) all_states += vec;
        clog << "All states: " << endl;
        for(auto & st : all_states) clog << st << endl;
        clog << "trg_edges: " << endl;
        for( auto e : trg_edges ) clog << "(" << e.a << "," << e.b << ")" << endl;
    }

    ULL all_states_cnt = 0;
    FBSS3 terminal_state;
    for(auto & vec : states) {
        all_states_cnt += vec.size();
        if(!vec.empty()) terminal_state = vec[0];
        vector<FBSS3>().swap(vec); // releasing space
    }
    // if constexpr (include_iter_cnt_in_hash) updateHashWithIterationCnt(terminal_state.hash );
    for(auto & vec : new_states) vector<FBSS3>().swap(vec); // releasing space


    assert( all_states_cnt <= 1 ); // there should be only one state - the one with node vt
    if(all_states_cnt == 1) {
        assert( terminal_state.vs_node == vt );
        assert( terminal_state.taken_masks.size() == 1 );
    }

    if( all_states_cnt == 0 ) return VBIG(L+1,0);

    if constexpr (debug) DEBUG(initial_state.hash.hash);

    int init_state_id = -1;
    int last_state_id = -1;

    int trgN;

    // using countsort is considerably faster, but requires copy of arcs, so roughly doubles memory peak...
    bool use_countsort = ( trg_edges_h.size() < 3e8 );

    if constexpr (compress_on_the_fly) {
        trgN = state_id_mapper.size();

        init_state_id = state_id_mapper[ initial_state.hash ];
        last_state_id = state_id_mapper[ terminal_state.hash ];

        if constexpr (debug) {
            DEBUG(free_ids);
            DEBUG(first_valid_edge_index);
        }
        clog << "Before removing invalid edges, trg_edges.size(): " << trg_edges.size() << endl;

        for( int i=(int)trg_edges.size()-1; i>=0; i-- ) {
            auto e = trg_edges[i];
            if( first_valid_edge_index[e.a] > i || first_valid_edge_index[e.b] > i ) {
                swap( trg_edges[i], trg_edges.back() );
                trg_edges.pop_back();
            }
        }

        // if constexpr (debug)
            clog << "After removing invalid edges, trg_edges.size(): " << trg_edges.size() << endl;
    }else {
        clog << "Remapping trg_edges_h" << endl;
        sw.start("remapping trg_edges_h");
        vector<HashType> all_hashes; all_hashes.reserve( 2 * trg_edges_h.size() );

        int BUCKETS_BITS = max(1, min(18, (int)ceil( log2(trg_edges_h.size()) ) ) );
        // constexpr int BUCKETS = (1 << BUCKETS_BITS);

        // for(auto &e : trg_edges_h){ all_hashes.push_back(e.a); all_hashes.push_back(e.b); }
        for(auto &e : trg_edges_h) {
            all_hashes.push_back(e.a); // #TEST
            if( e.b == terminal_state.hash ) all_hashes.push_back(e.b);
        }

        sw.start("remapping sort - all hashes");
         if (use_countsort) countSortTrgEdgesHash<HashType, 2>(BUCKETS_BITS, all_hashes);
         else sort( ALL(all_hashes) );

        // unsignedIntegerRadixSort(all_hashes);
        sw.stop("remapping sort - all hashes");
        all_hashes.resize( unique(ALL(all_hashes)) - all_hashes.begin() );
        // clog << "all_hashes: " << endl; for(auto e : all_hashes) clog << e.hash << endl;
        // assert( is_sorted(ALL(all_hashes)) );

        for( int i=0; i<all_hashes.size(); i++ ) {
            if( all_hashes[i] == initial_state.hash ) init_state_id = i;
            if( all_hashes[i] == terminal_state.hash ) last_state_id = i;
        }



        sw.start("remapping sort");
         if (use_countsort) countSortTrgEdgesHash<EdgeH, 0>(BUCKETS_BITS, trg_edges_h); // sort by a
         else sort( ALL(trg_edges_h), [](auto &a, auto &b){ return a.a < b.a; } );
        sw.stop("remapping sort");
        // clog << "trg_edges_h: " << endl; for(auto e : trg_edges_h) clog << "(" << e.a << "," << e.b << ")" << endl;
        // assert( is_sorted(ALL(trg_edges_h), [](auto a, auto b){ return a.a < b.a; } ) );

        int p = 0, q = 0;
        while( p < all_hashes.size() && q < trg_edges_h.size() ) {
            if( all_hashes[p] == trg_edges_h[q].a ) trg_edges_h[q++].a.hash = p;
            else if( all_hashes[p] < trg_edges_h[q].a ) p++;
            else q++;
        }



        sw.start("remapping sort");
        if (use_countsort) countSortTrgEdgesHash<EdgeH, 1>(BUCKETS_BITS, trg_edges_h); // sort by b
        else sort( ALL(trg_edges_h), [](auto &a, auto &b){ return a.b < b.b; } );
        sw.stop("remapping sort");
        // clog << "trg_edges_h: " << endl; for(auto e : trg_edges_h) clog << "(" << e.a << "," << e.b << ")" << endl;
        // assert( is_sorted(ALL(trg_edges_h), [](auto a, auto b){ return a.b < b.b; } ) );

        p = 0, q = 0;
        while( p < all_hashes.size() && q < trg_edges_h.size() ) {
            if( all_hashes[p] == trg_edges_h[q].b ) trg_edges_h[q++].b.hash = p;
            else if( all_hashes[p] < trg_edges_h[q].b ) p++;
            else q++;
        }


        trg_edges.clear();
        trgN = 0;
        for( auto &e : trg_edges_h ) {
            if(e.a.hash > all_hashes.size() || e.b.hash > all_hashes.size()) continue;
            trgN = max( trgN, (int)max( e.a.hash, e.b.hash ) );
            trg_edges.emplace_back(e.a.hash, e.b.hash, e.e1, e.e2);
        }
        trgN++;
        vector<EdgeH>().swap(trg_edges_h);

        sw.stop("remapping trg_edges_h");
    }


    if constexpr (debug) {
        DEBUG(trgN);
        DEBUG(init_state_id);
        DEBUG(last_state_id);
    }



    clog << "TRANSITION GRAPH ARCS CREATED" << endl;
    clog << "Before transition graph creation, it has state_id_mapper.size(): " << state_id_mapper.size() << " nodes and "
         << trg_edges.size() << " arcs, trgN: " << trgN << endl;




    if constexpr (compress_on_the_fly) {
        trgN = state_predecessors.size();

        vector<Pred>().swap(state_predecessors); // now clearing all data used to trim on the fly
        VI().swap(state_outdeg);
        vector<HashType>().swap(state_hash_remapper);
        VI().swap(first_valid_edge_index);
    }
    unordered_map<HashType,int, HashTypeHash>().swap(state_id_mapper); // releasing space

    if constexpr (debug) {
        clog << "aux_st_mapper keys:" << endl;
        for(auto [k,v] : aux_st_mapper) clog << k << " "; clog << endl;
    }

    auto constructTransitionGraph = [&](int trgN, auto & trg_edges, bool trg_edges_sorted_by_a = false) {
        sw.start("TRG creation");
        if( !trg_edges_sorted_by_a ) {
            // sort(ALL(trg_edges), []( auto & e1, auto & e2 ){ return PII( e1.a, e1.b ) < PII( e2.a, e2.b ); });
            if (use_countsort) Solver2FBS::countSortTrgEdges(trgN, trg_edges);
            else sort( ALL(trg_edges), [](auto &a, auto &b){ return a.a < b.a; } );
        }else {
            clog << "Abandoning sorting trg_edges, as it is told, that they are already sorted" << endl;
            assert( is_sorted(ALL(trg_edges), [](auto a, auto b){ return a.a < b.a; } ) );
        }
        VI first_of(trgN+1, trg_edges.size());
        for( int i=0; i<trg_edges.size(); i++ ) {
            auto& e = trg_edges[i];
            // first_of[e.a] = min(  i, first_of[e.a]);
            first_of[e.a] = ( i <= first_of[e.a] ? i : first_of[e.a] );
        }
        // for( int i=trgN-1; i>=0; i-- ) if( first_of[i] == trg_edges.size() ) first_of[i] = first_of[i+1];
        for( int i=trgN-1; i>=0; i-- ) first_of[i] = (first_of[i] == trg_edges.size() ? first_of[i+1] : first_of[i]);
        sw.stop("TRG creation");

        return first_of;
    };

    VI first_of;
    if constexpr (compress_on_the_fly) {
        first_of = constructTransitionGraph(trgN, trg_edges);
        trgN = first_of.size()-1;
    }

    clog << "Before trimming and remapping graph it has " << trgN << " nodes and "
         << trg_edges.size() << " arcs " << endl;

    auto trimTRG = [&constructTransitionGraph, &init_state_id, &last_state_id, &sw](int trgN, vector<Edge> & trg_edges,
        bool trg_edges_sorted_by_b = false) {
        sw.start("trimTRG");
        // vector<Edge> trg_edges_r = trg_edges;
        vector<unsigned short> outdeg(trgN,0);
        // for( auto &e : trg_edges_r ){ swap( e.a, e.b ); ++outdeg[e.b]; }
        for( auto &e : trg_edges ){ swap( e.a, e.b ); ++outdeg[e.b]; }
        // auto first_of_r = constructTransitionGraph( trgN, trg_edges_r );
        auto first_of_r = constructTransitionGraph( trgN, trg_edges, trg_edges_sorted_by_b );

        VB valid(trgN,true);

        VI to_trim; to_trim.reserve(trgN/2);
        for( int i=0; i<trgN; i++ ) if( outdeg[i] == 0 && i != last_state_id ) to_trim.push_back(i);
        for( int j=0; j<to_trim.size(); j++ ) {
            int num = to_trim[j];
            valid[num] = false;
            for( int i=first_of_r[num]; i < first_of_r[num+1]; i++ ) {
                // auto &e = trg_edges_r[i];
                auto &e = trg_edges[i];
                if( --outdeg[e.b] == 0 ) to_trim.push_back(e.b);
            }
        }
        VI().swap(to_trim);
        vector<unsigned short>().swap(outdeg);
        for( auto &e : trg_edges ){ swap( e.a, e.b ); }

        assert( valid[init_state_id] );
        assert( valid[last_state_id] );

        VI mapper(trgN,-1);
        int ind = 0;
        for( int i=0; i<trgN; i++ ) if( valid[i] ) mapper[i] = ind++;
        int r = 0;
        for(int i=0; i<trg_edges.size(); i++) {
            auto &e = trg_edges[i];
            if( mapper[e.a] != -1 && mapper[e.b] != -1 ) {
                e.a = mapper[e.a];
                e.b = mapper[e.b];
                trg_edges[r++] = e;
            }
        }
        last_state_id = mapper[last_state_id];
        init_state_id = mapper[init_state_id];
        while(trg_edges.size() > r) trg_edges.pop_back();

        VI first_of = constructTransitionGraph( ind, trg_edges);
        sw.stop("trimTRG");
        return first_of;
    };

    first_of = trimTRG(trgN, trg_edges, true);
    trgN = first_of.size()-1;

    clog << "After trimming (without remapping) transition graph has " << trgN << " nodes and " << trg_edges.size()
         << " arcs " << endl;
    int path_compressable_nodes = 0;
    int fully_compressable_nodes1 = 0;
    int fully_compressable_nodes2 = 0;

    auto calculateCompressableNodes = [&path_compressable_nodes, &fully_compressable_nodes1,
        &fully_compressable_nodes2, &sw] (int trgN, auto & trg_edges) {
        return 0;
        sw.start("calculateCompressableNodes");
        path_compressable_nodes = 0;
        fully_compressable_nodes1 = 0;
        fully_compressable_nodes2 = 0;
        VI indeg(trgN,0), outdeg(trgN,0);
        for(auto e : trg_edges){ outdeg[e.a]++; indeg[e.b]++; }
        for( int i=0; i<trgN; i++ ) {
            if( indeg[i] == 1 && outdeg[i] == 1 ) path_compressable_nodes++;
            if( indeg[i] >= 1 && outdeg[i] == 1 ) fully_compressable_nodes1++;
            if( indeg[i] == 1 && outdeg[i] >= 1 ) fully_compressable_nodes2++;
        }
        sw.stop("calculateCompressableNodes");
    };
    calculateCompressableNodes(trgN, trg_edges);



    auto compressTRG = [&](int trgN, VI & first_of, vector<Edge> & trg_edges) {
        sw.start("compressTRG");

        sw.start("forward_compression");
        vector<CompressedEdge> cmp_trg_edges;
        cmp_trg_edges.reserve( trgN - fully_compressable_nodes1 + 10 );

        VI cmp_trg_first_of;

        constexpr bool apply_backward_compression = false;

        {
            VB valid(trgN,true);

            vector<unsigned short> outdeg(trgN,0);
            for(auto e : trg_edges) ++outdeg[e.a];

            for( int i=0; i<trgN; i++ ) {
                for( int j=first_of[i]; j < first_of[i+1]; j++ ) {
                    Edge e = trg_edges[j];
                    if( e.a != init_state_id && outdeg[e.a] == 1  ) continue;

                    VI arcs;
                    if( e.e1 != -1 ) arcs.push_back(e.e1);
                    if( e.e2 != -1 ) arcs.push_back(e.e2);

                    while( e.b != last_state_id && outdeg[e.b] == 1 ) {
                        valid[e.b] = false;
                        auto ee = trg_edges[ first_of[e.b] ];
                        if( ee.e1 != -1 ) arcs.push_back(ee.e1);
                        if( ee.e2 != -1 ) arcs.push_back(ee.e2);
                        e.b = ee.b;
                    }

                    cmp_trg_edges.emplace_back( e.a, e.b );
                    swap(arcs, cmp_trg_edges.back().arcs);
                }
            }
            assert(valid[init_state_id]);
            assert(valid[last_state_id]);

            VI mapper(trgN,-1);
            int ind = 0;
            for( int i=0; i<trgN; i++ ) if( valid[i] ) mapper[i] = ind++;
            last_state_id = mapper[last_state_id];
            init_state_id = mapper[init_state_id];
            for( auto & e : cmp_trg_edges ) { e.a = mapper[e.a]; e.b = mapper[e.b]; }

            trgN = ind;

            if constexpr (!apply_backward_compression) {
                cmp_trg_first_of = VI(trgN+1, cmp_trg_edges.size());
                for( int i=0; i<cmp_trg_edges.size(); i++ ) {
                    auto& e = cmp_trg_edges[i];
                    cmp_trg_first_of[e.a] = ( i <= cmp_trg_first_of[e.a] ? i : cmp_trg_first_of[e.a] );
                }
                for( int i=trgN-1; i>=0; i-- ) {
                    cmp_trg_first_of[i] = (cmp_trg_first_of[i] == cmp_trg_edges.size() ? cmp_trg_first_of[i+1] : cmp_trg_first_of[i]);
                }
            }
        }
        sw.stop("forward_compression");

        clog << "After forward-compression graph has " << trgN << " nodes and " << cmp_trg_edges.size()
             << " arcs" << endl;
        calculateCompressableNodes(trgN, cmp_trg_edges);

        if constexpr (apply_backward_compression) {
            sw.start("backward_compression");
            for(auto & e : cmp_trg_edges) swap(e.a, e.b);
            auto cmp_trg_first_of_r = constructTransitionGraph( trgN, cmp_trg_edges );

            { // doing the same as before, but for reversed graph
                VB valid(trgN,true);

                vector<CompressedEdge> new_cmp_trg_edges;
                new_cmp_trg_edges.reserve( trgN - fully_compressable_nodes1 + 10 );

                VI outdeg(trgN,0);
                for(auto e : cmp_trg_edges) ++outdeg[e.a];

                for( int i=0; i<trgN; i++ ) {
                    for( int j=cmp_trg_first_of_r[i]; j < cmp_trg_first_of_r[i+1]; j++ ) {
                        CompressedEdge e = cmp_trg_edges[j];
                        if( e.a != last_state_id && outdeg[e.a] == 1  ) continue;

                        VI arcs = e.arcs;

                        while( e.b != last_state_id && outdeg[e.b] == 1 ) {
                            valid[e.b] = false;
                            auto & ee = cmp_trg_edges[ cmp_trg_first_of_r[e.b] ];
                            arcs += ee.arcs;
                            e.b = ee.b;
                        }

                        new_cmp_trg_edges.emplace_back( e.a, e.b );
                        swap(arcs, new_cmp_trg_edges.back().arcs);
                    }
                }
                assert(valid[init_state_id]);
                assert(valid[last_state_id]);

                VI mapper(trgN,-1);
                int ind = 0;
                for( int i=0; i<trgN; i++ ) if( valid[i] ) mapper[i] = ind++;
                last_state_id = mapper[last_state_id];
                init_state_id = mapper[init_state_id];
                for( auto & e : new_cmp_trg_edges ) { e.a = mapper[e.a]; e.b = mapper[e.b]; }
                trgN = ind;
                swap(new_cmp_trg_edges, cmp_trg_edges);
            }

            for(auto & e : cmp_trg_edges) swap(e.a, e.b);
            cmp_trg_first_of = constructTransitionGraph( trgN, cmp_trg_edges );

            sw.stop("backward_compression");
        }

        sw.stop("compressTRG");

        return make_tuple(cmp_trg_first_of, cmp_trg_edges);
    };


    vector<CompressedEdge> cmp_trg_arcs;
    tie(first_of, cmp_trg_arcs) = compressTRG(trgN, first_of, trg_edges);
    trgN = first_of.size()-1;
    vector<Edge>().swap(trg_edges); // releasing space
    clog << "After compression transition graph has " << trgN << " nodes and " << cmp_trg_arcs.size() << " arcs " << endl;

    clog << "Proceeding to result calculation" << endl;

    calculateCompressableNodes(trgN, cmp_trg_arcs);
    DEBUG(path_compressable_nodes);
    DEBUG(fully_compressable_nodes1);
    DEBUG(fully_compressable_nodes2);

    // vector<Lengths> lengths(trgN);

    was = VB(trgN,false);

    bool use_large_bigint = false;

    VBIG res(L+1,0);
    // auto calculateResult = [&]<typename BI, typename LNG>( vector<LNG> & lengths, vector<LNG> & all_arcs_lengths,
    auto calculateResult = [&]<typename BI, typename LNG>( vector<LNG*> & lengths, vector<LNG> & all_arcs_lengths,
        BI _bigint_, BI thr_value ) {

        constexpr bool solve_recursively = false;

        if constexpr (solve_recursively) {
            sw.start("Recursive result calculation");
            clog << "Solving recursively" << endl;

            lengths[last_state_id] = SP2Arc::mul_one;

            was = VB(trgN,false);
            was[last_state_id] = true;

            VB is_valid(trgN,false);
            is_valid[last_state_id] = true;

            int depth = 0;
            auto indent = [&](){ for(int i=0; i<depth; i++) clog << "   "; };
            function<void(int)> getRes = [&]( int num ) {
                depth++;
                if constexpr (debug) {
                    indent();
                    clog << "entering num: " << num << endl;
                }


                for( int i=first_of[num]; i < first_of[num+1]; i++ ) { // version for compressed graph
                    auto &e = cmp_trg_arcs[i];
                    int a = e.a, b = e.b;

                    if( !was[b] ) getRes(b);
                    if( !is_valid[b] ) continue;

                    Lengths temp = lengths[b];

                    Lengths mul = SP2Arc::mul_one;
                    for( int eid : e.arcs ) {
                        mul *= all_arcs_lengths[eid];
                        mul.trimToMax(L);
                    }
                    temp *= mul;
                    temp.trimToMax(L);

                    lengths[num] += temp;
                    lengths[num].trimToMax(L);
                    lengths[num].minimize();
                }

                depth--;
                was[num] = true;
                lengths[num].minimize();
                is_valid[num] = lengths[num].valid();
            };

            getRes(init_state_id);


            auto & lr = lengths[init_state_id];
            if(lr.valid()) for(int i=lr.minl(); i <= lr.maxl(); i++) res[i] = lr[i];

            DEBUG(lengths[init_state_id]);
            DEBUG(lengths[init_state_id].sum());

            int truly_used_states = 0;
            // for(int i=0; i<trgN; i++) if( was[i] && lengths[i].valid() ) truly_used_states++;
            for(int i=0; i<trgN; i++) if( was[i] && is_valid[i] ) truly_used_states++;
            // if constexpr (debug)
            clog << "Truly used states (somewhow compressed transition graph size): " << truly_used_states
                << ", that is " << (100.0 * truly_used_states / trgN ) << "% of all states" << endl;


            sw.stop("Recursive result calculation");
        }else {
            was = VB(trgN,false);
            sw.start("Toposort result calculation");
            clog << "Solving iteratively by toposort" << endl;

            sw.start("creating topological order");
            VI topo_order;
            VI indeg(trgN);
            for( auto & e : cmp_trg_arcs ) ++indeg[e.b];

            function<void(int)> toposort = [&](int num) {
                topo_order.push_back(num);
                was[num] = true;
                for( int i=first_of[num]; i < first_of[num+1]; i++ ) {
                    auto & e = cmp_trg_arcs[i];
                    if( --indeg[e.b] == 0 ) toposort(e.b);
                }
            };

            for( int i=0; i<trgN; i++ ) if( !was[i] && indeg[i] == 0 ) toposort(i);
            sw.stop("creating topological order");
            // for( int i=0; i<trgN; i++ ) if( indeg[i] == 0 ) toposort(i);

            DEBUG(trgN);
            DEBUG(topo_order.size());
            assert(topo_order.size() == trgN);

            // lengths[last_state_id] = LNG( vector<BI>(1,MULTIPLIER),0 );
            lengths[last_state_id] = new LNG( vector<BI>(1,1),0 );

            VI first_pred(trgN,-1);
            for( int v : topo_order ) {
                for( int i=first_of[v]; i < first_of[v+1]; i++ ) {
                    int b = cmp_trg_arcs[i].b;
                    if( first_pred[b] == -1 ) first_pred[b] = v;
                }
            }

            VI src_dst(trgN,inf);
            src_dst[init_state_id] = 0;
            for( int v : topo_order ) {
                for( int i=first_of[v]; i < first_of[v+1]; i++ ) {
                    auto & e = cmp_trg_arcs[i];
                    int suml = 0;
                    for( int eid : e.arcs ) suml += all_arcs_lengths[eid].minl();
                    src_dst[e.b] = min( src_dst[e.b], src_dst[v] + suml );
                }
            }

            sw.start("calculating for topological order");
            reverse(ALL(topo_order));
            VI to_clear;
            int dispensible_arcs = 0;
            int valid_arcs = 0;

            for( int v : topo_order ) {
                to_clear.clear();

                if( v != last_state_id ) lengths[v] = new LNG(); // zero

                for( int i=first_of[v]; i < first_of[v+1]; i++ ) {
                    auto & e = cmp_trg_arcs[i];
                    assert( lengths[e.b] != nullptr );
                    if( !lengths[e.b]->valid() ) continue;

                    int suml = 0;
                    for( int eid : e.arcs ) suml += all_arcs_lengths[eid].minl();

                    if( src_dst[e.a] + suml + lengths[e.b]->minl() > L ) {
                        dispensible_arcs++;
                        if( first_pred[e.b] == v ) to_clear.push_back(e.b);
                        continue;
                    }

                    LNG mul = LNG( vector<BI>(1,1), 0 ); // equivalent to SP2Arc::mul_one
                    for( int eid : e.arcs ) {
                        mul.trimToMax(L - all_arcs_lengths[eid].minl());
                        mul *= all_arcs_lengths[eid];
                    }

                    mul.trimToMax(L - lengths[e.b]->minl());
                    mul.minimize();
                    LNG temp = *lengths[e.b];
                    temp.trimToMax(L - mul.minl());

                    if( mul.valid() && temp.valid() ) {
                        valid_arcs++;
                        temp *= mul;

                        *lengths[v] += temp;
                        lengths[v]->trimToMax(L);
                        lengths[v]->minimize();
                    }

                    if( first_pred[e.b] == v ) to_clear.push_back(e.b);
                }

                for( auto & bi : lengths[v]->l ) if( bi > thr_value ) use_large_bigint = true;
                // for(int b : to_clear) vector<BI>().swap( lengths[b].l );
                for(int b : to_clear) if( lengths[b] != nullptr ){ delete lengths[b]; lengths[b] = nullptr; }
                if( v == init_state_id ) break;
            }

            DEBUG(dispensible_arcs);
            DEBUG(valid_arcs);
        }
    };

        auto toStringSmall = [&](small_bigint b) {
        string s = "";
        while( b > 0 ) {
            small_bigint rem = (b % 10);
            small_bigint temp = rem;
            for( int i=0; i<10; i++ ) {
                temp = i;
                if( temp == rem ) s += char( '0' + i );
            }
            b /= 10;
        }
        reverse(ALL(s));
        return s;
    };

    auto toStringLarge = [&](bigint b) {
        string s = "";
        while( b > 0 ) {
            bigint rem = (b % 10);
            bigint temp = rem;
            for( int i=0; i<10; i++ ) {
                temp = i;
                if( temp == rem ) s += char( '0' + i );
            }
            b /= 10;
        }
        reverse(ALL(s));
        return s;
    };

    auto convertLargeToSmall = [&](bigint large, small_bigint small) {
        auto s = toStringLarge(large);
        small = 0;
        for( auto c : s ) {
            small *= 10;
            small += (int)(c - '0');
        }
        return small;
    };

    auto convertSmallToLarge = [&](small_bigint small, bigint large) {
        auto s = toStringSmall(small);
        large = 0;
        for( auto c : s ) {
            large *= 10;
            large += (int)(c - '0');
        }
        return large;
    };


    bool use_small_bigint = true;

    if( use_small_bigint ) {
        clog << "Using small bigint" << endl;

        using BI = small_bigint;
        using LNG = LengthsT<BI>;
        vector<LNG*> lengths(trgN); // calculate using ULL
        vector<LNG> all_arcs;
        all_arcs.reserve( all_arcs_lengths.size() );
        for( auto & arc : all_arcs_lengths ) {
            vector<BI> t( arc.l.size() );
            for( int i=0; i<t.size(); i++ ) t[i] = convertLargeToSmall(arc[ arc.minl() + i], (BI)0);
            all_arcs.emplace_back(t, arc.minl());
        }
        small_bigint small_thr_value = 1;
        {
            int times;
#ifdef USE_BOOST_SMALL_BIGINT
            times = 60;
#else
            times = 32;
#endif
            for( int i=0; i<times; i++ ) small_thr_value *= 10;
        }
        calculateResult(lengths, all_arcs, (BI)0, small_thr_value);



        auto & lr = *lengths[init_state_id];
        if(lr.valid()) {
             for(int i=lr.minl(); i <= lr.maxl(); i++) res[i] = convertSmallToLarge( lr[i], (bigint)0 );
        }

        bigint thr_value = 1;

        int times;
#ifdef USE_BOOST_SMALL_BIGINT
        times = 60;
#else
        times = 32;
#endif

        for( int i=0; i<times; i++ ) thr_value *= 10;

        auto sum = accumulate( ALL(res), (bigint)0 );
        if( sum < 0 || sum > thr_value ) {
            clog << "result " << sum << " is < 0   or   > 10^" << times
                 << ", recalculating using arbitrary precision" << endl;
            res = VBIG(L+1,0);
            use_large_bigint = true;
        }

        for( auto ptr : lengths ) if( ptr != nullptr ){ delete ptr; ptr = nullptr; }
    }


    if(use_large_bigint){

#if defined(USE_BOOST_BIGINT)
        clog << "Using large bigint - BOOST" << endl;
#elif defined(USE_GMP_BIGINT)
        clog << "Using large bigint - GMP" << endl;
#else
        clog << "Using large bigint - InfInt" << endl;
#endif

        vector<Lengths*> lengths(trgN);
        calculateResult(lengths, all_arcs_lengths, (bigint)0, (bigint)0); // calculate using InfInt or any other multiple-precision

        auto & lr = *lengths[init_state_id];
        if(lr.valid()) for(int i=lr.minl(); i <= lr.maxl(); i++) res[i] = lr[i];

        for( auto ptr : lengths ) if( ptr != nullptr ){ delete ptr; ptr = nullptr; }
    }


    sw.stop("calculating for topological order");



    sw.stop("Toposort result calculation");


    sw.stop("solveFBS3ForNodeOrder");
    sw.writeAll();


    clog << "returning result" << endl;

    timer.stop("fbs");

    return res;
}

